#========================================================================================
# §package      DE_EPAGES::Catalog::API::Resource::CategoriesTree
# §state        public
# §base         DE_EPAGES::WebService::API::Resource::BaseObjectResource
# §description  implements the REST resource for a CategoriesTree
#========================================================================================
package DE_EPAGES::Catalog::API::Resource::CategoriesTree;
use base DE_EPAGES::WebService::API::Resource::BaseObjectResource;

use DE_EPAGES::Core::API::PerlTools qw ( JsonTrue );
use DE_EPAGES::Core::API::Url qw ( ParseQueryString );
use DE_EPAGES::Object::API::Factory qw ( LoadObjectByPath LoadClassByAlias );
use DE_EPAGES::Object::API::Object::Locale;
use DE_EPAGES::Shop::API::Resource::ShopRequestContext;
use DE_EPAGES::Database::API::Connection qw( GetCurrentDBHandle );
use JSON::XS;

use strict;



sub service {
    return LoadObjectByPath('/RestServices/CategoriesTree');
}



sub class {
    return 'Shop';
}



sub producedContentTypes {
    shift;
    my ($RequestMethod) = @_;

    return {
        GET => [
            ['get', 'application/json'],
            ['get', 'application/vnd.epages.v1+json'],
        ],
    }->{$RequestMethod};
}



sub match {
    my $self = shift;
    my ($Request) = @_;

    return 1 if $Request->uri->path =~ m|^/api/categories-tree$|;
    return undef;
}



sub requestContext {
    my $self = shift;
    my ( $Site, $GUID, $Request ) = @_;

    return $self->SUPER::requestContext( $Site, $GUID, $Request );
}



sub params {
    my $self = shift;
    my ( $ActionName, $Object, $Request ) = @_;

    if ($ActionName eq 'get') {
        my $hQuery = ParseQueryString( $Request->uri->query, 'utf-8' );
        my $IncludeInvisible = $hQuery->{'includeInvisible'}->[0];
        my $LocaleID = $hQuery->{'locale'}->[0];
        return [ $Object, $LocaleID, $IncludeInvisible ];
    } else {
        return [$Object];
    }
}


sub get {
    my $self = shift;
    my ( $Shop, $LocaleID, $IncludeInvisible ) = @_;

    my $Context = $self->_createShopRequestContext($Shop);
    $LocaleID = $Context->validLocale($LocaleID);

    my $Locale = $self->_loadLocale($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');

    $IncludeInvisible = JsonTrue($IncludeInvisible);


    my $DBI = GetCurrentDBHandle();

    #-- Find the root 'Categories' object of the shop, MUST always exist
    my $RootCat = $Shop->get('Categories');
    return {} unless $RootCat;

    my $CategoryClass = LoadClassByAlias('Category');
    my $ContentObjectClass = LoadClassByAlias('ContentObject');

    my $ShopAdminUrl = $Shop->get('WebUrlAdminSSL') . '/?ObjectPath=/Shops/' . $Shop->alias;

    #--               0           1       2        3          4             5                       6
    my $Sql = 'select o.objectid, o.guid, o.alias, lsa.value, co.isvisible, co.visibleinnavigation, o.parentid '.
        'from object o join contentobject co on o.objectid=co.contentobjectid '.
        'left join localizedstringattribute lsa on o.objectid=lsa.objectid and lsa.languageid=? and lsa.attributeid=? '.
        'where co.siteid=? and o.classid=? order by o.position';

    #-- Do NOT use something like GetLoopFromDoubleArray, which adds a couple hundred ms to the response time, we need to be FAST
    #-- But that means, be careful with the index of the array, it is just a plain array
    my $ahResult = $DBI->execute($Sql, [$LanguageID, $ContentObjectClass->attribute('Name')->id, $Shop->id, $CategoryClass->id]);

    my $aRoot = [
        $RootCat->get('ObjectID'),
        $RootCat->get('GUID'),
        $RootCat->get('Alias'),
        $RootCat->get('NameOrAlias', $LanguageID),
        $RootCat->get('IsVisible'),
        $RootCat->get('VisibleInNavigation'),
        $Shop->id,
    ];

    my $hResult = GetCatInfoAndRecurs($aRoot, $ahResult, $LanguageID, $IncludeInvisible, $ShopAdminUrl);
    return $hResult;
}

#-- Recursively build a category's tree based on it and its subcategories
sub GetCatInfoAndRecurs {
  my ($aCatToWorkOn, $aAllCats, $LanguageID, $IncludeInvisible, $CurrentMBOUrl) = @_;

  #-- $aCatToWorkOn
  #-- This should be an array with the data in the following order:
  #-- 0         1     2      3     4          5                    6
  #-- objectid, guid, alias, name, isvisible, visibleinnavigation, parentid
  #-- Why not use a hash with keys? Performance is important, do as little manipulation as possible.

  my $hInfo = {
    'categoryId' => $aCatToWorkOn->[1],
    'objectId' => $aCatToWorkOn->[0] + 0, #-- We want it typed as an integer
    'name' => $aCatToWorkOn->[3] ? $aCatToWorkOn->[3] : $aCatToWorkOn->[2],
    'visible' => $aCatToWorkOn->[4] ? JSON::XS::true : JSON::XS::false,
    'visibleInNavigation' => $aCatToWorkOn->[5] ? JSON::XS::true : JSON::XS::false,
    'mboUrl' => $CurrentMBOUrl . '/' . $aCatToWorkOn->[2],
  };

  #-- Find all sub categories of this current category, and remove them from our master list (performance?)
  my @Subs;
  for (my $i = 0;  $i < scalar(@$aAllCats);  $i++) {
      if ( $aAllCats->[$i]->[6] eq $aCatToWorkOn->[0] ) {
          push @Subs, $aAllCats->[$i];
          splice(@$aAllCats, $i, 1);
          $i--;
      }
  }

  my $aSubCatsInfo = [];
  foreach my $hSubCat ( @Subs ) {
    #-- If you have not asked for invisible categories and this one is invisible, skip!
    next if !$IncludeInvisible && !$hSubCat->[4];

    my $hSubCatInfo = GetCatInfoAndRecurs($hSubCat, $aAllCats, $LanguageID, $IncludeInvisible,
                                          $CurrentMBOUrl . '/' . $aCatToWorkOn->[2]);
    push @$aSubCatsInfo, $hSubCatInfo = $hSubCatInfo if scalar keys %$hSubCatInfo > 0;
  }
  $hInfo->{'subCategories'} = $aSubCatsInfo if scalar @$aSubCatsInfo > 0;

  return $hInfo;
}

sub _loadLocale {
    my $self = shift;
    my ($LocaleID) = @_;
    return DE_EPAGES::Object::API::Object::Locale->load($LocaleID);
}

sub _createShopRequestContext {
    my $self = shift;
    my ($Shop) = @_;
    return DE_EPAGES::Shop::API::Resource::ShopRequestContext->new($Shop);
}


1;
