#========================================================================================
# package      DE_EPAGES::Core::API::Object::DateTimeFormatter
# state        public
#----------------------------------------------------------------------------------------
# description  Formatter and parser for locale-dependent time and date strings. Based on
#               DateTime::Format::Strptime.
#----------------------------------------------------------------------------------------
# example      my $Formatter = DE_EPAGES::Core::API::Object::DateTimeFormatter->new(
#                   locale => 'de_DE',
#               );
#               my $Date   = $Formatter->parse_datetime( 'Donnerstag, 7. Oktober 2004', 'date' );
#               my $String = $Formatter->format_datetime( $Date, 'date', 'long' );
#========================================================================================
package DE_EPAGES::Core::API::Object::DateTimeFormatter;
use base DateTime::Format::Strptime;

use strict;

use DateTime::Locale;
use Time::Piece;
use DE_EPAGES::Core::API::Constants qw( PERL538 );
use DE_EPAGES::Core::API::Error qw (CallerError Error ExistsError GetError);
use DE_EPAGES::Core::API::PerlTools qw (GetLocalTimeZone GetLocalLocale GetCurrentDateTime IsA);
use DE_EPAGES::Core::API::TestParameter qw ( TestString );

# see  https://metacpan.org/pod/DateTime#CLDR-Patterns
# all values here must also be keys in _patternMap of
# DE_EPAGES/Dojo/Data/Public/javascript/epages/localize.js

my $FORMAT_PARSE_MAP = {
    '%{hour_12}'             => '%l',
    '%{hour}'                => '%H',
    '%{day}'                 => '%d',
    '%{month}'               => '%m',
    '%{time_zone_long_name}' => '%z',
    '%{ce_year}'             => '%Y',
    '%y'                     => '%Y',
};

# see  https://metacpan.org/pod/DateTime#CLDR-Patterns
# all values here must also be keys in _patternMap of
# DE_EPAGES/Dojo/Data/Public/javascript/epages/localize.js

my $CLDR_STRFTIME = [
    qr/G/    => '{era}',
    qr/yyyy/ => 'Y',
    qr/yy/   => 'Y', # always use 4 digits for year
    qr/y/    => 'Y', # always use 4 digits for year
    qr/u/    => 'Y',
    qr/MMMM/ => 'B',
    qr/MMM/  => 'b',
    qr/MM/   => 'm',
    qr/M/    => 'm', # was: {month}
    qr/dd/   => 'd',
    qr/d/    => 'd', # was: {day}
    qr/hh/   => 'l',
    qr/h/    => 'l', # was: {hour_12}
    qr/HH/   => 'H',
    qr/H/    => 'H', # was: {hour}
    qr/mm/   => 'M',
    qr/m/    => 'M', # was: {minute}
    qr/ss/   => 'S',
    qr/s/    => 'S', # was: {second}
    qr/S/    => 'N',
    qr/EEEE/ => 'A',
    qr/cccc/ => 'A',
    qr/E/    => 'a',
    qr/D/    => 'j',
    qr/F/    => '{weekday_of_month}',
    qr/w/    => 'V',
    qr/W/    => '{week_month}',
    qr/a/    => 'p',
    qr/k/    => 'k', # was: {hour_1}
    qr/K/    => 'l', # was: {hour_12_0}
    qr/zzzz/ => 'z', # was: {time_zone_long_name}
    qr/z/    => 'z', # was: {time_zone_long_name}
];

# use legacy format for all: date|time|datetime:short|medium time:long|full

my %LEGACY_FORMAT = (
    'ca_ES:date:full' => '%A %d de %B de %Y',
    'ca_ES:date:long' => '%d de %B de %Y',
    'ca_ES:date:medium' => '%d/%m/%Y',
    'ca_ES:date:short' => '%d/%m/%Y',
    'ca_ES:datetime:full' => '%A %d de %B de %Y %H:%M:%S %z',
    'ca_ES:datetime:long' => '%d de %B de %Y %H:%M:%S %z',
    'ca_ES:datetime:medium' => '%d/%m/%Y %H:%M:%S',
    'ca_ES:datetime:short' => '%d/%m/%Y %H:%M',
    'ca_ES:time:full' => '%H:%M:%S %z',
    'ca_ES:time:long' => '%H:%M:%S %z',
    'ca_ES:time:medium' => '%H:%M:%S',
    'ca_ES:time:short' => '%H:%M',
    'cs_CZ:date:full' => '%A, %d. %B %Y',
    'cs_CZ:date:long' => '%d. %B %Y',
    'cs_CZ:date:medium' => '%d.%m.%Y',
    'cs_CZ:date:short' => '%d.%m.%Y',
    'cs_CZ:datetime:full' => '%A, %d. %B %Y %H:%M:%S %z',
    'cs_CZ:datetime:long' => '%d. %B %Y %H:%M:%S %z',
    'cs_CZ:datetime:medium' => '%d.%m.%Y %H:%M:%S',
    'cs_CZ:datetime:short' => '%d.%m.%Y %H:%M',
    'cs_CZ:time:full' => '%H:%M:%S %z',
    'cs_CZ:time:long' => '%H:%M:%S %z',
    'cs_CZ:time:medium' => '%H:%M:%S',
    'cs_CZ:time:short' => '%H:%M',
    'da_DK:date:full' => '%A den %d. %B %Y',
    'da_DK:date:long' => '%d. %b %Y',
    'da_DK:date:medium' => '%d-%m-%Y',
    'da_DK:date:short' => '%d-%m-%Y',
    'da_DK:datetime:full' => '%A den %d. %B %Y %H.%M.%S %z',
    'da_DK:datetime:long' => '%d. %b %Y %H:%M:%S %z',
    'da_DK:datetime:medium' => '%d-%m-%Y %H:%M:%S',
    'da_DK:datetime:short' => '%d-%m-%Y %H:%M',
    'da_DK:time:full' => '%H:%M:%S %z',
    'da_DK:time:long' => '%H:%M:%S %z',
    'da_DK:time:medium' => '%H:%M:%S',
    'da_DK:time:short' => '%H:%M',
    'de_AT:date:full' => '%A, %d. %B %Y',
    'de_AT:date:long' => '%d. %B %Y',
    'de_AT:date:medium' => '%d.%m.%Y',
    'de_AT:date:short' => '%d.%m.%Y',
    'de_AT:datetime:full' => '%A, %d. %B %Y %H:%M:%S %z',
    'de_AT:datetime:long' => '%d. %B %Y %H:%M:%S %z',
    'de_AT:datetime:medium' => '%d.%m.%Y %H:%M:%S',
    'de_AT:datetime:short' => '%d.%m.%Y %H:%M',
    'de_AT:time:full' => '%H:%M:%S %z',
    'de_AT:time:long' => '%H:%M:%S %z',
    'de_AT:time:medium' => '%H:%M:%S',
    'de_AT:time:short' => '%H:%M',
    'de_BE:date:full' => '%A %d %B %Y',
    'de_BE:date:long' => '%d %B %Y',
    'de_BE:date:medium' => '%d.%m.%Y',
    'de_BE:date:short' => '%d/%m/%Y',
    'de_BE:datetime:full' => '%A %d %B %Y %H h %M min %S s %z',
    'de_BE:datetime:long' => '%d %B %Y %H:%M:%S %z',
    'de_BE:datetime:medium' => '%d.%m.%Y %H:%M:%S',
    'de_BE:datetime:short' => '%d/%m/%Y %H:%M',
    'de_BE:time:full' => '%H h %M min %S s %z',
    'de_BE:time:long' => '%H:%M:%S %z',
    'de_BE:time:medium' => '%H:%M:%S',
    'de_BE:time:short' => '%H:%M',
    'de_CH:date:full' => '%A, %d. %B %Y',
    'de_CH:date:long' => '%d. %B %Y',
    'de_CH:date:medium' => '%d.%m.%Y',
    'de_CH:date:short' => '%d.%m.%Y',
    'de_CH:datetime:full' => '%A, %d. %B %Y %H:%M:%S %z',
    'de_CH:datetime:long' => '%d. %B %Y %H:%M:%S %z',
    'de_CH:datetime:medium' => '%d.%m.%Y %H:%M:%S',
    'de_CH:datetime:short' => '%d.%m.%Y %H:%M',
    'de_CH:time:full' => '%H:%M:%S %z',
    'de_CH:time:long' => '%H:%M:%S %z',
    'de_CH:time:medium' => '%H:%M:%S',
    'de_CH:time:short' => '%H:%M',
    'de_DE:date:full' => '%A, %d. %B %Y',
    'de_DE:date:long' => '%d. %B %Y',
    'de_DE:date:medium' => '%d.%m.%Y',
    'de_DE:date:short' => '%d.%m.%Y',
    'de_DE:datetime:full' => '%A, %d. %B %Y %H:%M:%S %z',
    'de_DE:datetime:long' => '%d. %B %Y %H:%M:%S %z',
    'de_DE:datetime:medium' => '%d.%m.%Y %H:%M:%S',
    'de_DE:datetime:short' => '%d.%m.%Y %H:%M',
    'de_DE:time:full' => '%H:%M:%S %z',
    'de_DE:time:long' => '%H:%M:%S %z',
    'de_DE:time:medium' => '%H:%M:%S',
    'de_DE:time:short' => '%H:%M',
    'de_LU:date:full' => '%A, %d. %B %Y',
    'de_LU:date:long' => '%d. %B %Y',
    'de_LU:date:medium' => '%d.%m.%Y',
    'de_LU:date:short' => '%d.%m.%Y',
    'de_LU:datetime:full' => '%A, %d. %B %Y %H:%M:%S %z',
    'de_LU:datetime:long' => '%d. %B %Y %H:%M:%S %z',
    'de_LU:datetime:medium' => '%d.%m.%Y %H:%M:%S',
    'de_LU:datetime:short' => '%d.%m.%Y %H:%M',
    'de_LU:time:full' => '%H:%M:%S %z',
    'de_LU:time:long' => '%H:%M:%S %z',
    'de_LU:time:medium' => '%H:%M:%S',
    'de_LU:time:short' => '%H:%M',
    'en_AU:date:full' => '%A, %d %B %Y',
    'en_AU:date:long' => '%d %B %Y',
    'en_AU:date:medium' => '%d/%m/%Y',
    'en_AU:date:short' => '%d/%m/%Y',
    'en_AU:datetime:full' => '%A, %d %B %Y %l:%M:%S %p %z',
    'en_AU:datetime:long' => '%d %B %Y %l:%M:%S %p %z',
    'en_AU:datetime:medium' => '%d/%m/%Y %l:%M:%S %p',
    'en_AU:datetime:short' => '%d/%m/%Y %l:%M %p',
    'en_AU:time:full' => '%l:%M:%S %p %z',
    'en_AU:time:long' => '%l:%M:%S %p %z',
    'en_AU:time:medium' => '%l:%M:%S %p',
    'en_AU:time:short' => '%l:%M %p',
    'en_CA:date:full' => '%A, %B %d, %Y',
    'en_CA:date:long' => '%B %d, %Y',
    'en_CA:date:medium' => '%Y-%m-%d',
    'en_CA:date:short' => '%Y-%m-%d',
    'en_CA:datetime:full' => '%A, %B %d, %Y %l:%M:%S %p %z',
    'en_CA:datetime:long' => '%B %d, %Y %l:%M:%S %p %z',
    'en_CA:datetime:medium' => '%Y-%m-%d %l:%M:%S %p',
    'en_CA:datetime:short' => '%Y-%m-%d %l:%M %p',
    'en_CA:time:full' => '%l:%M:%S %p %z',
    'en_CA:time:long' => '%l:%M:%S %p %z',
    'en_CA:time:medium' => '%l:%M:%S %p',
    'en_CA:time:short' => '%l:%M %p',
    'en_CH:date:full' => '%A, %B %d, %Y',
    'en_CH:date:long' => '%B %d, %Y',
    'en_CH:date:medium' => '%d.%m.%Y',
    'en_CH:date:short' => '%d.%m.%Y',
    'en_CH:datetime:full' => '%A, %B %d, %Y %l:%M:%S %p %z',
    'en_CH:datetime:long' => '%B %d, %Y %l:%M:%S %p %z',
    'en_CH:datetime:medium' => '%d.%m.%Y %H:%M:%S',
    'en_CH:datetime:short' => '%d.%m.%Y %H:%M',
    'en_CH:time:full' => '%H:%M:%S %z',
    'en_CH:time:long' => '%H:%M:%S %z',
    'en_CH:time:medium' => '%H:%M:%S',
    'en_CH:time:short' => '%H:%M',
    'en_DK:date:full' => '%A, %B %d, %Y',
    'en_DK:date:long' => '%B %d, %Y',
    'en_DK:date:medium' => '%b %d, %Y',
    'en_DK:date:short' => '%m/%d/%Y',
    'en_DK:datetime:full' => '%A, %B %d, %Y %l:%M:%S %p %z',
    'en_DK:datetime:long' => '%B %d, %Y %l:%M:%S %p %z',
    'en_DK:datetime:medium' => '%b %d, %Y %l:%M:%S %p',
    'en_DK:datetime:short' => '%m/%d/%Y %l:%M %p',
    'en_DK:time:full' => '%l:%M:%S %p %z',
    'en_DK:time:long' => '%l:%M:%S %p %z',
    'en_DK:time:medium' => '%l:%M:%S %p',
    'en_DK:time:short' => '%l:%M %p',
    'en_GB:date:full' => '%A, %d %B %Y',
    'en_GB:date:long' => '%d %B %Y',
    'en_GB:date:medium' => '%d %b %Y',
    'en_GB:date:short' => '%d/%m/%Y',
    'en_GB:datetime:full' => '%A, %d %B %Y %H:%M:%S %z',
    'en_GB:datetime:long' => '%d %B %Y %H:%M:%S %z',
    'en_GB:datetime:medium' => '%d %b %Y %H:%M:%S',
    'en_GB:datetime:short' => '%d/%m/%Y %H:%M',
    'en_GB:time:full' => '%H:%M:%S %z',
    'en_GB:time:long' => '%H:%M:%S %z',
    'en_GB:time:medium' => '%H:%M:%S',
    'en_GB:time:short' => '%H:%M',
    'en_IE:date:full' => '%A %d %B %Y',
    'en_IE:date:long' => '%d %B %Y',
    'en_IE:date:medium' => '%d %b %Y',
    'en_IE:date:short' => '%d/%m/%Y',
    'en_IE:datetime:full' => '%A %d %B %Y %H:%M:%S %z',
    'en_IE:datetime:long' => '%d %B %Y %H:%M:%S %z',
    'en_IE:datetime:medium' => '%d %b %Y %H:%M:%S',
    'en_IE:datetime:short' => '%d/%m/%Y %H:%M',
    'en_IE:time:full' => '%H:%M:%S %z',
    'en_IE:time:long' => '%H:%M:%S %z',
    'en_IE:time:medium' => '%H:%M:%S',
    'en_IE:time:short' => '%H:%M',
    'en_NZ:date:full' => '%A, %d %B %Y',
    'en_NZ:date:long' => '%d %B %Y',
    'en_NZ:date:medium' => '%d/%m/%Y',
    'en_NZ:date:short' => '%d/%m/%Y',
    'en_NZ:datetime:full' => '%A, %d %B %Y %l:%M:%S %p %z',
    'en_NZ:datetime:long' => '%d %B %Y %l:%M:%S %p %z',
    'en_NZ:datetime:medium' => '%d/%m/%Y %l:%M:%S %p',
    'en_NZ:datetime:short' => '%d/%m/%Y %l:%M %p',
    'en_NZ:time:full' => '%l:%M:%S %p %z',
    'en_NZ:time:long' => '%l:%M:%S %p %z',
    'en_NZ:time:medium' => '%l:%M:%S %p',
    'en_NZ:time:short' => '%l:%M %p',
    'en_US:date:full' => '%A, %B %d, %Y',
    'en_US:date:long' => '%B %d, %Y',
    'en_US:date:medium' => '%b %d, %Y',
    'en_US:date:short' => '%m/%d/%Y',
    'en_US:datetime:full' => '%A, %B %d, %Y %l:%M:%S %p %z',
    'en_US:datetime:long' => '%B %d, %Y %l:%M:%S %p %z',
    'en_US:datetime:medium' => '%b %d, %Y %l:%M:%S %p',
    'en_US:datetime:short' => '%m/%d/%Y %l:%M %p',
    'en_US:time:full' => '%l:%M:%S %p %z',
    'en_US:time:long' => '%l:%M:%S %p %z',
    'en_US:time:medium' => '%l:%M:%S %p',
    'en_US:time:short' => '%l:%M %p',
    'es_AR:date:full' => '%A %d de %B de %Y',
    'es_AR:date:long' => '%d de %B de %Y',
    'es_AR:date:medium' => '%d/%m/%Y',
    'es_AR:date:short' => '%d/%m/%Y',
    'es_AR:datetime:full' => '%A %d de %B de %Y %Hh\'%M:%S %z',
    'es_AR:datetime:long' => '%d de %B de %Y %H:%M:%S %z',
    'es_AR:datetime:medium' => '%d/%m/%Y %H:%M:%S',
    'es_AR:datetime:short' => '%d/%m/%Y %H:%M',
    'es_AR:time:full' => '%Hh\'%M:%S %z',
    'es_AR:time:long' => '%H:%M:%S %z',
    'es_AR:time:medium' => '%H:%M:%S',
    'es_AR:time:short' => '%H:%M',
    'es_BO:date:full' => '%A %d de %B de %Y',
    'es_BO:date:long' => '%d de %B de %Y',
    'es_BO:date:medium' => '%d/%m/%Y',
    'es_BO:date:short' => '%d/%m/%Y',
    'es_BO:datetime:full' => '%A %d de %B de %Y %H:%M:%S %z',
    'es_BO:datetime:long' => '%d de %B de %Y %H:%M:%S %z',
    'es_BO:datetime:medium' => '%d/%m/%Y %H:%M:%S',
    'es_BO:datetime:short' => '%d/%m/%Y %H:%M',
    'es_BO:time:full' => '%H:%M:%S %z',
    'es_BO:time:long' => '%H:%M:%S %z',
    'es_BO:time:medium' => '%H:%M:%S',
    'es_BO:time:short' => '%H:%M',
    'es_CO:date:full' => '%A %d de %B de %Y',
    'es_CO:date:long' => '%d de %B de %Y',
    'es_CO:date:medium' => '%d/%m/%Y',
    'es_CO:date:short' => '%d/%m/%Y',
    'es_CO:datetime:full' => '%A %d de %B de %Y %H:%M:%S %z',
    'es_CO:datetime:long' => '%d de %B de %Y %H:%M:%S %z',
    'es_CO:datetime:medium' => '%d/%m/%Y %H:%M:%S',
    'es_CO:datetime:short' => '%d/%m/%Y %H:%M',
    'es_CO:time:full' => '%H:%M:%S %z',
    'es_CO:time:long' => '%H:%M:%S %z',
    'es_CO:time:medium' => '%H:%M:%S',
    'es_CO:time:short' => '%H:%M',
    'es_EC:date:full' => '%A %d de %B de %Y',
    'es_EC:date:long' => '%d de %B de %Y',
    'es_EC:date:medium' => '%d/%m/%Y',
    'es_EC:date:short' => '%d/%m/%Y',
    'es_EC:datetime:full' => '%A %d de %B de %Y %H:%M:%S %z',
    'es_EC:datetime:long' => '%d de %B de %Y %H:%M:%S %z',
    'es_EC:datetime:medium' => '%d/%m/%Y %H:%M:%S',
    'es_EC:datetime:short' => '%d/%m/%Y %H:%M',
    'es_EC:time:full' => '%H:%M:%S %z',
    'es_EC:time:long' => '%H:%M:%S %z',
    'es_EC:time:medium' => '%H:%M:%S',
    'es_EC:time:short' => '%H:%M',
    'es_ES:date:full' => '%A %d de %B de %Y',
    'es_ES:date:long' => '%d de %B de %Y',
    'es_ES:date:medium' => '%d/%m/%Y',
    'es_ES:date:short' => '%d/%m/%Y',
    'es_ES:datetime:full' => '%A %d de %B de %Y %H:%M:%S %z',
    'es_ES:datetime:long' => '%d de %B de %Y %H:%M:%S %z',
    'es_ES:datetime:medium' => '%d/%m/%Y %H:%M:%S',
    'es_ES:datetime:short' => '%d/%m/%Y %H:%M',
    'es_ES:time:full' => '%H:%M:%S %z',
    'es_ES:time:long' => '%H:%M:%S %z',
    'es_ES:time:medium' => '%H:%M:%S',
    'es_ES:time:short' => '%H:%M',
    'es_MX:date:full' => '%A %d de %B de %Y',
    'es_MX:date:long' => '%d de %B de %Y',
    'es_MX:date:medium' => '%d/%m/%Y',
    'es_MX:date:short' => '%d/%m/%Y',
    'es_MX:datetime:full' => '%A %d de %B de %Y %H:%M:%S %z',
    'es_MX:datetime:long' => '%d de %B de %Y %H:%M:%S %z',
    'es_MX:datetime:medium' => '%d/%m/%Y %H:%M:%S',
    'es_MX:datetime:short' => '%d/%m/%Y %H:%M',
    'es_MX:time:full' => '%H:%M:%S %z',
    'es_MX:time:long' => '%H:%M:%S %z',
    'es_MX:time:medium' => '%H:%M:%S',
    'es_MX:time:short' => '%H:%M',
    'es_US:date:full' => '%A %d de %B de %Y',
    'es_US:date:long' => '%d de %B de %Y',
    'es_US:date:medium' => '%b %d, %Y',
    'es_US:date:short' => '%m/%d/%Y',
    'es_US:datetime:full' => '%A %d de %B de %Y %l:%M:%S %p %z',
    'es_US:datetime:long' => '%d de %B de %Y %l:%M:%S %p %z',
    'es_US:datetime:medium' => '%b %d, %Y %l:%M:%S %p',
    'es_US:datetime:short' => '%m/%d/%Y %l:%M %p',
    'es_US:time:full' => '%l:%M:%S %p %z',
    'es_US:time:long' => '%l:%M:%S %p %z',
    'es_US:time:medium' => '%l:%M:%S %p',
    'es_US:time:short' => '%l:%M %p',
    'fi_FI:date:full' => '%A %d. %B %Y',
    'fi_FI:date:long' => '%d. %B %Y',
    'fi_FI:date:medium' => '%d.%m.%Y',
    'fi_FI:date:short' => '%d.%m.%Y',
    'fi_FI:datetime:full' => '%A %d. %B %Y %H.%M.%S %z',
    'fi_FI:datetime:long' => '%d. %B %Y %H.%M.%S %z',
    'fi_FI:datetime:medium' => '%d.%m.%Y %H.%M.%S',
    'fi_FI:datetime:short' => '%d.%m.%Y %H.%M',
    'fi_FI:time:full' => '%H.%M.%S %z',
    'fi_FI:time:long' => '%H.%M.%S %z',
    'fi_FI:time:medium' => '%H.%M.%S',
    'fi_FI:time:short' => '%H.%M',
    'fr_BE:date:full' => '%A %d %B %Y',
    'fr_BE:date:long' => '%d %B %Y',
    'fr_BE:date:medium' => '%d %b %Y',
    'fr_BE:date:short' => '%d/%m/%Y',
    'fr_BE:datetime:full' => '%A %d %B %Y %H h %M min %S s %z',
    'fr_BE:datetime:long' => '%d %B %Y %H:%M:%S %z',
    'fr_BE:datetime:medium' => '%d %b %Y %H:%M:%S',
    'fr_BE:datetime:short' => '%d/%m/%Y %H:%M',
    'fr_BE:time:full' => '%H h %M min %S s %z',
    'fr_BE:time:long' => '%H:%M:%S %z',
    'fr_BE:time:medium' => '%H:%M:%S',
    'fr_BE:time:short' => '%H:%M',
    'fr_CA:date:full' => '%A %d %B %Y',
    'fr_CA:date:long' => '%d %B %Y',
    'fr_CA:date:medium' => '%Y-%m-%d',
    'fr_CA:date:short' => '%Y-%m-%d',
    'fr_CA:datetime:full' => '%A %d %B %Y %H h %M min %S s %z',
    'fr_CA:datetime:long' => '%d %B %Y %H:%M:%S %z',
    'fr_CA:datetime:medium' => '%Y-%m-%d %H:%M:%S',
    'fr_CA:datetime:short' => '%Y-%m-%d %H:%M',
    'fr_CA:time:full' => '%H h %M min %S s %z',
    'fr_CA:time:long' => '%H:%M:%S %z',
    'fr_CA:time:medium' => '%H:%M:%S',
    'fr_CA:time:short' => '%H:%M',
    'fr_CH:date:full' => '%A, %d %B %Y',
    'fr_CH:date:long' => '%d %B %Y',
    'fr_CH:date:medium' => '%d %b %Y',
    'fr_CH:date:short' => '%d.%m.%Y',
    'fr_CH:datetime:full' => '%A, %d %B %Y %H.%M:%S h %z',
    'fr_CH:datetime:long' => '%d %B %Y %H:%M:%S %z',
    'fr_CH:datetime:medium' => '%d %b %Y %H:%M:%S',
    'fr_CH:datetime:short' => '%d.%m.%Y %H:%M',
    'fr_CH:time:full' => '%H.%M:%S h %z',
    'fr_CH:time:long' => '%H:%M:%S %z',
    'fr_CH:time:medium' => '%H:%M:%S',
    'fr_CH:time:short' => '%H:%M',
    'fr_FR:date:full' => '%A %d %B %Y',
    'fr_FR:date:long' => '%d %B %Y',
    'fr_FR:date:medium' => '%d %b %Y',
    'fr_FR:date:short' => '%d/%m/%Y',
    'fr_FR:datetime:full' => '%A %d %B %Y %H:%M:%S %z',
    'fr_FR:datetime:long' => '%d %B %Y %H:%M:%S %z',
    'fr_FR:datetime:medium' => '%d %b %Y %H:%M:%S',
    'fr_FR:datetime:short' => '%d/%m/%Y %H:%M',
    'fr_FR:time:full' => '%H:%M:%S %z',
    'fr_FR:time:long' => '%H:%M:%S %z',
    'fr_FR:time:medium' => '%H:%M:%S',
    'fr_FR:time:short' => '%H:%M',
    'fr_LU:date:full' => '%A %d %B %Y',
    'fr_LU:date:long' => '%d %B %Y',
    'fr_LU:date:medium' => '%d %b %Y',
    'fr_LU:date:short' => '%d/%m/%Y',
    'fr_LU:datetime:full' => '%A %d %B %Y %H:%M:%S %z',
    'fr_LU:datetime:long' => '%d %B %Y %H:%M:%S %z',
    'fr_LU:datetime:medium' => '%d %b %Y %H:%M:%S',
    'fr_LU:datetime:short' => '%d/%m/%Y %H:%M',
    'fr_LU:time:full' => '%H:%M:%S %z',
    'fr_LU:time:long' => '%H:%M:%S %z',
    'fr_LU:time:medium' => '%H:%M:%S',
    'fr_LU:time:short' => '%H:%M',
    'it_CH:date:full' => '%A, %d %B %Y',
    'it_CH:date:long' => '%d %B %Y',
    'it_CH:date:medium' => '%d.%m.%Y',
    'it_CH:date:short' => '%d.%m.%Y',
    'it_CH:datetime:full' => '%A, %d %B %Y %H.%M:%S h %z',
    'it_CH:datetime:long' => '%d %B %Y %H.%M.%S %z',
    'it_CH:datetime:medium' => '%d.%m.%Y %H:%M:%S',
    'it_CH:datetime:short' => '%d.%m.%Y %H:%M',
    'it_CH:time:full' => '%H:%M:%S %z',
    'it_CH:time:long' => '%H:%M:%S %z',
    'it_CH:time:medium' => '%H:%M:%S',
    'it_CH:time:short' => '%H:%M',
    'it_IT:date:full' => '%A %d %B %Y',
    'it_IT:date:long' => '%d %B %Y',
    'it_IT:date:medium' => '%d/%b/%Y',
    'it_IT:date:short' => '%d/%m/%Y',
    'it_IT:datetime:full' => '%A %d %B %Y %H.%M.%S %z',
    'it_IT:datetime:long' => '%d %B %Y %H.%M.%S %z',
    'it_IT:datetime:medium' => '%d/%b/%Y %H.%M.%S',
    'it_IT:datetime:short' => '%d/%m/%Y %H.%M',
    'it_IT:time:full' => '%H.%M.%S %z',
    'it_IT:time:long' => '%H.%M.%S %z',
    'it_IT:time:medium' => '%H.%M.%S',
    'it_IT:time:short' => '%H.%M',
    'nb_NO:date:full' => '%A %d. %B %Y',
    'nb_NO:date:long' => '%d. %B %Y',
    'nb_NO:date:medium' => '%d. %b %Y',
    'nb_NO:date:short' => '%d.%m.%Y',
    'nb_NO:datetime:full' => '%A %d. %B %Y kl. %H.%M.%S %z',
    'nb_NO:datetime:long' => '%d. %B %Y %H.%M.%S %z',
    'nb_NO:datetime:medium' => '%d. %b %Y %H.%M.%S',
    'nb_NO:datetime:short' => '%d.%m.%Y %H.%M',
    'nb_NO:time:full' => 'kl. %H.%M.%S %z',
    'nb_NO:time:long' => '%H.%M.%S %z',
    'nb_NO:time:medium' => '%H.%M.%S',
    'nb_NO:time:short' => '%H.%M',
    'nl_BE:date:full' => '%A %d %B %Y',
    'nl_BE:date:long' => '%d %B %Y',
    'nl_BE:date:medium' => '%d-%b-%Y',
    'nl_BE:date:short' => '%d/%m/%Y',
    'nl_BE:datetime:full' => '%A %d %B %Y %H:%M:%S %z',
    'nl_BE:datetime:long' => '%d %B %Y %H:%M:%S %z',
    'nl_BE:datetime:medium' => '%d-%b-%Y %H:%M:%S',
    'nl_BE:datetime:short' => '%d/%m/%Y %H:%M',
    'nl_BE:time:full' => '%H:%M:%S %z',
    'nl_BE:time:long' => '%H:%M:%S %z',
    'nl_BE:time:medium' => '%H:%M:%S',
    'nl_BE:time:short' => '%H:%M',
    'nl_NL:date:full' => '%A %d %B %Y',
    'nl_NL:date:long' => '%d %B %Y',
    'nl_NL:date:medium' => '%d %b %Y',
    'nl_NL:date:short' => '%d-%m-%Y',
    'nl_NL:datetime:full' => '%A %d %B %Y %H:%M:%S %z',
    'nl_NL:datetime:long' => '%d %B %Y %H:%M:%S %z',
    'nl_NL:datetime:medium' => '%d %b %Y %H:%M:%S',
    'nl_NL:datetime:short' => '%d-%m-%Y %H:%M',
    'nl_NL:time:full' => '%H:%M:%S %z',
    'nl_NL:time:long' => '%H:%M:%S %z',
    'nl_NL:time:medium' => '%H:%M:%S',
    'nl_NL:time:short' => '%H:%M',
    'no_NO:date:full' => '%A %d. %B %Y',
    'no_NO:date:long' => '%d. %B %Y',
    'no_NO:date:medium' => '%d. %b %Y',
    'no_NO:date:short' => '%d.%m.%Y',
    'no_NO:datetime:full' => '%A %d. %B %Y kl. %H.%M.%S %z',
    'no_NO:datetime:long' => '%d. %B %Y %H.%M.%S %z',
    'no_NO:datetime:medium' => '%d. %b %Y %H.%M.%S',
    'no_NO:datetime:short' => '%d.%m.%Y %H.%M',
    'no_NO:time:full' => 'kl. %H.%M.%S %z',
    'no_NO:time:long' => '%H.%M.%S %z',
    'no_NO:time:medium' => '%H.%M.%S',
    'no_NO:time:short' => '%H.%M',
    'pt_BR:date:full' => '%A, %d de %B de %Y',
    'pt_BR:date:long' => '%d de %B de %Y',
    'pt_BR:date:medium' => '%d/%m/%Y',
    'pt_BR:date:short' => '%d/%m/%Y',
    'pt_BR:datetime:full' => '%A, %d de %B de %Y %Hh%Mmin%Ss %z',
    'pt_BR:datetime:long' => '%d de %B de %Y %Hh%Mmin%Ss %z',
    'pt_BR:datetime:medium' => '%d/%m/%Y %H:%M:%S',
    'pt_BR:datetime:short' => '%d/%m/%Y %H:%M',
    'pt_BR:time:full' => '%Hh%Mmin%Ss %z',
    'pt_BR:time:long' => '%Hh%Mmin%Ss %z',
    'pt_BR:time:medium' => '%H:%M:%S',
    'pt_BR:time:short' => '%H:%M',
    'pt_PT:date:full' => '%A, %d de %B de %Y',
    'pt_PT:date:long' => '%d de %B de %Y',
    'pt_PT:date:medium' => '%d de %b de %Y',
    'pt_PT:date:short' => '%d/%m/%Y',
    'pt_PT:datetime:full' => '%A, %d de %B de %Y %Hh%Mmin%Ss %z',
    'pt_PT:datetime:long' => '%d de %B de %Y %Hh%Mmin%Ss %z',
    'pt_PT:datetime:medium' => '%d de %b de %Y %H:%M:%S',
    'pt_PT:datetime:short' => '%d/%m/%Y %H:%M',
    'pt_PT:time:full' => '%Hh%Mmin%Ss %z',
    'pt_PT:time:long' => '%Hh%Mmin%Ss %z',
    'pt_PT:time:medium' => '%H:%M:%S',
    'pt_PT:time:short' => '%H:%M',
    'sv_FI:date:full' => '%A %d %B %Y',
    'sv_FI:date:long' => '%d %B %Y',
    'sv_FI:date:medium' => '%d %b %Y',
    'sv_FI:date:short' => '%Y-%m-%d',
    'sv_FI:datetime:full' => '%A %d %B %Y kl. %H.%M.%S %z',
    'sv_FI:datetime:long' => '%d %B %Y %H.%M.%S %z',
    'sv_FI:datetime:medium' => '%d %b %Y %H.%M.%S',
    'sv_FI:datetime:short' => '%Y-%m-%d %H.%M',
    'sv_FI:time:full' => 'kl. %H.%M.%S %z',
    'sv_FI:time:long' => '%H.%M.%S %z',
    'sv_FI:time:medium' => '%H.%M.%S',
    'sv_FI:time:short' => '%H.%M',
    'sv_SE:date:full' => '%A %d %B %Y',
    'sv_SE:date:long' => '%d %B %Y',
    'sv_SE:date:medium' => '%d %b %Y',
    'sv_SE:date:short' => '%Y-%m-%d',
    'sv_SE:datetime:full' => '%A %d %B %Y kl. %H.%M.%S %z',
    'sv_SE:datetime:long' => '%d %B %Y %H.%M.%S %z',
    'sv_SE:datetime:medium' => '%d %b %Y %H.%M.%S',
    'sv_SE:datetime:short' => '%Y-%m-%d %H.%M',
    'sv_SE:time:full' => 'kl. %H.%M.%S %z',
    'sv_SE:time:long' => '%H.%M.%S %z',
    'sv_SE:time:medium' => '%H.%M.%S',
    'sv_SE:time:short' => '%H.%M',
);

#ERRORDEFINITION
sub ErrorDefinition { return {
    'FORMAT_NOT_DATETIME' => 'datetime \'#Dt\' doesn\'t match format \'#Format\'',
    'FORMAT_UNKNOWN' => 'unknown datetime format: #Type (allowed formats are: #Allowed)',
    }
}
#ERRORDEFINITION END

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Formatter = DE_EPAGES::Core::API::Object::DateTimeFormatter->new(%options);
#----------------------------------------------------------------------------------------
# description  Creates a formatter for dates and/or time strings.
#----------------------------------------------------------------------------------------
# input        %options | keys
#               <ul>
#                   <li>locale       - (optional) locale id, example: 'en_US',
#                                                 default: server locale - string</li>
#                   <li>time_zone    - (optional) time zone, example: '+01:00',
#                                                 default: server time zone - string</li>
#               </ul>| hash
# return       $Formatter | formatter object | object
#========================================================================================
sub new {
    my $class = shift;
    my (%options) = @_;
    $options{'time_zone'} = GetLocalTimeZone() unless defined $options{'time_zone'};
    $options{'locale'}    = GetLocalLocale()   unless defined $options{'locale'};
    $options{'pattern'}   = '%u';
    return bless DateTime::Format::Strptime->new(%options), $class;
}

#========================================================================================
# function     formatString
# state        public
#----------------------------------------------------------------------------------------
# syntax       $formatString = formatString($type, $formattype, $locale);
#----------------------------------------------------------------------------------------
# description  returns strptime pattern for $type and $formattype
#----------------------------------------------------------------------------------------
# input        $type | date/time/datetime | string
# input        $formattype | short/medium/long/full (optional short)| string
# input        $locale | locale id as 'en_US', default: server locale | string
# return       $formatString | pattern for DateTime::Format::Strptime | string
#========================================================================================
sub formatString { ##no critic qw(RequireArgUnpacking)
    my $self = ref $_[0] ? shift : undef;
    my ($type, $formattype, $locale) = @_;
    TestString('type',$type);

    if ($type !~ /^(?:date|time|datetime)$/) {
        CallerError('FORMAT_UNKNOWN', {
            'Type'   => $type,
            'Allowed' => 'date, datetime or time'
        })
    }

    my $LocaleID;
    if ( $] < PERL538 ) {
        $LocaleID = $self->locale;
        $locale = DateTime::Locale->load($LocaleID);
    } else {
        $locale = $self ? $self->locale : GetLocalLocale() unless defined $locale;
        $LocaleID = $locale;
    }
    $LocaleID =~ s,-,_,g;
    $LocaleID = 'no_NO' if $LocaleID eq 'no';

    $formattype = 'short' unless defined $formattype;
    my $functionname = $type.'_format_'.$formattype;

    my $formatString;
    # $locale->_cldr_to_strftime does not exist anymore with 5.38.0
    if ( defined $LEGACY_FORMAT{"$LocaleID:$type:$formattype"} ) {
        $formatString = $LEGACY_FORMAT{"$LocaleID:$type:$formattype"};
    } elsif ( $] < PERL538 ) {
        $formatString = $locale->_cldr_to_strftime($locale->$functionname()); ## no critic qw(ProhibitPrivateSubroutineUse)
        $formatString =~ s/\Q$_\E/$FORMAT_PARSE_MAP->{$_}/ foreach keys %$FORMAT_PARSE_MAP;
    } else {
        $formatString = _cldr_to_strftime($LocaleID, $type, $formattype);
    }

    return $formatString;
}

#========================================================================================
# function     format_datetime
# state        public
#----------------------------------------------------------------------------------------
# syntax       $FormattedDateTime = $Formatter->format_datetime($dt, $type, $formattype);
#----------------------------------------------------------------------------------------
# description  Formats datetime
#----------------------------------------------------------------------------------------
# input        $dt | datetime | object
# input        $type | date/time/datetime | string
# input        $formattype | (optional) short/medium/long/full
#                             or custom strftime format, for example '%Y-%m-%d'
#                             (default: short)| string
# return       $FormattedDateTime | formatted datetime | string
#========================================================================================
sub format_datetime {
    my $self = shift;
    my ($dt, $type, $formattype) = @_;
    die 'no DateTime object' unless IsA( $dt, 'DateTime' );
    TestString('type',$type);

    if ($type !~ /^(?:date|time|datetime)$/) {
        CallerError('FORMAT_UNKNOWN', {
            'Type'   => $type,
            'Allowed' => 'date, datetime or time'
        })
    }

    my $Locale = $self->locale;
    my $dtLocale = $] < PERL538 ? $Locale : $dt->locale->code;

    # $dt->set(locale => $Locale) does not work anymore with 5.38.0
    $dt->set(locale => $Locale) if $] < PERL538;
    $dt->set_locale($Locale) if $Locale ne $dtLocale;

    my $formatString;
    if (defined $formattype and $formattype =~ /%/) {
        $formatString = $formattype;
    } else {
        $formatString = $self->formatString($type, $formattype, $Locale);
    }

    if ($] >= PERL538) {
        # Time::Piece has problems with time zones
        $dt->add(seconds => $dt->offset) if $formatString =~ /%z/;
    }

    my $strftime = $dt->strftime( $formatString );
    $dt->set_locale($dtLocale) if $Locale ne $dtLocale;

    # strip leading and trailing spaces
    $strftime =~ s/^\s*(.*?)\s*$/$1/x;
    # strip spaces before numbers if $formatString contains %e/%k/%l
    $strftime =~ s,([:/.-]|\s)\s([1-9])\b,$1$2,gx if $formatString =~ /%[ekl]/;;

    # use legacy names for AM/PM
    $strftime =~ s,([ap])\.( |\x{202F})?m\.,\U$1M\E,gi if $Locale =~ /^en/;
    $strftime =~ s,( [ap]m),\U$1\E,gi if $Locale =~ /^en/;
    $strftime =~ s,\.\x{202F}m\.,.m.,g;
    return $strftime;
}

#========================================================================================
# function     parse_datetime
# state        public
#----------------------------------------------------------------------------------------
# syntax       $dt = $Formatter->parse_datetime($FormattedDateTime, $type, $formattype);
#----------------------------------------------------------------------------------------
# description  Return the datetime object of a formatted datetime
#----------------------------------------------------------------------------------------
# input        $FormattedDateTime | formatted datetime | string
# input        $type | date/time/datetime | string
# input        $formattype | (optional) short/medium/long/full
#                             or custom strftime format, for example '%Y-%m-%d'
#                             (default: short)| string
# return       $dt | datetime object | DateTime object
#========================================================================================
sub parse_datetime {
    my $self = shift;
    my ($string, $type, $formattype) = @_;
    TestString('type', $type);

    if ($type !~ /^(?:date|time|datetime)$/) {
        CallerError('FORMAT_UNKNOWN', {
            'Type'   => $type,
            'Allowed' => 'date, datetime or time'
        })
    }

    my $formatString;
    if (defined $formattype and $formattype =~ /%/) {
        $formatString = $formattype;
    } else {
        $formatString = $self->formatString($type, $formattype);
    }

    # time zones work only if datetime is given (not just time)
    if ($type eq 'time' and $formatString =~ /%z/) {
        my $Now = GetCurrentDateTime;
        $formatString = '%Y-%m-%d ' . $formatString
            unless $formatString =~ /^%Y-%m-%d /;
        $string = $Now->ymd('-') . ' ' . $string
            unless $string =~ /^(19|20)\d\d-[01]\d-[0-3]\d /;
    }

    my $dt;
    if ($] < PERL538) {
        # $self->pattern($formatString) does not work anymore with 5.38.0
        $self->pattern($formatString);
        $dt = eval { return $self->SUPER::parse_datetime($string); };
        Error( GetError() ) if ExistsError();
        Error('FORMAT_NOT_DATETIME', { 'Dt'=> $string, 'Format' => $formatString })
            if not defined $dt;
    } else {
        my $dtLocale = DateTime::Locale->load($self->locale);
        Time::Piece::mon_list(@{$dtLocale->month_format_abbreviated});
        Time::Piece::day_list(@{$dtLocale->day_format_abbreviated});
        Time::Piece::fullmon_list(@{$dtLocale->month_format_wide});
        Time::Piece::fullday_list(@{$dtLocale->day_format_wide});

        # a.m. -> AM ; p.m. -> PM
        $string =~ s,([ap])\.( |\x{202F})?m\.,\U$1M\E,gi;

        my $parsed_date = eval { return Time::Piece->strptime($string, $formatString); };
        Error('FORMAT_NOT_DATETIME', { 'Dt'=> $string, 'Format' => $formatString })
            if ExistsError() or not defined $parsed_date;

        $dt = DateTime->new(
            year       => $parsed_date->year,
            month      => $parsed_date->mon,
            day        => $parsed_date->mday,
            hour       => $parsed_date->hour,
            minute     => $parsed_date->minute,
            second     => $parsed_date->second,
            time_zone  => $self->time_zone,
            locale     => $self->locale,
            );
    }

    return $dt;
}

##############################################################################
# private functions from DateTime::Locale::Base for PERL538
##############################################################################

sub _cldr_to_strftime {
    my ($locale, $type, $formattype) = @_;

    # use legacy format if defined
    my $Format = "$locale:$type:$formattype";
    return $LEGACY_FORMAT{$Format} if defined $LEGACY_FORMAT{$Format};

    my $dtLocale = DateTime::Locale->load($locale);
    my $functionname = $type.'_format_'.$formattype;
    my $formatString = $dtLocale->$functionname;
    $formatString
        =~ s/(G+|y+|u+|M+|d+|h+|H+|m+|s+|S+|c+|E+|D+|F+|w+|W+|a+|k+|K+|z+)|'((?:[^']|'')*)'/
                        $2 ? _stringify($2) : $1 ? _convert($1) : "'"/eg;
    $formatString =~ s,\x{202F}, ,g;
    $formatString =~ s,%e,%d,g;

    return $formatString;
}

sub _convert {
    my ($simple) = @_;
    for ( my $x = 0; $x < @$CLDR_STRFTIME; $x += 2 ) {
        return '%' . $CLDR_STRFTIME->[ $x + 1 ] if $simple =~ /$CLDR_STRFTIME->[$x]/;
        return '%' . $CLDR_STRFTIME->[ $x + 1 ] if $simple =~ /$CLDR_STRFTIME->[$x]/;
    }
    die "**Dont know $simple***";
}

sub _stringify {
    my ($string) = @_;
    $string =~ s/%(?:[^%])/%%/g;
    $string =~ s/\'\'/\'/g;
    return $string;
}

1;
