#!/srv/epages/eproot/Perl/bin/perl
use strict;
use Git::Repository;
use Getopt::Long;
use Time::HiRes qw (time);
use Clone qw(clone);

use DE_EPAGES::Core::API::File qw( CreateDirectory DeleteDirectory WriteFile GetFileContentLines );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::Git qw( MapGitToPath ProjectUpdates4GitFingerprint );

sub usage {
    my ($ErrorMsg) = @_;
    print "Error:\n    $ErrorMsg\n\n" if $ErrorMsg;

    print <<END_USAGE;
Description:
    Creates a *.perl file containing the hash \$Fingerprint with paths and
    its sha1sums derived from Git repositories git\@github.com:ePages-de/* and
    (with --project) from git\@am-git.intern.epages.de:am-provider/*.

    Following Git repositories are used:

    o  git\@github.com:ePages-de/epages6-packages
    o  git\@github.com:ePages-de/node-binaries
    o  git\@github.com:ePages-de/node-epages6
    o  git\@github.com:ePages-de/epages-perl
    o  git\@github.com:ePages-de/epages-perl538
    o  git\@github.com:ePages-de/Cartridges
    o  git\@am-git.intern.epages.de:am-provider/ApplicationManagement (if --project)
    o  git\@am-git.intern.epages.de:am-provider/PROJECT (if --project)

    \$Fingerprint is a hash with \$Fingerprint->{ROLE}{PATH} = SHA1SUM

    Major ROLEs are:

    o  common (local-files)
    o  apache
    o  appsrv
    o  ascsrv
    o  reqsrv
    o  epages-j (j)
    o  mongo
    o  mongo-arbiter
    o  mysql

Usage:
    $0 [options] [flags]

Options:
    -git-base DIR  directory with local Git repository (default: env GIT_BASE)
    -version VRS   ePages version
    -write FILE    *.perl file (default: BASE/fingerprint-VRS-PROJ-HOST.perl)
                   If FILE is an existing directory, write
                   FILE/fingerprint-VRS-PROJ-HOST.perl
    -project PROJ  project in git\@am-git.intern.epages.de:am-provider/PROJ
    -system NAME.. list of hosts in am-provider/PROJ/Installation-Nodes/*/NAME
                   or all hosts in am-provider/PROJ/Installation-Nodes/NAME/*
Flags:
    -help            show the command line options
    -show            just show created files
    -nooutput        do not print any output

Example:
    $0 --git-base /tmp/git --version 7.11.0 --project PROJECT --system 'live my.example.com'
END_USAGE
    exit 2;
}

sub Main {

    local $| = 1;
    my %Options = ();
    my @GetOpt = qw(help nooutput show git-base=s version=s write=s project=s system=s);
    usage() unless GetOptions(\%Options, @GetOpt);

    usage() if $Options{help};
    usage('Option -version required') unless $Options{version};
    usage('Option -system requires -project') if $Options{system} and not $Options{project};

    my $Verbose = $Options{nooutput} || $Options{show} ? 0 : 1;
    my $Show = $Options{nooutput} ? 0 : 1;

    $Options{'git-base'} //= $ENV{GIT_BASE};
    usage('Option -git-base or environment veriable GIT_BASE required') unless defined $Options{'git-base'};

    my $time1 = time;

    my @RndRepos = qw(epages6-packages node-binaries node-epages6 epages-perl epages-perl538 Cartridges);
    my $RndUrl = 'git@github.com:ePages-de';

    my @AmRepos = ();
    if ($Options{project}) {
        @AmRepos = ($Options{project});
        unshift @AmRepos, '1und1Basic' if $Options{project} =~ /^1und1/;
        unshift @AmRepos, 'ApplicationManagement';
    }
    my $AmUrl = 'git@am-git.intern.epages.de:am-provider';

    CreateDirectory($Options{'git-base'}) unless -d $Options{'git-base'};

    _gitClone(\@RndRepos, $RndUrl, $Options{'git-base'}, $Verbose);
    _gitClone(\@AmRepos, $AmUrl, $Options{'git-base'}, $Verbose);

    my %Fingerprint;
    foreach my $RepoName (@RndRepos) {
        my $GitRepo = Git::Repository->new( work_tree => "$Options{'git-base'}/$RepoName" );
        print "$Options{'git-base'}/$RepoName: git fetch --update-head-ok origin release/v$Options{version}\n"
            if $Verbose;
        $GitRepo->run( fetch => '--update-head-ok', 'origin', "release/v$Options{version}" );
        print "$Options{'git-base'}/$RepoName: git ls-tree -r origin/release/v$Options{version}\n"
            if $Verbose;
        foreach my $Line ($GitRepo->run( 'ls-tree' => '-r', "origin/release/v$Options{version}" )) {
            my ($GitType, $Sha1, $File) = $Line =~ /^\s*\S+\s+(\S+)\s+(\S+)\s+(.*)/;
            next unless $GitType eq 'blob';
            my $PathList = MapGitToPath("$RndUrl/$RepoName", $File);
            next unless defined $PathList;
            $Fingerprint{$PathList->{$_}}{$_} = $Sha1 foreach keys %$PathList;
        }
    }
    print "\n" if $Verbose;

    foreach my $RepoName (@AmRepos) {
        my $GitRepo = Git::Repository->new( work_tree => "$Options{'git-base'}/$RepoName" );
        print "$Options{'git-base'}/$RepoName: git fetch --update-head-ok origin release/$Options{version}\n"
            if $Verbose;
        $GitRepo->run( fetch => '--update-head-ok', 'origin', "release/$Options{version}" );
    }

    my $Node = {
        localhost => {
            roles => [qw(common apache appsrv ascsrv reqsrv epages-j mongo mongo-arbiter mysql)],
            system => 'live',
        }
    };

    $Node = _getSystemNodes(\%Options) if $Options{system};

    foreach my $Host (keys %$Node) {
        my $NodeFingerprint = clone(\%Fingerprint);
        foreach my $RepoName (@AmRepos) {
            my $GitRepo = Git::Repository->new( work_tree => "$Options{'git-base'}/$RepoName" );
            ProjectUpdates4GitFingerprint($GitRepo, "origin/release/$Options{version}",
                                          $NodeFingerprint, $Node->{$Host},
                                          "$Options{'git-base'}/$RepoName", $Verbose);
        }

        my $Dump = Data::Dumper->new([$NodeFingerprint], ['Fingerprint']);
        my $FingerprintHash = $Dump->Dump;

        my $FingerprintFile = $Options{write};
        if (not defined $Options{write} or -d $Options{write}) {
            my $Dir = -d $Options{write} ? $Options{write} : $Options{'git-base'};
            $FingerprintFile = "$Dir/fingerprint-$Options{version}";
            $FingerprintFile .= "-$Options{project}" if $Options{project};
            $FingerprintFile .= "-$Node->{$Host}{host}" if defined $Node->{$Host}{host};
            $FingerprintFile .= '.perl';
        }

        print "WriteFile($FingerprintFile)\n" if $Show;
        print "\n" if $Verbose;
        WriteFile($FingerprintFile, \$FingerprintHash, {create_directory=>1});
    }

    printf "need %.1f seconds\n", time - $time1 if $Verbose;
}

RunScript(
    'Sub' => \&Main
);

sub _gitClone {
    my ($Repos, $RepoUrl, $GitBase, $Verbose) = @_;
    foreach my $RepoName (@$Repos) {
        next if -d "$GitBase/$RepoName/.git";
        DeleteDirectory("$GitBase/$RepoName", {'recursive' => 1});
        print "git clone $RepoUrl/$RepoName $GitBase/$RepoName\n" if $Verbose;
        Git::Repository->run( clone => "$RepoUrl/$RepoName", "$GitBase/$RepoName" );
    }
}

sub _getSystemNodes {
    my ($Options) = @_;
    my $Node = {};

    my %RoleFiles;
    foreach my $System (split(/\s+/, $Options->{system})) {
        my $InstallationNodes = "$Options->{'git-base'}/$Options->{project}/Installation-Nodes";
        $RoleFiles{$System} = $_ foreach glob("$InstallationNodes/*/$System/etc/epages-environment");
        unless (scalar keys %RoleFiles) {
            foreach (glob("$InstallationNodes/$System/*/etc/epages-environment")) {
                $RoleFiles{$1} = $_ if m,$InstallationNodes/$System/([^/]+)/,;
            }
        }
        unless (scalar keys %RoleFiles) {
            die "No system $Options->{system}, file missing: $InstallationNodes/*/$System/etc/epages-environment\n";
        }

        foreach my $Host (keys %RoleFiles) {
            $Node->{$Host}{system} = $1 if $RoleFiles{$Host} =~ m,$InstallationNodes/([^/]+)/,;
            $Node->{$Host}{roles} = [qw(common)];
            $Node->{$Host}{host} = $Host;
            foreach (@{GetFileContentLines($RoleFiles{$Host})}) {
                my ($Role) = split(/=/);
                next if $Role !~ /\S/ or $Role eq 'common';
                push @{$Node->{$Host}{roles}}, $Role;
            }
        }
    }
    return $Node;
}

1;
