#!perl -w
use Test::More;
use strict;
use DE_EPAGES::Core::API::PerlTools qw(
    AddDuration
    CheckIpFormat
    Compare
    CopyStructure
    ElementOf
    GenerateGUID
    GenerateShortGUID
    GetCallerPackage
    GetIPAddress
    GetMD5Digest
    IndexOf
    Ip2Hex
    SubtractDuration
    CreateSecureCode
    VerifySecureCode
    fcmp
    trim
    unique
    JsonTrue
    JsonFalse
);
use DE_EPAGES::Core::API::String qw ( ValidateNewStringLength );
use Data::Dumper qw ( Dumper );
use DateTime;
use utf8;
use JSON;

plan tests => 17;

## no critic qw(ProhibitMagicNumbers)

subtest "GetCallerPackage\n" => sub {
    ok( Test() eq 'main', 'GetCallerPackage main');
    ok( MyPackage::Test()  eq 'MyPackage',  'GetCallerPackage MyPackage');
    ok( MyPackage::Test2() eq 'MyPackage2',  'GetCallerPackage MyPackage2' );
};


subtest "JsonTrue/JsonFalse\n" => sub {
    is(JsonTrue(JSON::true), 1, 'JsonTrue(JSON::true)');
    is(JsonFalse(JSON::false), 1, 'JsonFalse(JSON::false)');

    is(JsonTrue(JSON::PP::true), 1, 'JsonTrue(JSON::PP::true)');
    is(JsonFalse(JSON::PP::false), 1, 'JsonFalse(JSON::PP::false)');

    is(JsonTrue(JSON::XS::true), 1, 'JsonTrue(JSON::XS::true)');
    is(JsonFalse(JSON::XS::false), 1, 'JsonFalse(JSON::XS::false)');

    is(JsonFalse(undef), 1, 'JsonFalse(undef)');
    is(JsonFalse('test'), 1, 'JsonFalse("test")');

    is(JsonTrue('true'), 1, 'JsonTrue("true")');
    is(JsonFalse('false'), 1, 'JsonFalse("false")');

    is(JsonTrue(1), 1, 'JsonTrue(1)');
    is(JsonFalse(0), 1, 'JsonFalse(0)');
};

subtest "trim/unique\n" => sub {
    ok( ('AAC'          eq join '', trim( '  A', ' A', "\t  C\n" )), 'trim' );
    ok( ("  A A\t  C\n" eq join '', unique( '  A', ' A', "\t  C\n" )), 'unique' );
    ok( ('AAC'          eq join '', trim unique( '  A', ' A', "\t  C\n" )), 'trim unique' );
    ok( ('AC'           eq join '', unique trim( '  A', ' A', "\t  C\n" )), 'unique trim' );
};

subtest "CopyStructure\n" => sub {
    my $Structure1 = GetStructure();
    my $Structure2 = GetStructure();
    local $Data::Dumper::Sortkeys = 1;
    ok (Dumper($Structure1) eq Dumper(CopyStructure($Structure2)), 'structure equal');
};

subtest "CheckIpFormat\n" => sub {
    ok( !defined CheckIpFormat('1.0.0.127'), 'CheckIpFormat IP4');
    ok( !defined CheckIpFormat('2001:4bd8:2:2:250:56ff:fe05:299'), 'CheckIpFormat IPv6');
    ok( !defined CheckIpFormat('::1'), 'CheckIpFormat localhost IPv6');
    cmp_ok( CheckIpFormat('100127'), 'eq', 'VOID_IP_ADDRESS', 'CheckIpFormat - no dots');
    cmp_ok( CheckIpFormat('x.0.0.127'), 'eq', 'VOID_IP_ADDRESS', 'CheckIpFormat - invalid characters');
    cmp_ok( CheckIpFormat('1.0..0.127'), 'eq', 'VOID_IPv4_ADDRESS', 'CheckIpFormat - invalid IPv4');
    cmp_ok( CheckIpFormat('1:1:2'), 'eq', 'VOID_IPv6_ADDRESS', 'CheckIpFormat - invalid IPv6');
};

subtest "GetIPAddress\n" => sub {
    my $LocalIP = GetIPAddress();
    is( CheckIpFormat($LocalIP), undef, "GetIPAddress(): $LocalIP");

    my $IP = GetIPAddress('255.255.255.255', 4);
    is( $IP, '255.255.255.255', 'GetIPAddress(255.255.255.255)' );

    $IP = GetIPAddress('localhost', 4);
    is( $IP, '127.0.0.1', 'GetIPAddress(localhost)' );

    SKIP: {
        skip 'not testing external systems', 2 if $ENV{'SKIP_EXTERNAL_TESTS'};
        $IP = GetIPAddress('www.google.com');
        isnt( $IP, $LocalIP, 'GetIPAddress(www.google.com)' );
        is( CheckIpFormat($IP), undef, "CheckIpFormat($IP)" );
    }
};

subtest "IP2Hex\n" => sub {
    cmp_ok( Ip2Hex('1.0.0.127'), 'eq', '0100007F', 'Ip2Hex localhost IP4');
    cmp_ok( Ip2Hex('::1'), 'eq', '00000000000000000000000000000001',  'Ip2Hex localhost IPv6');
    cmp_ok( Ip2Hex('2001:4bd8:2:2:250:56ff:fe05:299'), 'eq', '20014BD800020002025056FFFE050299', 'Ip2Hex IPv6');
    cmp_ok( Ip2Hex('2001:4bd8:2:2:250:56ff:fe05:299', 8), 'eq', 'FE050299', 'short Ip2Hex IPv6');
};

subtest "GenerateGUID\n" => sub {
    my $GUID1 = GenerateGUID();
    ok ($GUID1 =~ /^[0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12}$/x, 'GUID format'); ## no critic (RegularExpressions::ProhibitComplexRegexes)

    my $findDoubleGUID = 1;
    my %GUIDs;
    for(my $i=0;$i<100000;$i++) {
        my $GUID = GenerateGUID();
        $findDoubleGUID = 0 if exists $GUIDs{$GUID};
        $GUIDs{$GUID} = undef;
    }
    ok ($findDoubleGUID , '100000 created GUIDs not equal');
};

subtest "GenerateShortGUID\n" => sub {
    my $ShortGUID = GenerateShortGUID();
    ok ($ShortGUID =~ /^[0-9A-F]{20}$/, 'GenerateShortGUID format');
};

subtest "ElementOf\n" => sub {
    is( 2, IndexOf('ten', qw(twelve eleven ten)), 'IndexOf');
    ok( ElementOf('ten', qw(twelve eleven ten)), 'ElementOf');
    my $s1 = \'A'; #';
    my $s2 = 'B';
    my $s3 = 'X';
    my @a1= qw(A B C);
    my $a2= [qw(A B C)];
    ok( ElementOf($s2, @a1), 'ElementOf(string, array)');
    ok( ElementOf($s2, $a2), 'ElementOf(string, ref.array)');
    ok( ElementOf($s1, @a1), 'ElementOf(ref.string, array)');
    ok( ElementOf($s1, $a2), 'ElementOf(ref.string, ref.array)');
    ok( !ElementOf($s3, @a1), 'not ElementOf(string, array)');
    ok( !ElementOf($s3, $a2), 'not ElementOf(string, ref.array)');
};

subtest "fcmp\n" => sub {
    ok( 0  == fcmp(0.00011,1.10 - 1.09989), 'fcmp real');
    ok( -1 == fcmp(0,1), 'fcmp(0,1)');
    ok( 0   == fcmp(0.00011,0.00012, 4), 'fcmp equal with 4 decimal digits');
    ok( -1  == fcmp(0.00011,0.00012, 5), 'fcmp less with 5 decimal digits');
};

subtest "Compare\n" => sub {
    ok( 0 == Compare( 'String', undef, undef ), 'Compare undef to empty string' );
    ok( -1 == Compare( 'String', undef, '' ), 'Compare undef to empty string' );
    ok( 1 == Compare( 'String', '', undef ), 'Compare undef to empty string' );
    ok( 1 == Compare( 'String', 'a', 'B' ), 'Compare case sensitive' );
};

is( GetMD5Digest(\'test'), '098f6bcd4621d373cade4e832627b4f6', 'GetMD5Digest' );    #'

subtest "CreateSecureCode\n" => sub {
    is( CreateSecureCode("m\x{F6}llerfranzmoellder\@epages.de"), '218aec30aca416ddc49e04b4f903227c', 'CreateSecureCode' );
    is( VerifySecureCode("m\x{F6}llerfranzmoellder\@epages.de", '218aec30aca416ddc49e04b4f903227c'), 1, 'VerifySecureCode - true' );
    is( VerifySecureCode("m\x{F6}llerfranzmoellder\@epages.de", 'cc9e0e7699711ccb4d9e077690076a52'), 0, 'VerifySecureCode - false' );
};

subtest "AddDuration\n" => sub {
    my $dt = DateTime->new( year => 2010, month => 3, day => 28, hour => 0, minute => 30, time_zone => 'Europe/Berlin' );
    my $dt2 = AddDuration( $dt, 'hours', -2 );
    cmp_ok( "$dt",  'eq', '2010-03-28T00:30:00', 'original value has not changed' );
    cmp_ok( "$dt2", 'eq', '2010-03-27T22:30:00', 'substract 2 hours' );
    cmp_ok( $dt2->time_zone->name, 'eq', 'Europe/Berlin', 'time zone does not change' );

    my $dt3 = AddDuration( $dt, 'hours', 2 );
    cmp_ok( "$dt3", 'eq', '2010-03-28T03:30:00', 'skip invalid hour during daylight saving time switch' );

    $dt = DateTime->new( year => 2010, month => 3, day => 28, hour => 0, minute => 30 );
    my $dt4 = AddDuration( $dt, 'hours', 2 );
    cmp_ok( "$dt4", 'eq', '2010-03-28T02:30:00', 'add 2 hours to floating time' );
    cmp_ok( $dt4->time_zone->name, 'eq', 'floating', 'floating time zone does not change' );
};

subtest "SubtractDuration\n" => sub {
    my $dt = DateTime->new( year => 2010, month => 3, day => 28, hour => 0, minute => 30 );
    my $dt5 = SubtractDuration( $dt, 'hours', 2 );
    cmp_ok( "$dt5", 'eq', '2010-03-27T22:30:00', 'substract 2 hours from floating time' );
    cmp_ok( $dt, 'eq', $dt5, 'inpupt parameter has been modified' );
    cmp_ok( $dt5->time_zone->name, 'eq', 'floating', 'floating time zone does not change' );
};

subtest "ValidateNewStringLength\n" => sub {
    is(ValidateNewStringLength(undef,undef,15), 1, 'Validate undefined arg, allowed length');
    is(ValidateNewStringLength(undef,'Text is longer than the allowed length',15), 0, 'Validate undefined arg, new text too long');
    is(ValidateNewStringLength('Text is longer than the allowed length','<p>Text is longer than the allowed length</p>',15), 1, 'p tag special case, text to long');
    is(ValidateNewStringLength('Text', 'Text', 15), 1, 'Validate equal text, allowed length');
    is(ValidateNewStringLength('Text', "Text\r\n\a\e\f", 4), 1, 'Validate equal text but with non printable chars');
    is(ValidateNewStringLength('Text', 'txeT', 15), 1, 'Validate not equal text, allowed length');
    is(ValidateNewStringLength('Text is longer than the allowed length','Text is longer than the allowed length', 15), 1, 'Validate equal text, not allowed length');
    is(ValidateNewStringLength('Text is longer than the allowed length','Text is different and longer than the allowed length', 15), 0, 'Validate not equal text, not allowed length');
    is(ValidateNewStringLength('Text', '<p>Text</p>', 15), 1, 'Validate equal text with <p> tag added, allowed length');
    is(ValidateNewStringLength('Text', '<p>Text is longer than the allowed length</p>', 15), 0, 'Validate not equal text, not allowed length');
    is(ValidateNewStringLength('Text' . 'a' x 6, 'Test' . 'a' x 6, 9, 0.90), 1, 'Validate not equal text with exactly 90 % coverage');
    is(ValidateNewStringLength('Text' . 'b' x 6, 'Test' . 'b' x 6, 9, 0.95), 0, 'Validate not equal text with less than 95 % coverage');
};

# -----------------------------------------------------------------------------
# helper functions

sub Test {
    my $MyPackage = GetCallerPackage();
    return $MyPackage;
}
sub GetStructure {
    return {
        'Test1' => [1,3,4,5,6],
        'Test2' => {
            'AA' => 2,
            'BB' => 5,
        },
    };
}

##no critic (ProhibitMultiplePackages RequireCorrectImports)
package MyPackage; ## no critic (DE_EPAGES::Style::RequireCommentForExportedFunctions)
use DE_EPAGES::Core::API::PerlTools qw( GetCallerPackage );
sub Test {
    my $MyPackage = GetCallerPackage();
    return $MyPackage;
}
sub Test2 { return MyPackage2::Test(); }

package MyPackage2;
use DE_EPAGES::Core::API::PerlTools qw( GetCallerPackage );
sub Test {
    my $MyPackage = GetCallerPackage();
    return $MyPackage;
}
##use critic
