package DE_EPAGES::Coupon::API::CouponService;

use DE_EPAGES::Core::API::Error qw (ExistsError GetError Error);
use DE_EPAGES::Core::API::PerlTools qw (isValid JsonTrue);
use DE_EPAGES::Dictionary::API::Template;
use DE_EPAGES::Object::API::Factory qw (LoadClassByAlias Transaction);
use DE_EPAGES::Object::API::Object::Locale;
use DE_EPAGES::Object::API::Table::Language qw (GetInfoLanguage);
use DE_EPAGES::Shop::API::Resource::Formatter;
use DE_EPAGES::Shop::API::Resource::ShopRequestContext;

use strict;


sub new {
    my $class = shift;
    my $self  = bless {}, $class;
    return $self;
}

sub _isFeatureAvailable {
    my $self = shift;
    my ($Shop) = @_;
    if(!$Shop->get('FeatureMaxValue')->{'Coupon'}) {
       Error('FeatureUnavailable', {
          'Reason' => 'The Coupon feature is not available for the shop type!',
       });
    }
    return;
}

sub getCampaign {
    my $self = shift;
    my ($CouponCampaign, $LocaleID) = @_;

    my $Shop = $CouponCampaign->getSite();
    $self->_isFeatureAvailable($Shop);

    my $Context = $self->_createShopRequestContext($Shop);
    $LocaleID = $Context->validLocale($LocaleID);
    my $Locale = $self->_loadLocale($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');
    my $CouponCurrencyID = $CouponCampaign->get('CurrencyID');

    my $hCampaign = $self->_getCouponCampaignAttributeValues($CouponCampaign, $LanguageID);
    my $hCampaignData;
    my $hCampaignType->{'name'} = $hCampaign->{'Type'};

    my $Formatter = $self->_createFormatter($CouponCampaign->getSite(), $LocaleID);
    my $TaxModelUndefined = undef;

    if ($hCampaignType->{'name'} eq 'SHIPPING'){
        $hCampaignType->{'waiveShippingCosts'} = $Formatter->bool(1);
        $hCampaignType->{'formatted'} = $self->_getFreeDeliveryTranslation($LanguageID);
    } elsif($hCampaignType->{'name'} eq 'FIX'){
        $hCampaignType->{'amount'} = $hCampaign->{'Value'};
        $hCampaignType->{'formatted'} =
                $Formatter->price($hCampaign->{'Value'}, $CouponCurrencyID, $TaxModelUndefined)->{'formatted'};
        $hCampaignType->{'shippingCostsIncluded'} = $Formatter->bool($hCampaign->{'InclShippingCosts'});
        $hCampaignType->{'giftCouponWithoutTax'} = $Formatter->bool($hCampaign->{'WithoutTax'});
    } elsif($hCampaignType->{'name'} eq 'PERCENT'){
        $hCampaignType->{'percentage'} = $hCampaign->{'Value'};
        $hCampaignType->{'formatted'} =
                $Formatter->percentage($hCampaign->{'Value'})->{'formatted'};
        $hCampaignType->{'shippingCostsIncluded'} = $Formatter->bool($hCampaign->{'InclShippingCosts'});
    }

    $hCampaignData = {
        'campaignId'        => $hCampaign->{'GUID'},
        'identifier'        => $hCampaign->{'Alias'},
        'name'              => $hCampaign->{'Name'},
        'currency'          => $hCampaign->{'CurrencyID'},
        'minimumOrderPrice' => $Formatter->price($hCampaign->{'MinLineItemsSubTotal'}, $CouponCurrencyID, $TaxModelUndefined),
        'type'              => $hCampaignType,

    };

    return $hCampaignData;
}

sub createCampaign {
    my $self = shift;
    my ($Shop, $LocaleID, $hRequestBody) = @_;

    $self->_isFeatureAvailable($Shop);

    my $Parent  = $Shop->folder('CouponCampaigns');
    my $Class   = LoadClassByAlias ('CouponCampaign');

    my $Context = $self->_createShopRequestContext($Shop);
    $LocaleID = $Context->validLocale($LocaleID);
    my $Locale = $self->_loadLocale($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');

    my $Alias = $hRequestBody->{'identifier'};
    my $CurrencyID = $Context->validCurrency($hRequestBody->{'currency'});
    my $Position = $Parent->nextPosition();

    if(!isValid($Alias)) {
        Error('BadRequest', {
            'Reason' => 'The attribute identifier is mandatory!',
        });
    }
    if(!isValid($CurrencyID)) {
        Error('BadRequest', {
            'Reason' => 'The currency is mandatory!',
        });
    }
    my $CouponCampaign;
    eval {
        $self->_Transaction(
            sub {
                $CouponCampaign = $Class->insertObject( {
                    'Alias' => $Alias,
                    'Parent' => $Parent,
                });

                $CouponCampaign->set({
                    'Name' => $hRequestBody->{'name'},
                }, $LanguageID);

                $CouponCampaign->set({'CurrencyID' => $CurrencyID});

                $CouponCampaign->set({'Position' => $Position});

                if(isValid($hRequestBody->{'minimumOrderValue'})) {
                    $CouponCampaign->set({'MinLineItemsSubTotal' => $hRequestBody->{'minimumOrderValue'}});
                }

                $self->_setCouponCampaign($CouponCampaign, $hRequestBody);
            }
        );
    };
    if (ExistsError()) {
        my $Error = GetError();
        if ( $Error->code() eq 'OBJECT_EXISTS' ) {
            Error('BadRequest', {
                'Reason: CouponCampaign with given identifier exists already! identifier '=> $Alias,
            });
        } else {
            Error($Error);
        }
    }
    return $self->getCampaign($CouponCampaign, $LocaleID);
}



sub createCouponForCampaign {
    my $self = shift;
    my ($Campaign, $CouponCode) = @_;
    my $Shop = $Campaign->getSite();

    $self->_isFeatureAvailable($Shop);

    my $Folder = $Shop->folder('CouponCodes');
    my $Class  = LoadClassByAlias('CouponCode');
    eval {
        $self->_Transaction(
            sub {
                $Class->insertObject( {
                    'Alias' => $CouponCode,
                    'Parent' => $Folder,
                    'CouponCampaign' => $Campaign,
                    'Type'  => 'MANUAL'
                });
            });
    };
    if (ExistsError()) {
        my $Error = GetError();
        if ( $Error->code() eq 'OBJECT_EXISTS' ) {
            Error('BadRequest', {
                'Reason: Coupon with given code exists already! code '=> $CouponCode,
            });
        } else {
            Error($Error);
        }
    }
    return;
}



sub _loadLocale {
    my $self = shift;
    my ($LocaleID) = @_;
    return DE_EPAGES::Object::API::Object::Locale->load($LocaleID);
}



sub _Transaction {
    my $self = shift;
    my ($cCode, $cRollback) = @_;
    Transaction($cCode, $cRollback);
    return;
}

sub _setCouponCampaign {
    my $self = shift;
    my ($CouponCampaign, $hRequestBody) = @_;

    my $hRequestCouponType = $hRequestBody->{'type'};
    $CouponCampaign->set({'Type' => $hRequestCouponType->{'name'}});

    if($hRequestCouponType->{'name'} eq 'PERCENT'){
        $CouponCampaign->set({'Value' => $hRequestCouponType->{'percentage'} });
        my $ShippingCostsIncluded = JsonTrue($hRequestCouponType->{'shippingCostsIncluded'});
        $CouponCampaign->set({'InclShippingCosts' => $ShippingCostsIncluded });
    } elsif ($hRequestCouponType->{'name'} eq 'FIX') {
        $CouponCampaign->set({'Value' => $hRequestCouponType->{'amount'} });

        my $ShippingCostsIncluded = JsonTrue($hRequestCouponType->{'shippingCostsIncluded'});
        my $giftCouponWithoutTax = JsonTrue($hRequestCouponType->{'giftCouponWithoutTax'});

        $CouponCampaign->set({'InclShippingCosts' => $ShippingCostsIncluded });
        $CouponCampaign->set({'WithoutTax' => $giftCouponWithoutTax });
    }
}

sub _getCouponCampaignAttributeValues {
    my $self = shift;
    my ($CouponCampaign, $LanguageID) = @_;

    my @Attributes = qw (
        GUID
        Alias
        Name
        CurrencyID
        MinLineItemsSubTotal
        Type
        Value
        InclShippingCosts
        WithoutTax
    );

    return $CouponCampaign->get( \@Attributes, $LanguageID );
}


sub _createFormatter {
    my $self=shift;
    my ($Shop,$LocaleID)=@_;
    return DE_EPAGES::Shop::API::Resource::Formatter->new($Shop,$LocaleID);
}


sub _createShopRequestContext {
    my $self = shift;
    my ($Shop) = @_;
    return DE_EPAGES::Shop::API::Resource::ShopRequestContext->new($Shop);
}

sub _getFreeDeliveryTranslation {
    my $self = shift;
    my ($LanguageID) = @_;

    my $hLanguageInfo       = $self->_getInfoLanguage($LanguageID);
    my $LanguageCode2       = $hLanguageInfo->{'Code2'};
    my $DictionaryXml       = $ENV{'EPAGES_CARTRIDGES'}."/DE_EPAGES/Coupon/Templates/Dictionary.$LanguageCode2.xml";
    my $Template            = $self->_getTemplate();
    my $hDictionary         = $Template->dictionary( $DictionaryXml, $LanguageCode2 );
    my $hTranslationStrings = $hDictionary->{$LanguageCode2};

    return $hTranslationStrings->{'FreeShipping'};
}

sub _getInfoLanguage {
    my $self = shift;
    my ($LanguageID) = @_;

    return GetInfoLanguage($LanguageID);
}

sub _getTemplate {
    my $self = shift;

    return DE_EPAGES::Dictionary::API::Template->new();
}

1;
