use strict;
use Getopt::Long;
use Time::HiRes qw (time);
use DE_EPAGES::Cartridge::API::Table::Cartridge qw ( ExistsCartridgeByPackage );
use DE_EPAGES::Core::API::Error qw( ExistsError GetError );
use DE_EPAGES::Core::API::Log qw ( GetLog LogError);
use DE_EPAGES::Core::API::PerlTools qw( ElementOf );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::Warning qw ( Warning BlockWarnings);
use DE_EPAGES::Database::API::Connection qw( RunOnStore );
use DE_EPAGES::Ebay::API::Ebay qw ( UpdateAuctions UpdateShopAuctions );
use DE_EPAGES::Object::API::Factory qw ( LoadObjectByPath );
use DE_EPAGES::WebInterface::API::MessageCenter qw( SynchronizeCache );

sub Main {

    local $| = 1;

    my $Password;
    my $StoreName = 'Store';
    my $ShopAlias;
    my $Help;
    my $NoOutput;
    my $IgnoreErrors = 0;

    GetOptions(
        'help'          => \$Help,
        'nooutput'      => \$NoOutput,
        'ignore_error'  => \$IgnoreErrors,
        'passwd=s'      => \$Password,
        'storename=s'   => \$StoreName,
        'shop=s'        => \$ShopAlias,
    );

    usage() if $Help;

    my $time1 = time;

    RunOnStore(
        Store => $StoreName,
        DBPassword => $Password,
        Sub => sub  {

            # make sure that the cartridge is installed
            unless (ExistsCartridgeByPackage('DE_EPAGES::Ebay')) {
                print "Store: $StoreName : Cartridge Ebay is not installed.\n" unless ($NoOutput);
                return;
            }

            my $aWarnings;
            if ( defined $ShopAlias and $ShopAlias ne 'all' ) {
                my $Shop;
                eval { $Shop = LoadObjectByPath('/Shops')->child($ShopAlias); };
                if ( ExistsError() ) {
                    printf STDERR "[%s] WARN: Shop '%s' does not exist.\n", $StoreName, $ShopAlias;
                    return;
                }
                $aWarnings = BlockWarnings( sub {
                    UpdateShopAuctions( $Shop, ($NoOutput)?0:1 )
                });
            }
            else {
                $aWarnings = BlockWarnings( sub {
                        UpdateAuctions( ($NoOutput)?0:1 )
                    });
            }

            foreach my $Warning (@$aWarnings) {
                my $LogMessage = substr($Warning->shortMessage, 0, 4096); ## no critic qw(ProhibitMagicNumbers)
                if ($IgnoreErrors && ElementOf($Warning->code, qw (EbayTokenError EbayNoToken EbayRequestFailed UpdateTransactionListError))) {
                    GetLog('EBAY')->debug( $LogMessage );
                } else {
                    LogError($Warning);
                    print "WARN: $LogMessage\n";
                }
            }
        }
    );

    my $logmsg = sprintf ('need %.1f seconds', time - $time1);
    print $logmsg."\n" unless $NoOutput;
    GetLog()->debug("Update auctions: $logmsg");

    eval { SynchronizeCache(); };
    Warning( GetError() ) if ExistsError();
}

sub usage {
    print <<END_USAGE;
Usage:
    perl $0 [options] [flags]

options:
    -passwd       database user password              ()
    -storename    name of store                       (Store)
    -shop         shop alias               (optional) ()

flags:
    -nooutput     do not print any output
    -help         show the command line options
    -ignore_error ignores errors EbayTokenError EbayNoToken EbayRequestFailed, logs errors in category EBAY

Example:
    perl $0 -storename Store
END_USAGE
    exit 2;
}

RunScript(
    'Sub' => \&Main
);
