#========================================================================================
# package      DE_EPAGES::Ebay::UI::EbayAuctionWizard
# state        private
# base         DE_EPAGES::Presentation::UI::Object
# description  functions for ebay wizard
#========================================================================================
package DE_EPAGES::Ebay::UI::EbayAuctionWizard;
use base DE_EPAGES::Ebay::UI::Ebay;

use strict;
use DE_EPAGES::Cartridge::API::NonPermTemplate qw ( ProcessFileTemplate );
use DE_EPAGES::Core::API::Error qw ( GetError ExistsError Error);
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Ebay::API::Tools qw ( GetRecentCategories );
use DE_EPAGES::Ebay::API::Constants;
use DE_EPAGES::Ebay::API::Object::EbayCategoryTree;
use DE_EPAGES::Ebay::API::Object::RemoteRESTEbay;
use DE_EPAGES::Ebay::API::Table::EbayAuctionAttribute
    qw (InsertEbayAuctionAttribute DeleteEbayAuctionAttribute GetPKeyEbayAuctionAttributesByAuctionID GetInfoEbayAuctionAttributesByAuctionID);
use DE_EPAGES::Object::API::Factory qw ( ExistsObject LoadRootObject LoadClassByAlias LoadObjectByPath );
use DE_EPAGES::Object::API::Language qw ( GetPKeyLanguageByCode );
use DE_EPAGES::Presentation::API::Template qw ( Translate );
use DE_EPAGES::Product::API::Object::PagerSubProducts;
use DE_EPAGES::Search::API::SearchDefinition qw (GetSearchDefinition);
use feature qw ( switch );
use JSON;
use MIME::Base64;
use constant EBAY_ATTRIBUTESETS_VALUE_OTHER => -6;
use constant EBAY_ATTRIBUTESETS_VALUE_UNDEF => -10;
use constant MAX_CATEGORY_DEPTH => 10;


#==============================================================================
# Event handlers
#==============================================================================

sub View {
    my $self = shift;
    my ($Servlet) = @_;

    my $auction = $Servlet->object;
    my $ShopEnv = $auction->get('EbayShopEnvironment');

    if ($self->_checkAuctionDescription($auction)) {
        $Servlet->vars('AuctionDescriptionInvalid',1);
    }

    $Servlet->vars('EbayWarningNoEbayEnvironment',1)
        unless (defined $ShopEnv);

    $self->_TranslateDefaultAuctionPageTitle($Servlet);

    return $self->SUPER::View($Servlet);
}

sub _checkAuctionDescription {
    my $self = shift;
    my ($Auction) = @_;
    my $AuctionDescription = $Auction->get('Description');
    if (defined($AuctionDescription) && (
        $AuctionDescription =~ /window.open/i ||
        $AuctionDescription =~ /cookie\(/i ||
        $AuctionDescription =~ /replace\(/i ||
        $AuctionDescription =~ /<iframe /i ||
        $AuctionDescription =~ /<base /i
    )) {
        return 1;
    }
    return 0;
}

sub ViewVariationAssignmentStep {
    my $self      = shift;
    my ($Servlet) = @_;
    my $EbayAuction = $Servlet->object;

    my $Shop = $EbayAuction->get('Shop');

    my $Folder = $Shop->folder('EbayAuctions');
    my $SearchName = 'MBO-EbayAuction';
    my $SearchDefinition = GetSearchDefinition(LoadClassByAlias('EbayAuction'), $SearchName);
    my $Parameters = {
        'Shop.ID'   => $Shop->id,
        'ParentAuctionID' => $EbayAuction->id,
    };

    my $AllowWithoutStock = 0;
    if ($Shop->get('EbayStockDecreaseType') eq 'OnAuctionEnd' && $EbayAuction->get('IsReviseInventory')) {
        $AllowWithoutStock = 1;
    }

    # only search for variation items with quantity > 0 in case without stocklevel is not allowed
    unless ($AllowWithoutStock) {
        $Parameters->{'MinStockLevel'} = 0;
    }

    my %Options = (
        'Servlet' => $Servlet,
        'Object'  => $Folder,
        'SearchDefinition' => $SearchDefinition,
        'Parameters' => $Parameters,
    );

    my $Search = DE_EPAGES::Search::API::Object::PagerSearch->new(\%Options);
    return $Search->page;
}

sub ViewVariationsStep {
    my $self      = shift;
    my ($Servlet) = @_;
    my $EbayAuction = $Servlet->object;

    my $Shop = $EbayAuction->get('Shop');
    my $Folder = $Shop->folder('EbayAuctions');
    my $SearchName = 'MBO-EbayAuction';
    my $SearchDefinition = GetSearchDefinition(LoadClassByAlias('EbayAuction'), $SearchName);
    my %Options = (
        'Servlet' => $Servlet,
        'Object'  => $Folder,
        'SearchDefinition' => $SearchDefinition,
        'Parameters' => {
            'Shop.ID'   => $Shop->id,
            'ParentAuctionID' => $EbayAuction->id,
        },
    );

    my $Search = DE_EPAGES::Search::API::Object::PagerSearch->new(\%Options);
    return $Search->page;
}

sub UnityViewCategoryStep {
    my $self = shift;
    my ($Servlet) = @_;

    my $Auction = $Servlet->object;
    my $ShopEnv = $Auction->get('EbayShopEnvironment');

    $self->_TranslateDefaultAuctionPageTitle($Servlet);

    # no shop environment
    $Servlet->vars('EbayWarningNoEbayEnvironment',1) unless defined $ShopEnv;
    return unless defined $ShopEnv;

    # get parent category from query
    my $Form             = $Servlet->form;
    my $ParentCategoryID = $Form->queryValue('ParentCategoryID');

    # get info of category
    my $ExternalCategoryId  = $Auction->get('EbayCategoryId');
    my $EbayEnvironment     = $ShopEnv->get('EbayEnvironment');
    my $EbayCategoryTree    = DE_EPAGES::Ebay::API::Object::EbayCategoryTree->new($EbayEnvironment);
    my $FirstCategory = $EbayCategoryTree->getInfoEbayCategoryByExternalCategoryId($ExternalCategoryId);

    if (defined $FirstCategory) {
        $FirstCategory->{'Path'} = $EbayCategoryTree->getCategoryPath($FirstCategory->{'EbayCategoryID'});
        $Servlet->params('EbayCategory', $FirstCategory);

        # get parent category from auction
        if (not defined $ParentCategoryID and defined $FirstCategory) {
            $ParentCategoryID       = $FirstCategory->{$FirstCategory->{'IsLeaf'} ? 'ParentCategoryID' : 'EbayCategoryID'};
            my $InternalCategoryId  = $FirstCategory->{'EbayCategoryID'};
            $Servlet->params('InternalCategoryId', $InternalCategoryId);
        }
    }

    # set parent to root if not ParentCategoryID
    $ParentCategoryID = 0 unless $ParentCategoryID;

    # get the second category
    my $SecExternalCategoryId = $Auction->get('SecondCategoryId');
    my $SecondCategory        = $EbayCategoryTree->getInfoEbayCategoryByExternalCategoryId($SecExternalCategoryId);
    if (defined $SecondCategory) {
        $SecondCategory->{'Path'} = $EbayCategoryTree->getCategoryPath($SecondCategory->{'EbayCategoryID'});
        $Servlet->params('SecondCategory', $SecondCategory);

        my $InternalSecondCategoryId = $SecondCategory->{'EbayCategoryID'};
        $Servlet->params('InternalSecondCategoryId', $InternalSecondCategoryId);

        my $InternalSecondParentCategoryId = $SecondCategory->{$SecondCategory->{'IsLeaf'} ? 'ParentCategoryID' : 'EbayCategoryID'};
        $Servlet->params('InternalSecondParentCategoryId', $InternalSecondParentCategoryId);
    }


    #load categories
    my @Categories;
    if (not $ParentCategoryID) {
        #get root categories
        push @Categories, @{ $EbayCategoryTree->getInfoRootCategories() };
    } else {
        #get sub categories of parent
        my $Subs = $EbayCategoryTree->getInfoEbayCategoriesByParentCategoryID( $ParentCategoryID );
        push @Categories, @$Subs;
    }

    # load recent categories
    $Servlet->vars('EbayRecentCategories', GetRecentCategories( $Auction, MAX_CATEGORY_DEPTH ));

    my @CategoriesSorted;
    @CategoriesSorted = sort{ "\U".$a->{Name} cmp "\U".$b->{'Name'} }(@Categories);
    $Servlet->vars('EbayCategories', \@CategoriesSorted);

    # load current category

    my $ParentCategory = $ParentCategoryID ? $EbayCategoryTree->getInfoEbayCategory( $ParentCategoryID ) : {} ;
    $ParentCategory->{'Path'} = $EbayCategoryTree->getCategoryPath($ParentCategoryID );
    $Servlet->vars('EbayBrowserCategory', $ParentCategory);

    eval {
        # load EbayStore categories
        my $RemoteEbay = DE_EPAGES::Ebay::API::Object::RemoteRESTEbay->new('ShopEnvironment' => $ShopEnv,);

        my $hStoreOwner = $RemoteEbay->getUser(['StoreOwner']);
        my $StoreOwner = $hStoreOwner ? $hStoreOwner->{'StoreOwner'} : undef;
        if ($StoreOwner and $StoreOwner eq 'true') {
            my $aStoreCategories = $RemoteEbay->getStore()->{'CustomCategories'};
            $Servlet->vars('EbayStoreCategories', $aStoreCategories);
        }
   };

    if (ExistsError()) {
        $self->_handleEbayError($Servlet);
    }

    my $JsonCategoryObject = {
        'Success'         => JSON::true,
        'CategoryList'    => \@CategoriesSorted,
        'CurrentCategory' => $ParentCategory,
        'FirstCategory'   => $FirstCategory,
        'SecondCategory'  => $SecondCategory
    };

    my $JSON = to_json($JsonCategoryObject, {pretty => 1, indent => 2});
    $Servlet->vars('JsonCategoryObject', $JSON);
}


sub ViewCategoryStep {
    my $self = shift;
    my ($Servlet) = @_;

    my $auction = $Servlet->object;
    my $ShopEnv = $auction->get('EbayShopEnvironment');

    $self->_TranslateDefaultAuctionPageTitle($Servlet);

    $Servlet->vars('EbayWarningNoEbayEnvironment',1)
        unless (defined $ShopEnv);

    my $Form = $Servlet->form;

    # get parent category from query
    my $ParentCategoryID = $Form->queryValue('ParentCategoryID');

    return unless $ShopEnv;

    # get info of category
    my $ExternalCategoryId = $auction->get('EbayCategoryId');
    my $Env = $ShopEnv->get('EbayEnvironment');
    my $EbayCategoryTree = DE_EPAGES::Ebay::API::Object::EbayCategoryTree->new($Env);
    my $category = $EbayCategoryTree->getInfoEbayCategoryByExternalCategoryId($ExternalCategoryId);
    if (defined $category) {
        $category->{'Path'} = $EbayCategoryTree->getCategoryPath($category->{'EbayCategoryID'});
        $Servlet->params('EbayCategory', $category);

        # get parent category from auction
        if (not defined $ParentCategoryID and defined $category) {
            $ParentCategoryID = $category->{$category->{'IsLeaf'} ? 'ParentCategoryID' : 'EbayCategoryID'};
        }
    }

    # get the second category
    my $SecExternalCategoryId = $auction->get('SecondCategoryId');
    my $seccategory = $EbayCategoryTree->getInfoEbayCategoryByExternalCategoryId($SecExternalCategoryId);
    if (defined $seccategory) {
        $seccategory->{'Path'} = $EbayCategoryTree->getCategoryPath($seccategory->{'EbayCategoryID'});
        $Servlet->params('SecondCategory', $seccategory);
    }

    # set parent to root
    $ParentCategoryID = 0 unless $ParentCategoryID;

    #load categories
    my @categories;
    if (not $ParentCategoryID) {
        #get root categories
        push @categories, @{ $EbayCategoryTree->getInfoRootCategories() };
    } else {
        #get sub categories of parent
        my $subs = $EbayCategoryTree->getInfoEbayCategoriesByParentCategoryID( $ParentCategoryID );
        push @categories, @$subs;
    }

    # load recent categories
    $Servlet->vars('EbayRecentCategories', GetRecentCategories( $auction ));

    my @categories_sorted = sort{ "\U".$a->{Name} cmp "\U".$b->{'Name'} }(@categories);
    $Servlet->vars('EbayCategories', \@categories_sorted);

    my $ParentCategory;
    $ParentCategory = $EbayCategoryTree->getInfoEbayCategory( $ParentCategoryID ) if $ParentCategoryID;
    $ParentCategory->{'Path'} = $EbayCategoryTree->getCategoryPath($ParentCategoryID );
    $Servlet->vars('EbayBrowserCategory', $ParentCategory);

    eval {
        # load EbayStore categories
        my $RemoteEbay = DE_EPAGES::Ebay::API::Object::RemoteRESTEbay->new(
            'ShopEnvironment' => $ShopEnv,
        );

        my $hStoreOwner = $RemoteEbay->getUser(['StoreOwner']);
        my $StoreOwner = $hStoreOwner ? $hStoreOwner->{'StoreOwner'} : undef;
        if ($StoreOwner and $StoreOwner eq 'true') {
            my $aStoreCategories = $RemoteEbay->getStore()->{'CustomCategories'};
            $Servlet->vars('EbayStoreCategories', $aStoreCategories);
        }
   };

    if (ExistsError()) {
        $self->_handleEbayError($Servlet);
    }
}

# show the detail page and get informations about Attributes
sub ViewDetailStep {
    my $self = shift;
    my ($Servlet) = @_;

    my $auction = $Servlet->object;
    my $ShopEnv = $auction->get('EbayShopEnvironment');

    $self->_TranslateDefaultAuctionPageTitle($Servlet);

    eval {
        # load EbayStore categories
        my $RemoteEbay = DE_EPAGES::Ebay::API::Object::RemoteRESTEbay->new(
            'ShopEnvironment' => $ShopEnv,
        );

        my $hStoreOwner = $RemoteEbay->getUser(['StoreOwner']);
        my $StoreOwner = $hStoreOwner ? $hStoreOwner->{'StoreOwner'} : undef;
        if ((defined $StoreOwner) and ($StoreOwner eq 'true')) {
            $Servlet->vars('IsEbayStoreOwner', 1);
        }

        my $hresult         = $RemoteEbay->getEbayDetails('ReturnPolicyDetails');
        my $aReturnsWithin  = $self->getReturnsWithinOptionsCodes($hresult);
        $self->addReturnsWithinOptionsCodes($Servlet,$aReturnsWithin);

        my $CategoryID = $auction->get('EbayCategoryId');

        if ($CategoryID > 0) {
            eval { ## no critic (RequireExistsError)
                my $ConditionValues = $RemoteEbay->getCategoryFeatures($CategoryID,[qw(ProductRequiredEnabled ConditionValues ItemSpecificsEnabled)],0,1);
                if (defined $ConditionValues
                    and ref $ConditionValues eq 'HASH'
                    and ref $ConditionValues->{'ConditionValues'} eq 'HASH'
                    and defined($ConditionValues->{'ConditionValues'}->{'Condition'})) {
                        if (ref $ConditionValues->{'ConditionValues'}->{'Condition'} eq 'HASH') {
                            $Servlet->vars('ConditionValues', {
                                'Condition' => [$ConditionValues->{'ConditionValues'}->{'Condition'}]
                            });
                        } elsif (scalar @{$ConditionValues->{'ConditionValues'}->{'Condition'}}) {
                            $Servlet->vars('ConditionValues',$ConditionValues->{'ConditionValues'});
                        }
                }

                if ($ConditionValues->{'ItemSpecificsEnabled'} eq 'Enabled') {
                    my $Specs = $RemoteEbay->getItemAspectForCategory($CategoryID);

                    $Servlet->vars('UseItemSpecifics', 1);
                    if ($auction->get('Product')->get('IsMaster')) {
                        if ($Specs) {
                            # Remove the variation attributes from the item specifics
                            my $SpecsRecommendations = _getItemSpecificsWithoutVariationAttr($auction, $Specs);
                            $Servlet->vars('ItemSpecifics', $SpecsRecommendations);
                        }
                    } else {
                        $Servlet->vars('ItemSpecifics',$Specs) if defined($Specs);
                    }
                    $Servlet->vars('SelectedItemSpecifics',$auction->getItemSpecifics());
                }
            };
        }
    };

    if (ExistsError()) {
        $self->_handleEbayError($Servlet);
    }

    my $Env = $ShopEnv->get('EbayEnvironment');
    my $EbayCategoryTree = DE_EPAGES::Ebay::API::Object::EbayCategoryTree->new($Env);

    my $ExternalCategoryId = $auction->get('EbayCategoryId');
    my $hEbayCategory = $EbayCategoryTree->getInfoEbayCategoryByExternalCategoryId($ExternalCategoryId);

    # get the attribute values saved for the auction and map to a hash for adding to the attributeset loop
    my %UsedAttributes;
    my $ahEbayAuctionAttributes = GetInfoEbayAuctionAttributesByAuctionID($auction->id);

    # remap
    foreach my $hEAAttr (@$ahEbayAuctionAttributes) {

        $UsedAttributes{$hEAAttr->{'EbayAttributeSetID'}}{$hEAAttr->{'EbayAttributeID'}}{'ValueLiteral'} = $hEAAttr->{'EbayValueLiteral'};
        #add IDs
        push @{$UsedAttributes{$hEAAttr->{'EbayAttributeSetID'}}{$hEAAttr->{'EbayAttributeID'}}{'EbayValueIDs'}}, {'SelectedID' => $hEAAttr->{'EbayValueID'}};
    }

    # add the constants for the global attributesets
    $hEbayCategory->{'SiteConditionAttributeID'} = EBAY_ATTRIBUTEID_ITEM_CONDITION ;
    $hEbayCategory->{'SiteConditionValue_NEW'} = EBAY_VALUEID_ITEM_CONDITION_NEW ;
    $hEbayCategory->{'SiteConditionValue_USED'} = EBAY_VALUEID_ITEM_CONDITION_USED ;

    if (defined $Env->get('AttrSetIDCondition') && defined $UsedAttributes{$Env->get('AttrSetIDCondition')}{$hEbayCategory->{'SiteConditionAttributeID'}}) {
        my $selectedCondition = pop @{ $UsedAttributes{$Env->get('AttrSetIDCondition')}{$hEbayCategory->{'SiteConditionAttributeID'}}{'EbayValueIDs'} };
        if (defined($selectedCondition) && defined($selectedCondition->{'SelectedID'})) {
            $hEbayCategory->{'SiteConditionAttributeSelectedID'} = $selectedCondition->{'SelectedID'};
        }
    }

    # get attributeset from SQLite DB
    my $AttributeSetID = $hEbayCategory->{'AttributeSetID'};
    my $hAttributeSet  =  $EbayCategoryTree->getInfoAttributeSetByAttributeSetId($AttributeSetID);
    $hEbayCategory->{'AttributeSet'} = $hAttributeSet;

    # get attributes from SQLite DB
    my $ahAttributes  =  $EbayCategoryTree->getInfoAttributesByAttributeSetId($AttributeSetID);

    # get values for each attribute from SQLite DB
    foreach my $hAttribute (@$ahAttributes) {

        # add the values from SQLite DB
        my $ahAttributeValues = $EbayCategoryTree->getInfoAttributeValuesByAttributeId($hAttribute->{'AttributeID'});
        $hAttribute->{'Values'} = $ahAttributeValues;

        # set SelectedID and selected value if we have a textfield value used by the merchant
        $hAttribute->{'SelectedIDs'} = $UsedAttributes{$hAttributeSet->{'ExternalAttributeSetID'}}{$hAttribute->{'ExternalAttributeID'}}{'EbayValueIDs'};
        $hAttribute->{'SelectedValueLiteral'}=$UsedAttributes{$hAttributeSet->{'ExternalAttributeSetID'}}{$hAttribute->{'ExternalAttributeID'}}{'ValueLiteral'};

        # add the dependencies from SQLite DB
        my $ahAttributeDependencies = $EbayCategoryTree->getInfoAttributeValueDependenciesByAttributeId($hAttribute->{'AttributeID'});
        $hAttribute->{'Dependencies'} = $ahAttributeDependencies;

        # add the attributevalidationrules from SQLite DB
        my $ahAttributeValidationRules = $EbayCategoryTree->getInfoAttributeValidationRulesByAttributeId($hAttribute->{'AttributeID'});
        $hAttribute->{'ValidationRules'} = $ahAttributeValidationRules;
    }

    $hEbayCategory->{'AttributeSet'}{'Attributes'} = $ahAttributes;
    $Servlet->params('EbayCategory', $hEbayCategory);
}

sub _handleEbayError {
    my $self = shift;
    my ($Servlet) = @_;


    my $Error = GetError();
    if($Error->code eq 'EbayNoToken' or $Error->code eq 'EbayTokenError') {
        $Servlet->vars('EbayWarningNoToken', 1);
    } elsif ($Error->code eq 'EbayRequestNotSuccessful') {
        my $aEbayErrors = $Error->vars->{'EbayErrors'};

        # check if the sole error is "User has no shop"
        if (scalar @$aEbayErrors == 1 and grep {$_->{'ErrorCode'} == EBAY_ERROR_SELLER_NO_STORE} @$aEbayErrors) {
            GetLog->debug('Shop has no EbayStore.');
        } else {
            # all other warnings will be shown to the merchant
            $Servlet->vars('EbayErrors', $aEbayErrors);
        }
    } elsif ($Error->code eq 'NoEbayEnvironment') {
        $Servlet->vars('EbayWarningNoEbayEnvironment',1);
        return;
    } else {
        Error($Error);
    }

}

sub SaveSubAuctions {

    my $self = shift;
    my ($Servlet) = @_;

    my $auction = $Servlet->object;
    my $shop    = $auction->getSite();

    my $form   = $Servlet->form->form($auction, 'SaveSubAuctions');

    foreach my $SubAuctionEntry (@{$form->{'SubAuctions'}}) {
        my $SubAuction = $shop->loadObject($SubAuctionEntry->{'SubAuctionID'}, 'EbayAuction');
        $Servlet->testPermission($SubAuction, 'Save');
        delete $SubAuctionEntry->{'SubAuctionID'};
        $SubAuction->set($SubAuctionEntry);

    }

    my $MasterQuantity = 0;
    foreach my $SubAuction ( @{$auction->get('SubAuctions')} ) {
        $MasterQuantity += $SubAuction->get('Quantity');
    }
    $auction->set({'Quantity' => $MasterQuantity});

    return;

}

sub Save { ##no critic qw(ProhibitExcessComplexity)
    my $self = shift;
    my ($Servlet) = @_;

    my $auction = $Servlet->object;
    my $Form = $Servlet->form;
    # do not save if auction is running or ended
    return if $auction->get('Status') & (AUCTION_RUNNING | AUCTION_ENDED);

    my $oldShopEnv = $auction->get('EbayShopEnvironment');
    my $oldListingType = $auction->get('ListingType');
    my $oldCategoryID  = $auction->get('EbayCategoryId') // 0;

    $self->SUPER::Save($Servlet);


    # get info from the category if we have to display the Catalog asignment tab
    # do this only if we save the category page
    my $CategoryID = $auction->get('EbayCategoryId');

    if ( $Form->value('EbayCategoryId') and defined $CategoryID and ($CategoryID > 0) and $oldCategoryID ne $CategoryID ) {
        my $ShopEnv = $auction->get('EbayShopEnvironment');
        my $RemoteEbay = DE_EPAGES::Ebay::API::Object::RemoteRESTEbay->new(
            'ShopEnvironment' => $ShopEnv,
        );

        eval { ## no critic (RequireExistsError)
            my $hCategoryValues = {'ProductRequiredEnabled'=> 'Disabled'};
            $hCategoryValues = $RemoteEbay->getCategoryFeatures($CategoryID,[qw(ProductRequiredEnabled)],0,1,
                                                                {'Fields' => ['ProductRequiredEnabled']});

            # test if we have to map a product to an ebay catalog and save this on the auchtion
            my $EbayCatalogMappingRequired = ($hCategoryValues->{'ProductRequiredEnabled'} eq 'Enabled') ? 1 : 0;
            $auction->set({'EbayCatalogMappingRequired'      => $EbayCatalogMappingRequired,
                           'EbayCatalogBrandMappingRequired' => 0,
                           'EbayCatalogProductID'      => undef,
                           'EbayCatalogSearchGtin'     => undef,
                           'EbayCatalogUseStockImages' => 0,
                           'EbayCatalogAddShopProductDescription' => 0

                          });
        };
        if (ExistsError()) {
            $self->_handleEbayError($Servlet);
        }
    } elsif(not defined $CategoryID){
        # reset mapping values if we have no categories saved
        $auction->set({'EbayCatalogMappingRequired'      => 0,
                       'EbayCatalogBrandMappingRequired' => 0,
                       'EbayCatalogProductID'      => undef,
                       'EbayCatalogSearchGtin'     => undef,
                       'EbayCatalogUseStockImages' => 0,
                       'EbayCatalogAddShopProductDescription' => 0
                      });
    }

    my $Shop = $auction->getSite;
    my $ProtocolServerSSLName = $Shop->get('ProtocolAndServerSSL');

    # load and process default template from system for Unity shops - UNITY-2477
    if ($Shop->get('IsUnityRegistered') && $auction->alias ne 'EbayDefaultAuction') {
        my $DefaultTemplate = LoadObjectByPath(UNITY_DEFAULT_TEMPLATE);
        my $TemplateName = $DefaultTemplate->alias.'.html';
        my $ThemePath = UNITY_TEMPLATE_PATH . '/' . $TemplateName;
        my $hAuctionInfo = $auction->get([qw( Product TextBelowUnityTheme Gallery)]);
        my $Product = $hAuctionInfo->{'Product'};
        my $ProductWebPath = $Product->get('WebPath');
        my $hVars = { 'Product'             => $Product->tleHash,
                      'TextBelowUnityTheme' => $hAuctionInfo->{'TextBelowUnityTheme'},
                      'ImagePath'           => $ProductWebPath };
        my $ThemeContent = ProcessFileTemplate($ThemePath, $hVars);

        $auction->set({'Description' => $ThemeContent});

    }

    $auction->set({'Status' => AUCTION_EDITABLE});

    if ($self->_checkAuctionDescription($auction)) {
        $Servlet->form->executeFormError({
                'Name'       => 'Description',
                'Reason'     => 'InvalidAuctionDescription',,
                'ViewAction' => 'MBO-ViewEbayWizardTitleStep',
            }
        );
    }

    return unless $oldShopEnv;

    my $newShopEnv = $auction->get('EbayShopEnvironment');
    my $newListingType = $auction->get('ListingType');
    my $params = {};

    my $hValues = $Servlet->form->form($auction, 'AttributeSets');
    my $aLoop = $hValues->{'AttributeSet'};

    if (scalar(@$aLoop)) {
        # delete all saved attributes for this auction
        my $ahAttributeKeys = GetPKeyEbayAuctionAttributesByAuctionID($auction->id);
        DeleteEbayAuctionAttribute($_) foreach @$ahAttributeKeys ;

        my @NamedAttributes;

        while ( scalar @$aLoop ) {
            my $hLoop = shift @$aLoop;
            if ($hLoop->{'EbayAttributeSetID'} =~ /^\d+$/ && $hLoop->{'EbayAttributeID'} =~ /^-?\d+$/ && $hLoop->{'EbayValueID'}) {
                $hLoop->{'AuctionID'} = $auction->id;
                InsertEbayAuctionAttribute($hLoop);
            } elsif ($hLoop->{'EbayAttributeID'} =~ /\D/) {
                next if (!defined($hLoop->{'EbayValueLiteral'}) && (!defined($hLoop->{'EbayValueID'}) || $hLoop->{'EbayValueID'} eq EBAY_ATTRIBUTESETS_VALUE_UNDEF ));

                # If the select box value ( EbayValueID ) differs from EBAY_ATTRIBUTESETS_VALUE_OTHER and EBAY_ATTRIBUTESETS_VALUE_UNDEF
                # then ignore the value of the hidden text field ( EbayValueLiteral )! EPG-19401
                push @NamedAttributes, {
                    'Name'    => $hLoop->{'EbayAttributeID'},
                    'Value'   => $hLoop->{'EbayValueID'} eq EBAY_ATTRIBUTESETS_VALUE_OTHER ? $hLoop->{'EbayValueLiteral'} : $hLoop->{'EbayValueID'},
                    'IsOther' => $hLoop->{'EbayValueID'} eq EBAY_ATTRIBUTESETS_VALUE_OTHER };
            }
        }

        $auction->setItemSpecifics(\@NamedAttributes);
    }

    # if EbayEnvironment was changed delete all Shippings
    if ($newShopEnv->get('EbayEnvironment')->id != $oldShopEnv->get('EbayEnvironment')->id) {
        $auction->folder('AuctionShippingMethods')->deleteChildren;
        $auction->folder('AuctionPaymentMethods')->deleteChildren;
        $auction->folder('AuctionShippingLocations')->deleteChildren;

        #save prices again (since new env can cause new rounding rules)
        $params->{'StartPrice'} = $auction->get('StartPrice');
        $params->{'BuyItNowPrice'} = $auction->get('BuyItNowPrice');

        # delete EbayStoreCategoryId, EbayStoreCategory2Id
        $params->{'EbayStoreCategoryId'} = undef;
        $params->{'EbayStoreCategory2Id'} = undef;

        # reset "Get it Fast
        $params->{'IsGetItFast'} = 0;
        $params->{'ConditionID'} = undef;
        $params->{'IsNew'} = 0;
        $params->{'AcceptsReturnPolicy'} = 0;

        # reset category and second category
        $params->{'EbayCategoryId'} = 0;
        $params->{'SecondCategoryId'} = 0;
    }

    # if ListingType was changed
    if ($newListingType ne $oldListingType) {
        if ($newListingType eq 'FixedPriceItem') {
            $params->{'IsBuyItNowEnabled'} = 0;
            # ListingType 'FixedPriceItem' only has 'GoodUntilCanceled' option, which is the default option
            # Duration value fr 'GoodUntilCanceled' is 2147483647
            $params->{'Duration'} = 2147483647; ##no critic qw(ProhibitMagicNumbers)
        } elsif ($newListingType eq 'Chinese') {
            $params->{'Quantity'} = 1;
            $params->{'IsReviseInventory'} = 0;
            # ListingType 'Chinese' has default duration of 1
            $params->{'Duration'} = 1;
        }
    }

    # set qty to 1 if is default auction, a fixprice auction and stock level sync activated
    # then the auction must use the stocklevel from the product if defined otherwise 1
    if  ( ($newListingType eq 'FixedPriceItem') and ($auction->get('IsReviseInventory') == 1) ){
        if($auction->alias eq 'EbayDefaultAuction'){
            $params->{'Quantity'} = 1;
        } else{
            # correct the quantity if we have no stock level informations on the product
            my $AuctionProduct = $auction->get('Product');
            $params->{'Quantity'} = defined $AuctionProduct->get('StockLevel') ? $AuctionProduct->get('StockLevel') : 1;
        }
    }

    # set pictureURLs for all subauctions
    my $picVariationType = $auction->get('PicVariationType') // '';

    my $imageType;

    if  ($picVariationType eq 'ListView') {
        $imageType = 'ImageSmall';
    } elsif ($picVariationType eq 'DetailView') {
        $imageType = 'ImageMedium';
    } elsif ($picVariationType eq 'ZoomView') {
        $imageType = 'ImageLarge';
    } elsif ($picVariationType eq 'HotDealView') {
        $imageType = 'ImageHotDeal';
    }

    foreach my $SubAuction ( @{$auction->get('SubAuctions')} ) {
        my $image;
        $image = $SubAuction->get('Product')->getWithParent($imageType) if $imageType;
        if (not defined $image) {
            $SubAuction->set({'PictureUrl' => undef});
        } elsif ($image =~ /\/\//) { #contains //hostname
            $SubAuction->set({'PictureUrl' => $image});
        } else {
            $SubAuction->set({'PictureUrl' => $ProtocolServerSSLName . $SubAuction->get('Product')->get('WebPath') . '/' . $image});
        }
    }


    if ($auction->get('ScheduleTime')) {
        $params->{'StartTime'} = $auction->get('ScheduleTime');
    } elsif ($auction->get('ScheduleType') == 1) {
        $Servlet->form->addFormError({
                'Name'      => 'EbayScheduleNoStartTime',
                'Reason'    => 'EbayScheduleNoStartTime',
            });
        $Servlet->form->executeFormError;
    }

    $auction->set($params);

    # test if we have to go to the assignment step
    if($Form->value('TestAssignment') and ($auction->get('EbayCatalogMappingRequired') or $auction->get('EbayCatalogBrandMappingRequired'))){
        my $ViewActionName = scalar(@{$auction->get('SubAuctions')}) ? 'MBO-ViewEbayWizardVariationAssignmentStep' : 'MBO-ViewEbayWizardAssignmentStep';
        $Servlet->vars('ViewAction', $ViewActionName);
    }

}

sub VerifyAuction {
    my $self = shift;
    my ($Servlet) = @_;

    my $auction = $Servlet->object;
    # do not verify again, if auction if running or ended
    return if $auction->get('Status') & (AUCTION_RUNNING | AUCTION_ENDED);

    my $result;
    eval {
        $result = $auction->verify;

        # check if store owner
        my $ShopEnv = $auction->get('EbayShopEnvironment');
        my $RemoteEbay = DE_EPAGES::Ebay::API::Object::RemoteRESTEbay->new(
            'ShopEnvironment' => $ShopEnv,
        );

        my $hStoreOwner = $RemoteEbay->getUser(['StoreOwner']);
        my $StoreOwner = $hStoreOwner ? $hStoreOwner->{'StoreOwner'} : undef;
        $Servlet->vars('IsEbayStoreOwner', 1) if (defined $StoreOwner) and $StoreOwner eq 'true';

    };

    if (ExistsError()) {
        my $Error = GetError();
        $auction->set({'Status' => AUCTION_EDITABLE});

        if ($Error->code eq 'EbayRequestFailed') {
            $Servlet->form->addFormError({
                'Name'      => 'EbayRequestFailed',
                'Message'   => $Error->{'Vars'}->{'SOAPMessage'},
                'Reason'    => 'EbayRequestFailed',
            });
        } elsif ($Error->code eq 'EbayRequestNotSuccessful') {
            foreach my $error (@{$Error->{'Vars'}->{'EbayErrors'}}) {
                $Servlet->form->addFormError({
                    'Name'      => 'EbayRequestNotSuccessful',
                    'Message'   => $error->{'LongMessage'},
                    'Reason'    => 'EbayRequestNotSuccessful',
                    'Code'      => $error->{'ErrorCode'},
                    'MoreInfo'  => $error->{'Message'},
                });
                $Servlet->vars('PayPalError', 1) if $error->{'ErrorCode'} == EBAY_ERROR_WRONG_PAYPAL_ADDRESS;
            }
        } elsif ($Error->code eq 'EbayFeatureLimitExceeded') {
            $Servlet->form->addFormError({
                'Name'      => 'EbayFeatureLimitExceeded',
                'Reason'    => 'EbayFeatureLimitExceeded',
            });
        } elsif ($Error->code eq 'EbayNoToken') {
            $Servlet->vars('EbayWarningNoToken', 1);
        } else {
            Error($Error);
        }
    } else {

        # test if we have warnings and add they to the form
        if (exists $result->{'Warnings'}) {
            foreach my $warning (@{$result->{'Warnings'}}) {
                $Servlet->form->addFormError({
                    'Name'      => 'EbayRequestSuccessfulWithWarnings',
                    'Message'   => $warning->{'LongMessage'},
                    'Reason'    => 'EbayRequestSuccessfulWithWarnings',
                    'Code'      => $warning->{'ErrorCode'},
                    'MoreInfo'  => $warning->{'Message'},
                });
            }
        }

        my $Fees = $result->{'Fee'};
        @$Fees = grep {($_->{'Fee'} ne '0.0')} @$Fees;
        $Servlet->params('Fees', $Fees);

        my ($ListingFee) = grep {($_->{'Name'} eq 'ListingFee')} @$Fees;
        $Servlet->params('ListingFee', $ListingFee->{'Fee'});

        # set status
        if ($auction->get('ScheduleType') == 0) {
            $auction->set({'Status' => AUCTION_VERIFIED});
        } else {
            $auction->set({'Status' => AUCTION_SCHEDULED});
        }
    }

    $Servlet->form->executeFormError;
}

sub UploadAuction {
    my $self = shift;
    my ($Servlet) = @_;

    my $auction = $Servlet->object;

    # do upload only if auction is verified or scheduled
    return unless $auction->get('Status') & (AUCTION_VERIFIED | AUCTION_SCHEDULED);

    my $result;
    eval {
        $result = $auction->upload;
    };

    if (ExistsError()) {
        my $Error = GetError();
        $auction->set({'Status' => AUCTION_EDITABLE});

        if ($Error->code eq 'EbayRequestFailed') {
            $Servlet->form->addFormError({
                'Name'      => 'EbayRequestFailed',
                'Message'   => $Error->{'Vars'}->{'SOAPMessage'},
                'Reason'    => 'EbayRequestFailed',
            });
        } elsif ($Error->code eq 'EbayRequestNotSuccessful') {
            foreach my $error (@{$Error->{'Vars'}->{'EbayErrors'}}) {
                $Servlet->form->addFormError({
                    'Name'      => 'EbayRequestNotSuccessful',
                    'Message'   => $error->{'LongMessage'},
                    'Reason'    => 'EbayRequestNotSuccessful',
                    'Code'      => $error->{'ErrorCode'},
                });
            }
        } elsif ($Error->code eq 'EbayFeatureLimitExceeded') {
            $Servlet->form->addFormError({
                'Name'      => 'EbayFeatureLimitExceeded',
                'Reason'    => 'EbayFeatureLimitExceeded',
            });
        } else {
            Error($Error);
        }
    } else {

        # test if we have warnings and add they to the form
        if (exists $result->{'Warnings'}) {
            foreach my $warning (@{$result->{'Warnings'}}) {
                $Servlet->form->addFormError({
                    'Name'      => 'EbayRequestSuccessfulWithWarnings',
                    'Message'   => $warning->{'LongMessage'},
                    'Reason'    => 'EbayRequestSuccessfulWithWarnings',
                    'Code'      => $warning->{'ErrorCode'},
                    'MoreInfo'  => $warning->{'Message'},
                });
            }
        }

        my $Fees = $result->{'Fee'};
        my $itemId = $result->{'ItemID'};
        @$Fees = grep {($_->{'Fee'} ne '0.0')} @$Fees;
        $Servlet->params('Fees', $Fees);
        $Servlet->params('ItemID', $itemId);

        my ($ListingFee) = grep {($_->{'Name'} eq 'ListingFee')} @$Fees;
        $Servlet->params('ListingFee', $ListingFee->{'Fee'});
    }

    $Servlet->form->executeFormError;
}

sub NewAuctionShippingMethod {
    my $self = shift;
    my ($Servlet) = @_;

    my $auction = $Servlet->object;
    my $Shop = $auction->getSite();

    # validate input data and get values
    my $Form = $Servlet->form->form( $auction, 'UpdateShippingMethodsForm');

    my $renumberChildren = 0;
    foreach my $Entry (@{ $Form->{'ShippingMethods'}}) {
        my $Object = $Shop->loadObject($Entry->{'UpdateAuctionShippingMethodID'}, 'EbayAuctionShippingMethod');
        $Servlet->testPermission( $Object, 'Save' );
        next if (defined($Entry->{'OldPosition'}) && $Entry->{'UpdatePosition'} == $Entry->{'OldPosition'});
        $Object->set({'Position' => $Entry->{'UpdatePosition'}});
        $renumberChildren++;
    }

    $auction->folder('AuctionShippingMethods')->renumberChildren if ($renumberChildren);

    # new shipping method.
    return unless (
        $Servlet->form->value('EbayShippingMethodID') or
        defined ($Servlet->form->value('Price')) or
        defined ($Servlet->form->value('AdditionalPrice'))
    );

    # validate input data
    $Form = $Servlet->form->form( $auction, 'NewShippingMethodForm' );

    my $env = $auction->get('EbayShopEnvironment')->get('EbayEnvironment') ;
    my $Formatter = $auction->getSite()->formatter($env->get('CurrencyID'));
    my $Price = $Formatter->round( $Form->{'Price'} );
    my $AdditionalPrice;
    $AdditionalPrice = $Formatter->round( $Form->{'AdditionalPrice'} ) if (defined $Form->{'AdditionalPrice'});
    my $EbayShippingMethodID    = $Form->{'EbayShippingMethodID'};

    my $Folder = $auction->folder('AuctionShippingMethods');

    # test if the shipping method already exists
    my %SelectedIDs = map { $_->get('EbayShippingMethod')->id => 1 } @{ $Folder->children };
    if( exists $SelectedIDs{$EbayShippingMethodID} ) {
        $Servlet->vars( 'OBJECT_EXISTS', 1 );
        $Servlet->form->addFormError( {
            Name => 'EbayShippingMethodID',
            Reason => 'DuplicateShippingMethod'
        });
    }

    $Servlet->form->executeFormError;
    my $System = LoadRootObject();

    my $EbayShippingMethod = $System->loadObject( $EbayShippingMethodID, 'EbayShippingMethod' );
    die 'invalid ebay shippingmethod:'.$EbayShippingMethod->id.' used in Shop:'.$Shop->id if not $EbayShippingMethod->instanceOf('EbayShippingMethod');

    my $newAuctionSM = $Folder->insertChild('EbayAuctionShippingMethod',{
        'Alias'                 => $Folder->nextAlias('EbayAuctionShippingMethod%s'),
        'EbayShippingMethod'    => $EbayShippingMethod,
        'Price'                 => $Price,
        'AdditionalPrice'       => $AdditionalPrice,
        'EbayAuction'           => $auction,
        'Position'              => $Form->{'Position'}
    });

    $newAuctionSM->parent->renumberChildren;

    $auction->set({'Status' => AUCTION_EDITABLE});
    $Servlet->vars('NewShippingMethodSuccess',1);
    return;
}

sub NewAuctionPaymentMethod {
    my $self = shift;
    my ($Servlet) = @_;

    my $auction = $Servlet->object;

    my $Form = $Servlet->form->form($auction,'NewPaymentForm');

    my $EbayPaymentMethodID = $Form->{'EbayPaymentMethodID'};
    my $CODCost             = $Form->{'CODCost'};

    my $Folder = $auction->folder('AuctionPaymentMethods');

    # test if the payment method already exists
    my %SelectedIDs = map { $_->get('EbayPaymentMethod')->id => 1 } @{ $Folder->children };
    if( exists $SelectedIDs{$EbayPaymentMethodID} ) {
        $Servlet->vars( 'OBJECT_EXISTS', 1 );
        $Servlet->form->executeFormError( {
            Name => 'EbayPaymentMethodID',
            Reason => 'DuplicatePaymentMethod'
        });
    }

    my $System = LoadRootObject();
    my $EbayPaymentMethod = $System->loadObject( $EbayPaymentMethodID, 'EbayPaymentMethod' );

    # CODCost is mandatory for COD on ebay.it
    if ($EbayPaymentMethod->alias eq 'COD' && (!defined($CODCost)) && $auction->get('EbayShopEnvironment')->get('EbayEnvironment')->alias eq 'IT') {
        $Servlet->form->executeFormError({
            Name   => 'CODCost',
            Reason => 'CODCostMissing'
        });
    } else {
        if (defined $CODCost) {
            my $env = $auction->get('EbayShopEnvironment')->get('EbayEnvironment') ;
            my $Formatter = $auction->getSite()->formatter($env->get('CurrencyID'));
            $CODCost = $Formatter->round($CODCost);
        }
    }

    $Folder->insertChild('EbayAuctionPaymentMethod',{
        'Alias'                 => $Folder->nextAlias('EbayAuctionPaymentMethod%s'),
        'EbayPaymentMethod'     => $EbayPaymentMethod,
        'EbayAuction'           => $auction,
        'CODCost'               => $CODCost,
    });

    $auction->set({'Status' => AUCTION_EDITABLE});

    return;
}

sub NewAuctionShippingLocation {
    my $self = shift;
    my ($Servlet) = @_;

    my $auction = $Servlet->object;
    my $Shop = $auction->getSite();

    # validate input data
    my $Form = $Servlet->form->form( $auction, 'UpdateShippingLocationsForm');

    my $renumberChildren = 0;
    foreach my $Entry (@{ $Form->{'ShippingLocations'}}) {
        next unless ExistsObject($Entry->{'UpdateAuctionShippingLocationID'});
        my $Object = $Shop->loadObject($Entry->{'UpdateAuctionShippingLocationID'}, 'EbayAuctionShippingLocation');
        $Servlet->testPermission( $Object, 'Save' );
        next if (defined($Entry->{'OldPosition'}) && $Entry->{'UpdatePosition'} == $Entry->{'OldPosition'});
        $Object->set({'Position' => $Entry->{'UpdatePosition'}});
        $renumberChildren++;
    }

    $auction->folder('AuctionShippingLocations')->renumberChildren if ($renumberChildren);

    return unless ($Servlet->form->value('EbayShippingLocationID')
        or $Servlet->form->value('InternationalShippingMethodID')
        or defined($Servlet->form->value('InternationalPrice'))
        or defined($Servlet->form->value('InternationalAdditionalPrice')));

    # validate input data
    $Form = $Servlet->form->form( $auction, 'NewShippingLocationForm');

    my $env = $auction->get('EbayShopEnvironment')->get('EbayEnvironment') ;
    my $Formatter = $Shop->formatter($env->get('CurrencyID'));
    my $InternationalPrice = $Formatter->round( $Form->{'InternationalPrice'} );
    my $InternationalAdditionalPrice;
    $InternationalAdditionalPrice = $Formatter->round( $Form->{'InternationalAdditionalPrice'} ) if (defined $Form->{'InternationalAdditionalPrice'});
    my $EbayShippingLocationID        = $Form->{'EbayShippingLocationID'};
    my $InternationalShippingMethodID = $Form->{'InternationalShippingMethodID'};

    my $Folder = $auction->folder('AuctionShippingLocations');

    # test if shipping location with selected shipping service already exists
    my %SelectedIDs = map { $_->get('EbayShippingLocation')->id.$_->get('EbayShippingMethod')->id => 1 } @{ $Folder->children };
    my $ShippingLocationIDShippingMethodID = $EbayShippingLocationID.$InternationalShippingMethodID;
    if( exists $SelectedIDs{$ShippingLocationIDShippingMethodID} ) {
        $Servlet->vars( 'OBJECT_EXISTS', 1 );
        $Servlet->form->addFormError( {
            'Name'   => 'InternationalShippingMethodID',
            'Reason' => 'DuplicateShippingLocationAndMethod'
        });
        $Servlet->form->addFormError( {
            'Name'   => 'EbayShippingLocationID',
            'Reason' => 'DuplicateShippingLocationAndMethod'
        });
        $Servlet->form->executeFormError;
    }

    my $System = LoadRootObject();

    my $newAuctionShippingLocation = $Folder->insertChild('EbayAuctionShippingLocation',{
        'Alias'                 => $Folder->nextAlias('EbayAuctionShippingLocation%s'),
        'EbayShippingLocation'  => $System->loadObject( $EbayShippingLocationID, 'EbayShippingLocation' ),
        'EbayAuction'           => $auction,
        'EbayShippingMethod'    => $System->loadObject( $InternationalShippingMethodID, 'EbayShippingMethod' ),
        'Price'                 => $InternationalPrice,
        'AdditionalPrice'       => $InternationalAdditionalPrice,
        'Position'              => $Form->{'Position'}
    });

    $newAuctionShippingLocation->parent->renumberChildren;

    $auction->set({'Status' => AUCTION_EDITABLE});
    $Servlet->vars('NewShippingLocationSuccess',1);
    return;
}

sub _getItemSpecificsWithoutVariationAttr{
    my ($auction, $itemSpecifics) = @_;

    my @CategoryAspects = @{ $itemSpecifics->{'aspects'} };
    # Found the languageID of the selected ebay Environment
    my $LanguageID = GetPKeyLanguageByCode(lc $auction->get('EbayShopEnvironment')->get('EbayEnvironment')->get('LanguageID'));
    # Get all Variation Attributes
    my @aVariationAttributes = map { $_->{'Attribute'}->get('NameOrAlias',$LanguageID) } @{$auction->get('Product')->get('VariationTypeValues')};
    my %hVariationAttributes = map { $_ => undef } @aVariationAttributes;
    # Filter out all Variation Attributes
    @CategoryAspects = grep { !exists $hVariationAttributes{$_->{'localizedAspectName'}} } @CategoryAspects;
    $itemSpecifics->{'aspects'} = \@CategoryAspects;

    return $itemSpecifics;


}

sub UnityViewTitleStep {
    my $self = shift;
    my ($Servlet) = @_;

    my $auction = $Servlet->object;

    my $ImagePath;

    # load sample product in case of the default auction
    if ($Servlet->params->{'IsDefaultAuction'}) {
        my $DefaultTemplate = LoadObjectByPath(UNITY_DEFAULT_TEMPLATE);
        my $SampleProduct = $DefaultTemplate->child('SampleProduct');
        $Servlet->vars('Product', $SampleProduct->tleHash);
        $ImagePath = '/WebRoot';
    } else {
        $ImagePath = $auction->get('Product')->get('WebPath');
    }

    $Servlet->vars('ImagePath', $ImagePath);

    return $self->View($Servlet);
}

sub _TranslateDefaultAuctionPageTitle {
    shift;
    my ($Servlet) = @_;

    # set page title for default auction
    if ($Servlet->params('IsDefaultAuction')) {
        $Servlet->params('Title', Translate('DefaultAuction', undef, 'DE_EPAGES::Ebay'));
    }

    return;
}

sub SaveCatalogAssignment {
    my $self = shift;
    my ($Servlet) = @_;

    my $EbayAuction = $Servlet->object;
    my $SaveUPCEAN = $Servlet->form->formValue('SaveUPCEAN');
    my $UPCEAN     = $Servlet->form->formValue('UPCEAN');
    my $EPID       = $UPCEAN ? $Servlet->form->formValue('epid') : undef;
    my $EbayCatalogUseStockImages            = $Servlet->form->formValue('EbayCatalogUseStockImages');
    my $EbayCatalogAddShopProductDescription = $Servlet->form->formValue('EbayCatalogAddShopProductDescription');

    # save epid and UPCEAN on auction
    $EbayAuction->set({'EbayCatalogProductID'                 => $EPID,
                       'EbayCatalogSearchGtin'                => $UPCEAN,
                       'EbayCatalogUseStockImages'            => $EbayCatalogUseStockImages,
                       'EbayCatalogAddShopProductDescription' => $EbayCatalogAddShopProductDescription
                      });

    # save UPCEAN on Product if needed.
    my $Product = $EbayAuction->get('Product');
    $Product->set({'UPCEAN' => $UPCEAN}) if $SaveUPCEAN and defined $Product;

    $Servlet->params('SaveUPCEAN', $SaveUPCEAN ? $SaveUPCEAN : 0);

}

1;
