#========================================================================================
# package      DE_EPAGES::MultiStoreConnector::API::MSSynchronization
# state        public
# description  abstract synchronization class
#========================================================================================
package DE_EPAGES::MultiStoreConnector::API::MSSynchronization;

use strict;
use DE_EPAGES::WebInterface::API::MessageCenter qw ( SynchronizeCache );
use DE_EPAGES::DataCache::API::Cache qw (UpdateCache);
use DE_EPAGES::Core::API::Error qw ( GetError Error ExistsError);
use DE_EPAGES::Database::API::Connection qw (GetCurrentDBHandle);
use DE_EPAGES::Core::API::Warning qw( Warning );
use DE_EPAGES::Object::API::Factory qw (LoadObject LoadObjectByPath);
use DE_EPAGES::Presentation::API::PageCache qw (UpdatePageCache);
use DE_EPAGES::Database::API::DAL;
use Scalar::Util qw (looks_like_number);
use JSON::XS;

# We define a few protected methods which will be used in the specializations
## no critic (ProhibitUnusedPrivateSubroutines)

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Synchronization = DE_EPAGES::MultiStoreConnector::API::MSSynchronization->new(%options);
#----------------------------------------------------------------------------------------
# description  Creates a new synchronization object of DE_EPAGES::MultiStoreConnector::API::MSSynchronization.
#----------------------------------------------------------------------------------------
# input        %options | options for synchronization
#               <ul>
#                 <li>ROADWORKS - ROADWORKS
#               </ul>
#               | hash
# return       $Synchronization | synchronization object | object
#========================================================================================
sub new {
    my ($class, %options) = @_;

    #get db handle object
    my $dbh = GetCurrentDBHandle();

    #return object
    my $self =  bless {
        %options,
    }, $class;

    $self->{'AttributeParameters'} = {};     # is overwritten in sub modules
    $self->{'AttributeMandatory'} = {};    # is overwritten in sub modules

    $self->{'ParameterVerification'} = 1;
    $self->{'ShopCache'} = {};
    $self->{'DataCache'}->{'WeightUnitIDSQL'} = {};
    $self->{'DataCache'}->{'TaxClassAlias'} = {};
    $self->{'DBHandle'} = $dbh;
    $self->{ParameterCache}->{'Bool'} = {};
    $self->{ParameterCache}->{'Int'} = {};
    $self->{ParameterCache}->{'Number'} = {};
    $self->{ParameterCache}->{'String'} = {};


    return $self;
}

#========================================================================================
# function     dbh
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DBH = $HardCoreSynchronization->dbh();
# example      $self->dbh()->prepare('...');
#----------------------------------------------------------------------------------------
# description  return the SQLite database handler object
#----------------------------------------------------------------------------------------
# return       $DBH | SQL database handler | object
#========================================================================================
sub dbh{
    my $self = shift;

    return $self->dbi->dbh();
}

#========================================================================================
# function     dbi
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DBH = $HardCoreSynchronization->dbi();
# example      $self->dbi()->prepare('...');
#----------------------------------------------------------------------------------------
# description  return the SQLite database handler object
#----------------------------------------------------------------------------------------
# return       $DBI | SQL database interface | object
#========================================================================================
sub dbi{
    my $self = shift;

    return $self->{'DBHandle'};
}

#========================================================================================
# function     synchronizeCache
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = $Synchronization->synchronizeCache();
#----------------------------------------------------------------------------------------
# description  synchronize the cache for epages installation
#========================================================================================
sub synchronizeCache {
    my $self = shift;

    eval { SynchronizeCache(); };
    Warning( GetError() ) if ExistsError();
}

#========================================================================================
# function     resetObjectCache
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = $Synchronization->resetObjectCache();
#----------------------------------------------------------------------------------------
# description  removes complete cache for a object
#----------------------------------------------------------------------------------------
# input        $ObjectID | object id | integer
#========================================================================================
sub resetObjectCache {
    my $self = shift;
    my ($ObjectID) = @_;

    #check parameter
    Error('MissingParameter',{'Parameter' => 'ObjectID'}) unless ($ObjectID);

    #remove the object cache
    #possible additions see in DataCache.conf
    UpdateCache($self->dbi->cacheKey, 'Object', $ObjectID);
    UpdateCache($self->dbi->cacheKey, 'InfoObject', $ObjectID);
    UpdateCache($self->dbi->cacheKey, 'InfoContentObject', $ObjectID);
    UpdateCache($self->dbi->cacheKey, 'ObjectByParentID', $ObjectID);
}

#========================================================================================
# function     resetObjectDALCache
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = $Synchronization->resetObjectDALCache($ObjectID, $IsObjectDeleted);
#----------------------------------------------------------------------------------------
# description  removes DAL cache for a object
#----------------------------------------------------------------------------------------
# input        $ObjectID | object id | integer
# input        $IsObjectDeleted | 1 if object is deleted, 0 if object was updated or added | bool
#========================================================================================
sub resetObjectDALCache {
    my $self = shift;
    my ($ObjectID, $IsObjectDeleted) = @_;

    #check parameter
    Error('MissingParameter',{'Parameter' => 'ObjectID'}) unless ($ObjectID);

    #remove DAL Cache
    my $DAL_PACKAGE = DE_EPAGES::Database::API::DAL->load('DE_EPAGES::Object::API::Object::Object');
    $DAL_PACKAGE->cacheUpdate($ObjectID, undef, $IsObjectDeleted);
}

#========================================================================================
# function     resetObjectGUIDCache
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = $Synchronization->resetObjectGUIDCache($GUID);
#----------------------------------------------------------------------------------------
# description  removes complete cache for a object by accessing with guid
#----------------------------------------------------------------------------------------
# input        $GUID | object guid | string
#========================================================================================
sub resetObjectGUIDCache {
    my $self = shift;
    my ($GUID) = @_;

    #check parameter
    Error('MissingParameter',{'Parameter' => 'GUID'}) unless ($GUID);

    #remove the object cache
    #possible additions see in DataCache.conf
    UpdateCache($self->dbi->cacheKey, 'ObjectByGUID', $GUID);
}

#========================================================================================
# function     resetContentObjectCache
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = $Synchronization->resetContentObjectCache($ShopID);
#----------------------------------------------------------------------------------------
# description  removes complete cache for a object
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
#========================================================================================
sub resetShopContentObjectCache {
    my $self = shift;
    my ($ShopID) = @_;

    #check parameter
    Error('MissingParameter',{'Parameter' => 'ShopID'}) unless ($ShopID);

    #remove the content object cache
    UpdateCache($self->dbi->cacheKey, 'ContentObjectBySiteID', $ShopID);
}

#========================================================================================
# function     resetGlobalContentObjectCache
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = $Synchronization->resetGlobalContentObjectCache();
#----------------------------------------------------------------------------------------
# description  removes complete cache for all content objects
#----------------------------------------------------------------------------------------
# return       $Success | returns true if everything is correct | boolean
#========================================================================================
sub resetGlobalContentObjectCache {
    my $self = shift;

    #remove the content object cache
    UpdateCache($self->dbi->cacheKey, 'InfoContentObject');
}

#========================================================================================
# function     resetAttributeCache
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = $Synchronization->resetAttributeCache($AttributeClassID);
#----------------------------------------------------------------------------------------
# description  resets the cache of attributes for the given attribute class
#----------------------------------------------------------------------------------------
# input        $AttributeClassID | ID of the attribute's class | integer
#========================================================================================
sub resetAttributeCache {
    my $self = shift;
    my ($AttributeClassID) = @_;

    Error('MissingParameter',{'Parameter' => 'AttributeClassID'}) unless ($AttributeClassID);

    UpdateCache($self->dbi->cacheKey, 'AttributeByAttributeClassID', $AttributeClassID);
}

#========================================================================================
# function     updatePageCache
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = $Synchronization->updatePageCache($Object);
#----------------------------------------------------------------------------------------
# description  updates page cache for a object
#----------------------------------------------------------------------------------------
# input        $Object | epages 6 object | object
# return       $Success | returns true if everything is correct | boolean
#========================================================================================
sub updatePageCache {
    my $self = shift;
    my ($ObjectID) = @_;

    #check parameter
    Error('MissingParameter',{'Parameter' => 'ObjectID'}) unless ($ObjectID);

    #update page cache
    UpdatePageCache(LoadObject($ObjectID));
}

#========================================================================================
# function     clearPageCache
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = $Synchronization->clearPageCache($Shop,$Object);
#----------------------------------------------------------------------------------------
# description  clear page cache for a object in a shop
#----------------------------------------------------------------------------------------
# input        $Shop | shop name | string
# input        $Object | epages 6 object | object
#========================================================================================
sub clearPageCache {
    my $self = shift;
    my ($Shop,$Object) = @_;

    my $SitePageCache = DE_EPAGES::Object::API::PageCache->new( Site => $Shop );
    $SitePageCache->clearObjectCache($Object);
}

#========================================================================================
# function     clearSiteCache
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = $Synchronization->clearSiteCache($Shop);
#----------------------------------------------------------------------------------------
# description  clear site cache for a shop
#----------------------------------------------------------------------------------------
# input        $Shop | shop name | string
#========================================================================================
sub clearSiteCache {
    my $self = shift;
    my ($Shop) = @_;

    DE_EPAGES::Object::API::PageCache->new( Site => $Shop )->clearSiteCache();
}

#========================================================================================
# function     isParamVerificationActivated
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsActivated = $Synchronization->isParamVerificationActivated();
#----------------------------------------------------------------------------------------
# description  returns true if parameter verification is activated for synchronisation
#----------------------------------------------------------------------------------------
# return       $IsActivated | returns true if parameter verification is activated | boolean
#========================================================================================
sub isParamVerificationActivated{
    my $self = shift;

    return $self->{'ParameterVerification'};
}

#========================================================================================
# function     isInteger
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsInteger = $Synchronization->isInteger($Check);
#----------------------------------------------------------------------------------------
# description  returns true if variable is an integer
#----------------------------------------------------------------------------------------
# input        $Check | variable to check | ref/scalar/HASH/ARRAY
# return       $IsInteger | returns true if variable is an integer | boolean
#========================================================================================
sub isInteger {
    my $self = shift;
    my ($Check) = @_;

    #check variable
    return 0 if ($self->isString($Check));
    return ($Check =~ /^-?\d+$/);
}

#========================================================================================
# function     isString
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsString = $Synchronization->isString($Check);
#----------------------------------------------------------------------------------------
# description  returns true if variable is a string
#----------------------------------------------------------------------------------------
# input        $Check | variable to check | ref/scalar/HASH/ARRAY
# return       $IsString | returns true if variable is a string | boolean
#========================================================================================
sub isString {
    my $self = shift;
    my ($Check) = @_;

    #check variable
    return 1 if (defined $Check && $Check & ~$Check);
    return 1 if (defined $Check and (length $Check == 0));
    return 0;
}

#========================================================================================
# function     isNumber
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsNumber = $Synchronization->isNumber($Check);
#----------------------------------------------------------------------------------------
# description  returns true if variable is a number
#----------------------------------------------------------------------------------------
# input        $Check | variable to check | ref/scalar/HASH/ARRAY
# return       $IsNumber | returns true if variable is a number | boolean
#========================================================================================
sub isNumber {
    my $self = shift;
    my ($Check) = @_;

    #return 0
    return 0 if ($self->isString($Check));

    #check variable
    return looks_like_number($Check);
}

#========================================================================================
# function     isBool
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsBool = $Synchronization->isBool($Check);
#----------------------------------------------------------------------------------------
# description  returns true if variable is a boolean
#----------------------------------------------------------------------------------------
# input        $Check | variable to check | ref/scalar/HASH/ARRAY
# return       $IsBool | returns true if variable is a boolean | boolean
#========================================================================================
sub isBool {
    my $self = shift;
    my ($Check) = @_;

    return (($Check eq 'false') or ($Check eq 'true') or ($Check eq '0') or ($Check eq '1'));
}

#========================================================================================
# function     isObject
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsObject = $Synchronization->isObject($Check);
#----------------------------------------------------------------------------------------
# description  returns true if variable is a JSON object (hash reference)
#----------------------------------------------------------------------------------------
# input        $Check | variable to check | ref/scalar/HASH/ARRAY
# return       $IsObject | returns true if variable is an object | boolean
#========================================================================================
sub isObject {
    my $self = shift;
    my ($Check) = @_;

    return (ref($Check) eq 'HASH');
}

#========================================================================================
# function     isArray
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsObject = $Synchronization->isArray($Check);
#----------------------------------------------------------------------------------------
# description  returns true if variable is a JSON array (array reference)
#----------------------------------------------------------------------------------------
# input        $Check | variable to check | ref/scalar/HASH/ARRAY
# return       $isArray | returns true if variable is an array | boolean
#========================================================================================
sub isArray {
    my $self = shift;
    my ($Check) = @_;

    return (ref($Check) eq 'ARRAY');
}

#========================================================================================
# function     verifySubshops
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Synchronization->verifySubshops($aMultiStoreSubShops, $StoreName);
#----------------------------------------------------------------------------------------
# description  Checks if the given shop aliases are in the database (store)
#               If at least one alias is invalid, an error is thrown, otherwise the
#               method returns without error.
#----------------------------------------------------------------------------------------
# input        $aMultiStoreSubShops | shop aliases | ref.array.string
# input        $StoreName | store name | string
#========================================================================================
sub verifySubshops {
    my $self = shift;
    my ($aMultiStoreSubShops, $StoreName) = @_;

    foreach my $hShopInfo (@$aMultiStoreSubShops){
        #get info
        my $ShopAlias = $hShopInfo->{'alias'};

        #check if shop exists
        eval{
            LoadObjectByPath('/Shops/' . $ShopAlias);
        };
        if(ExistsError){
            my $MyError = GetError();

            # if subshop not found in store
            if($MyError->code eq 'PATH_NOT_FOUND'){
                Error('ResourceNotFound', {
                    'Reason' => 'empty result set for shop \'' . $ShopAlias . '\'',
                    'GUID' => $ShopAlias,
                    'Class' => 'Shop',
                    'Site' => $StoreName
                });
            }
            else{
                Error($MyError);
            }
        }
    }
}

#========================================================================================
# function     getAttributeParameterType
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ParameterType = $self->getAttributeParameterType($AttributeName);
#----------------------------------------------------------------------------------------
# description  get parameter type of attribute name
#----------------------------------------------------------------------------------------
# input        $AttributeName | name of attribute | string
# return       $ParameterType | parameter type of product attribute | string
#========================================================================================
sub getAttributeParameterType{
    my $self = shift;
    my ($AttributeName) = @_;

    return $self->{'AttributeParameters'}->{$AttributeName};
}

#========================================================================================
# function     getAttributeMandatory
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Mandatory = $self->getAttributeMandatory($AttributeName);
#----------------------------------------------------------------------------------------
# description  get parameter type of attribute name
#----------------------------------------------------------------------------------------
# input        $AttributeName | name of attribute | string
# return       $Mandatory | true ifproduct attribute is mandatory  | bool
#========================================================================================
sub getAttributeMandatory{
    my $self = shift;
    my ($AttributeName) = @_;

    return $self->{'AttributeMandatory'}->{$AttributeName};
}

#========================================================================================
# function     verifyContent
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->verifyContent($hContent);
#----------------------------------------------------------------------------------------
# description  verify attributes parameter, throws error if verification fails
#----------------------------------------------------------------------------------------
# input        $hContent | ref.hash
#========================================================================================
sub verifyContent{
    my $self = shift;
    my ($hContent) = @_;

    #get info
    my $aAttributeNames = $self->getAttributeNames();
    my $Alias = $hContent->{'alias'};

    #for each attribute
    foreach my $AttributeName (@$aAttributeNames) {

        #get info
        my $ParameterType = $self->getAttributeParameterType($AttributeName);
        my $IsMandatory = $self->getAttributeMandatory($AttributeName);
        my $AttributeValue = $hContent->{$AttributeName};
        my $RaiseError;

        #check parameter type
        my $IsEmptyString = (($ParameterType eq 'string') and ($AttributeValue eq ''));

        if ($IsMandatory and (not(defined $AttributeValue) or $IsEmptyString)) {
            Error('MissingMandatoryAttribute', {
                'Message' => 'mandatory attribute not exists',
                'Attribute' => $AttributeName,
                'Object' => $Alias,
            })
        }
        elsif ((not defined $AttributeValue) and (not $IsMandatory)){$RaiseError = 0}
        elsif($ParameterType eq 'string') { $RaiseError = 1 unless ($self->isString($AttributeValue))}
        elsif($ParameterType eq 'integer') { $RaiseError = 1 unless ($self->isInteger($AttributeValue))}
        elsif($ParameterType eq 'number') { $RaiseError = 1 unless ($self->isNumber($AttributeValue))}
        elsif($ParameterType eq 'boolean') { $RaiseError = 1 unless ($self->isBool($AttributeValue))}
        elsif($ParameterType eq 'object') { $RaiseError = 1 unless ($self->isObject($AttributeValue))}
        elsif($ParameterType eq 'array') { $RaiseError = 1 unless ($self->isArray($AttributeValue))}
        else { Error('AttributeNameNotValid', {'Message' => $AttributeName . ' \'' . $AttributeValue . '\' is not a valid attribute name. \''} ) }

        #raise error if not correct
        Error('AttributeValueVerificationError', {
            'Message' => 'wrong parameter type for attribute',
            'Object' => $Alias,
            'Attribute' => $AttributeName,
            'Value' => $AttributeValue,
            'ParameterType' => $ParameterType
        } ) if ($RaiseError);
    }
}

#========================================================================================
# function     getAttributeNames
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aAttributeNames = $self->getAttributeNames();
#----------------------------------------------------------------------------------------
# description  get all existing attribute names
#----------------------------------------------------------------------------------------
# return       $aAttributeNames | allexisting attribute names  | ref.array
#========================================================================================
sub getAttributeNames {
    my $self = shift;

    my @Return = keys %{$self->{'AttributeParameters'}};

    return \@Return;

}

#========================================================================================
# function     transformInputParameters
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->transformInputParameters($hParameters);
#----------------------------------------------------------------------------------------
# description  transform attributes e.g. 'true' -> 1 etc.
#----------------------------------------------------------------------------------------
# input        $hParameters | hash.ref
#========================================================================================
sub transformInputParameters{
    my $self = shift;
    my ($hParameters) = @_;

    # for each given parameter
    foreach my $ParameterName (keys %{$hParameters}) {
        my $AttributeParameterType = $self->getAttributeParameterType($ParameterName);
        if (defined($AttributeParameterType) && $AttributeParameterType eq 'boolean') {
            # if it is supposed to be boolean, transform it!
            $self->_transformInputBooleanParameter($hParameters, $ParameterName);
        }
    }
}

#========================================================================================
# function     transformOutputParametersGet
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->transformOutputParametersGet($hParameters);
#----------------------------------------------------------------------------------------
# description  Transform attributes for GET response e.g. 1 -> true etc. The given hashref is
#               modified directly.
#----------------------------------------------------------------------------------------
# input        $hParameters | parameters | ref.hash
#========================================================================================
sub transformOutputParametersGet{
    my $self = shift;
    my ($hParameters) = @_;

    # for each given parameter
    foreach my $ParameterName (keys %{$hParameters}) {
        my $AttributeParameterType = $self->getAttributeParameterType($ParameterName);

        if (defined($AttributeParameterType)) {

            if($AttributeParameterType eq 'boolean') {
                # if it is supposed to be boolean, transform it!
                $self->_transformOutputBooleanParameter($hParameters, $ParameterName);
            }
            elsif ($AttributeParameterType eq 'integer') {
                $self->_transformOutputIntegerParameter($hParameters, $ParameterName);
            }
            elsif ($AttributeParameterType eq 'number') {
                $self->_transformOutputNumberParameter($hParameters, $ParameterName);
            }
        }
    }
}

#========================================================================================
# function     _transformOutputBooleanParameter
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->_transformOutputBooleanParameter($hParameters, $Key);
#----------------------------------------------------------------------------------------
# description  Transform boolean attribute of incoming POST request e.g. true -> 1 etc. The given hashref is
#               modified directly.
#----------------------------------------------------------------------------------------
# input        $hParameters | parameters | ref.hash
# input        $Key | key of the parameter to transform | string
#========================================================================================
sub _transformInputBooleanParameter{
    my $self = shift;
    my ($hParameters, $Key) = @_;

    if ($hParameters->{$Key} eq 'true') {$hParameters->{$Key} = 1}
    elsif ($hParameters->{$Key} eq 'false') {$hParameters->{$Key} = 0}
}

#========================================================================================
# function     _transformOutputBooleanParameter
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->_transformOutputBooleanParameter($hParameters, $Key);
#----------------------------------------------------------------------------------------
# description  Transform boolean attribute for GET response e.g. 1 -> true etc. The given hashref is
#               modified directly.
#----------------------------------------------------------------------------------------
# input        $hParameters | parameters | ref.hash
# input        $Key | key of the parameter to transform | string
#========================================================================================
sub _transformOutputBooleanParameter{
    my $self = shift;
    my ($hParameters, $Key) = @_;

    if (not defined($hParameters->{$Key})) {delete $hParameters->{$Key}}
    elsif ($hParameters->{$Key} == 1) {$hParameters->{$Key} = JSON::XS::true}
    elsif ($hParameters->{$Key} == 0) {$hParameters->{$Key} = JSON::XS::false}
}

#========================================================================================
# function     _transformOutputIntegerParameter
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->_transformOutputIntegerParameter($hParameters, $Key);
#----------------------------------------------------------------------------------------
# description  Transform integer attribute for GET response. e.g. '1' -> 1. The given hashref is
#               modified directly.
#----------------------------------------------------------------------------------------
# input        $hParameters | parameters | ref.hash
# input        $Key | key of the parameter to transform | string
#========================================================================================
sub _transformOutputIntegerParameter{
    my $self = shift;
    my ($hParameters, $Key) = @_;

    if (not defined($hParameters->{$Key})) {delete $hParameters->{$Key}}
    else {$hParameters->{$Key} += 0}
}

#========================================================================================
# function     _transformOutputIntegerParameter
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->_transformOutputNumberParameter($hParameters, $Key);
#----------------------------------------------------------------------------------------
# description  Transform float attribute for GET response. e.g. '1.1' -> 1.1. The given hashref is
#               modified directly.
#----------------------------------------------------------------------------------------
# input        $hParameters | parameters | ref.hash
# input        $Key | key of the parameter to transform | string
#========================================================================================
sub _transformOutputNumberParameter{
    my $self = shift;
    my ($hParameters, $Key) = @_;

    $hParameters->{$Key} += 0.0;
}

#========================================================================================
# function     _getIDFromGUID
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ID = $self->_getIDFromGUID($GUID);
#----------------------------------------------------------------------------------------
# description  get the objectid of object with guid
#----------------------------------------------------------------------------------------
# input        $GUID | object guid | string
# return       $ID | object id | integer
#========================================================================================
sub _getIDFromGUID{
    my $self = shift;
    my ($GUID) = @_;

    #check parameter
    Error('MissingParameter',{'Parameter' => 'GUID'}) unless $GUID;

    #get object id from cache
    my $ID = $self->{'DataCache'}->{'IDFromGUID'}->{$GUID}; #get info from cache

    unless ($ID){ #unless info exist get info from db and cache info
        #execute on data base
        my @Row = $self->dbh->selectrow_array(
            'SELECT objectid FROM object WHERE guid=?',
            {},
            $GUID
        );


        #set taxclass id
        $ID = $Row[0];

        #check parameter
        Error('NoResultForGUID',{
            'GUID' => $GUID,
        }) unless $ID;

        #set cache taxclass id
        $self->{'DataCache'}->{'IDFromGUID'}->{$GUID} = $ID;
    }

    return $ID;

}

1;
