#==============================================================================
# package      DE_EPAGES::Object::API::TestXML
# state        public
# description  Unit test for XML export and import
#==============================================================================
package DE_EPAGES::Object::API::TestXML;
use base Exporter;
use strict;

our @EXPORT_OK = qw (
    SixTestsXML
);

use Data::Dumper qw( Dumper );
use Test::Simple;
use DE_EPAGES::Core::API::File qw( DeleteFile );
use DE_EPAGES::Database::API::Connection qw( ConnectTestSite DisconnectSite );
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias LoadRootObject LoadObjectByPath ExistsObjectByPath );
use DE_EPAGES::Object::API::XML::ObjectDeleteHandler qw ( DeleteObjects );
use DE_EPAGES::Object::API::XML::ObjectExportDriver  qw ( ExportObjects );
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw ( ImportObjects );
use DE_EPAGES::XML::API::SemanticDiff qw (DiffFiles);
use DE_EPAGES::Object::API::XMLCompare qw ( CompareXML );

use constant ROOTPATH => '/Test';

our $TEST_OK;

#========================================================================================
# function     SixTestsXML
# state        public
#----------------------------------------------------------------------------------------
# syntax       SixTestsXML( $aObjectAliases, $Encoding );
# example      SixTestsXML( [ 'TestObject' ], 'utf-8' );
#----------------------------------------------------------------------------------------
# description  This function tests the XML import handler, export driver and delete
#               handler for a certain class of objects. It performs the following 6
#               operations:
#               <ol>
#                   <li>Init - delete the object /Test if it still exists and then
#                               creates the object again</li>
#                   <li>Import - import the objects from the file import.xml</li>
#                   <li>Export - exports the child objects of /Test with the alias names
#                               objects given in $aObjectAliases to the file
#                               export.xml using encoding $Encoding</li>
#                   <li>Diff - compares the files import.xml and export.xml igoring
#                               whitespaces and differences in the attributes GUID,
#                               Position and 'delete'.</li>
#                   <li>Delete - deletes the objects that are marked with 'delete="1"'
#                               in the import.xml file and chacks that all children of
#                               /Test with alis names in $aObjectAliases are gone.</li>
#                   <li>Cleanup - deletes the /Test object and the export.xml file</li>
#               </ol>
#               Note: all objects that are createed during the test should be children
#               or gand children of /Test
#----------------------------------------------------------------------------------------
# input        $aObjectAliases | list of object aliases | ref.array.string
# input        $Encoding | encoding for Export, default: iso-8859-1 | string
#========================================================================================
sub SixTestsXML ($;$) {
    my ($aObjects, $Encoding) = @_;
    ConnectTestSite();
    local $TEST_OK = 1;
    ok(( Init() or $TEST_OK = 0), 'initialization' );
    ok(( Import() or $TEST_OK = 0), 'import');
    ok(( Export($aObjects, $Encoding) or $TEST_OK = 0), 'export');
    #ok(( Diff() or $TEST_OK = 0), 'diff');
    ok((CompareXML('import.xml', 'export.xml',{
        'printmessage' => 1,
        'checkdelete' => 0,
        'checkreference' => 1,
        'checkShopFeature' => 0,
    }) or $TEST_OK = 0), 'diff');
    ok(( Delete($aObjects) or $TEST_OK = 0), 'delete');
    ok( Cleanup(), 'cleanup' );
    DisconnectSite();
}

sub Import () {
    ImportObjects('import.xml', ROOTPATH);
    return 1;
}

sub Export ($) {
    my ($aObjects, $Encoding) = @_;
    $Encoding = 'iso-8859-1' unless defined $Encoding;
    ExportObjects(ROOTPATH, 'export.xml', {'encoding'=>$Encoding, 'withdefaults' => 1 }, $aObjects);
    return 1;
}

sub Diff () {
    my $aDifferences = DiffFiles('import.xml','export.xml', ['GUID', 'delete', 'Position', 'CreationDate', 'Inherit']);
    return 1 unless scalar @$aDifferences;
    print "XML Differences:\n" . Dumper($aDifferences);
    return 0;
}

sub Delete ($) {
    my ($aObjects) = @_;
    DeleteObjects('import.xml', ROOTPATH);
    foreach my $Path (grep { ExistsObjectByPath($_) } map { ROOTPATH . '/'. $_ } @$aObjects) {
        return 0;
    }
    return 1;
}

sub Cleanup () {
    LoadObjectByPath(ROOTPATH)->delete if ExistsObjectByPath(ROOTPATH);
    DeleteFile('export.xml') if $TEST_OK;
    return 1;
}

sub Init () {
    Cleanup();
    LoadClassByAlias('Object')->insertObject({'Alias'=>'Test', 'Parent'=>LoadRootObject()});
    return 1;
}

1;
