#==============================================================================
# §package      DE_EPAGES::OpenAI::API::CompletionAPI
# §state        private
#------------------------------------------------------------------------------
# §description  Request methods for the Completions API
#==============================================================================
package DE_EPAGES::OpenAI::API::CompletionAPI;
use base Exporter;

use strict;
use Readonly;
use JSON::XS;

use DE_EPAGES::Core::API::TestParameter qw ( TestInteger TestString );
use DE_EPAGES::Object::API::Factory qw ( LoadRootObject );
use DE_EPAGES::OpenAI::API::Constants qw ( $API_COMPLETIONS_URL $API_COMPLETIONS_DEFAULT_MODEL $API_COMPLETIONS_DEFAULT_TEMPERATURE );
use DE_EPAGES::OpenAI::API::Tools qw ( DoCurlRequest UpdateTokenUsage GetRemainingTokens );

our @EXPORT_OK = qw (
    PostCompletionAPI
);

#========================================================================================
# §function     PostCompletionAPI
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       PostCompletionAPI($Shop, $Question, $Model, $Temperature);
#----------------------------------------------------------------------------------------
# §description  Post a request to the completions API
#----------------------------------------------------------------------------------------
# §input        $Shop | Shop Object | object
# §input        $ahMessages | An array of messages to pass to the API, with 'role' and 'content' keys | ref.array.hash
# §input        $Model | The open ai model to use | string
# §input        $Temperature | The temperature parameter of the api | string
#----------------------------------------------------------------------------------------
# §return       $Res | HTTP Response object | HTTP::Response
#========================================================================================
sub PostCompletionAPI {
    my ($Shop, $ahMessages, $Model, $Temperature) = @_;

    my $TokensRemaining = GetRemainingTokens($Shop);
    return undef unless $TokensRemaining > 0;

    my $System = LoadRootObject();
    my $OpenAI_AccessToken = $System->get('OpenAI_AccessToken');
    my $hHeaders = {
        'Authorization' => "Bearer $OpenAI_AccessToken",
        'Content-Type' => 'application/json',
    };

    TestString('OpenAI_AccessToken', $OpenAI_AccessToken);

    $Model = $API_COMPLETIONS_DEFAULT_MODEL unless $Model;
    $Temperature = $API_COMPLETIONS_DEFAULT_TEMPERATURE unless $Temperature;

    TestString('Model', $Model);
    TestInteger('Temperature', $Temperature);

    my $hContent = {
        'model' => $Model,
        #'seed'  => time(), optional
        'messages' => $ahMessages,
        'temperature' => $Temperature,
        #'max_tokens'  => 350,
    };
    my $Content = encode_json $hContent;

    my $Res = DoCurlRequest('POST', $API_COMPLETIONS_URL, $Content, $hHeaders);

    UpdateTokenUsage($Shop, $Res);

    return $Res;
}



1;
