#==============================================================================
# §package      DE_EPAGES::OpenAI::API::Tools
# §state        private
#------------------------------------------------------------------------------
# §description  Tools
#==============================================================================
package DE_EPAGES::OpenAI::API::Tools;
use base Exporter;

use strict;
use Readonly;
use JSON::XS;
use Net::Curl::Easy qw(:constants);

use DE_EPAGES::Core::API::Error qw ( ExistsError GetError);
use DE_EPAGES::Core::API::Log qw ( LogError );
use DE_EPAGES::ThirdPartyConfig::API::Log qw ( Log3rdParty );
use DE_EPAGES::Core::API::PerlTools qw ( GenerateShortGUID GetCurrentDateTime );
use DE_EPAGES::Object::API::Factory qw ( LoadObjectByPath LoadRootObject );

our @EXPORT_OK = qw (
    DoCurlRequest
    UpdateTokenUsage
    GetRemainingTokens
    LogOpenAI
);

#========================================================================================
# §function     DoCurlRequest
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       DoCurlRequest($Method, $Url, $Content, $hHeaders);
#----------------------------------------------------------------------------------------
# §description  Perform a HTTP request with the given parameters, using libcurl
#----------------------------------------------------------------------------------------
# §input        $Method | HTTP method e.g. GET, POST | string
# §input        $Url | URL to connect to | string
# §input        $Content | Request body content | string
# §input        $hHeaders | Headers to add to the request | ref.hash
# §input        $hAdditionalInfo | Additional information to log | ref.hash
#----------------------------------------------------------------------------------------
# §return       $Res | HTTP Response object | HTTP::Response
#========================================================================================
sub DoCurlRequest {
    my ( $Method, $Url, $Content, $hHeaders, $hAdditionalInfo ) = @_;

    #-- https://gms.tf/when-curl-sends-100-continue.html
    $hHeaders->{'Expect'} = '';

    #-- Data structures (would appear to) get mangled by Data Dumper, so make a copy for the logging side and keep the original as it was
    my $hLogHeaders = defined $hHeaders ? {%$hHeaders} : undef;

    #-- Remove the authorization header from the log
    $hLogHeaders->{'Authorization'} = '--- removed ---' if exists $hLogHeaders->{'Authorization'};

    my $LogName = 'OpenAI';
    $LogName .= ' ' . $hAdditionalInfo->{'LogName'} if $hAdditionalInfo->{'LogName'};
    my $LogSuffix = GenerateShortGUID();
    delete $hAdditionalInfo->{'LogName'};

    my $hReqLog = {
      Method          => $Method,
      Url             => $Url,
      Content         => $Content,
      hHeaders        => $hLogHeaders,
      hAdditionalInfo => $hAdditionalInfo,
    };

    my $Easy = Net::Curl::Easy->new();
    $Easy->setopt( CURLOPT_URL, $Url );
    $Easy->setopt( CURLOPT_CUSTOMREQUEST, $Method );

    my $aHeaders;
    push @$aHeaders, $_ . ':' . $hHeaders->{$_} foreach keys %$hHeaders;
    $Easy->setopt( CURLOPT_HTTPHEADER, $aHeaders );

    if ( $Content ) {
      #$Content = encode( 'utf-8', $Content );
      $Easy->setopt( CURLOPT_POSTFIELDS, $Content );
    }

    $Easy->setopt( CURLOPT_WRITEHEADER(), \$Easy->{Headers} );
    $Easy->setopt( CURLOPT_FILE(),        \$Easy->{Body} );

    LogOpenAI( "Request $LogName $LogSuffix", $hReqLog );
    eval {
        $Easy->perform();
    };

    if ( ExistsError() ) {
        my $Error = GetError();
        LogError($Error);

        return undef;
    } else {

        my $ResponseCode = $Easy->getinfo(CURLINFO_RESPONSE_CODE);

        #-- We tried to log a 'nice' variant of the response by decoding the json, but it flattens the structure after about 3 levels, so ...

        my $hResLog = {
          Method         => $Method,
          Url            => $Url,
          ResponseCode   => $ResponseCode,
          Headers        => $Easy->{Headers},
          Body           => $Easy->{Body},
        };
        LogOpenAI( "Response $LogName $LogSuffix", $hResLog );

        my $Res = HTTP::Response->parse( $Easy->{Headers} );
        $Res->content( $Easy->{Body} );

        return $Res;
    }
}

#========================================================================================
# §function     LogOpenAI
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       LogOpenAI($Name, $Stuff);
#----------------------------------------------------------------------------------------
# §description  Log to communication.log
#----------------------------------------------------------------------------------------
# §input        $Name | The name to log under | string
# §input        $Stuff | The stuff to actually log | *
#========================================================================================
sub LogOpenAI {
    my ($Name, $Stuff) = @_;

    my $System = LoadRootObject();
    my $OpenAI_Logging = $System->get('OpenAI_Logging');
    return unless $OpenAI_Logging;

    Log3rdParty( $Name, $Stuff );
}

#========================================================================================
# §function     UpdateTokenUsage
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       UpdateTokenUsage($Shop, $Res);
#----------------------------------------------------------------------------------------
# §description  Update the shop's tokens used counter for this month for an OpenAI request
#----------------------------------------------------------------------------------------
# §input        $Shop | Shop object | object
# §input        $Res | Response of the request that was made | HTTP::Response
#========================================================================================
sub UpdateTokenUsage {
    my ($Shop, $Res) = @_;

    return unless $Res && $Res->is_success;

    my $ResContent = $Res->decoded_content;
    my $hResponse;
    eval {
        $hResponse = decode_json $ResContent;
    };
    return if $@;

    my $InputTokens = $hResponse->{'usage'}->{'prompt_tokens'};
    my $OutputTokens = $hResponse->{'usage'}->{'completion_tokens'};

    return unless defined $InputTokens;
    return unless defined $OutputTokens;

    #--
    #--  {
    #--    'YYYYMM' => {
    #--        'InputTokens' => integer,
    #--        'OutputTokens' => integer,
    #--    }
    #--  }
    #--
    my $hOpenAITokensUsage = $Shop->get('OpenAITokensUsage');

    if ( ! defined $hOpenAITokensUsage || ref $hOpenAITokensUsage ne 'HASH' ) {
        $hOpenAITokensUsage = {};
        $Shop->set({'OpenAITokensUsage' => $hOpenAITokensUsage});
        $hOpenAITokensUsage = $Shop->get('OpenAITokensUsage');
    }

    my $Now = GetCurrentDateTime();
    my $Year = $Now->year();
    my $Month = $Now->month();
    my $Key = "$Year$Month";

    if ( exists $hOpenAITokensUsage->{$Key} && exists $hOpenAITokensUsage->{$Key}->{'InputTokens'} && exists $hOpenAITokensUsage->{$Key}->{'OutputTokens'} ) {
    } else {
        $hOpenAITokensUsage->{$Key} = {
            'InputTokens' => 0,
            'OutputTokens' => 0,
        };
    }

    $hOpenAITokensUsage->{$Key}->{'InputTokens'} += $InputTokens;
    $hOpenAITokensUsage->{$Key}->{'OutputTokens'} += $OutputTokens;
    $Shop->set({'OpenAITokensUsage' => $hOpenAITokensUsage});

    my $TotalTokensUsedThisMonth = $hOpenAITokensUsage->{$Key}->{'InputTokens'} + $hOpenAITokensUsage->{$Key}->{'OutputTokens'};

    my $OpenAITokensFeature = LoadObjectByPath("/Features/OpenAITokens");
    $Shop->featureCurrentValue( $OpenAITokensFeature, $TotalTokensUsedThisMonth, 1 );
}

#========================================================================================
# §function     GetRemainingTokens
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       GetRemainingTokens($Method, $Url, $Content, $hHeaders);
#----------------------------------------------------------------------------------------
# §description  Retrieve the remaining tokens available for a shop
#----------------------------------------------------------------------------------------
# §input        $Shop | Shop object | object
#----------------------------------------------------------------------------------------
# §return       $RemainingTokens | The number of remaining tokens available | integer
#========================================================================================
sub GetRemainingTokens {
    my ($Shop) = @_;

    #--
    #--  {
    #--    'YYYYMM' => {
    #--        'InputTokens' => integer,
    #--        'OutputTokens' => integer,
    #--    }
    #--  }
    #--
    my $hOpenAITokensUsage = $Shop->get('OpenAITokensUsage');

    #-- Initialise the attribute if not already
    if ( ! defined $hOpenAITokensUsage || ref $hOpenAITokensUsage ne 'HASH' ) {
        $hOpenAITokensUsage = {};
        $Shop->set({'OpenAITokensUsage' => $hOpenAITokensUsage});
        $hOpenAITokensUsage = $Shop->get('OpenAITokensUsage');
    }

    my $Now = GetCurrentDateTime();
    my $Year = $Now->year();
    my $Month = $Now->month();
    my $Key = "$Year$Month";

    #-- Is there data already for this month? If not, initialise this month's hash
    if ( exists $hOpenAITokensUsage->{$Key} && exists $hOpenAITokensUsage->{$Key}->{'InputTokens'} && exists $hOpenAITokensUsage->{$Key}->{'OutputTokens'} ) {
    } else {
        $hOpenAITokensUsage->{$Key} = {
            'InputTokens' => 0,
            'OutputTokens' => 0,
        };
    }

    my $TotalTokensUsed = $hOpenAITokensUsage->{$Key}->{'InputTokens'} + $hOpenAITokensUsage->{$Key}->{'OutputTokens'};
    $TotalTokensUsed = $TotalTokensUsed >= 0 ? $TotalTokensUsed : 0; #-- Guard against negative values

    my $OpenAITokensFeature = LoadObjectByPath("/Features/OpenAITokens");
    my $MaxPossible = $Shop->featureMaxValue($OpenAITokensFeature);

    #-- It must have a value, if it is less than 0, assume 0 (which does NOT mean unlimited, 0 means 0)
    $MaxPossible = defined $MaxPossible && $MaxPossible >= 0 ? $MaxPossible : 0;

    #-- Most of the time it will not be exact because we won't know how many tokens openai will consume
    #-- So if it is < 0, return 0
    my $RemainingTokens = $MaxPossible - $TotalTokensUsed;
    $RemainingTokens = 0 if $RemainingTokens < 0;

    return $RemainingTokens;
}

1;
