#========================================================================================
# package      DE_EPAGES::Order::API::Attributes::LineItemPayment_Ext
# state        private
# base         DE_EPAGES::Object::API::Attributes::Reference
# description  Attribute interface for LineItemPayment.
#========================================================================================
package DE_EPAGES::Order::API::Attributes::LineItemPayment_Ext;
use base qw( DE_EPAGES::Object::API::BaseAttribute );

use strict;
use DE_EPAGES::Core::API::Constants qw( PERL538 );
use DE_EPAGES::Order::API::HirePurchase qw(CalculateRate CalculateEffectiveInterestRate);

#========================================================================================
# function     getAttributes
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = $Package->getAttributes( $LineItemPayment, $aNames );
# example      # never call getAttributes directly!
#               $Value = $Object->get( $AttributeName );
#----------------------------------------------------------------------------------------
# description  Returns attribute values of an object.
#               Used by <function DE_EPAGES::Object::Object::Object::get>
#----------------------------------------------------------------------------------------
# input        $LineItemPayment | LineItemPayment object | object
# input        $aNames | list of attribute names | ref.array
# return       $hValues | attribute name => value hash | ref.hash
#========================================================================================
sub getAttributes {
    my $class = shift;
    my ($LineItemPayment, $aNames) = @_;
    my $hInfo = {};

    foreach my $Name ( @$aNames ) {
        if ($Name eq 'IsHirePurchaseAllowed') {
            $hInfo->{$Name} = _IsHirePurchaseAllowed($LineItemPayment);
        } elsif ($Name eq 'HirePurchaseInstallments') {
            $hInfo->{$Name} = _GetHirePurchaseInstallments($LineItemPayment);
        } elsif ($Name eq 'HirePurchaseEffectiveInterestRateValue') {
            $hInfo->{$Name} = _HirePurchaseEffectiveInterestRateValue($LineItemPayment);
        } elsif ($Name eq 'HirePurchaseBasePrice') {
            my $Container = $LineItemPayment->container;
            $hInfo->{$Name} = $Container->get('GrandTotal');
        } elsif ($Name eq 'HirePurchaseGrandTotal') {
            my $HirePurchaseNoOfMonths = $LineItemPayment->get('HirePurchaseNoOfMonths');
            my $HirePurchaseInstallmentAmount = $LineItemPayment->get('HirePurchaseInstallmentAmount');
            $hInfo->{$Name} = defined $HirePurchaseNoOfMonths && defined $HirePurchaseInstallmentAmount ? $HirePurchaseNoOfMonths * $HirePurchaseInstallmentAmount : 0;
        } elsif ($Name eq 'Taxes') {
            $hInfo->{$Name} = $class->_paymentTaxes($LineItemPayment);
        }
    }
    return $hInfo;
}

sub _HirePurchaseEffectiveInterestRateValue {
    my ($LineItemPayment) = @_;

    my $Container = $LineItemPayment->container;
    my $LocaleID = $Container->get('LocaleID');
    my $Locale = $Container->getSite()->siteLocale($LocaleID);

    return unless $LineItemPayment->get('HirePurchaseBasePrice');
    return unless $LineItemPayment->get('HirePurchaseInstallmentAmount');
    return unless $LineItemPayment->get('HirePurchaseNoOfMonths');
    return if $LineItemPayment->get('HirePurchaseNoOfMonths') == 0;

    my $effectiveInterestRateValue = CalculateEffectiveInterestRate($LineItemPayment->get('HirePurchaseBasePrice'),
                                                                    $LineItemPayment->get('HirePurchaseInstallmentAmount'),
                                                                    $LineItemPayment->get('HirePurchaseNoOfMonths'),
                                                                    $LineItemPayment->get('HirePurchaseInterestRateValue')
                                                                    );
    return $Locale->roundMoney($effectiveInterestRateValue, $Container->get('CurrencyID'));
}

sub _IsHirePurchaseAllowed {
    my ($LineItemPayment) = @_;

    # line item has month saved
    return 1 if $LineItemPayment->get('HirePurchaseNoOfMonths');

    my $PaymentMethod = $LineItemPayment->get('PaymentMethod');
    return 0 unless defined $PaymentMethod;

    # check if payment method supports hirepurchase
    return 0 unless($PaymentMethod->get('IsHirePurchaseEnabled'));

    # check if we have to offer hire purchase
    my $Container  = $LineItemPayment->container();
    my $CurrencyID = $Container->get('CurrencyID');

    # get interest rate matrix
    my $Folder = $LineItemPayment->getSite()->folder('InterestRateMatrices');
    return 0 unless $Folder->existsChild($CurrencyID);

    # get the interest rates
    my @aInterestRates = sort {$a->get('MinimumAmount') <=> $b->get('MinimumAmount')} @{$Folder->child($CurrencyID)->get('InterestRates')};# currencyid is alias in folder
    return 0 if scalar (@aInterestRates) < 1;

    # check if the basket amount is at least the smalles minimum amount of the interest rates
    my $basketAmount = $Container->get('ShippingSubTotal');
    return 0 if ($basketAmount < $aInterestRates[0]->get('MinimumAmount'));

    return 1;
}

sub _GetHirePurchaseInstallments {
    my ($LineItemPayment) = @_;

    return [] unless $LineItemPayment->get('PaymentMethod')->get('IsHirePurchaseEnabled');

    my $Container  = $LineItemPayment->container();
    my $Shop       = $LineItemPayment->getSite();
    my $CurrencyID = $Container->get('CurrencyID');

    my @InterestRatesLoop = ();

    my $Folder = $Shop->folder('InterestRateMatrices');
    my $InterestRateMatrix = $Folder->existsChild($CurrencyID) # currencyid is alias
        ? $Folder->child($CurrencyID)
        : undef;
    if (defined $InterestRateMatrix) {
        # get the interest rates
        my $aInterestRates = $InterestRateMatrix->get('InterestRates');

        # build hash
        my %hRates = ();
        foreach my $Rate (@$aInterestRates) {
            $hRates{$Rate->get('MinimumAmount')*1}{$Rate->get('NumberOfMonths')*1} = $Rate->get('InterestRateValue');
        }

        # find matching ( <= basket ) maximum amount
        my $basketAmount = $LineItemPayment->get('HirePurchaseBasePrice');
        my ($minAmount) = sort { $b <=> $a } grep { $_ <= $basketAmount } keys %hRates;

        # add possible interest rates
        if (defined $minAmount) {
            my $LocaleID = $Container->get('LocaleID');
            my $Locale = $Container->getSite()->siteLocale($LocaleID);
            foreach my $noOfMonths(sort {$a<=>$b} keys %{$hRates{$minAmount}}) {
                my $installmentAmount = CalculateRate($basketAmount, $noOfMonths, $hRates{$minAmount}{$noOfMonths});
                $installmentAmount = $Locale->roundMoney($installmentAmount, $CurrencyID, 'ceil');
                my $effectiveInterestRateValue = CalculateEffectiveInterestRate($basketAmount,
                    $installmentAmount, $noOfMonths,
                    $hRates{$minAmount}{$noOfMonths}*100 ); ## no critic qw(ProhibitMagicNumbers)

                push @InterestRatesLoop, {
                    'NumberOfMonths'    => $noOfMonths,
                    'InterestRateValue' => 100 * $hRates{$minAmount}{$noOfMonths},
                    'EffectiveInterestRateValue' => $Locale->roundMoney($effectiveInterestRateValue, $CurrencyID),
                    'InstallmentAmount' => $installmentAmount,
                    'Sum'               => $noOfMonths * $installmentAmount,
                    'BasePrice'         => $basketAmount,
                };
            }
        }
    }

    return \@InterestRatesLoop;
}


# get the payment prices in a loop for each tax rate
# e.g. if we have 5,- payment and products with 7% and 19% tax,
# payment is splitted to the relation of the 7% and 19% produts of th total price
# from each payment price the tax is calculated
sub _paymentTaxes {
    my $self = shift;
    my ($LineItemPayment, $ItemsName) = @_;
    my $Deprecated = $] < PERL538 ? 1 : 0;
    $Deprecated = 0 if $Deprecated and $ENV{USE_NEW_LINEITEMCONTAINER};

    # get the TaxLineItems from Container
    my $Container = $LineItemPayment->container;
    my $aTaxLineItems = $Container->get('Taxes');
    my $aTaxDigitalLineItems = $Container->get('TaxesDigital');
    my $NoTax = $Container->get('NoTax');

    # if the customer has removed a TaxClass, we do not calculate the splitten tax amounts
    return [] if $Container->get('IsTaxClassRemoved');

    my $hTaxRates = $Container->netPricesForTax();
    my $hPaymentTaxItems = $hTaxRates->{'LineItemPayment'};
    my $Formatter = $Container->formatter;

    my $aRet;

    foreach my $TaxLineItem (@$aTaxLineItems, @$aTaxDigitalLineItems) {
        my $IsDigital = $TaxLineItem->instanceOf('LineItemTaxDigital') ? 1:undef;
        my $Postfix = $IsDigital ? 'Digital' : '';

        my $TaxArea = $TaxLineItem->get('TaxArea' . $Postfix);
        my $TaxMatrix = $TaxLineItem->get('TaxMatrix' . $Postfix);
        my $TaxClass = $TaxLineItem->get('TaxClass' . $Postfix);
        return [] unless defined $TaxClass;

        my $TaxRate;
        if ($IsDigital) {
            $TaxRate = defined $TaxArea
            ? ($NoTax ? 0 : $TaxMatrix->taxRateDigital($TaxArea, $TaxClass))
            : $TaxLineItem->get('TaxRateDigital');
        }
        else {
            $TaxRate = defined $TaxArea
            ? ($NoTax ? 0 : $Container->taxRate($TaxClass))
            : $TaxLineItem->get('TaxRate');
        }

        my $Value;
        if ($Deprecated) {
            $Value = $hPaymentTaxItems->{$TaxClass->id} // 0;
        } else {
            my @TaxItem = grep { $_->[0] eq $TaxClass->id } @$hPaymentTaxItems;
            $Value = @TaxItem ? $TaxItem[0][1] : 0;
        }
        next if ($Value eq '0') || !(defined $Value);

        push @$aRet, {
            'TaxClass' . $Postfix => $TaxClass,
            'TaxRate'  . $Postfix => $TaxRate,
            'Value'     => $Formatter->round($Value),
            'TaxAmount' => $TaxLineItem->calculateTaxAmount($Value, $TaxRate, $TaxRate)
        };
    }

    return $aRet;
}

1;
