#========================================================================================
# package      DE_EPAGES::Order::API::Object::LineItemContainer
# state        public
# base         DE_EPAGES::Shop::API::Object::ShopObject
# description  object interface for LineItemContainer.
#========================================================================================
package DE_EPAGES::Order::API::Object::LineItemContainer;
use base DE_EPAGES::Shop::API::Object::ShopObject;

use strict;
use Scalar::Util qw (blessed);
use DE_EPAGES::Core::API::Constants qw( PERL538 );
use DE_EPAGES::Core::API::Error qw( Error );
use DE_EPAGES::Core::API::Log qw(LogDebug);
use DE_EPAGES::Core::API::PerlTools qw ( Compare GenerateGUID GetMD5Digest);
use DE_EPAGES::Core::API::TestParameter qw ( TestObject );
use DE_EPAGES::Core::API::Warning qw ( Warning );
use DE_EPAGES::Database::API::Connection qw (Transaction);
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias LoadObject );
use DE_EPAGES::Tax::API::Shop qw ( ShouldSplitShippingAndPaymentTax );
use DE_EPAGES::Shop::API::Constants qw (Text2TaxModel TaxModel2Text TAX_MODEL_GROSS TAX_MODEL_NET);
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Payment::API::Constants qw ( CASHDISCOUNTMODE_PERCENT_DISCOUNT );


sub ErrorDefinition {
    return {
        'LineItemPriceNotRounded' => 'line item (class #Class, package #Package, lineitem #LineItemID #LineItemPath, shop #ShopAlias) doesn\'t round the lineitem price (#Price).',
        'TaxAmountNotRounded' => 'line item (class #Class, package #Package, lineitem #LineItemID #LineItemPath, shop #ShopAlias) doesn\'t round the tax amount (#TaxAmount).',
        'ContainerNoCustomer' => 'Can\'t get the customer for lineitem container #Path.',
        'ContainerNoUser' => 'Can\'t get the user for lineitem container #Path.',
        'RecursiveRecalculate' => 'LineItemContainer->recalculate was called recursively',
    };
}

#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $LineItemContainer->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className  { 'LineItemContainer' }

#========================================================================================
# function     dbPackage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DbPackage = $LineItemContainer->dbPackage;
#----------------------------------------------------------------------------------------
# description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemContainer = DE_EPAGES::Order::API::Object::LineItemContainer->insert($hInfo);
# example      my $Class = LoadClassByAlias('LineItemContainer');
#               my $LineItemContainer = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertLineItemContainer' is triggered.
#----------------------------------------------------------------------------------------
# input        $Package | class package | string
# input        $hInfo | attributes of new object | ref.hash
# return       $LineItemContainer | new object | object
# hook         OBJ_InsertLineItemContainer | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemContainerID - lineitem container identifier - int
#               </ul>
#========================================================================================
sub insert {
    my ($Package, $hInfo) = @_;

    my $self = $Package->SUPER::insert($hInfo);

    TriggerHook('OBJ_InsertLineItemContainer', {'Object' => $self, 'LineItemContainerID' => $self->id });
    delete $hInfo->{$_} foreach qw ( LocaleID CurrencyID TaxArea TaxModel GrandTotal TotalBeforeTax TotalTax LastShippingMethod LastPaymentMethod TaxAreaDigital ); ## no critic qw(ProhibitLongLines)
    return $self;
}

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemContainer->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteLineItemContainer' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteLineItemContainer | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemContainerID - lineitem container identifier - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteLineItemContainer', {'Object' => $self, 'LineItemContainerID' => $self->id });
    LoadClassByAlias('LineItem')->deleteObjects('ByLineItemContainerID', [$self->id]);
    $self->SUPER::delete;
    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemContainer->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateLineItemContainer' and 'OBJ_AfterUpdateLineItemContainer'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateLineItemContainer | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemContainerID - lineitem container identifier - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateLineItemContainer | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemContainerID - lineitem container identifier - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateLineItemContainer', {'Object'=>$self, 'Values'=> $hValues, 'LineItemContainerID' => $self->id, 'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdateLineItemContainer',  {'Object'=>$self, 'Values'=> $hValues, 'LineItemContainerID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}

#========================================================================================
# function     importXML
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Object = __PACKAGE__->importXML($Handler, $hElement, $Class, $hAttributes);
#----------------------------------------------------------------------------------------
# description  Imports an object from an XML file. This function is called at the XML
#               start tag. The end tag triggers the <function importXMLEnd> function.
#----------------------------------------------------------------------------------------
# input        $Handler | xml import handler with support function addObject | object
# input        $hElement | XML element structure, see XML::SAX | ref.hash
# input        $Class | class object; the new object will be an instance of this class | object
# input        $hAttributes | attributes hash | ref.hash
# return       $Object | the imported object | object
#========================================================================================
sub importXML {
    my $class = shift;
    my ($Handler, $hElement, $Class, $hAttributes) = @_;

    # map TaxModel name to Model id
    if (exists $hAttributes->{'TaxModel'}) {
        $hAttributes->{'TaxModel'} = Text2TaxModel($hAttributes->{'TaxModel'});
    }
    # max TaxArea name to area object
    if (exists $hAttributes->{'TaxArea'}) {
        my $XMLParent = $Handler->topID('Object');
        my $Shop = $XMLParent->getSite();
        my $AreaID = $Shop->get('TaxMatrix')->findChild($hAttributes->{'TaxArea'});
        $hAttributes->{'TaxArea'} = LoadObject($AreaID) if defined $AreaID; # new style EU instead of old style /TaxMatrixGermany/EU
    }

    return $class->SUPER::importXML($Handler, $hElement, $Class, $hAttributes );
}

#========================================================================================
# function     exportXMLAttributes
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hAttributes = $Object->exportXMLAttributes( $Driver, $hAllAttributes );
#----------------------------------------------------------------------------------------
# description  Returns plain attributes of object to XML export driver. This function
#               removes bit attributes with value 0.
#----------------------------------------------------------------------------------------
# input        $Driver | xml export driver with support function addObject | object
# input        $hAllAttributes | all attributes of object,
#                                 see $Object-><function exportableAttributes> | ref.hash
# return       $hAttributes | plain attributes | ref.hash.string
#========================================================================================
sub exportXMLAttributes {
    my $self = shift;
    my ($Driver, $hAttributes) = @_;
    local $DE_EPAGES::Object::API::Language::LANGUAGEID = $self->get('LanguageID');
    $hAttributes = $self->SUPER::exportXMLAttributes($Driver, $hAttributes);
    $hAttributes->{'TaxModel'} = TaxModel2Text($hAttributes->{'TaxModel'}) if exists $hAttributes->{'TaxModel'};
    $hAttributes->{'TaxArea'} = $self->get('TaxArea')->alias if exists $hAttributes->{'TaxArea'};
    return $hAttributes;
}

#========================================================================================
# function     needShipping
# state        public
#----------------------------------------------------------------------------------------
# syntax       $NeedShippingLineItem = $Container->needShipping( $LineItems );
#----------------------------------------------------------------------------------------
# description  Returns true if the container need a shipping lineitem.
#----------------------------------------------------------------------------------------
# input        $LineItems | list of lineitems (used attribute IsShippingRequired) (optional) | ref.array.object
# return       $NeedShippingLineItem | container need a shipping lineitem | boolean
#========================================================================================
sub needShipping {
    my $self = shift;
    my ($LineItems) = @_;
    return 0 unless $self->id == $self->businessContainer->id;
    $LineItems = $self->get('LineItems') unless defined $LineItems;
    return grep { $_->get('IsShippingRequired') } @$LineItems;
}

#========================================================================================
# function     needPayment
# state        public
#----------------------------------------------------------------------------------------
# syntax       $NeedPaymentLineItem = $Container->needPayment( $LineItems );
#----------------------------------------------------------------------------------------
# description  Returns true if the container need a payment lineitem (LineItemPrice !=0).
#----------------------------------------------------------------------------------------
# input        $LineItems | list of lineitems (used attribute LineItemPrice) (optional) | ref.array.object
# return       $NeedPaymentLineItem | container need a payment lineitem | boolean
#========================================================================================
sub needPayment {
    my $self = shift;
    my ($LineItems) = @_;
    return 0 unless $self->id == $self->businessContainer->id;

    # test if we have lineitems and lineitemprices
    if (not defined $LineItems) {
        $LineItems = $self->get('LineItems');
        my $Shipping = $self->get('Shipping');
        push @$LineItems, $Shipping if defined $Shipping;
    }
    return grep { $_->get('LineItemPrice')} @$LineItems;
}

#========================================================================================
# function     insertLineItem
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LineItem = $Container->insertLineItem($Class, $hLineItem);
# example      $LineItem = $Container->insertLineItem(LoadClassByAlias('LineItemProduct'), {
#                       'Quantity'      => 4,
#                       'Name'          => 'epages merchant 5.0',
#                       'LineItemPrice' => 20000,
#                       'TaxRate'       => 0.16,
#               });
#----------------------------------------------------------------------------------------
# description  Returns the inserted lineitem. Generates the Alias and GUID of new
#               lineitem.
#----------------------------------------------------------------------------------------
# input        $Class | line item class | object
# input        $hLineItem | data | ref.hash
# return       $LineItem | line item | object
#========================================================================================
sub insertLineItem {
    my $self = shift;
    my ($Class, $hLineItem ) = @_;

    $hLineItem->{'Parent'} = $self;
    $hLineItem->{'GUID'} = GenerateGUID();
    $hLineItem->{'Alias'} = $Class->alias. '_' . GetMD5Digest(\$hLineItem->{'GUID'}) unless defined $hLineItem->{'Alias'};
    $hLineItem->{'Name'} = substr($hLineItem->{'Name'},0,255) if $hLineItem->{'Name'}; ## no critic (ProhibitMagicNumbers)

    my $LineItem = $Class->insertObject($hLineItem);


    # UNITY-6568 test if the tax class fits to the tax matrix of the lineitem container
    # and correct tax class
    # we need this handling if we deal with and backup tax matrix for an order
    my $LineItemTaxClass = $LineItem->get('TaxClass');
    my $ContainerTaxMatrix = $self->get('TaxMatrix');
    if ((defined $LineItemTaxClass) and (defined $ContainerTaxMatrix)) {
        unless($LineItemTaxClass->isChildOf($ContainerTaxMatrix)){
            $LineItemTaxClass = $ContainerTaxMatrix->child($LineItemTaxClass->alias);
            $LineItem->set({'TaxClass' => $LineItemTaxClass});
        }
    }

    # add tax line item
    unless( $LineItem->instanceOf('LineItemTax') || $LineItem->instanceOf('LineItemTaxDigital') ) {
        $self->_insertTaxForLineItem($LineItem);
    }

    my $aLineItems = $self->get('LineItems');

    # shipping line item
    if( $LineItem->instanceOf('LineItemShipping') ) {
        push @$aLineItems, $LineItem;
    }
    else {
        my $ShippingLineItem = $self->addLineItemShipping( $aLineItems );
        push @$aLineItems, $ShippingLineItem if defined $ShippingLineItem;
    }

    # payment line item
    unless( $LineItem->instanceOf('LineItemPayment') ) {
        $self->addLineItemPayment( $aLineItems );
    }

    return $LineItem;
}

#========================================================================================
# function     setLineItem
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LineItem = $Container->setLineItem($Class, $hLineItem);
# example      $LineItem = $Container->setLineItem(LoadClassByAlias('LineItemProduct'), {
#                       'Quantity'      => 4,
#                       'Name'          => 'epages merchant 5.0',
#                       'LineItemPrice' => 20000,
#                       'TaxRate'       => 0.16,
#               });
#----------------------------------------------------------------------------------------
# description  Creates a line item if this line item not exist, all keys of the line item
#               will be compared with existing items.
#----------------------------------------------------------------------------------------
# input        $Class | line item class | object
# input        $hLineItem | data | ref.hash
# return       $LineItem | line item | object
#========================================================================================
sub setLineItem {
    my $self = shift;
    my ($Class, $hLineItem) = @_;

    my @LineItems = grep { $_->instanceOf($Class) } @{$self->children};
    foreach my $Child (@LineItems) {
        my $AllEqual = 1;
        foreach my $Attribute (keys %$hLineItem) {
            my $Type = $Child->class->attribute($Attribute)->get('Type');
            if (defined $Child->get($Attribute) and
                Compare($Type, $hLineItem->{$Attribute}, $Child->get($Attribute))) {
                $AllEqual = 0;
                last;
            }
        }
        return $Child if $AllEqual;
    }
    return $self->insertLineItem($Class, $hLineItem);
}

#========================================================================================
# function     user
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Container->user();
#----------------------------------------------------------------------------------------
# description  returns user of this container
#----------------------------------------------------------------------------------------
# return        $user | related user | object
#========================================================================================
sub user {
    my $self = shift;

    my $Parent = $self->parent;
    if ( $Parent->instanceOf('Basket') || $Parent->instanceOf('CustomerOrder') ) {
        return $Parent->get('User');
    }
    elsif ( $Parent->instanceOf('OrderDocument') ) {
        return $Parent->get('Order')->get('User');
    }
    elsif ( $Parent->instanceOf('LineItemProduct') ) {
        return $Parent->get('LineItemContainer')->user;
    }
    else {
        Error('ContainerNoUser', { 'Path' => $self->get('Path') } );
    }
}

#========================================================================================
# function     customer
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Container->customer();
#----------------------------------------------------------------------------------------
# description  returns customer of this container
#----------------------------------------------------------------------------------------
# return        $customer | related customer | object
#========================================================================================
sub customer {
    my $self = shift;

    my $Parent = $self->parent;
    if ( $Parent->instanceOf('Basket') || $Parent->instanceOf('CustomerOrder') ) {
        return $Parent->get('Customer');
    }
    elsif ( $Parent->instanceOf('OrderDocument') ) {
        return $Parent->get('Order')->get('Customer');
    }
    elsif ( $Parent->instanceOf('LineItemProduct') ) {
        return $Parent->get('LineItemContainer')->customer;
    }
    else {
        Error('ContainerNoCustomer', { 'Path' => $self->get('Path') } );
    }
}

#========================================================================================
# function     recalculate
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Container->recalculate($hValues);
#----------------------------------------------------------------------------------------
# description  Call this function if lineitem prices has been changed, to recalculate
#               GrandTotal, TotalTax and TotalBeforeTax. Before all lineitems of the container
#               will be recalculated. The hValues will be forwarded to these lineitems.
#               <br>GrandTotal = SUM (children.LineItemPrice)
#               <br>TotalTax   = SUM (children.TaxAmount)
#               <br>TotalBeforeTax = GrandTotal - TotalTax
#----------------------------------------------------------------------------------------
# input        $hValues | changed values (optional) | ref.hash
#========================================================================================
sub recalculate {
    my $self = shift;
    my ($hValues) = @_;
    $hValues = {} unless defined $hValues;

    if( ($self->{'recalculate'} // 0) > 1 ) { # EPG-23639, EPG-24067: allow one level of recusrion for basket discount
        Error('RecursiveRecalculate', {
            'LineItemContainer' => $self->pathString,
        });
    }

    local $self->{'recalculate'} = ($self->{'recalculate'} // 0) + 1;

    TriggerHook( 'BeforeRecalculateLineItemContainer', {
        'Object' => $self,
        'Values' => $hValues
    });

    # recalculate items
    my @TYPES = qw ( LineItems SalesDiscounts Discounts Shipping Payment PaymentDiscount Statics Taxes TaxesDigital );
    my $hLineItems = $self->get([@TYPES]);
    my @LineItems;
    foreach my $Items (grep { defined $hLineItems->{$_} } @TYPES) {
        if (ref $hLineItems->{$Items} ne 'ARRAY') {
            my $Item = $hLineItems->{$Items};
            $Item->recalculate($self, $hValues);
            $self->_addLineItemTax($Item);
        }else {
            my $TypeItems = $hLineItems->{$Items};
            foreach my $Item (@$TypeItems) {
                $Item->recalculate($self, $hValues) ;
                $self->_addLineItemTax($Item);
            }
        }
    }

    $hLineItems = $self->get([@TYPES]);
    foreach my $Items (grep { defined $hLineItems->{$_} } @TYPES) {
        if (ref $hLineItems->{$Items} ne 'ARRAY') {
            my $Item = $hLineItems->{$Items};
            push @LineItems, $Item;
        }else {
            my $TypeItems = $hLineItems->{$Items};
            push @LineItems, @$TypeItems;
        }
    }

    TriggerHook( 'AfterRecalculateLineItems', {
        'Object' => $self,
        'Values' => $hValues
    });

    # calculate prices
    my $Formatter = $self->formatter;
    my ($GrandTotal, $TotalTax) = (0,0);
    foreach my $Child (@LineItems) {
        my $Class = $Child->class;
        next unless $Class->existsAttribute('LineItemPrice');

        my $Price;
        my %Vars = (
            'Object' => $Child,
            'Price' => undef
        );
        # change the price of a lineitem e.g. set shipping to 0, if we have a coupon for free shipping
        TriggerHook( 'LineItemPrice',\%Vars);
        if (defined $Vars{'Price'}) {
            $Price = $Vars{'Price'};
        } else {
            # else standard list price
            $Price = $Child->get('LineItemPrice') // 0;
        }

        my $RoundPrice = $Formatter->round($Price);
        Warning('LineItemPriceNotRounded', {
            'Class'       => $Child->class->alias,
            'Package'     => blessed ($Child),
            'Price'       => $Price,
            'LineItemID'  => $Child->id,
            'LineItemPath'=> $Child->get('Path'),
            'ShopAlias'   => $Child->getSite()->alias,
        }) if $Formatter->compare($Price, $RoundPrice);
        $GrandTotal += $RoundPrice;

        next unless $Class->existsAttribute('TaxAmount');
        my $TaxAmount = $Child->get('TaxAmount');
        if( defined $TaxAmount ) {
            my $RoundTaxAmount = $Formatter->round($TaxAmount);
            Warning('TaxAmountNotRounded', {
                'Class'       => $Child->class->alias,
                'Package'     => blessed ($Child),
                'TaxAmount'   => $TaxAmount,
                'LineItemID'  => $Child->id,
                'LineItemPath'=> $Child->get('Path'),
                'ShopAlias'   => $Child->getSite()->alias,
            }) if $Formatter->compare($TaxAmount, $RoundTaxAmount);
            $TotalTax += $RoundTaxAmount;
        }
    }

    my $aLineItems = $self->get('LineItems');
    # add paymentdiscount
    push @$aLineItems, $self->get('PaymentDiscount') if $self->get('PaymentDiscount') and
                                                        $self->get('PaymentDiscount')->get('DiscountMode') == CASHDISCOUNTMODE_PERCENT_DISCOUNT;
    # determine if the discount column is required
    my $HasDiscounts = 0;
    foreach my $LineItem (@$aLineItems){
        my $Discount = $LineItem->get('Discount');
        if(defined $Discount && $Discount > 0){
            $HasDiscounts = 1;
            last;
        }
    }

    # set all values
    $self->set({
        'GrandTotal'     => $Formatter->roundTotal($GrandTotal),
        'TotalTax'       => $Formatter->round($TotalTax),
        'TotalBeforeTax' => $Formatter->round($GrandTotal - $TotalTax),
        'HasDiscounts'   => $HasDiscounts,
    });

    # substract pre paid lineitems from GrandTotal (like gift coupons wich do not affect the tax calculation)
    my $aPrePaymentLineItems = $self->get('PrePayments');
    my $PrePaymentValue = 0;
    foreach my $LineItem (@$aPrePaymentLineItems){
        $LineItem->recalculate($self, $hValues);
        $PrePaymentValue = $LineItem->get('LineItemPrice');
    }

    # set the grand total again because now we have the correct lineitemprice for
    $self->set({'GrandTotal' => $Formatter->roundTotal($self->get('GrandTotal') + $PrePaymentValue)}) if $PrePaymentValue;


    my $Payment = $self->get('Payment');
    if ($Payment){
        # delete payment only if we have nothing to pay
        # (100% paymentdiscounts are ignored)
        $Payment->delete if ($self->get('PaymentSubTotal') <= 0) and ($self->get('GrandTotal') <= 0);
    }

    TriggerHook( 'AfterRecalculateLineItemContainer', {
        'Object' => $self,
        'Values' => $hValues
    });

    return;
}

# EPG-16054 (also EPG-13856): have to add Tax if lineitem already exists
sub _addLineItemTax {
    my $self = shift;
    my ($LineItem) = @_;
    # add tax line item
    unless( $LineItem->instanceOf('LineItemTax') || $LineItem->instanceOf('LineItemTaxDigital') ) {
        my $TaxLineItem = $self->_insertTaxForLineItem($LineItem);
        # recalculate, because the TaxLineitem was maybe added and not recalculated before
        $TaxLineItem->recalculate($self->container()) if defined $TaxLineItem;
    }
}

# inset tax for a lineitem
# choose lineitemtaxdigital if we have a taxclass digital set on the item
sub _insertTaxForLineItem {
    my $self = shift;
    my ($LineItem) = @_;

    my $TaxLineItem;

    # add tax lineitem digital
    if ($LineItem->get('TaxClassDigital')) {
        my $TaxClassDigital = $LineItem->get('TaxClassDigital');
        my $Price = $LineItem->get('LineItemPrice');
        $TaxLineItem = $self->addLineItemTaxDigital($TaxClassDigital) if defined $TaxClassDigital and $Price and $self->id == $self->businessContainer->id;
    } else{
    # add normal tax lineitem
        my $TaxClass = $LineItem->get('TaxClass');
        my $Price = $LineItem->get('LineItemPrice');
        $TaxLineItem = $self->addLineItemTax($TaxClass) if defined $TaxClass and $Price and $self->id == $self->businessContainer->id;
    }

    return $TaxLineItem;

}

#========================================================================================
# function     formatter
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Formatter = $Container->formatter;
#----------------------------------------------------------------------------------------
# description  Returns the shop money formatter for the currency of container.
#----------------------------------------------------------------------------------------
# return       $Formatter | money formatter | object
#========================================================================================
sub formatter {
    my $self = shift;
    return $self->getSite()->formatter($self->get('CurrencyID'));
}

#========================================================================================
# function     container
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Me = $Container->container;
#----------------------------------------------------------------------------------------
# description  Returns the container it self. Used for line item method
#               <function DE_EPAGES::Order::API::Object::LineItem::container>.
#----------------------------------------------------------------------------------------
# return       $Me | self (current object) | object
#========================================================================================
sub container { return shift; }

#========================================================================================
# function     businessContainer
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Me = $businessContainer->container;
#----------------------------------------------------------------------------------------
# description  Returns the container it self if container is not child of a lineitem.
#               <function DE_EPAGES::Order::API::Object::LineItem::container>.
#----------------------------------------------------------------------------------------
# return       $Me | self (current object) | object
#========================================================================================
sub businessContainer {
    my $self = shift;
    return $self->parent->instanceOf('LineItem') ? $self->parent->container->businessContainer : $self;
}

#========================================================================================
# function     taxRate
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Rate = $Container->taxRate($TaxClass);
#----------------------------------------------------------------------------------------
# description  returns the tax rate for the given class, the area is saved at this container
#----------------------------------------------------------------------------------------
# input        $TaxClass | tax class | object
# input        $UseDefaultTaxArea | use default tax area | boolean
# return       $Rate | tax rate | float
#========================================================================================
sub taxRate {
    my $self = shift;
    my ($TaxClass,$UseDefaultTaxArea) = @_;

    TestObject('TaxClass', $TaxClass);

    my $TaxMatrix = $self->get('TaxMatrix');
    my $TaxArea   = $UseDefaultTaxArea ? $TaxMatrix->defaultTaxArea : $self->get('TaxArea');

    # test if we have the correct tax class for the matrix
    # tax not same can be the error.
    # correct the tax class with the backup tax class by using the alias
    unless($TaxClass->isChildOf($TaxMatrix)){
        $TaxClass = $TaxMatrix->child($TaxClass->alias);
    }

    unless (defined $TaxArea) {
        $TaxArea = $TaxMatrix->fallbackTaxArea;
        $self->set({TaxArea => $TaxArea});
    }
    return $TaxMatrix->taxRate($TaxArea, $TaxClass, $self);
}

#========================================================================================
# function     clone
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Clone = $Container->clone( $hInfo );
# syntax       $Clone = $Container->clone( $hInfo, $Recursive );
#----------------------------------------------------------------------------------------
# description  Inserts a new lineitem container into the database as copy from old lineitem container.
#----------------------------------------------------------------------------------------
# input        $hInfo | hash with mandatory attribute values.
#                       <ul>
#                         <li>ClassID - class id - int</li>
#                         <li>Alias - alias - varchar(255)</li>
#                         <li>ParentID - parent objectid - int</li>
#                         <li>Inherit - (optional) inherit permissions from parent object - boolean</li>
#                       </ul> | hash ref
# input        $Recursive | clone child object too (optional, default is false) | boolean
# return       $Clone | the new object | object
#========================================================================================
sub clone {
    my $self = shift;
    my ($hInfo, $Recursive) = @_;
    $Recursive = 0 unless defined $Recursive;

    my $Clone = $self->SUPER::clone($hInfo, $Recursive);

    $Clone->container->set({'TaxArea' => $self->container->get('TaxArea')}) if $self->container->get('TaxArea');
    return $Clone;
}


#========================================================================================
# function     taxRateDigital
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Rate = $Container->taxRateDigital($TaxClassDigital);
#----------------------------------------------------------------------------------------
# description  returns the tax rate digital for the given class, the area is saved at this container
#----------------------------------------------------------------------------------------
# input        $TaxClassDigital | tax class digital | object
# return       $Rate | tax rate | float
#========================================================================================
sub taxRateDigital {
    my $self = shift;
    my ($TaxClassDigital) = @_;
    TestObject('TaxClassDigital', $TaxClassDigital);

    my $TaxMatrixDigital = $self->get('TaxMatrixDigital');
    my $TaxAreaDigital   = $self->get('TaxAreaDigital');
    unless (defined $TaxAreaDigital) {
        $TaxAreaDigital = $TaxMatrixDigital->defaultTaxAreaDigital;
        $self->set({'TaxAreaDigital' => $TaxAreaDigital});
    }

    return $TaxMatrixDigital->taxRateDigital($TaxAreaDigital, $TaxClassDigital);
}

#========================================================================================
# function     netPricesForTax
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hNetPrices = $Container->netPricesForTax;
#----------------------------------------------------------------------------------------
# description  get prices for all lineitems, sorted by TaxClassID into $hNetPrices->{'Total'}
#               $hNetPrices->{'Total'} is used to calculate the tax amount for each TaxClass
#               other informations in $hNetPrices can be used for debug informations
#----------------------------------------------------------------------------------------
# return       $hNetPrices | hash ref of line item prices for tax rates
#                             see example in t/unit/SplitableTaxes
#               <ul>
#                   <li>Total            - all prices for calculating tax for key TaxClassID - ref.array/hash
#                   <li>LineItem         - prices for taxable items other than payment, shipping - ref.array/hash
#                   <li>LineItemPayment  - prices for payment lineitems - ref.array/hash
#                   <li>LineItemPaymentDiscount - prices for payment discount lineitems - ref.array/hash
#                   <li>LineItemShipping - prices for shipping lineitems - ref.array/hash
#                   <li>LineItemOther    - total price for lineitems with no TaxClassID - float
#                   <li>LineItemDiscount - prices for discount lineitems - ref.array/hash
#                   <li>LineItemStatic   - prices for static lineitems - ref.array/hash
#               </ul> | ref.hash
#========================================================================================
sub netPricesForTax {
    my $self = shift;
    my $Deprecated = $] < PERL538 ? 1 : 0;
    $Deprecated = 0 if $Deprecated and $ENV{USE_NEW_LINEITEMCONTAINER};

    my $Shop = $self->getSite();

    # can we split shipping and payment taxes? This is dependent from the target country if we have active del. thresholds
    my ($SplitShippingTaxes, $SplitPaymentTaxes) = @{ ShouldSplitShippingAndPaymentTax($Shop, $self) };

    # do we have a taxmethod from an external tax cartridge on the shop? Then use settings from the method.
    my $TaxMethod = $Shop->get('TaxMethod');
    my $DisableShippingTaxesByTaxMethod = 0;
    my $DisablePaymentTaxesByTaxMethod = 0;

    if (defined $TaxMethod) {
        $SplitShippingTaxes = 0 if $TaxMethod->disableShippingSplit;
        $SplitPaymentTaxes = 0 if $TaxMethod->disablePaymentSplit;
        $DisableShippingTaxesByTaxMethod = 1 unless $TaxMethod->isShippingTaxable($self);
        $DisablePaymentTaxesByTaxMethod = 1 unless $TaxMethod->isPaymentTaxable($self);
    }

    my $hNetPrices = {};     # container for all taxable values (see return)
    my $hNetPricesTotal = {}; # container for total net values for calculating the tax amount later (TaxClassID => NetPrice)
    my %ItemNetPrices;       # LineItemSales prices (TaxClassID => NetPrice)
    my %ItemNetPricesWithDisabledSplitting; # LineItemSales prices from products, wich are excluded from shipping tax splitting (TaxClassID => NetPrice)
    my %DiscountNetPrices;   # LineItemDiscount and LineItemSalesDiscount prices (TaxClassID => NetPrice)
    my %CouponWithShippingNetPrices;      # coupon prices if coupon is also valid for shipping costs
    my %StaticNetPrices;      # LineItemStatic prices (TaxClassID => NetPrice)
    my $OtherNetPrice = 0;    # net prices without a tax class
    my $DisableShipping = 0;  # shipping can be disabled by a LineItemSalesDiscount object e.g. a coupon for free shipping
    my $TaxClassLineItem = {}; # net prices by lineitemid for reducing prices by discounts

    # PRODUCTS / COUPONS / BASKET DISCOUNTS:  calculate tax for products, coupons and discounts
    my $aTaxables = $self->get('TaxableLineItems');

    #splitting is only enabled, if we have lineitems with taxrates > 0% (EPG-21596)
    my $ShippingAndPaymentSplitting=0 ;
    foreach my $LineItem (@$aTaxables) {
        next unless $LineItem->instanceOf('LineItemSales')
                    or $LineItem->instanceOf('LineItemSalesDiscount')
                    or $LineItem->instanceOf('LineItemDiscount');

        # add sales and discount line item prices etc. to the hNetPricesTotal and the debug hash
        # also test if we have a discount to disable the shipping; shipping and discount is then ignored for the tax calculation
        if ($LineItem->instanceOf('LineItemDiscount') or $LineItem->instanceOf('LineItemSalesDiscount')) {
            if (not $DisableShipping and $LineItem->instanceOf('LineItemSalesDiscount')) {
                $DisableShipping = $LineItem->get('IsForShipping');
                next if $DisableShipping;
            }

            # add coupon with enabled shipping to an extra hash and add to the globals later
            if((not $DisableShipping) and $LineItem->instanceOf('LineItemCoupon') and $LineItem->get('InclShippingCosts')) {
                _AddNetPrice($LineItem,{},\%CouponWithShippingNetPrices);
            } else {
            # add normal coupon or discount
                _AddNetPrice($LineItem,$hNetPricesTotal,\%DiscountNetPrices,\$OtherNetPrice);
            }

        } else {
            # only add items by id here to TaxClassLineItem hash
            $TaxClassLineItem->{$LineItem->id} = $LineItem->get('LineItemPrice');
        }
    }

    LogDebug('CouponWithShippingNetPrices: ', \%CouponWithShippingNetPrices );

    my $Formatter = $self->formatter;
    my $SplitedDiscountPrices = $TaxClassLineItem;

    if ($OtherNetPrice) {
        # UNDEFINED: split undefined tax rates to all existing
        #use TaxClassLineItem hash for split discounts because we have to reduce also the ItemNetPricesWithDisabledSplitting values!
        $SplitedDiscountPrices = _SplitDiscount($TaxClassLineItem,$TaxClassLineItem, $OtherNetPrice, $Formatter);

        # reset other price wich was splitted before over the lineitems
        $hNetPrices->{'LineItemOther'} = $OtherNetPrice;
        $OtherNetPrice = 0;
    }

    # walk trough the items and add items after discounts to the hashes for calculating shipping and payment taxes
    # we have to add lineites not earlier because discounts are not added already
    foreach my $LineItem (@$aTaxables) {
        next unless exists $SplitedDiscountPrices->{$LineItem->id};

        # add items with disabled shipping split ot tax rates=0 for later reducing the price
        if(not $LineItem->get('IsShippingRequired') ){# or ($LineItem->get('TaxRate') == 0) ){
            _AddNetPrice($LineItem,{},\%ItemNetPricesWithDisabledSplitting, {}, $SplitedDiscountPrices->{$LineItem->id});
        }

        # add normal item prices for tax classes
        _AddNetPrice($LineItem,$hNetPricesTotal,\%ItemNetPrices, \$OtherNetPrice, $SplitedDiscountPrices->{$LineItem->id});

        # roll back EPG-21596 (this was related only for selling gold)
        #$ShippingAndPaymentSplitting = 1 if $LineItem->get('TaxRate') != 0;

        # (UNITY-8435) enables again bc. the tax items follows the main items in a correct way
        $ShippingAndPaymentSplitting = 1;
    }

    # debug shipping and payment splitting
    LogDebug('splitting for shipping and payment: ',{'splitting for shipping enabled by shop country: '=> $SplitShippingTaxes,
                                                     'splitting for payment enabled by shop country: ' => $SplitPaymentTaxes ,
                                                     'splitting for shipping and payment enabled by basket items: '=> $ShippingAndPaymentSplitting,
                                                     'taxation of shipping disabled by external tax method' => $DisableShippingTaxesByTaxMethod,
                                                     'taxation of payment disabled by external tax method' => $DisablePaymentTaxesByTaxMethod,
                                                    });

    # EPG-21596 Pro-rata tax for delivery and payment: Do not prorate in case of products with 0% VAT
    # disable splitting for payment and shipping if only products with 0% tax rate are in the basket
    if ($ShippingAndPaymentSplitting == 0) {
        $SplitShippingTaxes = 0;
        $SplitPaymentTaxes = 0;
    }

    # SHIPPING: add shipping line item and split shipping for tax classes
    # ignore shipping if disabled by an other lineitem e.g. coupon for shipping
    my %ShippingNetPrices;   # shipping tax values

    if (not $DisableShippingTaxesByTaxMethod){
        $self->_addLineItemShippingNetPrice($hNetPricesTotal,\%ShippingNetPrices,$SplitShippingTaxes,\$OtherNetPrice,\%ItemNetPricesWithDisabledSplitting) unless $DisableShipping;
    }

    #COUPON WHICH ARE VALID FOR SHIPPING COSTS
    # we split coupon here because the coupon price was calculated with shipping, so we have to reduce the shipping also for tax calculation
    if (exists $CouponWithShippingNetPrices{'NoTaxClass'}){
        $hNetPricesTotal = _SplitDiscount($hNetPricesTotal,$hNetPricesTotal, $CouponWithShippingNetPrices{'NoTaxClass'}, $Formatter);
    }

    # PAYMENT: add payment and paymentdiscount afterwards
    my %PaymentNetPrices;    # payment tax values
    my %PaymentDiscountNetPrices; # payment discount tax values
    # enable splitting over tax classes for payments
    if (not $DisablePaymentTaxesByTaxMethod){
        $self->_addLineItemPaymentNetPrice(\%ItemNetPrices,$hNetPricesTotal,\%PaymentNetPrices,$SplitPaymentTaxes,\$OtherNetPrice, \%ItemNetPricesWithDisabledSplitting);
    }

    # PAYMENTDISCOUNT: payment discount is for the whole basket and is splitted over tax classes, because this lineitem has no tax class
    $self->_addLineItemNetPrice($hNetPricesTotal,'PaymentDiscount',\%PaymentDiscountNetPrices,\$OtherNetPrice);

    # STATIC LINEITEMS: add all static lineitems like eco paricipation
    foreach my $LineItem (@$aTaxables) {
        next unless $LineItem->instanceOf('LineItemStatic');
        _AddNetPrice($LineItem,$hNetPricesTotal,\%StaticNetPrices, \$OtherNetPrice);
    }

    # UNDEFINED: split unknown items with undefined tax rates to all existing like prices of elements with no tax class
    if ($OtherNetPrice) {
        $hNetPricesTotal = _SplitDiscount($hNetPricesTotal,\%ItemNetPrices, $OtherNetPrice, $Formatter);
        $hNetPrices->{'LineItemOther'} += $OtherNetPrice;
    }

    # build hash with all items included
    $hNetPrices->{'LineItem'} = _sortKeys(\%ItemNetPrices, $Deprecated);
    $hNetPrices->{'LineItemCouponWithShipping'} = _sortKeys(\%CouponWithShippingNetPrices, $Deprecated);
    $hNetPrices->{'LineItemDiscount'} = _sortKeys(\%DiscountNetPrices, $Deprecated);
    $hNetPrices->{'LineItemPayment'} = _sortKeys(\%PaymentNetPrices, $Deprecated);
    $hNetPrices->{'LineItemPaymentDiscount'} = _sortKeys(\%PaymentDiscountNetPrices, $Deprecated);
    $hNetPrices->{'LineItemShipping'} = _sortKeys(\%ShippingNetPrices, $Deprecated);
    $hNetPrices->{'LineItemStatic'} = _sortKeys(\%StaticNetPrices, $Deprecated);
    $hNetPrices->{'LineItemWithDisabledSplitting'} = _sortKeys(\%ItemNetPricesWithDisabledSplitting, $Deprecated);
    $hNetPrices->{'TaxClassLineItem'} = $TaxClassLineItem;
    $hNetPrices->{'TaxClassLineItemsAfterDiscounts'} = _sortKeys($SplitedDiscountPrices, $Deprecated);
    $hNetPrices->{'Total'} = _sortKeys($hNetPricesTotal, $Deprecated);

    LogDebug('prices for tax calculation for each tax class: ', $hNetPrices );
    LogDebug('shipping for basket disabled by a lineitem : ', $DisableShipping );

    return $hNetPrices;
}

# add taxes to the hTotal hash for a shipping line item
# shipping values can be splitted to different tax classes!
sub _addLineItemShippingNetPrice {
    my $self = shift;
    my ($hNetPricesTotal,$hNet,$SplitShippingTaxes,$rOtherNetPrice,$hItemNetPricesWithDisabledSplitting) = @_;
    my $Deprecated = $] < PERL538 ? 1 : 0;
    $Deprecated = 0 if $Deprecated and $ENV{USE_NEW_LINEITEMCONTAINER};

    my $ShippingLineItem = $self->get('Shipping');
    return unless $ShippingLineItem;

    # basket sum with discounts, coupons but without shipping and payment
    #my $DiscountsSubTotal = $self->get('DiscountsSubTotal');
    my $DiscountsSubTotal;
    $DiscountsSubTotal += $_ foreach values( %{$hNetPricesTotal});

    if ($DiscountsSubTotal and $SplitShippingTaxes) {

        # do not use shipping free items for split shipping calculation
        # reduce the DiscountsSubTotal, only relevant items for splitting are used
        my $ReducedDiscountsSubTotal = $DiscountsSubTotal;
        my $NetPricesTotalKeys = _getKeys($hNetPricesTotal, $Deprecated);
        foreach my $TaxClassID (@$NetPricesTotalKeys) {
            $ReducedDiscountsSubTotal -= $hItemNetPricesWithDisabledSplitting->{$TaxClassID} if $hItemNetPricesWithDisabledSplitting->{$TaxClassID};
        }

        foreach my $TaxClassID (@$NetPricesTotalKeys) {
            # split lineitemprice only for known TaxClassIDs
            next unless $TaxClassID;
            # reduce the sum for tis tax class, if we have items with disabled shipping splitting
            my $TaxClassPrice = $hNetPricesTotal->{$TaxClassID};
            $TaxClassPrice -= $hItemNetPricesWithDisabledSplitting->{$TaxClassID} if $hItemNetPricesWithDisabledSplitting->{$TaxClassID};

            # calculate splited shipping price
            next if $ReducedDiscountsSubTotal == 0; # prevent division by zero
            my $ShippingTaxRate = $ShippingLineItem->get('LineItemPrice') * $TaxClassPrice / $ReducedDiscountsSubTotal;
            $hNet->{$TaxClassID} += $ShippingTaxRate;
            _setKeys($hNet, $TaxClassID) unless $Deprecated;

            # add calculated shipping rate to the global hash for a tax calculation later
            $hNetPricesTotal->{$TaxClassID} += $ShippingTaxRate;
       }
    } else {
        $self->_addLineItemNetPrice($hNetPricesTotal,'Shipping',$hNet,$rOtherNetPrice);
    }

    return;
}

# add taxes to the hTotal hash for a payment line item
# payment values can be splitted to different tax classes!
sub _addLineItemPaymentNetPrice {
    my $self = shift;
    my ($ItemNetPrices, $hNetPricesTotal,$hNet,$SplitPaymentTaxes,$rOtherNetPrice, $hItemNetPricesWithDisabledSplitting) = @_;
    my $Deprecated = $] < PERL538 ? 1 : 0;
    $Deprecated = 0 if $Deprecated and $ENV{USE_NEW_LINEITEMCONTAINER};

    my $PaymentLineItem = $self->get('Payment');
    return unless $PaymentLineItem;

    # calculate sum for all added elements until now
    my $Sum = 0;

    # calculate splitting ratio only for original line items
    $Sum += $_ foreach values %{$ItemNetPrices};

    if ($Sum and $SplitPaymentTaxes) {
        my $NetPricesTotalKeys = _getKeys($hNetPricesTotal, $Deprecated);
        foreach my $TaxClassID (@$NetPricesTotalKeys) {
            # split lineitemprice only for known TaxClassIDs
            next unless $TaxClassID;
            next unless defined $PaymentLineItem->get('LineItemPrice');
            next if $Sum == 0; #prevent division by zero

            # calculate splitting ratio only for original line items
            my $PaymentTaxRate = $PaymentLineItem->get('LineItemPrice') * $ItemNetPrices->{$TaxClassID} / $Sum;
            $hNet->{$TaxClassID} += $PaymentTaxRate;
            _setKeys($hNet, $TaxClassID) unless $Deprecated;
            $hNetPricesTotal->{$TaxClassID} += $PaymentTaxRate;
       }
    } else {
        $self->_addLineItemNetPrice($hNetPricesTotal,'Payment',$hNet,$rOtherNetPrice);
    }

    return;
}

# add taxes to the hTotal hash for lineitems by name
sub _addLineItemNetPrice {
    my $self = shift;
    my ($hTotal,$Name,$hNet,$rOtherNetPrice) = @_;

    my $LineItem = $self->get($Name);
    return unless $LineItem;
    _AddNetPrice($LineItem,$hTotal,$hNet,$rOtherNetPrice);
    return;
}

# add taxes to the hTotal hash
sub _AddNetPrice {
    my ($LineItem,$hTotal,$hNet,$rOtherNetPrice, $Price) = @_;
    my $Deprecated = $] < PERL538 ? 1 : 0;
    $Deprecated = 0 if $Deprecated and $ENV{USE_NEW_LINEITEMCONTAINER};

    return unless $LineItem;
    my $TaxClass = $LineItem->get('TaxClassDigital') ? $LineItem->get('TaxClassDigital') : $LineItem->get('TaxClass');

    $Price = defined $Price ? $Price: $LineItem->get('LineItemPrice');

    if (defined $TaxClass) {
        my $TaxClassID = $TaxClass->id;
        _setKeys($hNet, $TaxClassID) unless $Deprecated;
        _setKeys($hTotal, $TaxClassID) unless $Deprecated;

        $hTotal->{$TaxClassID} //= 0;

        if ($Price) {
            $hNet->{$TaxClassID} += $Price;
            $hTotal->{$TaxClassID} += $Price;
        }
    } elsif ($Price) {
        _setKeys($hNet, 'NoTaxClass') unless $Deprecated;
        $$rOtherNetPrice += $Price;
        $hNet->{'NoTaxClass'} += $Price;
    }
    return;
}

# split discounts with no tax information over all other tax classes
sub _SplitDiscount {
    my ($hNetPricesTotal, $OriginalNetPrices, $SplitPrice, $Formatter) = @_;
    my %NetPrices = %{$hNetPricesTotal};
    my $Deprecated = $] < PERL538 ? 1 : 0;
    $Deprecated = 0 if $Deprecated and $ENV{USE_NEW_LINEITEMCONTAINER};

    return $hNetPricesTotal if $SplitPrice == 0;

    my $NetPricesTotalKeys = _getKeys($hNetPricesTotal, $Deprecated);
    my $OriginalNetPricesKeys = _getKeys($OriginalNetPrices, $Deprecated);

    # sum of line items to split
    if (scalar @{$NetPricesTotalKeys} == 1 ) {
        $NetPrices{$NetPricesTotalKeys->[0]} += $SplitPrice;
    } else {
        my $Sum = 0;
        # calculate splitting ratio only for original line items
        $Sum += $_ foreach values %{$OriginalNetPrices};

        # split sum
        my %Split;
        if( $Sum != 0 ) {
            # calculate splitting ratio only for original line items
            $Split{$_} = ($SplitPrice * $OriginalNetPrices->{$_}) / $Sum foreach @$OriginalNetPricesKeys;
        }
        else {
            $Split{$_} = 0 foreach @$NetPricesTotalKeys;
        }
        my $SumSplit = 0;
        $SumSplit += $_ foreach values %Split;

        # possible split diff (through round) save to most value of tax amount
        my $SumDiff = $SplitPrice - $SumSplit;
        if ($SumDiff) {
            my @keysSplit = keys %Split;
            # get max
            my $maxKey = shift @keysSplit;
            foreach my $Key (@keysSplit) {
                $maxKey = $Key if $Split{$maxKey} < $Split{$Key};
            }
            $Split{$maxKey} += $SumDiff;
        }
        $NetPrices{$_} += $Split{$_} foreach keys %Split;
    }
    return \%NetPrices;
}

#========================================================================================
# function     addLineItemTax
#----------------------------------------------------------------------------------------
# syntax       $TaxLineItem = $Container->addLineItemTax($TaxClass);
#----------------------------------------------------------------------------------------
# description  Adds a tax lineitem for the given tax class, if not exists yet.
#----------------------------------------------------------------------------------------
# input        $TaxClass | tax class | object
# return       $TaxLineItem | tax line item | object
#========================================================================================
sub addLineItemTax {
    my $self = shift;
    my ($TaxClass) = @_;
    my $Shop = $self->getSite();
    my $Parent = $self->parent;
    return if $Shop->get('VATreleased') && $Shop->get('TaxModel') == TAX_MODEL_GROSS;
    return if $Shop->get('TaxCalculationAfterAddress') && $Shop->get('TaxModel') == TAX_MODEL_NET && $Parent->instanceOf('Basket') && !$Parent->get('IsAddressOK');

    my $TaxRate = $self->taxRate($TaxClass);

    return unless $TaxRate;

    my $LineItemClass = $TaxClass->get('TaxMatrix')->get('LineItemClass');
    $LineItemClass = LoadClassByAlias('LineItemTax') unless defined $LineItemClass;
    my $TaxLineItem = $self->setLineItem($LineItemClass, {
        'TaxClass' => $TaxClass
    });
    local $DE_EPAGES::Object::API::Language::LANGUAGEID = $self->get('LanguageID');
    $TaxLineItem->set({
        'Name'    => $TaxClass->get('NameOrAlias'),
        'TaxRate' => $TaxRate,
    });

    return $TaxLineItem;
}

#========================================================================================
# function     addLineItemTaxDigital
#----------------------------------------------------------------------------------------
# syntax       $DigitalTaxLineItem = $Container->addLineItemTaxDigital($TaxClassDigital);
#----------------------------------------------------------------------------------------
# description  Adds a tax lineitem digital for the given tax class digital, if not exists yet.
#----------------------------------------------------------------------------------------
# input        $TaxClassDigital | tax class digital | object
# return       $DigitalTaxLineItem | tax line item digital | object
#========================================================================================
sub addLineItemTaxDigital {
    my $self = shift;
    my ($TaxClassDigital) = @_;

    my $Shop = $self->getSite();
    my $Parent = $self->parent;

    # we have gross and vatreleased -> no tax
    return if $Shop->get('VATreleased') && $Shop->get('TaxModel') == TAX_MODEL_GROSS;

    # tax calculation after address input and net
    return if $Shop->get('TaxCalculationAfterAddress') && $Shop->get('TaxModel') == TAX_MODEL_NET && $Parent->instanceOf('Basket') && !$Parent->get('IsAddressOK');

    # get rate
    my $TaxRateDigital = $self->taxRateDigital($TaxClassDigital);

    # add the lineitem with the correct class
    my $LineItemClass = $TaxClassDigital->get('TaxMatrixDigital')->get('LineItemClass');
    $LineItemClass = LoadClassByAlias('LineItemTaxDigital') unless defined $LineItemClass;
    my $TaxLineItemDigital = $self->setLineItem($LineItemClass, {
        'TaxClassDigital' => $TaxClassDigital
    });
    local $DE_EPAGES::Object::API::Language::LANGUAGEID = $self->get('LanguageID');
    $TaxLineItemDigital->set({
        'Name'    => $TaxClassDigital->get('NameOrAlias'),
        'TaxRate' => $TaxRateDigital,
    });

    return $TaxLineItemDigital;
}

sub addTaxLineItem {
    warn 'LineItemContainer->addTaxLineItem() is obsolete remove from it simple from your code, LineItemContainer->insertLineItem() does it.';
}

#========================================================================================
# function     addLineItemShipping
#----------------------------------------------------------------------------------------
# syntax       $ShippingLineItem = $Container->addLineItemShipping($LineItems);
#----------------------------------------------------------------------------------------
# description  Adds a shipping lineitem, if not exists yet and shipping is required.
#----------------------------------------------------------------------------------------
# input        $LineItems | list of lineitems (optional) | ref.array.SalesLineItem
# return       $ShippingLineItem | shipping line item | object
#========================================================================================
sub addLineItemShipping {
    my $self = shift;
    my ($LineItems) = @_;

    # EPG-21726: recursion check
    local $self->{'in_addLineItemShipping'} = $self->{'in_addLineItemShipping'} // 0;
    if( $self->{'in_addLineItemShipping'} > 1 ) {
        Error('LineItemShipping_Recursion', { LineItemContainer => $self->pathString });
    }
    else {
        $self->{'in_addLineItemShipping'}++;
    }

    $LineItems = $self->get('LineItems') unless defined $LineItems;

    # shipping line item
    my $ShippingLineItem = $self->get('Shipping');
    if ($self->needShipping($LineItems)) {
        # insert shipping method, find default
        my $aShippingMethods;
        if($self->parent->instanceOf('Basket')){
            $aShippingMethods = $self->get('ShippingMethods');
        } else {
            $aShippingMethods = $self->getSite()->get('ShippingMethods');
        }
        if (scalar @$aShippingMethods) {
            my $ShippingMethod = defined $ShippingLineItem ? $ShippingLineItem->get('ShippingMethod') : $self->get('LastShippingMethod');
            $ShippingMethod = undef if defined $ShippingMethod and not grep { $_->id == $ShippingMethod->id } @$aShippingMethods;
            if (not defined $ShippingLineItem or not defined $ShippingMethod) {
                # select default
                if(not defined $ShippingMethod){
                    ($ShippingMethod) = grep { $_->get('IsDefault') } @$aShippingMethods;
                }
                # if not default select first
                $ShippingMethod = $aShippingMethods->[0] unless defined $ShippingMethod;
                # add new shipping
                Transaction(sub {
                    $ShippingLineItem->delete if defined $ShippingLineItem;
                    $ShippingLineItem = $ShippingMethod->addToContainer($self, {});
                });
            } # else can use old shipping
        } else {
            Warning('NoShipping');
        }
    } elsif (defined $ShippingLineItem) {
        $ShippingLineItem->delete;
        $ShippingLineItem = undef;
    }
    return $ShippingLineItem;
}

#========================================================================================
# function     addLineItemPayment
#----------------------------------------------------------------------------------------
# syntax       $PaymentLineItem = $Container->addLineItemPayment($LineItems);
#----------------------------------------------------------------------------------------
# description  Adds a payment lineitem, if not exists yet and payment is required.
#----------------------------------------------------------------------------------------
# input        $LineItems | list of lineitems (optional) | ref.array.SalesLineItem + ShippingLineItem
# return       $PaymentLineItem | payment line item | object
#========================================================================================
sub addLineItemPayment {
    my $self = shift;
    my ($LineItems) = @_;
    if (not defined $LineItems) {
        $LineItems = $self->get('LineItems');
        my $Shipping = $self->get('Shipping');
        push @$LineItems, $Shipping if defined $Shipping;
    }
    my $PaymentLineItem =$self->get('Payment');
    if ($self->needPayment($LineItems)) {
        # insert payment method, find default
        my $aPaymentMethods;
        if($self->parent->instanceOf('Basket')){
            $aPaymentMethods = $self->get('PaymentMethods');
        } else {
            $aPaymentMethods = $self->getSite()->get('PaymentMethods');
        }
        if (scalar @$aPaymentMethods) {
            my $PaymentMethod = defined $PaymentLineItem ? $PaymentLineItem->get('PaymentMethod') : $self->get('LastPaymentMethod');
            # is current available ?
            $PaymentMethod = undef if defined $PaymentMethod and not grep { $_->id == $PaymentMethod->id } @$aPaymentMethods;
            # can be current pre-assigned ? (payment methods with special flows can be problematic if pre-assigned)
            $PaymentMethod = undef if defined $PaymentMethod and not $PaymentMethod->get('CanBeDefaultPaymentMethod');
            if (not defined $PaymentLineItem or not defined $PaymentMethod) {
                # select default
                if (not defined $PaymentMethod) {
                    ($PaymentMethod) = grep { $_->get('IsDefault') } @$aPaymentMethods;
                }
                # if not default select first
                $PaymentMethod = $aPaymentMethods->[0] unless defined $PaymentMethod;
                Transaction(sub {
                    $PaymentLineItem->delete if defined $PaymentLineItem;
                    $PaymentLineItem = $PaymentMethod->addToContainer($self, {});
                });
            }
        } else {
            Warning('NoPayment');
        }
    } elsif (defined $PaymentLineItem) {
        $PaymentLineItem->delete;
        $PaymentLineItem = undef;
    }
    return $PaymentLineItem;
}

sub _setKeys {
    my ($Hash, $Key) = @_;

    # find key with '/' => $Sequence: '/z/t/r/'
    my $Sequence = (grep {m,/,} keys %$Hash)[0];
    # do nothing if $Sequence already contains $Key
    return if $Sequence =~ m,/$Key/,;

    delete $Hash->{$Sequence};
    # add new key 'k' to $Sequence: '/z/t/r/k/'
    $Sequence //= '/';
    $Sequence .= "$Key/";
    $Hash->{$Sequence} = 0;
}

sub _getKeys {
    my ($Hash, $Deprecated) = @_;
    my @Array;

    return [keys %$Hash] if $Deprecated;

    # find key with '/' => $Sequence: '/z/t/r/'
    my $Sequence = (grep {m,/,} keys %$Hash)[0];
    # split $Sequence into @Keys: ('z', 't', 'r')
    my @Keys = $Sequence ? split(/\//, $Sequence) : sort keys %$Hash;

    # create @Array with elements: $Key
    foreach my $Key (@Keys) {
        next unless defined $Hash->{$Key};
        push @Array, $Key;
    }

    return \@Array;
}

sub _sortKeys {
    my ($Hash, $Deprecated) = @_;
    my @Array;

    return $Hash if $Deprecated;

    # find key with '/' => $Sequence: '/z/t/r/'
    my $Sequence = (grep {m,/,} keys %$Hash)[0];
    # split $Sequence into @Keys: ('z', 't', 'r')
    my @Keys = $Sequence ? split(/\//, $Sequence) : sort keys %$Hash;

    # create @Array with elements: [$Key, $Hash->{$Key}]
    foreach my $Key (@Keys) {
        next unless defined $Hash->{$Key};
        push @Array, [$Key, $Hash->{$Key}];
    }

    return \@Array;
}

1;
