#========================================================================================
# package      DE_EPAGES::Order::API::Object::LineItemTaxDigital
# state        public
# base         DE_EPAGES::Order::API::Object::LineItem
# description  object interface for LineItemTaxDigital.
#========================================================================================
package DE_EPAGES::Order::API::Object::LineItemTaxDigital;
use base DE_EPAGES::Order::API::Object::LineItem;

use strict;
use DE_EPAGES::Core::API::Constants qw( PERL538 );
use DE_EPAGES::Shop::API::Constants qw ( TAX_MODEL_GROSS TAX_MODEL_NET );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemTaxDigital->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteLineItemTaxDigital' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteLineItemPayment | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemID - line item tax id - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteLineItemTaxDigital', {'Object' => $self, 'LineItemID' => $self->id });
    $self->SUPER::delete;
    return;
}

#========================================================================================
# function     recalculate
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LineItem->recalculate($Container, $hValues);
#----------------------------------------------------------------------------------------
# description  Re-calculates the tax of the line item container (e.g. Basket) after any
#               other line items or line item container attributes have changed.
#----------------------------------------------------------------------------------------
# input        $Container | line item container | object
# input        $hValues | values that have changed | ref.hash
#========================================================================================
sub recalculate {
    my $self = shift;
    my ($Container, $hValues) = @_;

    my $TaxClass  = $self->get('TaxClassDigital');
    return unless defined $TaxClass;

    my $BasePrice = $self->calculateBasePrice;
    # all items (products and shipping) of this tax rate are removed

    unless ($BasePrice) {
        $self->delete;
        return;
    }
    my $TaxMatrix = $self->get('TaxMatrixDigital');
    #my $DefaultTaxRate = $TaxMatrix->taxRateDigital($TaxMatrix->defaultTaxAreaDigital, $TaxClass);
    my $TaxArea = $self->get('TaxAreaDigital');
    my $NoTax = $Container->get('NoTax');

    my $ShipingCountryID =  $Container->get('ShippingCountryID');
    if (($Container->get('TaxModel') == TAX_MODEL_NET) and $ShipingCountryID) {
        $NoTax = 1 unless $TaxMatrix->taxAreaDigitalByCountry($ShipingCountryID);
    }

    # test if we have an exclusion by address and set prices to 0
    $NoTax = 1 unless $Container->get('BasketOrOrder')->taxAreaDigitalByAddress();


    my $TaxRate = defined $TaxArea ?
        ($NoTax ? 0 : $TaxMatrix->taxRateDigital($TaxArea, $TaxClass)) :
        $self->get('TaxRate');

    my $TaxAmount = $NoTax ? 0 : $self->calculateTaxAmount($BasePrice, $TaxRate);

    my $LineItemPrice = $self->calculateLineItemPrice($BasePrice, $TaxRate);

    $self->set({
        'BasePrice'    => $BasePrice,
        'TaxAmount'    => $TaxAmount,
        'LineItemPrice'=> $LineItemPrice,
        'TaxRate'      => $TaxRate,
    });
    return;
}

#========================================================================================
# function     calculateBasePrice
# state        private
#----------------------------------------------------------------------------------------
# syntax       $BasePrice = $LineItem->calculateBasePrice;
#----------------------------------------------------------------------------------------
# description  calculates sum of line item prices for dependend tax rate
#----------------------------------------------------------------------------------------
# return       $BasePrice | sum of line item prices with dependend tax rate | money
#========================================================================================
sub calculateBasePrice {
    my $self = shift;
    my $Deprecated = $] < PERL538 ? 1 : 0;
    $Deprecated = 0 if $Deprecated and $ENV{USE_NEW_LINEITEMCONTAINER};

    my $Container = $self->container;
    return 0 if $Container->getSite()->get('VATreleased') && $Container->getSite()->get('TaxModel') == TAX_MODEL_GROSS;

    # get all tax amounts and shipping tx amounts
    my $hNetPrices = $Container->netPricesForTax();

    my $NetPrice;
    if ($Deprecated) {
        $NetPrice = $hNetPrices->{'Total'}{$self->get('TaxClassDigital')->id} // 0;
    } else {
        my @TaxItem = grep { $_->[0] eq $self->get('TaxClassDigital')->id } @{$hNetPrices->{'Total'}};
        $NetPrice = @TaxItem ? $TaxItem[0][1] : 0;
    }

    return $NetPrice ? $self->formatter->round($NetPrice) : 0;
}

#========================================================================================
# function     calculateTaxAmount
# state        private
#----------------------------------------------------------------------------------------
# syntax       $TaxAmount = $LineItem->calculateTaxAmount($BasePrice, $TaxRate, $DefaultTaxRate);
#----------------------------------------------------------------------------------------
# description  calculates the tax amount for a tax rate
#----------------------------------------------------------------------------------------
# input        $BasePrice | sum of line item prices with dependend tax rate | money
# input        $TaxRate | tax rate for tax area | money
# input        $DefaultTaxRate | default tax rate for tax area | money
# return       $TaxAmount | tax amount for tax rate  | money
#========================================================================================
sub calculateTaxAmount {
    my $self = shift;
    my ($BasePrice, $TaxRate, $DefaultTaxRate) = @_;
    $DefaultTaxRate = $TaxRate unless defined $DefaultTaxRate;
    return 0 if $TaxRate == 0;
    return $self->formatter->round($BasePrice * $TaxRate) if $self->container->get('TaxModel') == TAX_MODEL_NET;
    return $self->formatter->round($BasePrice * $TaxRate * (1- ( $DefaultTaxRate/(1 + $DefaultTaxRate))));
}

#========================================================================================
# function     calculateLineItemPrice
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemPrice = $LineItem->calculateLineItemPrice($BasePrice, $TaxRateDigital, $DefaultTaxRate);
#----------------------------------------------------------------------------------------
# description  calculates the lineitem price tax
#----------------------------------------------------------------------------------------
# input        $BasePrice | sum of line item prices with dependend tax rate | money
# input        $TaxRate | tax rate for tax area | money
# input        $DefaultTaxRate | default tax rate for tax area | money
# return       $LineItemPrice | price for tax | money
#========================================================================================
sub calculateLineItemPrice {
    my $self = shift;
    my ($BasePrice, $TaxRate, $DefaultTaxRate) = @_;
    $DefaultTaxRate = $TaxRate unless defined $DefaultTaxRate;
    return $self->formatter->round($BasePrice * $TaxRate) if $self->container->get('TaxModel') == TAX_MODEL_NET;

    # prevent rounding problems with LineItemPrice (EPG-19270)
    my $Factor = $self->formatter->round((( 1- $DefaultTaxRate/( 1+$DefaultTaxRate ) )*( 1+$TaxRate ) -1),6);
    return $self->formatter->round( $BasePrice * $Factor);

}

#========================================================================================
# function     onDeleteClass
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Object->onDeleteClass;
#----------------------------------------------------------------------------------------
# description  class of this object will be deleted. Make an normal tax line item.
#========================================================================================
sub onDeleteClass {
    my $self = shift;
    $self->set({'Class'=>$self->class->baseClass});
}

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemTax = DE_EPAGES::Order::API::Object::LineItemTax->insert($hInfo);
# example      my $Class = LoadClassByAlias('LineItemTax');
#               my $LineItemTax = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertLineItemTax' is triggered.
#----------------------------------------------------------------------------------------
# input        $Package | class package | string
# input        $hInfo | attributes of new object | ref.hash
# return       $LineItemTax | new object | object
# hook         OBJ_InsertLineItemTax | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemID    - line item tax id - int
#               </ul>
#========================================================================================
sub insert {
    my ($Package, $hInfo) = @_;

    my $self = $Package->SUPER::insert($hInfo);
    TriggerHook('OBJ_InsertLineItemTaxDigital', {'Object' => $self, 'LineItemID' => $self->id });
    return $self;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemTaxDigital->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateLineItemTaxDigital' and 'OBJ_AfterUpdateLineItemTaxDigital'.
#----------------------------------------------------------------------------------------
# input        $hValues | values hash | ref.hash
# input        $LanguageID | language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateLineItemTaxDigital | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemTaxDigitalID - line item tax id - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateLineItemTaxDigital | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemTaxDigitalID - line item tax id - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set { ##no critic qw(ProhibitAmbiguousNames)
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateLineItemTaxDigital', {'Object'=>$self, 'Values'=> $hValues, 'LineItemTaxDigitalID' => $self->id, 'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdateLineItemTaxDigital',  {'Object'=>$self, 'Values'=> $hValues, 'LineItemTaxDigitalID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}

1;
