/**
 * Widget to administrate AI content for the scribe editor or a callback function. It gets the content of the editor or
 * as property *description* and provides the possibility to optimize it through AI (OpenAI)
 *
 * The `ui.presentationAiContent()` widget opens a dialog where the optimization is handled. After a successful optimization the result can be used to replace the currenct content inside the scribe editor.
 *
 * ### Examples
 * Add `ui.presentationAiConten` functionality to a cta button.
 *
 * JavaScript:
 *
 *
 *  ep('.ep-product-ai-description-btn').presentationAiContent({
 *      // Variant 1: Configuration using a callback function
 *      // onApply: function(content) {
 *      //   console.log('Write new content to console with a callback function:', content);
 *      // },
 *      // description: "#Description",
 *
 *      // Variant 2: Configuration using a scribe instance
 *      scribe: scribeEditor,
 *
 *.     // Configuration for both variants
 *      changeActionAI: "JSONOptimizeAiContent",
 *      changeActionConfirm: "JSONSaveSeenAIDialog",
 *      isConfirmed: #IF(#Shop.IsAIDialogConfirmed EQ "0")false#ELSE#SPACE[0]true#ENDIF
 *  });
 *
 * HTML:
 *
 *  <button class="ep-product-ai-description-btn">Optimize Description</button>
 *
 *
 * @class jQuery.ui.presentationAiContent
 * @extends jQuery.widget
 *
 * @uses jQuery
 * @uses jQuery.tmpl
 * @uses jQuery.ui.widget
 * @uses jquery/metaparse
 * @uses ep
 * @uses ep.fn.busy
 * @uses ep.ui.tooltip
 * @uses de_epages
 * @since 7.90.0
 */

/**
 * Opens an existing dialog after cleaning up its content (re-render, title ...) or
 * creates and initializes a dialog if not exists.
 *
 * @method open
 * @member jQuery.ui.presentationAiContent
 *
 * @since 7.90.0
 */

/**
 * Method to close optimization dialog after the new content was written to a existing
 * scribe instance or after callback function was called with the new content as argument.
 *
 * @method close
 * @member jQuery.ui.presentationAiContent
 *
 * @since 7.90.0
 */

/**
 * Cleanup method to remove events, elements from DOM and destroy used widgets.
 *
 * @method destroy
 * @member jQuery.ui.presentationAiContent
 *
 * @since 7.90.0
 */

/**
 * See `jQuery.ui.presentationAiContent` for details.
 *
 * @param none
 *
 * @method open
 * @member jQuery
 *
 * @since 7.90.0
 */

/**
 * See `jQuery.ui.presentationAiContent` for details.
 *
 * @param none
 *
 * @method close
 * @member jQuery
 *
 * @since 7.90.0
 */

/**
 * See `jQuery.ui.presentationAiContent` for details.
 *
 * @param none
 *
 * @method destroy
 * @member jQuery
 *
 * @since 7.90.0
 */

/**
 * @cfg {Object} scribe An object containing a scribe instance. (optional)
 */

/**
 * @cfg {Function} onApply Callback function executed on close with the new content as argument. (optional)
 */

/**
 * @cfg {String} description First content. (optional)
 */

/**
 * @cfg {Boolean} isConfirmed Indicates, if flag is already set, not to show the confirmation dialog again.
 */

/**
 * @cfg {String} changeActionAI ChangeAction to request server for optimized OpenAI content.
 */

/**
 * @cfg {Boolean} changeActionConfirm ChangeAction to send and save  *isConfirmed* functionality.
 */

/**
 * @cfg {Objet} dict Dictionary with all relevant UI text inside the dialogs (for structure see source code). (optional)
 */

/*
 * @copyright       © Copyright 2006-2023, epages GmbH, All Rights Reserved.
 *
 * @module          ep.ui.presentationAiContent
 */

/*jslint nomen: true*/
/*global define, setTimeout*/
define("de_epages/presentation/ui/ai-content", [
    "jquery",
    "ep/ui/dialog",
    "de_epages",
    "$dict!./ai-content",
    "$tmpl!./ai-content-optimize-dialog",
    "$tmpl!./ai-content-confirm-dialog",

    "jquery/i18n",
    "jquery/tmpl",
    "jquery/metaparse",
    "jquery/ui/widget",
    "ep/fn/busy",
    "ep/ui/tooltip"
], function($, ep, de_epages, aiContentDict, tmplOptimizeDialog, tmplConfirmDialog) {
    'use strict';

    /*
     * @dictionary       de_epages.presentation.ui.ai-content
     *
     * @translation      {BubbleHelpGenerateCurrentDescription}
     *                   {BubbleHelpGenerateNewDescription}
     *                   {BubbleHelpOptimizeCurrentDescription}
     *                   {BubbleHelpOptimizeNewDescription}
     *                   {CTAGenerateApply}
     *                   {CTAOptimizeApply}
     *                   {CTAReplaceApply}
     *                   {CTACopy}
     *                   {CTACopied}
     *                   {CTAGenerateDescription}
     *                   {CTAOptimizeDescription}
     *                   {DescriptionGenerate}
     *                   {DescriptionOptimize}
     *                   {DescriptionReplace}
     *                   {LabelGenerateDescription}
     *                   {LabelGenerateNewDescription}
     *                   {LabelOptimizeDescription}
     *                   {LabelOptimizeNewDescription}
     *                   {LabelReplaceCheckbox}
     *                   {TitelGenerateDescription}
     *                   {TitelOptimizeDescription}
     *                   {TitelReplaceDescription}
     *                   {RemainingTokens}
     *                   {RemainingTokensDescription}
     */

    /*
     * @dictionary       ep.dict
     *
     * @translation      {Cancel}
     *                   {Back}
     */

    var translations = {
            tCancel: aiContentDict.translate('Cancel'),
            tBack: aiContentDict.translate('Back'),
            tBubbleHelpGenerateCurrentDescription: aiContentDict.translate('BubbleHelpGenerateCurrentDescription'),
            tBubbleHelpGenerateNewDescription: aiContentDict.translate('BubbleHelpGenerateNewDescription'),
            tBubbleHelpOptimizeCurrentDescription: aiContentDict.translate('BubbleHelpOptimizeCurrentDescription'),
            tBubbleHelpOptimizeNewDescription: aiContentDict.translate('BubbleHelpOptimizeNewDescription'),
            tCTAGenerateApply: aiContentDict.translate('CTAGenerateApply'),
            tCTAOptimizeApply: aiContentDict.translate('CTAOptimizeApply'),
            tCTAReplaceApply: aiContentDict.translate('CTAReplaceApply'),
            tCTACopy: aiContentDict.translate('CTACopy'),
            tCTACopied: aiContentDict.translate('CTACopied'),
            tCTAGenerateDescription: aiContentDict.translate('CTAGenerateDescription'),
            tCTAOptimizeDescription: aiContentDict.translate('CTAOptimizeDescription'),
            tDescriptionGenerate: aiContentDict.translate('DescriptionGenerate'),
            tDescriptionOptimize: aiContentDict.translate('DescriptionOptimize'),
            tDescriptionReplace: aiContentDict.translate('DescriptionReplace'),
            tLabelGenerateDescription: aiContentDict.translate('LabelGenerateDescription'),
            tLabelGenerateNewDescription: aiContentDict.translate('LabelGenerateNewDescription'),
            tLabelOptimizeDescription: aiContentDict.translate('LabelOptimizeDescription'),
            tLabelOptimizeNewDescription: aiContentDict.translate('LabelOptimizeNewDescription'),
            tLabelReplaceCheckbox: aiContentDict.translate('LabelReplaceCheckbox'),
            tTitelGenerateDescription: aiContentDict.translate('TitelGenerateDescription'),
            tTitelOptimizeDescription: aiContentDict.translate('TitelOptimizeDescription'),
            tTitelReplaceDescription: aiContentDict.translate('TitelReplaceDescription'),
            tRemainingTokens: aiContentDict.translate('RemainingTokens'),
            tRemainingTokensDescription: aiContentDict.translate('RemainingTokensDescription')
        };



    $.widget('ui.presentationAiContent', {
    //  optimizeDialog,
    //  confirmDialog,
    //  optimizeDialogClassName,
    //  confirmDialogClassName,

        tmplElements: {},
        buttons: {},
        optimizeButtonClass: "ep-ai-content-add-button",
        confirmButtonClass: "ep-ai-content-confirm-button",
        states: {
            1: 'generate',
            2: 'optimize'
        },
        state: 1,

        copyStates: {
            DISABLED: 1,
            COPY: 2,
            COPIED: 3
        },

        options : {
            scribe: undefined,
            onApply: function(content) {
                console.log(content);
            },
            description: "",
            isConfirmed: false,
            changeActionAI: "JSONOptimizeAiContent",
            changeActionConfirm: "JSONSaveSeenAIDialog",
            dict: {
                generate: {
                    remainingTokens: translations.tRemainingTokens,
                    description: translations.tDescriptionGenerate,
                    label: {
                        current: translations.tLabelGenerateDescription,
                        new: translations.tLabelGenerateNewDescription
                    },
                    tooltip: {
                        remainingTokens: translations.tRemainingTokensDescription,
                        current: translations.tBubbleHelpGenerateCurrentDescription,
                        new: translations.tBubbleHelpGenerateNewDescription
                    },
                    title: translations.tTitelGenerateDescription,
                    button: {
                        apply: translations.tCTAGenerateApply,
                        description: translations.tCTAGenerateDescription
                    }
                },
                optimize: {
                    remainingTokens: translations.tRemainingTokens,
                    description: translations.tDescriptionOptimize,
                    label: {
                        current: translations.tLabelOptimizeDescription,
                        new: translations.tLabelOptimizeNewDescription
                    },
                    tooltip: {
                        remainingTokens: translations.tRemainingTokensDescription,
                        current: translations.tBubbleHelpOptimizeCurrentDescription,
                        new: translations.tBubbleHelpOptimizeNewDescription
                    },
                    title: translations.tTitelOptimizeDescription,
                    button: {
                        apply: translations.tCTAOptimizeApply,
                        description: translations.tCTAOptimizeDescription
                    }
                },
                confirm: {
                    title: translations.tTitelReplaceDescription,
                    text: translations.tDescriptionReplace,
                    label: translations.tLabelReplaceCheckbox,
                    button: {
                        apply: translations.tCTAReplaceApply,
                        back: translations.tBack
                    }
                },
                common: {
                    copy: translations.tCTACopy,
                    copied: translations.tCTACopied
                }
            }
        },


        /*
         * Method called when the widget gets initialized. Used to configure basic data for the widget and
         * to create needed properties/elements and so on.
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        _create: function() {
            var self = this,
                o = self.options,
                elem = self.element;

            self.optimizeDialogClassName = 'ep-ai-content-optimize-dialog-' + self.uuid;
            self.confirmDialogClassName = 'ep-ai-content-confirm-dialog-' + self.uuid;

            if (o.scribe) {
                o.description = o.scribe.element.text();
            }

            // open optimize dialog on click
            elem.on('click.openOptimizeDialog', function(evt) {
                evt.preventDefault();

                self.open();
            });

            self.state = o.description.length === 0 ? 1 : 2;
        },


        /*
         * Method can later be used to set further options (see comment for an example).
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        _setOption: function(key, value) {
            var self = this,
                o = self.options;

            switch (key) {
                // add scribe instance
                case 'scribe':
                    o.scribe = value;
                    o.description = value.element.text();
                    break;

            }

            self._superApply(arguments);
        },


        /*
         * Method to initialize and build the optimize dialog to handle the content that should be
         * optimized with AI (OpenAI).
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        _buildOptimizeDialog: function() {
            var _self = this,
                o = _self.options,
                elem = _self.element,
                renderedTmpl = _self._renderOptimizeDialog();

            // Initialize dialog widget
            _self.optimizeDialog = $('<div/>').html(renderedTmpl).uiDialog({
                width: '610px',
                modal: true,
                title: o.dict[_self.states[_self.state]].title,
                classes: {
                  'ui-dialog': _self.optimizeDialogClassName
                },
                buttons: [
                    {
                        text: translations.tCancel,
                        click: function() {
                            ep(this).uiDialog('close');
                        },
                        cssClass: 'ep-button ep-button-close'
                    },
                    {
                        text: o.dict[_self.states[_self.state]].button.apply,
                        disabled: !self.tmplElements || (self.tmplElements && self.tmplElements.descrTarget.val().length === 0),
                        click: function() {
                            // confirm flag already set
                            if (o.isConfirmed) {
                                _self.close();
                            } else {
                                // hide optimize dialog
                                _self._toggleOptimizeDialog(false);
                                _self._confirm();
                            }
                        },
                        cssClass: 'ep-button ep-button-add ' + _self.confirmButtonClass
                    }
                ],
                autoOpen: true
            });
        },


        /*
         * Renders the content of the optimize dialog depending on the current state.
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        _renderOptimizeDialog: function() {
            var self = this,
                o = self.options,
                elem = self.element,
                renderedTmpl;

            /*
             * Clean-up before teplate gets rendered
             */
            // remove input events from description textareas
            self.tmplElements.descrSource && self.tmplElements.descrSource.off('input.aiContentDescrSource');
            self.tmplElements.descrTarget && self.tmplElements.descrTarget.off('input.aiContentDescrTarget');
            // remove click event from buttons if present
            self.tmplElements.copyButton && self.tmplElements.copyButton.off('click.aiContentCopy');
            self.tmplElements.optimizeButton && self.tmplElements.optimizeButton.off('click.aiContentOptimize');

            // render template
            renderedTmpl = tmplOptimizeDialog([{
                descrSource: o.description,
                remainingTokens: o.remainingTokens,
                copyState: self.copyStates.DISABLED,
                dict: $.extend({}, o.dict[self.states[self.state]], {common: o.dict.common})
            }]);

            // delete self.tmplElements;

            // Get DOM-Elements from template
            $.extend(self.tmplElements, renderedTmpl.tmplItem('elements'));

            // meta parse rendered template
            renderedTmpl.find('.ep-js').metaparse();

            /*
             * After-render routines
             */
            // register input events on description textareas
            self.tmplElements.descrSource.on('input.aiContentOptimize', function() {
                var $this = $(this);

                self.tmplElements.optimizeButton.prop('disabled', $(this).val() === "");
            });
            self.tmplElements.descrTarget.on('input.aiContentDescrTarget', function() {
                var isFilled = $(this).val().length ? true : false;

                self._setCopyState(isFilled ? self.copyStates.COPY : self.copyStates.DISABLED);
                self._onAfterGenerate(!isFilled);
            });

            // register click event on copy button
            self.tmplElements.copyButton.on('click.aiContentCopy', function() {
                var newContent = self.tmplElements.descrTarget.val();

                self._copyToClipboard(newContent);
                self._setCopyState(newContent.length ? self.copyStates.COPIED : self.copyStates.DISABLED);
            });

            // register click event on optimize button
            self.tmplElements.optimizeButton.on('click.aiContentOptimize', function() {
                self._openAiRequest();
            })

            return renderedTmpl;
        },


        /*
         * Method to send the current description and requests a OpenAI generated/optimized response.
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        _openAiRequest: function() {
            var self = this,
                o = self.options;

            // show busy layer
            self.optimizeDialog.closest('.' + self.optimizeDialogClassName).busy('show');

            // send request to backend which handels the communication with OpenAI
            ep.ajax({
                type: "POST",
                dataType: "json",
                url: '?',
                data: {
                    ObjectID: o.objectID,
                    ViewAction: "JSONViewResponse",
                    ChangeAction: o.changeActionAI,
                    RequestedLanguageID: o.requestedLanguageId,
                    Content: self.tmplElements.descrSource.val()
                }
            })
            .always(function() {
                self.optimizeDialog.closest('.' + self.optimizeDialogClassName).busy('hide');
            })
            .done(function(data, textStatus, jqXHR) {
                $('.RemainingTokensBlock .RemainingTokens').html($.i18n.formatNumber(data.remaining_tokens, 'n0'));
                if ( data.remaining_tokens <= 0) {
                    $('.RemainingTokensBlock').css({'color':'#c00'});
                }
                // request was successful
                if (data.success) {
                    // add optimized description to target text area
                    self.tmplElements.descrTarget.val(data.content);
                    // enable/disabel apply button depending on the length of the optimized description
                    self._onAfterGenerate(data.content.length ? false : true);
                    self._setCopyState(data.content.length ? self.copyStates.COPY : self.copyStates.DISABLED);
                } else if (data.error) {
                    console.log('Presentation/Data/Public/lib/de_epages/presentation/ui/ai-content.js: ', data.error);
                }
            })
            .fail(function(jqXHR, textStatus, errorThrown) {
                // uncomment the following line for debugging if an error occurs
                // console.log(arguments);
            });
        },


        /*
         * Method to initialize a confirmation dialog if not already exists.
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        _confirm: function() {
            var self = this,
                o = self.options,
                renderedTmpl;

            // dialog already exists => open it
            if (self.confirmDialog) {
                self.confirmDialog.uiDialog('open');
            } else {
                // render template for confirmation dialog
                renderedTmpl = tmplConfirmDialog([{
                    descrSource: o.description,
                    dict: $.extend({}, {confirm: o.dict.confirm})
                }]);

                // Get DOM-Elements from template
                $.extend(self.tmplElements, renderedTmpl.tmplItem('elements'));

                // meta parse rendered template
                renderedTmpl.find('.ep-js').metaparse();

                // initialize and auto open the dialog
                self.confirmDialog = $('<div/>').html(renderedTmpl).uiDialog({
                    width: '50%',
                    modal: true,
                    title: o.dict.confirm.title,
                    classes: {
                      'ui-dialog': self.confirmDialogClassName
                    },
                    buttons: [
                        {
                            text: o.dict.confirm.button.back,
                            click: function() {
                                self.confirmDialog.uiDialog('close');
                                self._toggleOptimizeDialog(true);
                            },
                            cssClass: 'ep-button ep-button-close ep-confirm-back'
                        },
                        {
                            text: o.dict.confirm.button.apply,
                            click: function() {
                                // checkbox to confirm not to show the dialog again is checked
                                if (self.tmplElements["confirmInput"].prop('checked')) {
                                    self._confirmRequest();
                                } else {
                                    // close confirmation dialog
                                    self.confirmDialog.uiDialog('close');
                                    // apply new description and close optimize dialog
                                    self.close();
                                }
                            },
                            cssClass: 'ep-button ep-button-add ' + self.confirmButtonClass
                        }
                    ],

                    close: function(evt) {
                        var target = $(evt.originalTarget);

                        if (target.hasClass('ui-dialog-titlebar-close') || evt.keyCode && evt.keyCode === 27) {
                            self._toggleOptimizeDialog(true);
                        }
                    },
                    autoOpen: true
                });
            }
        },


        /*
         * Method to send a confirmation flag to the server that saves that the confirmation dialog
         * won't be shown again.
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        _confirmRequest: function() {
            var self = this,
                o = self.options;

            ep.ajax({
                type: "POST",
                dataType: "json",
                url: '?',
                data: {
                    ViewAction: "JSONViewResponse",
                    ChangeAction: o.changeActionConfirm,
                    IsAIDialogConfirmed: 1
                }
            })
            .always(function() {})
            .done(function(data, textStatus, jqXHR) {
                if (data.success) {
                    // set confirmation flag
                    o.isConfirmed = true;
                    // close confirmation dialog
                    self.confirmDialog.uiDialog('close');
                    // apply new description and close optimize dialog
                    self.close();
                } else if (data.error) {
                    console.log('Presentation/Data/Public/lib/de_epages/presentation/ui/ai-content.js: ', data.error);
                }
            })
            .fail(function(jqXHR, textStatus, errorThrown) {
                // uncomment the following line for debugging if an error occurs
                // console.log(arguments);
            });
        },


        /*
         * Method to add the apply button inside the optimization dialog to button collection.
         * And additionaly this method enables/disables several element states depending of the argument.
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        _onAfterGenerate: function(disabled) {
            var self = this,
                o = self.options;

            // add apply button to buttons collection if necessary
            if (self.buttons.apply === undefined) {
                self.buttons.apply = self.optimizeDialog.data('uiUiDialog').uiButtonSet.find('.' + self.confirmButtonClass);
            }

            // set state of button
            self.buttons.apply
                .prop('disabled', disabled)
                .html(o.dict[self.states[self.state]].button.apply);

            // set state of textarea
            self.tmplElements.descrTarget.prop('disabled', disabled);
        },


        /*
         * Method to manage the state of the copy button.
         *
         * copyState: 0 = disable, 1 = enabled && not copied, 2 = enabled && copied
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        _setCopyState: function(copyState) {
            var self = this,
                o = self.options;

            self.tmplElements.copyButton
                .prop('disabled', copyState < self.copyStates.COPY)
                .find('span')
                    .html(o.dict.common[copyState < self.copyStates.COPIED ? 'copy' : 'copied']);
        },


        /*
         * Method to show/hide the optimize dialog (show = true => show, show = false => hide).
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        _toggleOptimizeDialog: function(show) {
            var self = this,
                o = self.options;

            self.optimizeDialog.closest('.' + self.optimizeDialogClassName)[show ? 'show' : 'hide']();
        },


        /*
         * Method to write content to the clipboard.
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        _copyToClipboard: function(content) {
            var self = this,
                o = self.options;


                // if browser knows *navigator.permissions*
                try {
                    // request needed permission
                    navigator.permissions.query({ name: "clipboard-write" }).then(function(result) {
                        // check if permission is sufficient
                        if (result.state === "granted" || result.state === "prompt") {
                            // write to the clipboard now
                            navigator.clipboard.writeText(content).then(function(data) {
                                self._setCopyState(self.copyStates.COPIED);
                            });
                        }
                    // exception for Firefox (and more?)
                    }).catch(function(error) {
                        // write to the clipboard now without permissions
                        navigator.clipboard.writeText(content).then(function(data) {
                            self._setCopyState(self.copyStates.COPIED);
                        }).catch(function(err) {
                            self._handleExecCommand(self.tmplElements.descrTarget[0]);
                        })
                    });
                // remaining exceptions
                } catch(e) {
                    self._handleExecCommand(self.tmplElements.descrTarget[0]);
                }
        },


        /*
         * Method to write to select content in a text area and write it to clip board.
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        _handleExecCommand: function(textArea) {
            textArea.focus();
            textArea.setSelectionRange(0, textArea.value.length);
            document.execCommand("copy");
            textArea.blur();
        },


        /*
         * Opens an existing dialog after cleaning up its content (re-render, title ...) or
         * creates and initializes a dialog if not exists.
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        open: function() {
            var self = this,
                elem = self.element,
                o = self.options;

            // get description from scribe editor
            if (o.scribe) {
                o.description = o.scribe.element.text();
            }

            self.state = o.description.length === 0 ? 1 : 2;

            if (self.optimizeDialog) {
                // render template and replace dialog content
                self.optimizeDialog.html(self._renderOptimizeDialog());
                // update apply button
                // self.buttons.apply.html(o.dict[self.states[self.state]].button.apply);
                self._onAfterGenerate(true);
                // update dialog title
                self.optimizeDialog.uiDialog('option', 'title', o.dict[self.states[self.state]].title);
                // open dialog
                self.optimizeDialog.uiDialog('open');
            } else {
                self._buildOptimizeDialog();
            }
        },


        /*
         * Method to close optimization dialog after the new content was written to a existing
         * scribe instance or after callback function was called with the new content as argument.
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        close: function() {
            var self = this,
                o = self.options,
                content = self.tmplElements.descrTarget.val();

            // write new content back to scribe editor
            if (o.scribe) {
                o.scribe.scribe.setHTML(content);
            } else if (o.onApply && typeof o.onApply === "function") {
                o.onApply(content);
            }

            // close dialog
            self.optimizeDialog.uiDialog('close');
        },


        /*
         * Cleanup method to remove events, elements from DOM and destroy used widgets.
         *
         * @member jQuery.ui.presentationAiContent
         *
         * @since 7.90.0
         */
        destroy: function() {
            var self = this,
                elem = self.element;

            // default garbage collection
            self._superApply(arguments);

            // ToDo: remove events
            elem.off('click.openOptimizeDialog');
            self.tmplElements.descrSource && self.tmplElements.descrSource.off('input.aiContentDescrSource');
            self.tmplElements.descrSource && self.tmplElements.descrSource.off('input.aiContentOptimize');
            self.tmplElements.descrTarget && self.tmplElements.descrTarget.off('input.aiContentDescrTarget');
            self.tmplElements.optimizeButton && self.tmplElements.optimizeButton.off('click.aiContentOptimize');
            self.tmplElements.copyButton && self.tmplElements.copyButton.off('click.aiContentCopy');

            // ToDo: remove nodes/elements from DOM

            // ToDo: destroy used widgets
            self.optimizeDialog && self.optimizeDialog.uiDialog('destroy');
            self.confirmDialog && self.confirmDialog.uiDialog('destroy');
        }
    });

    return de_epages;
});
