#========================================================================================
# §package      DE_EPAGES::Product::API::ProductService
# §state        public
#----------------------------------------------------------------------------------------
# §description  Collects the products of a shop
#========================================================================================
package DE_EPAGES::Product::API::ProductService;

use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::PerlTools qw ( isValid trim JsonTrue );
use DE_EPAGES::Object::API::Factory qw ( LoadObject LoadClassByAlias Transaction ExistsObject );
use DE_EPAGES::Object::API::Object::Locale;
use DE_EPAGES::Object::DAL::UnitOfMeasurement qw ( DAL_GetInfoUnitOfMeasurementByAbbreviation );
use DE_EPAGES::Shop::API::Resource::ShopRequestContext;
use DE_EPAGES::Shop::API::Resource::Formatter;
use DE_EPAGES::Shop::API::ShopTaxClass;
use DE_EPAGES::Product::API::Product qw ( TestUPCEAN GetVariationAttributes );
use DE_EPAGES::Product::API::Resource::CustomAttributes;
use DE_EPAGES::Product::API::Resource::ProductSlideshow;
use DE_EPAGES::Product::API::Resource::ProductVariation;
use DE_EPAGES::Product::API::Resource::ProductLowestPrice;
use DE_EPAGES::Product::API::Resource::ProductCategories;
use DE_EPAGES::Product::API::Resource::ProductType;
use DE_EPAGES::Product::DAL::Product qw ( DAL_GetUpdatedProducts );
use DE_EPAGES::Product::API::Tools qw ( MergeUpdatedProductsResults ConvertDateTimeForEpJ );
use DE_EPAGES::Tax::API::Shop qw ( TaxInfoBySellingCountryID );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Presentation::API::Table::Session qw (ExistsSessionBySessionID GetInfoSessionBySessionID);
use DE_EPAGES::Product::API::CustomAttributesPatch;
use JSON qw ( from_json );

use strict;

no warnings 'redefine'; ##no critic qw(ProhibitNoWarnings)

#========================================================================================
# §function     new
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       new( );
#----------------------------------------------------------------------------------------
# §description  Constructor
#----------------------------------------------------------------------------------------
# §return       $self | self | DE_EPAGES::Product::API::ProductService
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {}, $class;
    return $self;
}

sub updateProductCustomAttributePartially {
    my $self = shift;
    my($Product, $LanguageID, $ahRequestBody) = @_;

    my $Patch = $self->_createProductCustomAttributesPatch($Product, $LanguageID);
    foreach my $hPatchOperation (@{$ahRequestBody}) {
        $Patch->patch($hPatchOperation, $LanguageID) ;
    }

    return;
}

sub _createProductCustomAttributesPatch {
    my $self = shift;
    my ($Product, $LanguageID) = @_;
    return DE_EPAGES::Product::API::CustomAttributesPatch->new($Product, $LanguageID);
}

sub getAllUpdatedProducts {
    my $self = shift;
    my ($Shop, $AttributesUpdatedSince, $StockLevelUpdatedSince) = @_;

    my $ShopId = $Shop->id();
    my $aProductIDsWithChangedAttributes = [];
    my $aProductIDsWithChangedStocklevel = [];

    if(defined $AttributesUpdatedSince) {
        $aProductIDsWithChangedAttributes = $self->_DAL_GetUpdatedProducts($ShopId, 'RestApiResponseAttributes', $AttributesUpdatedSince);
    }

    if(defined $StockLevelUpdatedSince) {
        $aProductIDsWithChangedStocklevel = $self->_DAL_GetUpdatedProducts($ShopId, 'RestApiStockLevel', $StockLevelUpdatedSince);
    }

    my $aUpdatedProducts = MergeUpdatedProductsResults($aProductIDsWithChangedAttributes, $aProductIDsWithChangedStocklevel);

    my @UpdatedProducts = map { [$self->_LoadObject($_->[0]), ConvertDateTimeForEpJ($_->[1])] } @$aUpdatedProducts;

    return \@UpdatedProducts;
}

sub getProductDetailsData {
    my $self = shift;
    my ($Shop, $Product, $LocaleID, $CurrencyID, $SessionID, $SellingCountryID) = @_;
    my $Context = $self->_getShopContext($Shop);
    my $APIBaseURL = $Context->apiBaseUrl();
    my $Locale = $self->_loadLocale($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');

    my $hShopData = $self->getShopData($Shop, $LocaleID, $CurrencyID);
    my $hSellingCountryData = TaxInfoBySellingCountryID($Shop, $SellingCountryID );
    my $hProductData = $self->getProductData($Product, $hShopData, $SessionID, $hSellingCountryData);

    my $aLinks = $hProductData->{'links'};
    $hProductData->{'links'} = $self->_addProductDetailsLinks($Product, $APIBaseURL, $LanguageID, $aLinks);
    $hProductData->{'shippingMethodsRestrictedTo'} = $self->_getShippingMethods($Product->get('ShippingMethods'), $APIBaseURL, $LanguageID),

    my $aCategories = $Product->get('AllProductCategories');
    my $MainCategory = $aCategories->[0];
    if (defined $MainCategory ) {
        push @{$hProductData->{'links'}}, {
            'rel' => 'main-category',
            'href' => $self->_getCategoryHref($APIBaseURL, $MainCategory->guid()),
            'title' => $MainCategory->get( 'NavigationString', $LanguageID ),
        };
    }

    return $hProductData;
}

sub getShopData {
    my $self = shift;
    my ($Shop, $LocaleID, $CurrencyID) = @_;

    my $Context = $self->_getShopContext($Shop);
    my $Locale = $self->_loadLocale($LocaleID);

    my $aAttributesToGet = [ qw( Address DeliveryThresholdIsActive TaxModel TaxMatrix UseProductZoom IsProductNoVisible ) ];
    my $hShopData = $Shop->get($aAttributesToGet);

    $hShopData->{'Formatter'}    = $self->_createFormatter($Shop, $LocaleID);
    $hShopData->{'Context'}      = $Context;
    $hShopData->{'Locale'}       = $Locale;
    $hShopData->{'CurrencyID'}   = $CurrencyID;
    $hShopData->{'LanguageID'}   = $Locale->get('LanguageID');
    $hShopData->{'SiteLocale'}   = $Shop->siteLocale($LocaleID);
    $hShopData->{'APIBaseURL'}   = $Context->apiBaseUrl();
    $hShopData->{'ImageBaseURL'} = $Context->imageBaseUrl();
    $hShopData->{'CustomizableProductHeadlineDefaultText'} = $Shop->get('CustomizableProductHeadlineDefaultText',
                                                                        $hShopData->{'LanguageID'});

    return $hShopData;
}

sub getProductData {
    my $self = shift;
    my ($Product, $hShopData, $SessionID, $hSellingCountryData) = @_;

    my $SessionUserID   = ($SessionID and ExistsSessionBySessionID($SessionID)) ?
        GetInfoSessionBySessionID($SessionID)->{'UserID'} : undef;

    my $Formatter = $hShopData->{'Formatter'};
    # use the Taxmodel from selling country if needed
    my $TaxModel = 0;
    if ($Product->getSite()->get('TaxModel') == 1) {
        $TaxModel = $hSellingCountryData->{'TaxModel'};
    }
    my $LanguageID = $hShopData->{'LanguageID'};
    my $SiteLocale = $hShopData->{'SiteLocale'};
    my $APIBaseURL = $hShopData->{'APIBaseURL'};
    my $ImageBaseURL = $hShopData->{'ImageBaseURL'};
    my $CurrencyID = $hShopData->{'CurrencyID'};
    my $TaxMatrix = $hShopData->{'TaxMatrix'};
    my $UseProductZoom = $hShopData->{'UseProductZoom'};
    my $IsProductNoVisible = $hShopData->{'IsProductNoVisible'};
    my $CustomizableProductHeadlineDefaultText = $hShopData->{'CustomizableProductHeadlineDefaultText'};

    my @Attributes = qw(
        IsVisible
        AvailabilityText
        AvailabilityIcon
        DeliveryPeriod
        Description
        DisplayTitle
        IsSpecialOffer
        Weight WeightUnit
        Text NameOrAlias
        ImageHotDeal ImageMedium ImageLarge ImageSmall ImageThumbnail
        Path GUID
        PriceQuantity OrderUnit
        RefUnit RefAmount RefContentAmount
        ShippingMethods
        WebPath
        WebUrl
        EnergyLabelsString
        EnergyLabelHighestEfficiencyClass
        EnergyLabelSourceFile
        ProductDataSheet
        Alias
        UPCEAN
        Manufacturer ManufacturerSKU
        MainCharacteristics
        TaxClassDigital
        TaxClass
        Keywords
        StockLevel StockLevelAlert
        IsMaster
        SuperProduct
        VariationTypeValues
        Length Width Height
        ProductVideoProvider
        IsManufacturerPriceRRP
        IsShippingLinkVisible
        IsCustomizable
        CustomizableTextLength
        CustomizableHeadlineText
        IsVATInfoVisible
        IsNew
        NewnessDate
    );

    my $hProduct = $Product->getWithParent(\@Attributes, $LanguageID);
    $ImageBaseURL = $ImageBaseURL . $hProduct->{'WebPath'};

    $TaxMatrix = $hSellingCountryData->{'TaxMatrix'} ? $hSellingCountryData->{'TaxMatrix'} : $TaxMatrix ;

    #-- Assuming the tax matrix passed is right, this price is correct with respect to delivery thresholds
    my $ListPrice = $Product->getListPrice($CurrencyID, $TaxModel, undef, undef, $TaxMatrix);

    my $ManufacturerPrice = $Product->getManufacturerPrice($CurrencyID, $TaxModel);
    my $EcoParticipation = $Product->getEcoParticipation($CurrencyID, $TaxModel);
    my $Deposit = $Product->getDeposit($CurrencyID, $TaxModel, undef, $TaxMatrix);
    my $PriceWithDeposits = defined $ListPrice ? $ListPrice + ($EcoParticipation // 0) + ($Deposit // 0) : undef;
    my $ReferencePrice = $Product->getReferencePrice($CurrencyID, $ListPrice);

    my $RefPriceFormatted;
    my $RefUnit = $hProduct->{'RefUnit'};
    if( defined $ReferencePrice && defined $RefUnit ) {
        $RefPriceFormatted = sprintf '%d %s = %s',
            $hProduct->{'RefAmount'},
            $RefUnit->get('Abbreviation') // $RefUnit->get('NameOrAlias', $LanguageID),
            $SiteLocale->formatMoney($ReferencePrice, $CurrencyID);
    }

    my $hQuantity = $Formatter->quantity( $hProduct->{'PriceQuantity'}, $hProduct->{'OrderUnit'} );
    my $hAmount   = $Formatter->quantity( $hProduct->{'RefContentAmount'}, $hProduct->{'RefUnit'} );
    my $hImageInformation = $Product->get('ImageInformations');

    if(not defined $hImageInformation ){
        $Product->updateImageInformations(undef);
        $hImageInformation = $Product->get('ImageInformations');
    }

    my $hTaxData = {
        'LanguageID' => $LanguageID,
        'TaxClassDigital' => $hProduct->{'TaxClassDigital'},
        'TaxClass' => $Product->getMappedTaxClassByTaxMatrix($TaxMatrix),
        'TaxMatrix' => $TaxMatrix,
    };

    my $aSearchKeywords = $self->_createSearchKeywords($hProduct->{'Keywords'});

    my $ProductVariationType = 'regular';
    my $ProductVariationMasterName = undef;
    my $ProductVariationMasterGUID = undef;
    my $aProductVariationSelection = undef;
    if($hProduct->{'IsMaster'}){
        $ProductVariationType = 'master';
    }elsif($hProduct->{'SuperProduct'}){
        $ProductVariationType = 'variation';
        $aProductVariationSelection = GetVariationAttributes($Product, $LanguageID, $hProduct->{'VariationTypeValues'});
        $ProductVariationMasterName = $hProduct->{'SuperProduct'}->get('NameOrAlias', $LanguageID);
        $ProductVariationMasterGUID = $hProduct->{'SuperProduct'}->guid;
    }

    my $LowestPrice;
    my $HighestPrice;
    my $LowestPriceSubProduct;
    my $Shop = $Product->getSite();
    #-- These prices do not take into account delivery thresholds (at the time this comment was written :D)
    #-- "!$hProduct->{'IsMaster'}" shouldn't be here, why should it run for normal products? BUT, some tests call this with a master product and the check fails,
    #-- data inconsistency probably. So whatever, keep it here. For 'REAL' normal products, it won't do anything anyway.
    if ( !$hProduct->{'IsMaster'} || ($hProduct->{'IsMaster'} && ! $Shop->get('DeliveryThresholdIsActive')) ) {
        $LowestPrice = $Product->lowestSubProductPrice($CurrencyID, $TaxModel, 1);
        $LowestPriceSubProduct = $Product->subProductWithLowestPriceThatHasOwnPrice($CurrencyID);

        $HighestPrice = $Product->highestSubProductPrice($CurrencyID, $TaxModel);
    }

    #-- Where the product is a master, loop through visible/available variations to find customer specific prices for quantity 1 and set the 'lowestPrice' key to this value
    #-- Only apply this for Unity shops
    if ( $hProduct->{'IsMaster'} && $Shop->get('IsUnityRegistered') ) {

      #-- Scenario: Master product does not have any delivery threshold exceptions ('TaxClassMapping' attribute)
      #--             But the variations do
      #--             The master's list may be wrong, we need to compare the list price across all variations to find the right prices
      #--             OR the prices are just different between the master and variations
      if ( $Shop->get('DeliveryThresholdIsActive') ) {
        my $LSubProduct = $Product->subProductWithLowestPriceThatHasOwnPrice($CurrencyID);
        if ( $LSubProduct ) {
            my $LSubListPrice = $LSubProduct->getListPrice($CurrencyID, $TaxModel, undef, undef, $TaxMatrix);
            $LowestPrice = $LSubListPrice;
            $LowestPriceSubProduct = $LSubProduct;
        }
        my $HSubProduct = $Product->subProductWithHighestPriceThatHasOwnPrice($CurrencyID);
        if ( $HSubProduct ) {
            my $HSubListPrice = $HSubProduct->getListPrice($CurrencyID, $TaxModel, undef, undef, $TaxMatrix);
            $HighestPrice = $HSubListPrice;
        }
      }

      if (defined $SessionUserID) {
          my $aSessionUserCustomers = ExistsObject($SessionUserID) ? $Shop->loadObject($SessionUserID, 'User')->get('Customers') : [];
          if (scalar @$aSessionUserCustomers) {
              my $Customer = $aSessionUserCustomers->[0];
              if( (defined $Customer) and (defined $Customer->get('CustomerGroup')) ){
                  my $CustomerPriceListAlias = $Customer->get('CustomerGroup')->alias;
                  if ($Shop->folder('PriceLists')->existsChild($CustomerPriceListAlias)){
                      my $CustomerPriceList = $Shop->folder('PriceLists')->child($CustomerPriceListAlias);

                      my $aSubProducts = $Product->get('SubProducts');
                      foreach my $SubProduct ( grep { $_->canAddToBasket() } @$aSubProducts ) {

                          #-- Scale price and discount are stored in separate tables, and we only care about the row where quantity == 1,
                          #-- since we can't use any other quantity to calculate a unit price for informative purposes
                          # as there is no guarantee that a customer will buy more than 1
                          my @Prices = grep { $_->{'Quantity'} == 1 }
                          @{$CustomerPriceList->getScalePricesByProductID($SubProduct->id,
                                                                          $TaxModel, $SubProduct->getMappedTaxClassByTaxMatrix($TaxMatrix))
                          };
                          my @Discounts = grep { $_->{'Quantity'} == 1 } @{$CustomerPriceList->getQuantityDiscountsByProductID($SubProduct->id)};

                          #-- Originally considered that there would be rows with only discount and no scale price like in Base,
                          #   but apparently not possible, scale price is set to 0 automatically
                          #-- Scale Price | Discount
                          #-- Y             Y         Ok
                          #-- Y             N         Ok
                          #-- N             Y         If no scale price entered, it is set to 0 via JS in the UnityMBO
                          #-- N             N         Ok
                          if ( scalar @Prices || scalar @Discounts ) {
                              my $ScalePrice = $Prices[0]{'Price'};
                              my $Discount = $Discounts[0]{'Discount'};

                              #-- Not sure if this can ever be < 0. JS prevents bad user input (e.g. 200% discount, or negative price)
                              #   but if bad values are added to DB via other means?
                              #-- So only use this calculated price if it is non-negative
                              my $CalculatedFromPrice = $ScalePrice - (($ScalePrice * $Discount) / 100); ##no critic qw(ProhibitMagicNumbers)

                              if ($CalculatedFromPrice >= 0 && $CalculatedFromPrice < $LowestPrice) {
                                  $LowestPrice = $SiteLocale->roundMoney($CalculatedFromPrice, $CurrencyID);
                                  $LowestPriceSubProduct = $SubProduct;
                              }
                          }
                      }
                  }
              }
          }
      } else {
          #-- Standard bulk prices at quantity 1 are always hardcoded and set to the list price (from what I can see), so there is no scope for differing prices here
          #-- There is therefore, no need for a different set of logic where the user is not logged in as the standard logic that sets $LowestPrice already
          #   accounts for this since that works on list price
      }

    }

    #-- Convert this to UTC so that there is no ambiguity about the timezone. Also remove the time element by setting it to midnight, it is
    #-- irrelevant for the expiry date, which is based on the scheduler run time. Set the right expectations ^.^
    my $ReturnedNewnessDate = $hProduct->{'NewnessDate'} ? $hProduct->{'NewnessDate'}->set_time_zone('UTC')->set(hour=>00,minute=>00,second=>00)->iso8601 . 'Z' : undef;

    # this is common information for all types of products, including master and variations
    my $hResult = {
        'productVariationType' => $ProductVariationType,
        'productVariationMasterName' => $ProductVariationMasterName,
        'productVariationMasterId' => $ProductVariationMasterGUID,
        'productVariationSelection' => $aProductVariationSelection,
        'visible' => $Formatter->bool($hProduct->{'IsVisible'}),
        'availabilityText' => $hProduct->{'AvailabilityText'},
        'availability' => $self->_getAvailability($hProduct->{'AvailabilityIcon'}),
        'deliveryWeight' => $Formatter->quantity($hProduct->{'Weight'}, $hProduct->{'WeightUnit'}),
        'deliveryPeriod' => $hProduct->{'DeliveryPeriod'},
        'description' => $hProduct->{'Text'},
        'title' => $hProduct->{'DisplayTitle'},
        'forSale' => $Formatter->bool($Product->isMaster ? 0 : $Product->canAddToBasket),
        'productImage' => $hProduct->{'ImageLarge'},
        'images' => $self->_getImages($hProduct, $ImageBaseURL, {
            'UseProductZoom' => $UseProductZoom,
            'Formatter' => $Formatter,
            'ImageInformations' => $hImageInformation
        }),
        'specialOffer' => $Formatter->bool($hProduct->{'IsSpecialOffer'}), # TODO: move to Catalog cartridge
        'name' => $hProduct->{'NameOrAlias'},
        'priceInfo' => {
            'basePrice' => !defined($RefPriceFormatted) ? undef : {
                'formatted' => $RefPriceFormatted,
                'quantity' => $hAmount,
                'refPrice' => $Formatter->price($ReferencePrice, $CurrencyID, $TaxModel),
                'refQuantity' => $Formatter->quantity($hProduct->{'RefAmount'}, $hProduct->{'RefUnit'}),
            },
            'depositPrice' => $Formatter->price($Deposit, $CurrencyID, $TaxModel),
            'ecoParticipationPrice' => $Formatter->price($EcoParticipation, $CurrencyID, $TaxModel),
            'manufacturerPrice' => $Formatter->price($ManufacturerPrice, $CurrencyID, $TaxModel),
            'price' => $Formatter->price($ListPrice, $CurrencyID, $TaxModel),
            'priceWithDeposits' => $Formatter->price($PriceWithDeposits, $CurrencyID, $TaxModel),
            'lowestPrice' => $Formatter->price($LowestPrice, $CurrencyID, $TaxModel),
            'highestPrice' => $Formatter->price($HighestPrice, $CurrencyID, $TaxModel),
            'quantity' => $hQuantity,
            'taxClass' => $self->_calculateTaxClass($hTaxData),
        },
        'productId' => $hProduct->{'GUID'},
        'sfUrl' => $hProduct->{'WebUrl'},
        'shortDescription' => $hProduct->{'Description'},
        'energyLabelsString' => $hProduct->{'EnergyLabelSourceFile'} ? $hProduct->{'EnergyLabelsString'} : undef,
        'energyLabelsMaxEfficiencyString' => $hProduct->{'EnergyLabelSourceFile'} ? $hProduct->{'EnergyLabelHighestEfficiencyClass'} : undef,
        'energyLabelSourceFile' => $hProduct->{'EnergyLabelSourceFile'} ? $ImageBaseURL . '/' . $hProduct->{'EnergyLabelSourceFile'} : undef,
        'productDataSheet' => $hProduct->{'ProductDataSheet'} ? $ImageBaseURL . '/' . $hProduct->{'ProductDataSheet'} : undef,
        'productNumber' => $hProduct->{'Alias'},
        'upcEan' => $hProduct->{'UPCEAN'},
        'manufacturer' => $hProduct->{'Manufacturer'},
        'manufacturerProductNumber' => $hProduct->{'ManufacturerSKU'},
        'essentialFeatures' => $hProduct->{'MainCharacteristics'},
        'searchKeywords' => $aSearchKeywords,
        'stocklevel' => $hProduct->{'StockLevel'},
        'minStocklevel' => $hProduct->{'StockLevelAlert'},
        'productLength' => $hProduct->{'Length'},
        'productHeight' => $hProduct->{'Height'},
        'productWidth' => $hProduct->{'Width'},
        'links' => $self->_createLinks($Product, $APIBaseURL, $LanguageID),
        'videos' => [],
        'isProductNumberVisible' => $Formatter->bool($IsProductNoVisible),
        'isManufacturerPriceRRP' => $Formatter->bool($hProduct->{'IsManufacturerPriceRRP'}),
        'isShippingLinkVisible' => $Formatter->bool($hProduct->{'IsShippingLinkVisible'}),
        'isCustomizable' => $Formatter->bool($hProduct->{'IsCustomizable'}),
        'customizableTextLength' => $hProduct->{'CustomizableTextLength'},
        'customizableHeadlineText' => $hProduct->{'CustomizableHeadlineText'} // $CustomizableProductHeadlineDefaultText,
        'isVATInfoVisible' => $Formatter->bool($hProduct->{'IsVATInfoVisible'}),
        'isNew'  => $Formatter->bool($hProduct->{'IsNew'}),
        'newnessDate' => $ReturnedNewnessDate,
        'refAmount' => $hProduct->{'RefAmount'},
        'refUnit' => $hProduct->{'RefUnit'} ? $hProduct->{'RefUnit'}->alias : undef,
        'refContentAmount' => $hProduct->{'RefContentAmount'},
    };

    my $LRefPriceFormatted;
    if ( $hProduct->{'IsMaster'} && $LowestPriceSubProduct ) {
      my $LReferencePrice = $LowestPriceSubProduct->getReferencePrice($CurrencyID, $LowestPrice);
      my $LRefUnit = $LowestPriceSubProduct->get('RefUnit');
      if( defined $LReferencePrice && defined $LRefUnit ) {
          $LRefPriceFormatted = sprintf '%d %s = %s',
              $LowestPriceSubProduct->get('RefAmount'),
              $LRefUnit->get('Abbreviation') // $LRefUnit->get('NameOrAlias', $LanguageID),
              $SiteLocale->formatMoney($LReferencePrice, $CurrencyID);
      }
    }
    $hResult->{priceInfo}->{lowestRefPriceFormatted} = $LRefPriceFormatted;

    # add videos in an array of hashes for upcomming changes (maybe we have more than one video and platform)
    # so we have to change only this resource and the perl backend and not the REST part in epagesj
    if($hProduct->{'ProductVideoProvider'}){
        my $aVideoInfo  = from_json($hProduct->{'ProductVideoProvider'}, { utf8  => 1  });
        push @{$hResult->{'videos'}}, {'type' => $_->{'VideoType'}, 'source' => $_->{'VideoID'}} foreach @$aVideoInfo;
    }

    # trigger hook for adding/changing product data
    TriggerHook('API_ProductService_GetProductData', {
                                                        'hResult'      => $hResult,
                                                        'Product'      => $Product,
                                                        'ProductInfo'  => $hProduct,
                                                        'ProductData'  => $hShopData,
                                                        'SessionUserID' => $SessionUserID,
                                                        'TaxCountryData' => $hSellingCountryData,
                                                     });

    #-- Case: A master product HAS customer specific pricing (csp), variations do NOT
    #--       The 'bulkPriceInfo' key contains this csp information for the master, and the 'Now' sf ui will use this to display the csp price for the master
    #--       BUT, this price is NOT used for the variations in either the product display OR basket display/calculation logic
    #--
    #--       It is therefore, incorrectly shown and should show the lowest variation LIST price or lowest variation csp price
    #--       Delete this key for master products so that the 'Now' sf ui does not render this information
    #--       Can't just ignore the hook trigger either in this case, someone else might be doing other things besides bulk pricing calculations
    if ( $hProduct->{'IsMaster'} && $Shop->get('IsUnityRegistered') ) {
      delete $hResult->{'bulkPriceInfo'} if $hResult->{'bulkPriceInfo'};
      delete $hResult->{'customerSpecificPriceInfo'} if $hResult->{'customerSpecificPriceInfo'};
    }

    return $hResult;
}

sub _calculateTaxClass {
    my $self = shift;
    my ($hTaxData) = @_;
    my $TaxClassDigital = $hTaxData->{'TaxClassDigital'};

    if(defined ($TaxClassDigital)) {
        return $self->_calculateTaxClassDigital($TaxClassDigital, $hTaxData->{'LanguageID'});
    } else {
        return $self->_calculateTaxClassNormal($hTaxData);
    }
}

sub _calculateTaxClassDigital {
    my $self = shift;

    my ($TaxClassDigital, $LanguageID) = @_;
    return {
        'taxClassId' => $TaxClassDigital->guid(),
        'name' => $TaxClassDigital->get('NameOrAlias', $LanguageID),
        'percentage'=> undef,
    };
}

sub _calculateTaxClassNormal {
    my $self = shift;
    my ($hTaxData) = @_;
    my $TaxClass = $hTaxData->{'TaxClass'};
    my $TaxMatrix = $hTaxData->{'TaxMatrix'};

    return unless defined ($TaxClass) && defined ($TaxMatrix);

    my $TaxArea = $TaxMatrix->defaultTaxArea();
    return {
        'taxClassId' => $TaxClass->guid(),
        'name' => $TaxClass->get('NameOrAlias', $hTaxData->{'LanguageID'}),
        'percentage'=> $TaxMatrix->taxRate($TaxArea, $TaxClass) * 100, ## no critic (ProhibitMagicNumbers)
    };
}

sub createProduct {
    my $self = shift;
    my ($Shop, $LocaleID, $CurrencyID, $hRequestBody) = @_;

    my $Locale = $self->_loadLocale($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');
    my $TaxModel = $Shop->get('TaxModel');

    my $Alias = $hRequestBody->{'productNumber'};
    if(!isValid($Alias)) {
        Error('BadRequest', {
            'Reason' => 'The attribute productNumber is mandatory!',
            'ShopAlias' => $Shop->alias(),
        });
    }

    my $ProductFolderClass = $self->_LoadClassByAlias('ProductFolder');
    my $ProductFolder = $Shop->folder('Products', { 'Class' => $ProductFolderClass } );
    my $ProductClass = $Shop->get('ProductClass');

    my $Product;
    eval {
        $self->_Transaction(
            sub {

                $Product = $ProductClass->insertObject( {
                    'Alias' => $Alias,
                    'Parent' => $ProductFolder,
                });

                my $hDeliveryWeight = $self->_getDeliveryWeight($hRequestBody->{'deliveryWeight'}, $Shop);
                my $aSearchKeywords = $hRequestBody->{'searchKeywords'} // [];
                my $Keywords = join(', ', @$aSearchKeywords);

                $Product->set({
                    'Name' => $hRequestBody->{'name'},
                    'Text' => $hRequestBody->{'description'},
                    'Weight' => $hDeliveryWeight->{'amount'},
                    'WeightUnit' => $hDeliveryWeight->{'unit'},
                    'DeliveryPeriod' => $hRequestBody->{'deliveryPeriod'},
                    'Description' => $hRequestBody->{'shortDescription'},
                    'Manufacturer' => $hRequestBody->{'manufacturer'},
                    'MainCharacteristics' => $hRequestBody->{'essentialFeatures'},
                    'IsVisible' => JsonTrue($hRequestBody->{'visible'}),
                    'Keywords' => $Keywords,
                    'StockLevel' => $hRequestBody->{'stocklevel'},
                    'ManufacturerSKU' => $hRequestBody->{'manufacturerProductNumber'},
                    'Length' => $hRequestBody->{'productLength'},
                    'Width' => $hRequestBody->{'productWidth'},
                    'Height' => $hRequestBody->{'productHeight'},
                    'StockLevelAlert' => $hRequestBody->{'minStocklevel'},
                }, $LanguageID);

                if(isValid($hRequestBody->{'upc'})) {
                    TestUPCEAN('Value', $hRequestBody->{'upc'});
                    $Product->set({'UPCEAN' => $hRequestBody->{'upc'}});
                }
                if(isValid($hRequestBody->{'ean'})) {
                    TestUPCEAN('Value', $hRequestBody->{'ean'});
                    $Product->set({'UPCEAN' => $hRequestBody->{'ean'}});
                }
                if(isValid($hRequestBody->{'taxClassId'})) {
                    my $TaxClass = $self->_loadTaxClass($hRequestBody->{'taxClassId'}, $Shop);
                    if(defined $TaxClass) {
                        $self->_setTaxClass($Product, $TaxClass);
                    } else {
                        Error('TaxClass could not be found!');
                    }
                }

                $Product->setListPrice($CurrencyID, $TaxModel, $hRequestBody->{'price'});
                $Product->setManufacturerPrice($CurrencyID, $TaxModel, $hRequestBody->{'manufacturerPrice'});
                $Product->setDeposit($CurrencyID, $TaxModel, $hRequestBody->{'depositPrice'});
            }
        );
    };
    if (ExistsError()) {
        my $Error = GetError();
        if ( $Error->code() eq 'OBJECT_EXISTS' ) {
            Error('BadRequest', {
                'Reason' => 'Product with given alias exists already!',
                'ShopAlias' => $Shop->alias(),
            });
        } elsif ( $Error->code() eq 'FORMAT_NOT_UPCEAN' ) {
            Error('BadRequest', {
                'Reason' => 'UPC or EAN is not valid!',
                'ShopAlias' => $Shop->alias(),
            });
        } else {
            Error($Error);
        }
    }

    my $hShopData = $self->getShopData($Shop, $LocaleID, $CurrencyID);
    return $self->getProductData($Product, $hShopData);
}

sub _addProductDetailsLinks {
    my $self = shift;
    my ($Product, $APIBaseURL, $LanguageID, $Links) = @_;

    # now add specific information that is not available for all products

    my @Links = @{$Links};
    if (defined $Product->getWithParent('ImagesSlideShowString')) {
        push @Links, {
            'rel' => 'slideshow',
            'href' => DE_EPAGES::Product::API::Resource::ProductSlideshow->href($APIBaseURL, $Product->guid()),
        };
    }

    if (scalar grep { $_->{'Attribute'}->get('IsVisible') &&
                      $_->{'Attribute'}->type ne 'PreDefCustomerString' }
        @{$Product->get('UserDefinedAttributeValues', $LanguageID)}) {
        push @Links, {
            'rel' => 'custom-attributes',
            'href' => DE_EPAGES::Product::API::Resource::CustomAttributes->href($APIBaseURL, $Product->guid),
        };
    }

    return \@Links;
}

sub _createLinks {
    my $self = shift;
    my ($Product, $APIBaseURL, $LanguageID) = @_;

    my @Links = ();

    # add a link to self.
    push @Links, {
        'rel' => 'self',
        'href' => $self->_buildHref($APIBaseURL, $Product->guid),
    };

    # now add specific information that is not available for all products

    my $SuperProduct = $Product->superProduct();
    if( $SuperProduct ) {
        # this is a sub product - add a link to variations selection.
        push @Links, {
            'rel' => 'variations',
            'href' => DE_EPAGES::Product::API::Resource::ProductVariation->href($APIBaseURL, $SuperProduct->guid),
        };

        # add a link to master product.
        push @Links, {
            'rel' => 'master',
            'href' => $self->_buildHref($APIBaseURL, $SuperProduct->guid),
        };
    }
    elsif ($Product->isMaster) {
        # add a link to variations selection
        push @Links, {
            'rel' => 'variations',
            'href' => DE_EPAGES::Product::API::Resource::ProductVariation->href($APIBaseURL, $Product->guid()),
        };

        # add a link to the lowest price
        if ($Product->existSubProductPrices) {
            push @Links, {
                'rel' => 'lowest-price',
                'href' => DE_EPAGES::Product::API::Resource::ProductLowestPrice->href($APIBaseURL, $Product->guid()),
            };
        }
    }
    # add a link to categories.
    push @Links, {
        'rel' => 'categories',
        'href' => DE_EPAGES::Product::API::Resource::ProductCategories->href($APIBaseURL, $Product->guid),
    };

    # add a link to crossselling products.
    # use cross-selling products of the master product for variations (UNITY-4850)
    if(scalar(@{$Product->getWithParent('CrossSelling')})){
        push @Links, {
            'rel' => 'cross-selling',
            'href' => DE_EPAGES::Product::API::Resource::Product->hrefToProductCrossSelling($APIBaseURL, $Product->guid),
        };
    }

    # add a link to the producttype.
    push @Links, {
        'rel' => 'producttype',
        'href' => DE_EPAGES::Product::API::Resource::ProductType->hrefToProductType($APIBaseURL, $Product->class->guid),
    } if $Product->get('Class');


    return \@Links;
}

sub _loadTaxClass {
    my $self = shift;
    my ($GUID, $Shop) = @_;

    my $ShopTaxClass = $self->_createShopTaxClass($Shop);
    my $TaxClass = $ShopTaxClass->loadTaxClassbyGUID($GUID);

    return $TaxClass;
}

sub _setTaxClass {
    my $self = shift;
    my ($Product, $TaxClass) = @_;

    if($TaxClass->instanceOf('TaxClass') || $TaxClass->instanceOf('TaxClassUS')) {
        $Product->set({'TaxClass' => $TaxClass});

        # Note: if there is a TaxClassDigital, the TaxClass of a product will be ignored!
        $Product->set({'TaxClassDigital' => undef});
    }

    if($TaxClass->instanceOf('TaxClassDigital')) {
        $Product->set({'TaxClassDigital' => $TaxClass});
    }

    return;
}

sub _getAvailability {
    my $self = shift;
    my ($AvailabilityIcon) = @_;

    return undef unless (defined $AvailabilityIcon);
    $AvailabilityIcon =~ s/^Product//;
    $AvailabilityIcon =~ s/Icon$//;
    return $AvailabilityIcon;
}

sub _getShippingMethods {
    my $self = shift;
    my ($aShippingMethods, $APIBaseURL, $LanguageID) = @_;

    # empty list is equal to no restriction. This should be reflected as null in the REST API
    if(!scalar @$aShippingMethods) {
        return undef;
    }
    return [
        map {
            {
                href => "$APIBaseURL/api/shipping-methods/".$_->guid,
                rel => 'related',
                title => $_->get('NameOrAlias', $LanguageID)
            }
        }
        @$aShippingMethods
    ];
}

sub _createSearchKeywords {
    my $self = shift;
    my ($KeywordString) = @_;

    return [] unless defined $KeywordString;

    my @SplittedKeywords = split(/,/, $KeywordString); # split by ','
    my @TrimmedKeywords = map {trim($_)} @SplittedKeywords; # remove leading and trailing spaces
    my @Keywords = grep {length($_) > 0} @TrimmedKeywords; # remove empty strings
    my $aKeywords = [@Keywords];

    return $aKeywords;
}

sub _getImages {
    my $self = shift;
    my ($hProduct, $BaseURL, $hParams) = @_;

    return [
        map { $self->_getImage($_, $hProduct->{$_}, $BaseURL, $hParams) }
        grep { defined $hProduct->{$_} }
        qw ( ImageHotDeal ImageMedium ImageLarge ImageSmall ImageThumbnail )
    ];
}

sub _getImage {
    my $self = shift;
    my ($Classifier, $Value, $BaseUrl, $hParams) = @_;

    my $hSizes = $hParams->{'ImageInformations'}->{$Classifier}->{$Value};

    # Remove 'Image'-prefix (ImageHotdeal -> Hotdeal)
    $Classifier =~ s/^Image//;

    return {
        'isZoomEnabled' => $hParams->{'Formatter'}->bool($hParams->{'UseProductZoom'}),
        'classifier'    => $Classifier,
        'url'           => ($Value =~ /^https?:\/\//) ? $Value : $BaseUrl . '/' . $Value,
        'width'         => defined $hSizes->{'Width'} && $hSizes->{'Width'} =~ m/\d+/ && $hSizes->{'Width'} > 0 ? $hSizes->{'Width'} : undef,
        'height'        => defined $hSizes->{'Height'} && $hSizes->{'Height'} =~ m/\d+/ && $hSizes->{'Height'} > 0 ? $hSizes->{'Height'} : undef,
        'fileSize'      => defined $hSizes->{'Size'} && $hSizes->{'Size'} =~ m/\d+/ && $hSizes->{'Size'} > 0 ? $hSizes->{'Size'} : undef,
    };
}

sub _DAL_GetUpdatedProducts {
    my $self = shift;
    my ($ShopId, $AttributeSet, $UpdatedAfter) = @_;
    return DAL_GetUpdatedProducts($ShopId, $AttributeSet, $UpdatedAfter);
}

sub _LoadObject {
    my $self = shift;
    my ($ObjectID) = @_;
    return LoadObject($ObjectID);
}

sub _createFormatter {
    my $self = shift;
    my ($Shop, $LocaleID) = @_;
    return DE_EPAGES::Shop::API::Resource::Formatter->new($Shop, $LocaleID);
}

sub _buildHref {
    my $self = shift;
    my ($APIBaseURL, $ProductGUID) = @_;
    return "$APIBaseURL/api/products/$ProductGUID";
}

sub _loadLocale {
    my $self = shift;
    my ($LocaleID) = @_;
    return DE_EPAGES::Object::API::Object::Locale->load($LocaleID);
}

sub _getShopContext {
    my $self = shift;
    my ($Shop) = @_;
    return DE_EPAGES::Shop::API::Resource::ShopRequestContext->new($Shop);
}

sub _getCategoryHref {
    my $self = shift;
    my ( $APIBaseURL, $GUID ) = @_;
    return "$APIBaseURL/api/categories/$GUID";
}

sub _getDeliveryWeight {
    my $self = shift;
    my ($hDeliveryWeightData, $Shop) = @_;

    return {} unless (
            defined($hDeliveryWeightData) &&
            defined($hDeliveryWeightData->{'amount'}) &&
            defined($hDeliveryWeightData->{'unit'})
    );

    my $aUnitOfMeasurementInfo = DAL_GetInfoUnitOfMeasurementByAbbreviation($hDeliveryWeightData->{'unit'});
    Error('BadRequest', {
          'Reason' => "delivery weight unit is not valid: $hDeliveryWeightData->{'unit'}",
          'ShopAlias' => $Shop->alias(),
    }) unless ( (scalar @{$aUnitOfMeasurementInfo} > 0) && ($aUnitOfMeasurementInfo->[0]->{'Type'} eq 'mass') );

    return {
        'amount' => $hDeliveryWeightData->{'amount'},
        'unit' => LoadObject($aUnitOfMeasurementInfo->[0]->{'UnitID'})
    };
}

#========================================================================================
# §function     Function
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       Function( $Product );
#----------------------------------------------------------------------------------------
# §description  Deletes the given product.
#
#               If the product IS a sub-product, only the the sub-product will be deleted,
#               the main product automatically does not contain this sub-product anymore
#               without explicitly removing it from the main product as well.
#
#               If the product HAS some sub-products, they will be deleted as well.
#
#               All mappings between the product and shippings will be deleted as well.
#
#               The RemoteSearch cartridge implememts the hook OnDeleteProduct and
#               updates the search index automatically when deleting the product.
#----------------------------------------------------------------------------------------
# §input        $Product | A product | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub deleteProduct {
    my $self = shift;
    my ($Product) = @_;

    $Product->delete();

    return;
}

sub _LoadClassByAlias {
    my $self = shift;
    my ($Alias) = @_;
    return LoadClassByAlias($Alias);
}

sub _Transaction {
    my $self = shift;
    my ($cCode, $cRollback) = @_;
    Transaction($cCode, $cRollback);
    return;
}

sub _createShopTaxClass {
    my $self = shift;
    my ($Shop) = @_;
    return DE_EPAGES::Shop::API::ShopTaxClass->new($Shop);
}


1;
