#========================================================================================
# package      DE_EPAGES::Product::API::Resource::ProductVariation
# state        public
# base         DE_EPAGES::WebService::API::Resource::BaseObjectResource
# description  implements the REST resource for getting variation attributes and sub
#               products from a master product.
#========================================================================================
package DE_EPAGES::Product::API::Resource::ProductVariation;
use base DE_EPAGES::WebService::API::Resource::BaseObjectResource;

use DE_EPAGES::Core::API::Error qw (Error);
use DE_EPAGES::Core::API::PerlTools qw (ElementOf JsonTrue);
use DE_EPAGES::Core::API::Url qw (ParseQueryString);
use DE_EPAGES::Object::API::Factory qw (LoadObjectByPath);
use DE_EPAGES::Object::API::Object::Locale;
use DE_EPAGES::WebService::API::ResourcePager;
use DE_EPAGES::Shop::API::Resource::ShopRequestContext;
use DE_EPAGES::Product::API::Product qw (UpdateSelectedVariationsOfMasterProduct);
use DE_EPAGES::Product::API::ProductVariationService;

use strict;

use constant DEFAULT_PAGE_SIZE => 100;
use constant MAX_PAGE_SIZE => 1000;



sub service {
    return LoadObjectByPath('/RestServices/ProductVariation');
}



sub class {
    return 'Product';
}



sub producedContentTypes {
    shift;
    my ($RequestMethod) = @_;

    return {
        GET => [
            ['get', 'application/vnd.epages.v1+json'],
            ['get', 'application/json'],
        ],
        POST => [
            ['post', 'application/vnd.epages.v1+json'],
            ['post', 'application/json'],
        ],
    }->{$RequestMethod};
}



sub match {
    my $self = shift;
    my ($Request) = @_;

    return unless $Request->uri->path =~ m|^/api/products/([^/]+)/variations/?$|;

    return $1;
}



sub href {
    my $self = shift;
    my ($APIBaseURL, $ProductGUID) = @_;
    return "$APIBaseURL/api/products/$ProductGUID/variations";
}



sub params {
    my $self = shift;
    my ($ActionName, $Object, $Request) = @_;

    if( $ActionName eq 'get' ) {
        my $hQuery = ParseQueryString( $Request->uri->query, 'utf-8' );
        my $LocaleID = $hQuery->{'locale'}->[0];
        my $Page = $hQuery->{'page'}->[0];
        my $ResultsPerPage = $hQuery->{'resultsPerPage'}->[0];
        my $IncludeInvisible = $hQuery->{'includeInvisible'}->[0];
        return [ $Object, $LocaleID, $Page, $ResultsPerPage, $IncludeInvisible ];
    }
    elsif( $ActionName eq 'post' ) {
        my $hQuery = ParseQueryString( $Request->uri->query, 'utf-8' );
        my $LocaleID = $hQuery->{'locale'}->[0];
        my $CurrencyID = $hQuery->{'currency'}->[0];
        return [ $Object, $LocaleID, $CurrencyID ];
    }
    else {
        return [ $Object ];
    }
}

sub post {
    my $self = shift;
    my ($MasterProduct, $LocaleID, $CurrencyID, $hRequestBody) = @_;

    my $Shop = $MasterProduct->getSite();
    my $Context = DE_EPAGES::Shop::API::Resource::ShopRequestContext->new($Shop);
    $LocaleID = $Context->validLocale($LocaleID);
    my $Locale= DE_EPAGES::Object::API::Object::Locale->load($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');

    # storage for objects for creating variations later on
    my $hObjectsForVariation = {};

    # 1st: validate $hRequestBody
    # test if the producttype exists
    Error('BadRequest', {
        'Reason' => 'The given productType does not exists!',
        'productType' => $hRequestBody->{'productType'},
    })   unless $Shop->get('ProductTypes')->existsChild($hRequestBody->{'productType'});

    my $ProductType = $Shop->get('ProductTypes')->child($hRequestBody->{'productType'});
    die unless $ProductType->instanceOf('ProductType');

    # test if there are some sub products and if we have the same producttype
    my $aExistingSubProducts = $MasterProduct->folder('SubProducts')->children;

    if (scalar @{$aExistingSubProducts}){
        Error('BadRequest', {
            'Reason' => 'Product has already sub products from a different productType!',
            'productType' => $hRequestBody->{'productType'},
        }) unless $MasterProduct->class->alias eq $hRequestBody->{'productType'};

        # test if we have already sub products and the same amount of their attributes of attributes in the body
        my $aVariationTypeValues = $MasterProduct->get('VariationTypeValues');
        my $aAttributeNames = [];
        push @$aAttributeNames, $_->{'name'} foreach (@{$hRequestBody->{'variationAttributes'}});
        foreach my $Entry (@$aVariationTypeValues){
            my $Attribute = $Entry->{'Attribute'};
            Error('BadRequest', {
                    'Reason' => 'Not all attributes for existing variation products provided in variationAttributes',
                }) unless ElementOf($Attribute->alias,@$aAttributeNames);
        }
    }

    # store for not getting the objects again
    $hObjectsForVariation->{'ProductType'} = $ProductType;
    $hObjectsForVariation->{'VisibleAfterCreation'} = JsonTrue($hRequestBody->{'visibleAfterCreation'});

    # test feature (max count of attributes we can use)
    my $MaxAttributeCount = $Shop->featureMaxValue( 'Variations' );

    Error('BadRequest', {
        'Reason' => 'Max count of attributes for creating variations exceeded.',
        'MaxCount' => $MaxAttributeCount,
    }) if scalar @{$hRequestBody->{'variationAttributes'}} > $MaxAttributeCount;

    foreach my $hAttribute (@{$hRequestBody->{'variationAttributes'}}){
        Error('BadRequest', {
            'Reason' => 'The name for variationAttribute does not exists!',
        })   unless $hAttribute->{'name'};

        # test if PreDefAttribute exists
        Error('BadRequest', {
            'Reason' => 'The given variationAttribute does not exists!',
            'name' => $hAttribute->{'name'},
        })   unless $ProductType->findChild('Attributes') and
                    $ProductType->folder('Attributes')->existsChild($hAttribute->{'name'});

        # store for not getting the objects again
        my $hAttributeObjects = {};
        $hAttributeObjects->{'Attribute'} = $ProductType->folder('Attributes')->child($hAttribute->{'name'});

        # test Values
        my $AttributeValueFolder = $ProductType->folder('Attributes')->child($hAttribute->{'name'})->folder('PreDefAttributes');

        # store for not getting the objects again
        my $DefaultValue = $hAttribute->{'defaultForExistingProducts'} // undef;
        Error('BadRequest', {
            'Reason' => 'No values provided!',
            'attributeName' => $hAttribute->{'name'},
        }) unless ($hAttribute->{'values'} and scalar @{$hAttribute->{'values'}}) || $DefaultValue;

        if(($hAttribute->{'values'} and scalar @{$hAttribute->{'values'}})){
            foreach my $hValue (@{$hAttribute->{'values'}}) {
                # is value empty or not?
                Error('BadRequest', {
                    'Reason' => 'The value field is empty!',
                    'attributeName' => $hAttribute->{'name'},
                }) unless $hValue->{'value'};

                # exists Value and value in shop?
                Error('BadRequest', {
                    'Reason' => 'Attribute value not exists in attributes for the given producttype!',
                    'value' => $hValue->{'value'},
                }) unless $AttributeValueFolder->existsChild($hValue->{'value'});

                push @{$hAttributeObjects->{'Values'}}, $AttributeValueFolder->child($hValue->{'value'});
            }
        }

        if($DefaultValue){
            # exists Value and value in shop?
            Error('BadRequest', {
                'Reason' => 'value for defaultForExistingProducts not exists in attributes for the given producttype!',
                'defaultForExistingProducts' => $DefaultValue,
            }) unless $AttributeValueFolder->existsChild($DefaultValue);
            $hAttributeObjects->{'defaultForExistingProducts'} = $AttributeValueFolder->child($DefaultValue);
        }

        push @{$hObjectsForVariation->{'Attributes'}}, $hAttributeObjects;
    }

    # set the wanted product type on the product
    $MasterProduct->set({'Class' => $ProductType});

    # assign given product type to product and create variations with given values and set visibility
    my $ProductVariationService = DE_EPAGES::Product::API::ProductVariationService->new($MasterProduct, $LanguageID);

    # test if we have already sub products and missing attributes on it
    # If we have already variations in place and create new PT attributes with new values
    # then the existing variations must be updated with the new attribute default values
    # this must be done by providing a field 'defaultForExistingProducts' in the attribute section
    # for each new unused attribute
    my $UpdateSelectedVariationsIsRequired = undef;
    foreach my $SubProduct (@$aExistingSubProducts) {
        foreach my $hAttributeStore (@{$hObjectsForVariation->{'Attributes'}}) {
            my $DefaultValueForExistingProducts = $hAttributeStore->{'defaultForExistingProducts'};
            # test if the attribute is set on each sub product  otherwise set a default value
            my $Attribute = $hAttributeStore->{'Attribute'};
            unless($SubProduct->get($Attribute->alias)){
                Error('BadRequest', {
                    'Reason' => 'existing sub products needs to be updated with a default value for the new attribute; field defaultForExistingProducts not provided!',
                }) unless $DefaultValueForExistingProducts;
                $SubProduct->set({ $Attribute->alias => $DefaultValueForExistingProducts });
                $UpdateSelectedVariationsIsRequired = 1;
            }
        }
    }
    # update after setting new defaults if needed
    UpdateSelectedVariationsOfMasterProduct( [$MasterProduct] ) if $UpdateSelectedVariationsIsRequired;

    # now create products and return success
    $ProductVariationService->createSubProducts($hObjectsForVariation);

    return undef;
}

sub get {
    my $self = shift;
    my ($MasterProduct, $LocaleID, $Page, $PageSize, $IncludeInvisible) = @_;

    $self->_checkIfProductIsMasterProduct($MasterProduct);

    my $Shop = $MasterProduct->getSite();
    my $Context = DE_EPAGES::Shop::API::Resource::ShopRequestContext->new($Shop);
    $LocaleID = $Context->validLocale($LocaleID);
    my $Locale= DE_EPAGES::Object::API::Object::Locale->load($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');

    my $Pager = DE_EPAGES::WebService::API::ResourcePager->new({
        'Page' => $Page,
        'PageSize' => $PageSize,
        'DefaultPageSize' => DEFAULT_PAGE_SIZE,
        'MaxPageSize' => MAX_PAGE_SIZE,
    });

    my $ProductVariationService = DE_EPAGES::Product::API::ProductVariationService->new($MasterProduct, $LanguageID);

    my $aSubProducts = $ProductVariationService->getSubProducts($IncludeInvisible);
    my $aVariationAttributes = $ProductVariationService->getVariationAttributes();

    my $APIBaseURL = $Context->apiBaseUrl();
    my $hPage = $Pager->pagedItems($aSubProducts, sub {
        return $ProductVariationService->getVariationsForSubproduct($_, $APIBaseURL);
    });
    $hPage->{'variationAttributes'} = $aVariationAttributes;
    $hPage->{'links'} = [];

    return $hPage;
}

sub _checkIfProductIsMasterProduct {
    my $self = shift;
    my ($Product) = @_;

    my $aVariationTypeValues = $Product->get('VariationTypeValues');
    Error('ResourceNotFound', {
        MasterProductGUID => $Product->guid }
    ) if 0 == scalar @$aVariationTypeValues;

    return;
}

1;
