#========================================================================================
# package      DE_EPAGES::Product::API::SlideshowService
# state        public
#----------------------------------------------------------------------------------------
# description  Manages the product slideshow
#========================================================================================
package DE_EPAGES::Product::API::SlideshowService;

use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Product::API::Constants qw ( UNITY_IMAGE_ATTRIBUTES );
use DE_EPAGES::Shop::API::Resource::Formatter;
use List::MoreUtils qw ( uniq );

use strict;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       new( $Product, $ImageBaseUrl );
#----------------------------------------------------------------------------------------
# description  Constructor
#----------------------------------------------------------------------------------------
# input        $Product | Product | DE_EPAGES::Product::API::Object::Product
# input        $ImageBaseUrl | ImageBaseUrl | String
# return       $self | self | DE_EPAGES::Product::API::SlideshowService
#========================================================================================
sub new {
    my $class = shift;
    my ($Product, $ImageBaseUrl) = @_;
    my $Shop = $Product->getSite();
    my $self  = bless {
        'Product' => $Product,
        'ImageBaseUrl' => $ImageBaseUrl,
        'isZoomEnabled' => _createFormatter($Shop)->bool($Shop->get('UseProductZoom')),
    }, $class;
    return $self;
}

sub _createFormatter {
    my ($Shop) = @_;
    return DE_EPAGES::Shop::API::Resource::Formatter->new($Shop, $Shop->get('LocaleID'));
}

sub getSlideshow {
    my $self = shift;

    my $Product = $self->{'Product'};
    my ($hInfo) = $Product->get(['ImageInformations', 'ImagesSlideShow']);

    if(not defined $hInfo->{'ImageInformations'} ){
        $Product->addImageInformations([UNITY_IMAGE_ATTRIBUTES], undef);
        $Product->extendImageInformationsSlideShowString(undef);
        $hInfo->{'ImageInformations'} = $Product->get('ImageInformations');
    }

    my $aSlideshowItems = [];
    for my $SlideshowImage (@{$hInfo->{'ImagesSlideShow'}}) {
        push @$aSlideshowItems, {
            'name'     => $SlideshowImage->{'ImageName'},
            'sizes'    => $self->_createListOfImageSizes($SlideshowImage, $hInfo->{'ImageInformations'}),
        };
    }

    my $hSlideshow = {
        'items' => $aSlideshowItems,
    };

    return $hSlideshow;
}

sub getSlideshowImage {
    my $self = shift;
    my ($ImageName) = @_;

    my $Product = $self->{'Product'};
    my ($hInfo) = $Product->get(['ImageInformations', 'ImagesSlideShow']);

    my $hSlideshowItems = {};
    for my $SlideshowImage (@{$hInfo->{'ImagesSlideShow'}}) {
        my $SlideshowImageName = $SlideshowImage->{'ImageName'};
        $hSlideshowItems->{$SlideshowImageName} = {
            'name' => $SlideshowImage->{'ImageName'},
            'sizes' => $self->_createListOfImageSizes($SlideshowImage, $hInfo->{'ImageInformations'}),
        };
    }

    return $hSlideshowItems->{$ImageName};
}

sub getSlideshowSequence {
    my $self = shift;

    my $hSlideshow = $self->getSlideshow();
    my $aSlideshowItems = $hSlideshow->{'items'};

    my @SlideshowImageSequence = map { $_->{'name'} } @$aSlideshowItems;
    my $aSlideshowImageSequence = \@SlideshowImageSequence;

    return $aSlideshowImageSequence;
}

sub setSlideshowSequence {
    my $self = shift;
    my ($aSlideshowImageSequence) = @_;

    my $IsValid = $self->_checkIfSequenceIsValid($aSlideshowImageSequence);
    if(!$IsValid) {
        Error('BadRequest', {'Reason' => 'New sequence has to be a permutation of the old sequence!'});
    }

    my $ImagesSlideShowString = join(';', @$aSlideshowImageSequence);

    my $Product = $self->{'Product'};
    $Product->set({'ImagesSlideShowString' => $ImagesSlideShowString});

    $Product->updatePageCache();

    return;
}

sub _checkIfSequenceIsValid {
    my $self = shift;
    my ($aSlideshowImageSequence) = @_;

    my $aCurrentSlideshowImages = $self->getSlideshowSequence();

    # check if new slideshow sequence contains the same amount of images as the old one.
    my $nCurrentImages = @$aCurrentSlideshowImages;
    my $nNewImages = @$aSlideshowImageSequence;
    if($nNewImages != $nCurrentImages) {
        return 0;
    }

    # check if new slideshow sequence does not contain duplicates
    my $nNewImagesWithoutDuplicates = uniq @$aSlideshowImageSequence;
    if($nNewImagesWithoutDuplicates != $nNewImages) {
        return 0;
    }

    # check if a set of both sequence elements contains the same size as one sequence.
    my @Union = (@$aSlideshowImageSequence, @$aCurrentSlideshowImages);
    my @UniqesOfUnion = uniq @Union;
    my $nUniqesOfUnion = @UniqesOfUnion;
    if($nUniqesOfUnion != $nCurrentImages) {
        return 0;
    }

    return 1;
}

sub _createListOfImageSizes {
    my $self = shift;
    my ($SlideshowImage, $hImageInformations) = @_;

    my $ImageWebPath = $self->_getImageWebPath();

    my $hSlideshowImageSizes = $SlideshowImage->{'Sizes'};
    my @SizeNames = keys %$hSlideshowImageSizes;

    # get resized slideshow image informations
    my $hSlideShowImageInfo = {};
    if (exists $hImageInformations->{'ImagesSlideShowString'} ) {
        $hSlideShowImageInfo = $hImageInformations->{'ImagesSlideShowString'}->{$SlideshowImage->{'ImageName'}}->{'Sizes'};
    }

    my $aSizes = [];
    for my $SizeName (@SizeNames) {
        my $ImageUrl = $ImageWebPath.'/'.$hSlideshowImageSizes->{$SizeName};

        my $FullClassifier = 'Image'. $SizeName;
        my $hImageInfo = $hSlideShowImageInfo->{$FullClassifier}->{$hSlideshowImageSizes->{$SizeName}};

        push @$aSizes, {
            'url' => $ImageUrl,
            'classifier' => $SizeName,
            'isZoomEnabled' => $self->{'isZoomEnabled'},
            'width'         => defined $hImageInfo->{'Width'} && $hImageInfo->{'Width'} =~ m/\d+/ && $hImageInfo->{'Width'} > 0 ? $hImageInfo->{'Width'} : undef,
            'height'        => defined $hImageInfo->{'Height'} && $hImageInfo->{'Height'} =~ m/\d+/ && $hImageInfo->{'Height'} > 0 ? $hImageInfo->{'Height'} : undef,
            'fileSize'      => defined $hImageInfo->{'Size'} && $hImageInfo->{'Size'} =~ m/\d+/ && $hImageInfo->{'Size'} > 0 ? $hImageInfo->{'Size'} : undef,
        };
    }

    return $aSizes;
}



sub _getImageWebPath {
    my $self = shift;

    my $ImageBaseURL = $self->{'ImageBaseUrl'};
    my $Product = $self->{'Product'};
    my $WebPath = $Product->get('WebPath');
    my $ImageWebPath = $ImageBaseURL.$WebPath;

    return $ImageWebPath;
}

1;