package DE_EPAGES::Product::t::ProductResource;
use base DE_EPAGES::Shop::API::Test;

use strict;
use Test::More tests => 3;
use JSON;

use DE_EPAGES::Object::API::XML::ObjectImportHandler qw( ImportObjects );
use DE_EPAGES::Product::API::Resource::CustomAttributes;

## no critic qw(ProhibitMagicNumbers)

sub test {
    my $self = shift;

    $self->test_get_master_de;
    $self->test_get_master_en;
    $self->test_get_sub_de;
}

sub test_get_master_de {
    my $self = shift;

    my $Resource = DE_EPAGES::Product::API::Resource::CustomAttributes->new;
    my $ProductFolder = $self->shop->get('ProductFolder');

    my $Product = $ProductFolder->child('MyJacket');
    my $SSLDomain = $self->shop->get('ProtocolAndServerSSL');
    my $ImageBasePath = $SSLDomain.$Product->get('WebPath');

    my $hFound = _sortProducts($Resource->get( $Product, 'de_DE' )->{items});
    my $hExpected = $self->_createExpected('test_get_master_de', $ImageBasePath);

    is_deeply( $hFound, $hExpected, 'GET MyJacket de_DE' );
}

sub test_get_master_en {
    my $self = shift;

    my $Resource = DE_EPAGES::Product::API::Resource::CustomAttributes->new;
    my $ProductFolder = $self->shop->get('ProductFolder');

    my $Product = $ProductFolder->child('MyJacket');
    my $SSLDomain = $self->shop->get('ProtocolAndServerSSL');
    my $ImageBasePath = $SSLDomain.$Product->get('WebPath');

    my $hFound = _sortProducts($Resource->get( $Product, 'en_AU' )->{items});
    my $hExpected = $self->_createExpected('test_get_master_en', $ImageBasePath);

    is_deeply( $hFound, $hExpected, 'GET MyJacket en_AU' );
}


sub test_get_sub_de {
    my $self = shift;

    my $Resource = DE_EPAGES::Product::API::Resource::CustomAttributes->new;
    my $ProductFolder = $self->shop->get('ProductFolder');

    my $Product = $ProductFolder->child('MyJacket')->child('SubProducts')->child('JacketRedL');
    my $SSLDomain = $self->shop->get('ProtocolAndServerSSL');
    my $ImageBasePath = $SSLDomain.$Product->get('WebPath');

    my $hFound = _sortProducts($Resource->get( $Product, 'de_DE' )->{items});
    my $hExpected = $self->_createExpected('test_get_sub_de', $ImageBasePath);

    is_deeply( $hFound, $hExpected, 'GET JacketRedL d' );
}

sub init {
    my $self = shift;

    $self->SUPER::init;
    $self->shop->set({
        DomainName => 'test-domain',
        HasSSLCertificate => 1
    });

    note 'importing CustomAttributesResource.xml';
    ImportObjects('CustomAttributesResource.xml', $self->shop->pathString);
}

sub _createExpected {
    my $self = shift;
    my ($Test, $ImageBasePath) = @_;
    my $Boolean = $Test =~ /test_get_master/ ? 'true' : 'false';

    return [
        {
            'displayKey' => $Test =~ /en/ ? 'Boolean' : 'Ja/Nein',
            'singleValue' => JSON::true,
            'values' => [
                {
                    'displayValue' => $Boolean,
                    'value' => $Boolean,
                }
            ],
            'type' => 'boolean',
            'key' => 'Boolean'
        },
        {
            'displayKey' => $Test =~ /en/ ? 'Date' : 'Datum',
            'singleValue' => JSON::true,
            'values' => [
                {
                    'displayValue' => $Test =~ /en/ ? '22/07/2014' : '22.07.2014',
                    'value' => '2014-07-22'
                }
            ],
            'type' => 'date',
            'key' => 'Date'
        },
        {
            'displayKey' => $Test =~ /en/ ? 'DateTime' : 'Datum und Zeit',
            'singleValue' => JSON::true,
            'values' => [
                {
                    'displayValue' => $Test =~ /en/ ? '22/07/2014 9:58 AM' : '22.07.2014 09:58',
                    'value' => '2014-07-22T09:58:05+02:00'
                }
            ],
            'type' => 'datetime',
            'key' => 'DateTime'
        },
        {
            'displayKey' => $Test =~ /en/ ? 'File' : 'Datei',
            'singleValue' => JSON::true,
            'values' => [
                {
                    'displayValue' => $ImageBasePath.'/image.png',
                    'value' => $ImageBasePath.'/image.png'
                }
            ],
            'type' => 'url',
            'key' => 'File'
        },
        {
            'displayKey' => $Test =~ /en/ ? 'Float' : 'Dezimalzahl',
            'singleValue' => JSON::true,
            'values' => [
                {
                    'displayValue' => $Test =~ /en/ ? '12,345.6789' : '12.345,6789',
                    'value' => '12345.6789'
                }
            ],
            'type' => 'number',
            'key' => 'Float'
        },
        {
            'displayKey' => $Test =~ /en/ ? 'Integer' : 'Ganze Zahl',
            'singleValue' => JSON::true,
            'values' => [
                {
                    'displayValue' => $Test =~ /en/ ? '123,456,789' : '123.456.789',
                    'value' => '123456789'
                }
            ],
            'type' => 'number',
            'key' => 'Integer'
        },
        {
            'displayKey' => $Test =~ /en/ ? 'LocalizedFile' : "Datei, sprachabh\x{e4}ngig",
            'singleValue' => JSON::true,
            'values' => [
                {
                    'displayValue' => $Test =~ /en/ ? '' : $ImageBasePath.'/Handbuch.pdf',
                    'value' => $Test =~ /en/ ? undef : $ImageBasePath.'/Handbuch.pdf'
                }
            ],
            'type' => 'url',
            'key' => 'LocalizedFile'
        },
        {
            'displayKey' => $Test =~ /en/ ? 'LocalizedString' : 'Text, sprachabhngig',
            'singleValue' => JSON::true,
            'values' => [
                {
                    'displayValue' => $Test =~ /en/ ? '' : 'Achtung: Restposten. Jetzt zugreifen!',
                    'value' => $Test =~ /en/ ? undef : 'Achtung: Restposten. Jetzt zugreifen!'
                }
            ],
            'type' => 'string',
            'key' => 'LocalizedString'
        },
        {
            'displayKey' => $Test =~ /en/ ? 'Money' : 'Geldbetrag',
            'singleValue' => JSON::true,
            'values' => [
                {
                    'displayValue' => $Test =~ /en/ ? '12,345.6789' : '12.345,6789',
                    'value' => '12345.6789'
                }
            ],
            'type' => 'number',
            'key' => 'Money'
        },
        {
            'displayKey' =>  $Test =~ /en/ ? 'PreDefMultiLocalizedString' : "Mehrfachauswahl, sprachabh\x{e4}ngig",
            'singleValue' => JSON::false,
            'values' => [
                {
                    'displayValue' => $Test =~ /en/ ? 'Value1' : 'Wert 1',
                    'value' => 'Value1'
                },
                {
                    'displayValue' => $Test =~ /en/ ? 'Value 3' : 'Wert 3',
                    'value' => 'Value 3'
                }
            ],
            'type' => 'string',
            'key' => 'PreDefMultiLocalizedString'
        },
        {
            'displayKey' => $Test =~ /en/ ? 'PreDefMultiString' : "Mehrfachauswahl, nicht sprachabh\x{e4}ngig",
            'singleValue' => JSON::false,
            'values' => [
                {
                    'displayValue' => 'Value 1',
                    'value' => 'Value1'
                },
                {
                    'displayValue' => 'Value 2',
                    'value' => 'Value 2'
                }
            ],
            'type' => 'string',
            'key' => 'PreDefMultiString'
        },
        {
            'displayKey' => $Test =~ /en/ ? 'String' : 'Text',
            'singleValue' => JSON::true,
            'values' => [
                {
                    'displayValue' => 'Blue Angel',
                    'value' => 'Blue Angel'
                }
            ],
            'type' => 'string',
            'key' => 'String'
        },
        {
            'displayKey' => $Test =~ /en/ ? 'Time' : 'Zeit',
            'singleValue' => JSON::true,
            'values' => [
                {
                    'displayValue' =>  $Test =~ /en/ ? '9:58 AM' : '09:58',
                    'value' => '09:58:00'
                }
            ],
            'type' => 'time',
            'key' => 'Time'
        },
    ];
}

sub _sortProducts {
    my ($Products) = @_;
    my @SortedProducts = sort { $a->{key} cmp $b->{key} } @$Products;
    return \@SortedProducts;
}

__PACKAGE__->runScript;
