#========================================================================================
# package      DE_EPAGES::Product::t::unit::ProductService
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for DE_EPAGES::Product::API::ProductService
#               testing methods:
#               - getAllProducts()
#               - getAllUpdatedProducts()
#========================================================================================
package DE_EPAGES::Product::t::unit::ProductService;
use base DE_EPAGES::Core::API::Test;

use DE_EPAGES::Product::API::ProductService;
use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use DE_EPAGES::Test::Mock::Mockify;
use DE_EPAGES::Test::Mock::Mockery;
use Test::More;

use strict;

##no critic qw(ProhibitMagicNumbers)

sub test {
    my $self = shift;

    $self->test_getAllUpdatedProducts_noDatesDefined();
    $self->test_getAllUpdatedProducts_attributes_noProductInInterval();
    $self->test_getAllUpdatedProducts_attributes_oneProductInInterval();
    $self->test_getAllUpdatedProducts_attributes_twoProductsInInterval();
    $self->test_getAllUpdatedProducts_stocklevel_noProductInInterval();
    $self->test_getAllUpdatedProducts_stocklevel_oneProductInInterval();
    $self->test_getAllUpdatedProducts_stocklevel_twoProductsInInterval();
    $self->test_getAllUpdatedProducts_stocklevelAndAttributes();

    done_testing();
}

sub test_getAllUpdatedProducts_noDatesDefined {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks();
    my $ProductService = $Mockery->get('SUT');
    my $Shop = $Mockery->get('Shop');
    my $AttributesUpdatedSince = undef;
    my $StockLevelUpdatedSince = undef;

    my $aProducts = $self->_getUpdatedProductAttributes(
        $ProductService->getAllUpdatedProducts($Shop, $AttributesUpdatedSince, $StockLevelUpdatedSince));

    is_deeply($aProducts, [], "$SubTestName - test if no products were returned.");
}



sub test_getAllUpdatedProducts_attributes_noProductInInterval {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks();
    my $ProductService = $Mockery->get('SUT');
    my $Shop = $Mockery->get('Shop');
    my $AttributesUpdatedSince = $self->_createDateTime(7);
    my $StockLevelUpdatedSince = undef;

    my $aProducts = $self->_getUpdatedProductAttributes(
        $ProductService->getAllUpdatedProducts($Shop, $AttributesUpdatedSince, $StockLevelUpdatedSince));

    is_deeply($aProducts, [], "$SubTestName - test if no products were returned.");
}



sub test_getAllUpdatedProducts_attributes_oneProductInInterval {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks_oneProduct('Product1');
    my $ProductService = $Mockery->get('SUT');
    my $Shop = $Mockery->get('Shop');
    my $AttributesUpdatedSince = $self->_createDateTime(5);
    my $StockLevelUpdatedSince = undef;

    my $aProducts = $self->_getUpdatedProductAttributes(
        $ProductService->getAllUpdatedProducts($Shop, $AttributesUpdatedSince, $StockLevelUpdatedSince));

    my $aExpectedProducts = [
        [
            {
                'Alias' => 'Product1',
                'IsVisible' => 1
            },
            '2016-01-20T14:23:26Z',
        ]
    ];
    is_deeply($aProducts, $aExpectedProducts, "$SubTestName - test if one product was returned.");
}



sub test_getAllUpdatedProducts_attributes_twoProductsInInterval {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks_twoProducts('Product1', 'Product2');
    my $ProductService = $Mockery->get('SUT');
    my $Shop = $Mockery->get('Shop');
    my $AttributesUpdatedSince = $self->_createDateTime(2);
    my $StockLevelUpdatedSince = undef;

    my $aProducts = $self->_getUpdatedProductAttributes(
        $ProductService->getAllUpdatedProducts($Shop, $AttributesUpdatedSince, $StockLevelUpdatedSince));

    my $aExpectedProducts = [
        [
            {
                'Alias' => 'Product1',
                'IsVisible' => 1
            },
            '2016-01-20T14:23:26Z',
        ],
        [
            {
                'Alias' => 'Product2',
                'IsVisible' => 1
            },
            '2016-01-12T12:56:14Z',
        ]
    ];
    is_deeply($aProducts, $aExpectedProducts, "$SubTestName - test if two products were returned.");
}



sub test_getAllUpdatedProducts_stocklevel_noProductInInterval {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks();
    my $ProductService = $Mockery->get('SUT');
    my $Shop = $Mockery->get('Shop');
    my $AttributesUpdatedSince = undef;
    my $StockLevelUpdatedSince = $self->_createDateTime(7);

    my $aProducts = $self->_getUpdatedProductAttributes(
        $ProductService->getAllUpdatedProducts($Shop, $AttributesUpdatedSince, $StockLevelUpdatedSince));

    is_deeply($aProducts, [], "$SubTestName - test if no products were returned.");
}



sub test_getAllUpdatedProducts_stocklevel_oneProductInInterval {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks_oneProduct('Product3');
    my $ProductService = $Mockery->get('SUT');
    my $Shop = $Mockery->get('Shop');
    my $AttributesUpdatedSince = undef;
    my $StockLevelUpdatedSince = $self->_createDateTime(4);

    my $aProducts = $self->_getUpdatedProductAttributes(
        $ProductService->getAllUpdatedProducts($Shop, $AttributesUpdatedSince, $StockLevelUpdatedSince));

    my $aExpectedProducts = [
        [
            {
                'Alias' => 'Product3',
                'IsVisible' => 1
            },
            '2016-01-20T14:23:26Z',
        ],
    ];
    is_deeply($aProducts, $aExpectedProducts, "$SubTestName - test if one product was returned.");
}



sub test_getAllUpdatedProducts_stocklevel_twoProductsInInterval {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks_twoProducts('Product1', 'Product3');
    my $ProductService = $Mockery->get('SUT');
    my $Shop = $Mockery->get('Shop');
    my $AttributesUpdatedSince = undef;
    my $StockLevelUpdatedSince = $self->_createDateTime(2);

    my $aProducts = $self->_getUpdatedProductAttributes(
        $ProductService->getAllUpdatedProducts($Shop, $AttributesUpdatedSince, $StockLevelUpdatedSince));
    my $aExpectedProducts = [
        [
            {
                'Alias' => 'Product1',
                'IsVisible' => 1
            },
            '2016-01-20T14:23:26Z',
        ],
        [
            {
                'Alias' => 'Product3',
                'IsVisible' => 1
            },
            '2016-01-12T12:56:14Z',
        ]
    ];
    is_deeply($aProducts, $aExpectedProducts, "$SubTestName - test if two products were returned.");
}



sub test_getAllUpdatedProducts_stocklevelAndAttributes {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks_threeProducts('Product1', 'Product2', 'Product3');
    my $ProductService = $Mockery->get('SUT');
    my $Shop = $Mockery->get('Shop');
    my $AttributesUpdatedSince = $self->_createDateTime(2);
    my $StockLevelUpdatedSince = $self->_createDateTime(2);

    my $aProducts = $self->_getUpdatedProductAttributes(
        $ProductService->getAllUpdatedProducts($Shop, $AttributesUpdatedSince, $StockLevelUpdatedSince));

    my $aExpectedProducts = [
        [
            {
                'Alias' => 'Product1',
                'IsVisible' => 1
            },
            '2016-01-20T14:23:26Z',
        ],
        [
            {
                'Alias' => 'Product2',
                'IsVisible' => 1
            },
            '2016-01-12T12:56:14Z',
        ],
        [
            {
                'Alias' => 'Product3',
                'IsVisible' => 1
            },
            '2015-11-14T17:35:12Z',
        ],
        ];

    is_deeply($aProducts, $aExpectedProducts, "$SubTestName - test if all products were returned.");
}



sub _createMocks {
    my $self = shift;

    my $Mockery = DE_EPAGES::Test::Mock::Mockery->new();

    $Mockery->extend($self->_createCommonMocks());

    $Mockery->register('SUT', ['ProductClass'], sub {
        my ($ProductClass) = @_;
        my $Mockify = DE_EPAGES::Test::Mock::Mockify->new('DE_EPAGES::Product::API::ProductService', []);
        $Mockify->addMock('_DAL_GetUpdatedProducts', sub {
            return ();
        });
        return $Mockify->getMockObject();
    });

    return $Mockery->assemble();
}



sub _createMocks_oneProduct {
    my $self = shift;
    my ($ProductMockName) = @_;

    my $Mockery = DE_EPAGES::Test::Mock::Mockery->new();

    $Mockery->extend($self->_createCommonMocks());

    $Mockery->register('SUT', [$ProductMockName], sub {
        my ($Product1) = @_;
        my $Mockify = DE_EPAGES::Test::Mock::Mockify->new('DE_EPAGES::Product::API::ProductService', []);
        $Mockify->addMock('_DAL_GetUpdatedProducts', sub {
            return ([[3145, '2016-01-20 15:23:26']]); ## no critic (ProhibitMagicNumbers)
        });
        $Mockify->addMockWithReturnValueAndParameterCheck('_LoadObject', $Product1, [{'int' => 3145}]);
        return $Mockify->getMockObject();
    });

    return $Mockery->assemble();
}



sub _createMocks_twoProducts {
    my $self = shift;
    my ($Product1MockName, $Product2MockName) = @_;

    my $Mockery = DE_EPAGES::Test::Mock::Mockery->new();

    $Mockery->extend($self->_createCommonMocks());

    $Mockery->register('SUT', [$Product1MockName, $Product2MockName], sub {
        my ($Product1, $Product2) = @_;
        my $Mockify = DE_EPAGES::Test::Mock::Mockify->new('DE_EPAGES::Product::API::ProductService', []);
        $Mockify->addMock('_DAL_GetUpdatedProducts', sub {
            return [[3145, '2016-01-20 15:23:26'], [6423, '2016-01-12 13:56:14']]; ## no critic (ProhibitMagicNumbers)
        });
        $Mockify->addMock('_LoadObject', sub {
            my $MockedSelf = shift;
            my ($ObjectId) = @_;
            if ( $ObjectId == 3145) { ## no critic (ProhibitMagicNumbers)
                return $Product1;
            }
            elsif ( $ObjectId == 6423) { ## no critic (ProhibitMagicNumbers)
                return $Product2;
            }
        });
        return $Mockify->getMockObject();
    });

    return $Mockery->assemble();
}



sub _createMocks_threeProducts {
    my $self = shift;
    my ($Product1MockName, $Product2MockName, $Product3MockName) = @_;

    my $Mockery = DE_EPAGES::Test::Mock::Mockery->new();

    $Mockery->extend($self->_createCommonMocks());

    $Mockery->register('SUT', [$Product1MockName, $Product2MockName, $Product3MockName], sub {
        my ($Product1, $Product2, $Product3) = @_;
        my $Mockify = DE_EPAGES::Test::Mock::Mockify->new('DE_EPAGES::Product::API::ProductService', []);
        $Mockify->addMock('_DAL_GetUpdatedProducts', sub {
            return [[3145, '2016-01-20 15:23:26'], [5653, '2015-11-14 18:35:12'], [6423, '2016-01-12 13:56:14']]; ## no critic (ProhibitMagicNumbers)
        });
        $Mockify->addMock('_LoadObject', sub {
            my $MockedSelf = shift;
            my ($ObjectId) = @_;
            if ( $ObjectId == 3145) { ## no critic (ProhibitMagicNumbers)
                return $Product1;
            }
            elsif ( $ObjectId == 6423) { ## no critic (ProhibitMagicNumbers)
                return $Product2;
            }
            elsif ( $ObjectId == 5653) { ## no critic (ProhibitMagicNumbers)
                return $Product3;
            }
        });
        return $Mockify->getMockObject();
    });

    return $Mockery->assemble();
}



sub _createCommonMocks {
    my $self = shift;

    my $Mockery = DE_EPAGES::Test::Mock::Mockery->new();

    $Mockery->register('Shop', ['Product1', 'Product2', 'Product3', 'InvisibleProduct'], sub {
        my ($Product1, $Product2, $Product3, $InvisibleProduct) = @_;
        my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Shop::API::Object::Shop', {
            'ID' => '32165',
            'Products' => [$Product1, $Product2, $Product3, $InvisibleProduct],
        });
        return $Mockify->getMockObject();
    });
    $Mockery->register('Product1', [], sub {
        my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Product::API::Object::Product', {
            'Alias' => 'Product1',
            'IsVisible' => 1,
        });
        return $Mockify->getMockObject();
    });
    $Mockery->register('Product2', [], sub {
        my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Product::API::Object::Product', {
            'Alias' => 'Product2',
            'IsVisible' => 1,
        });
        return $Mockify->getMockObject();
    });
    $Mockery->register('Product3', [], sub {
        my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Product::API::Object::Product', {
            'Alias' => 'Product3',
            'IsVisible' => 1,
        });
        return $Mockify->getMockObject();
    });
    $Mockery->register('InvisibleProduct', [], sub {
        my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Product::API::Object::Product', {
            'Alias' => 'InvisibleProduct',
            'IsVisible' => 0,
        });
        return $Mockify->getMockObject();
    });
    $Mockery->register('ProductClass', ['LastUpdateProfileForAttributes', 'LastUpdateProfileForStocklevel'], sub {
        my ($LastUpdateProfileForAttributes, $LastUpdateProfileForStocklevel) = @_;
        my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Object::API::Object::Class', {
            'Alias' => 'ProductClass',
        });
        $Mockify->addMock('lastUpdateProfile', sub {
            my $MockedSelf = shift;
            my ($ProfileName) = @_;
            if ($ProfileName eq 'RestApiResponseAttributes') {
                return $LastUpdateProfileForAttributes;
            }
            elsif ($ProfileName eq 'RestApiStockLevel') {
                return $LastUpdateProfileForStocklevel;
            }
        });
        return $Mockify->getMockObject();
    });
    $Mockery->register('LastUpdateProfileForAttributes', [], sub {
        my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Object::API::Object::LastUpdateProfile', {
            'Alias' => 'LastUpdateProfileForAttributes',
        });
        $Mockify->addMock('getLastUpdate', sub {
            my $MockedSelf = shift;
            my ($Product) = @_;
            my $ProductAlias = $Product->alias();
            if ($ProductAlias eq 'Product1') {
                return $self->_createDateTime(6);
            }
            elsif ($ProductAlias eq 'Product2') {
                return $self->_createDateTime(3);
            }
            elsif ($ProductAlias eq 'Product3') {
                return undef;
            }
        });
        return $Mockify->getMockObject();
    });
    $Mockery->register('LastUpdateProfileForStocklevel', [], sub {
        my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Object::API::Object::LastUpdateProfile', {
            'Alias' => 'LastUpdateProfileForStocklevel',
        });
        $Mockify->addMock('getLastUpdate', sub {
            my $MockedSelf = shift;
            my ($Product) = @_;
            my $ProductAlias = $Product->alias();
            if ($ProductAlias eq 'Product1') {
                return $self->_createDateTime(3);
            }
            elsif ($ProductAlias eq 'Product2') {
                return undef
            }
            elsif ($ProductAlias eq 'Product3') {
                return $self->_createDateTime(4);
            }
        });
        return $Mockify->getMockObject();
    });

    return $Mockery;
}

sub _getUpdatedProductAttributes {
    my $self = shift;
    my ($aProducts) = @_;
    return [
        sort { $a->[0]->{Alias} cmp $b->[0]->{Alias} }
        map {[{Alias => $_->[0]->alias(), IsVisible => $_->[0]->get('IsVisible')}, $_->[1]]}
        @$aProducts
    ];
}



sub _createDateTime {
    my $self = shift;
    my ($Day) = @_;
    my $DateTime = DateTime->new(
        year       => 2015,
        month      => 4,
        day        => $Day,
        hour       => 14,
        minute     => 0,
        second     => 0,
        nanosecond => 0,
        time_zone  => 'Europe/Berlin',
    );
    return $DateTime;
}

__PACKAGE__->runScript();
