#========================================================================================
# package      DE_EPAGES::Product::t::unit::SlideshowService
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for DE_EPAGES::Product::API::SlideshowService
#========================================================================================
package DE_EPAGES::Product::t::unit::SlideshowService;
use base DE_EPAGES::Core::API::Test;

use DE_EPAGES::Product::API::SlideshowService;
use DE_EPAGES::Test::Mock::EpagesObject;
use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use DE_EPAGES::Test::Mock::Mockify qw (WasCalled);
use Test::Exception;
use Test::More;
use JSON;

use strict;



sub test {
    my $self = shift;

    $self->test_getSlideshow();

    $self->test_getSlideshowImage();

    $self->test_getSlideshowSequence();

    $self->test_setSlideshowSequence();
    $self->test_setSlideshowSequence_TooLessElements();
    $self->test_setSlideshowSequence_TooManyElement();
    $self->test_setSlideshowSequence_DuplicatedElements();
    $self->test_setSlideshowSequence_InvalidElements();

    done_testing();
}



sub test_getSlideshow {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Product = $self->_createProduct(1);
    my $ImageBaseUrl = 'http://base.url';
    my $SlideshowService = DE_EPAGES::Product::API::SlideshowService->new($Product, $ImageBaseUrl);

    my $aSlideshow = _sortProducts($SlideshowService->getSlideshow()->{items});
    my $aExpectedSlideshow = $self->_createExpectedSlideshow();

    is_deeply($aSlideshow, $aExpectedSlideshow, "$SubTestName - test if slideshow hash is correct");
}



sub test_getSlideshowImage {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Product = $self->_createProduct(0);
    my $ImageBaseUrl = 'http://base.url';
    my $SlideshowService = DE_EPAGES::Product::API::SlideshowService->new($Product, $ImageBaseUrl);

    my $aSlideshow = _sortProducts([$SlideshowService->getSlideshowImage('AnotherImage')]);
    my $aExpectedSlideshow = [$self->_createExpectedSlideshowImage()];

    is_deeply($aSlideshow, $aExpectedSlideshow, "$SubTestName - test if slideshow hash is correct");
}



sub test_getSlideshowSequence {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Product = $self->_createProduct(1);
    my $ImageBaseUrl = 'http://base.url';
    my $SlideshowService = DE_EPAGES::Product::API::SlideshowService->new($Product, $ImageBaseUrl);

    my $aSlideshowSequence = $SlideshowService->getSlideshowSequence();

    my $aExpectedSlideshowSequence = ['TestImage1', 'AnotherImage'];
    is_deeply($aSlideshowSequence, $aExpectedSlideshowSequence, "$SubTestName - test if slideshow sequence array is correct");
}



sub test_setSlideshowSequence {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Product = $self->_createProduct(1);
    my $ImageBaseUrl = 'http://base.url';
    my $SlideshowService = DE_EPAGES::Product::API::SlideshowService->new($Product, $ImageBaseUrl);

    my $aNewSequence = ['AnotherImage', 'TestImage1'];
    $SlideshowService->setSlideshowSequence($aNewSequence);

    my $aSlideshowSequence = $SlideshowService->getSlideshowSequence();
    my $aExpectedSlideshowSequence = ['TestImage1', 'AnotherImage'];
    is_deeply($aSlideshowSequence, $aExpectedSlideshowSequence, "$SubTestName - test if slideshow sequence array is correct");
    is($aSlideshowSequence->[0], 'TestImage1', "$SubTestName - test if the 1st element is the 1st of the new sequence");
    is($aSlideshowSequence->[1], 'AnotherImage', "$SubTestName - test if the 2nd element is the 2nd of the new sequence");
    ok(WasCalled($Product, 'updatePageCache'), "$SubTestName - test if updatePageCache was called.");
}



sub test_setSlideshowSequence_TooLessElements {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Product = $self->_createProduct(1);
    my $ImageBaseUrl = 'http://base.url';
    my $SlideshowService = DE_EPAGES::Product::API::SlideshowService->new($Product, $ImageBaseUrl);

    my $aNewSequence = ['TestImage1'];

    throws_ok( sub {
        $SlideshowService->setSlideshowSequence($aNewSequence);
    }, qr/BadRequest: Reason=New sequence has to be a permutation of the old sequence!/, ## no critic (ProhibitComplexRegexes)
    "$SubTestName - test if error is thrown if new sequence has too less elements");
}



sub test_setSlideshowSequence_TooManyElement {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Product = $self->_createProduct();
    my $ImageBaseUrl = 'http://base.url';
    my $SlideshowService = DE_EPAGES::Product::API::SlideshowService->new($Product, $ImageBaseUrl);

    my $aNewSequence = ['AnotherImage', 'TestImage1', 'TestImage2'];

    throws_ok( sub {
        $SlideshowService->setSlideshowSequence($aNewSequence);
    }, qr/BadRequest: Reason=New sequence has to be a permutation of the old sequence!/, ## no critic (ProhibitComplexRegexes)
    "$SubTestName - test if error is thrown if new sequence has too many elements");
}



sub test_setSlideshowSequence_DuplicatedElements {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Product = $self->_createProduct();
    my $ImageBaseUrl = 'http://base.url';
    my $SlideshowService = DE_EPAGES::Product::API::SlideshowService->new($Product, $ImageBaseUrl);

    my $aNewSequence = ['TestImage1', 'TestImage1'];

    throws_ok( sub {
        $SlideshowService->setSlideshowSequence($aNewSequence);
    }, qr/BadRequest: Reason=New sequence has to be a permutation of the old sequence!/, ## no critic (ProhibitComplexRegexes)
    "$SubTestName - test if error is thrown if new sequence has duplicated");
}



sub test_setSlideshowSequence_InvalidElements {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Product = $self->_createProduct();
    my $ImageBaseUrl = 'http://base.url';
    my $SlideshowService = DE_EPAGES::Product::API::SlideshowService->new($Product, $ImageBaseUrl);

    my $aNewSequence = ['AnImage', 'YetAnotherImage'];

    throws_ok( sub {
        $SlideshowService->setSlideshowSequence($aNewSequence);
    }, qr/BadRequest: Reason=New sequence has to be a permutation of the old sequence!/, ## no critic (ProhibitComplexRegexes)
    "$SubTestName - test if error is thrown if new sequence contains images which do no belong to the product");
}



sub _createProduct {
    my $self = shift;
    my ($UseProductZoom) = @_;

    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Product::API::Object::Product', {
        'WebPath' => '/test/web/path',
        'ImagesSlideShow' => $self->_createImageSlideShow(),
        'Site' => $self->_createShop($UseProductZoom),
        'ImageInformations' => {},
    });
    $Mockify->addMockWithReturnValue('updatePageCache', undef);

    return $Mockify->getMockObject();
}

sub _createShop {
    my $self = shift;
    my ($UseProductZoom) = @_;

    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Shop::API::Object::Shop', {
        'UseProductZoom' => $UseProductZoom,
        'LocaleID' => 'de_DE',
    });
    $Mockify->addMockWithReturnValueAndParameterCheck('siteLocale', 1, ['string']);

    return $Mockify->getMockObject();
}



sub _createImageSlideShow {
    my $self = shift;
    return [
        {
            'ImageName' => 'TestImage1',
            'Sizes' => {
                'reallybig' => 'testimg1_notsmall.jpg',
                'smaller' => 'testimg1_smaller.jpg',
                'small' => 'testimg1_small.jpg',
            },
        },
        {
            'ImageName' => 'AnotherImage',
            'Sizes' => {
                'big' => 'testimg_big.jpg',
                'small' => 'img_small.jpg',
            },
        },
    ];
}



sub _createExpectedSlideshow {
    my $self = shift;
    return [
        {
            'name' => 'TestImage1',
            'sizes' => [
                {
                    'url' => 'http://base.url/test/web/path/testimg1_notsmall.jpg',
                    'classifier' => 'reallybig',
                    'isZoomEnabled' => JSON::true,
                    'width' => undef,
                    'height' => undef,
                    'fileSize' => undef

                },
                {
                    'url' => 'http://base.url/test/web/path/testimg1_small.jpg',
                    'classifier' => 'small',
                    'isZoomEnabled' => JSON::true,
                    'width' => undef,
                    'height' => undef,
                    'fileSize' => undef

                },
                {
                    'url' => 'http://base.url/test/web/path/testimg1_smaller.jpg',
                    'classifier' => 'smaller',
                    'isZoomEnabled' => JSON::true,
                    'width' => undef,
                    'height' => undef,
                    'fileSize' => undef
                }
           ]
        },
        {
            'name' => 'AnotherImage',
            'sizes' => [
                {
                    'url' => 'http://base.url/test/web/path/testimg_big.jpg',
                    'classifier' => 'big',
                    'isZoomEnabled' => JSON::true,
                    'width' => undef,
                    'height' => undef,
                    'fileSize' => undef
                },
                {
                    'url' => 'http://base.url/test/web/path/img_small.jpg',
                    'classifier' => 'small',
                    'isZoomEnabled' => JSON::true,
                    'width' => undef,
                    'height' => undef,
                    'fileSize' => undef
                },
            ]
        }
    ];
}



sub _createExpectedSlideshowImage {
    my $self = shift;
    return {
        'name' => 'AnotherImage',
        'sizes' => [
            {
                'url' => 'http://base.url/test/web/path/testimg_big.jpg',
                'classifier' => 'big',
                'isZoomEnabled' => JSON::false,
                'width' => undef,
                'height' => undef,
                'fileSize' => undef
            },
            {
                'url' => 'http://base.url/test/web/path/img_small.jpg',
                'classifier' => 'small',
                'isZoomEnabled' => JSON::false,
                'width' => undef,
                'height' => undef,
                'fileSize' => undef
            },
        ]
    };
}



sub _sortProducts {
    my ($Products) = @_;
    my @SortedProducts;
    foreach my $Item (@$Products) {
        my @sizes = sort { $a->{classifier} cmp $b->{classifier} } @{$Item->{sizes}};
        delete $Item->{sizes};
        $Item->{sizes} = \@sizes;
        push @SortedProducts, $Item;
    };
    return \@SortedProducts
}

__PACKAGE__->runScript();
