#========================================================================================
# package      DE_EPAGES::Shop::API::ShopUriBuilder
# state        public
# description  provides functionallity to generate uris on a shop basis.
#========================================================================================
package DE_EPAGES::Shop::API::ShopUriBuilder;

use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::PerlTools qw ( IsA isValid );
use DE_EPAGES::Core::API::TestParameter qw ( TestObject );
use DE_EPAGES::Object::API::Factory qw ( LoadRootObject );
use DE_EPAGES::Shop::API::Object::Shop;
use URI;
use URI::Escape;
use URI::URL;

use strict;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $ShopUriBuilder = ShopUriBuilder->new( $ShopOrObject );
#----------------------------------------------------------------------------------------
# description  Creates a ShopUriBuilder.
#----------------------------------------------------------------------------------------
# input        $Object | a shop or a object refering to a shop | string
# return       ShopUriBuilder instance | DE_EPAGES::Shop::API::ShopUriBuilder
#========================================================================================
sub new {
    my $class = shift;
    my ($Object, $System)  = @_;
    my $self  = bless {}, $class;
    $System //= LoadRootObject();
    $self->_initialize($Object, $System);

    return $self;
}

#========================================================================================
# function     buildUri
# state        public
#----------------------------------------------------------------------------------------
# syntax       buildUri( $PathTemplate, $hPathParams, $hQueryParams );
#----------------------------------------------------------------------------------------
# description  generates a Public Uri for the builders specific shop using the
#               passed resource path and replacing all key/value pairs
#               from the input parameters.
#----------------------------------------------------------------------------------------
# input        $PathTemplate | resource path, may contain path variables | string
# input        $hPathParams | paramaters to replace in the resource path | refhash
# input        $hQueryParams | paramaters to add query parameters to url (optional) | refhash
# return       Full path uri | string
#========================================================================================
sub buildUri {
    my $self = shift;
    my ($PathTemplate, $hPathParams, $hQueryParams) = @_;

    $self->_checkSslCertificateForPublicBaseUrl();

    my $ResolvedPath = $self->_resolvePath($PathTemplate, $hPathParams);

    my $PathWithQueryParams = $self->_addQueryParameters($ResolvedPath, $hQueryParams);

    return sprintf('%s/%s', $self->{'PublicBaseUrl'}, $PathWithQueryParams);
}

#========================================================================================
# function     buildPrivateUri
# state        public
#----------------------------------------------------------------------------------------
# syntax       buildPrivateUri( $PathTemplate, $hPathParams, $hQueryParams );
#----------------------------------------------------------------------------------------
# description  generates a Private Uri for the builders specific shop using the
#               passed resource path and replacing all key/value pairs
#               from the input parameters.
#----------------------------------------------------------------------------------------
# input        $PathTemplate | resource path, may contain path variables | string
# input        $hPathParams | paramaters to replace in the resource path | refhash
# input        $hQueryParams | paramaters to add query parameters to url (optional) | refhash
# return       Full path uri | string
#========================================================================================
sub buildPrivateUri {
    my $self = shift;
    my ($PathTemplate, $hPathParams, $hQueryParams) = @_;

    my $ResolvedPath = $self->_resolvePath($PathTemplate, $hPathParams);

    my $PathWithQueryParams = $self->_addQueryParameters($ResolvedPath, $hQueryParams);

    return sprintf('%s/%s', $self->{'PrivateBaseUrl'}, $PathWithQueryParams);
}
#========================================================================================
# function     _initialize
# state        private
#----------------------------------------------------------------------------------------
# description  initializes the builder.
#----------------------------------------------------------------------------------------
# input        $Object | shop or shop object | DE_EPAGES::Shop::API::Object::Shop
#========================================================================================
sub _initialize {
    my $self = shift;
    my ($Object, $System) = @_;

    my $PrivateBaseUrl = $System->get('EPagesJBaseURL');
    $self->_setBasesUrl($PrivateBaseUrl, 'PrivateBaseUrl');

    if(IsA($Object, 'DE_EPAGES::Shop::API::Object::Shop')) {

        my $PublicBaseUrl = $Object->get('RESTAPIURL');
        $self->_setBasesUrl($PublicBaseUrl, 'PublicBaseUrl');

        return;
    }

    TestObject('Object', $Object, 'DE_EPAGES::Object::API::Object::Object');

    my $Shop = $Object->getSite();
    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');

    my $PublicBaseUrl = $Shop->get('RESTAPIURL');
    $self->_setBasesUrl($PublicBaseUrl, 'PublicBaseUrl');

    return;
}


#========================================================================================
# function     _setBasesUrl
# state        private
#----------------------------------------------------------------------------------------
# description  sets the PublicBaseUrl or PrivateBaseUrl of the builder.
#----------------------------------------------------------------------------------------
# input        $BaseUrl | base url to set | string
#----------------------------------------------------------------------------------------
# input        $TypeUrl | private or public url to set | string
#========================================================================================
sub _setBasesUrl {
    my $self = shift;
    my ($BaseUrl, $TypeUrl) = @_;

    $self->{$TypeUrl} = $self->_stripTrailingSlashes($BaseUrl);
}

#========================================================================================
# function     _stripTrailingSlashes
# state        private
#----------------------------------------------------------------------------------------
# description  strips all trailing slashes of the passed string.
#----------------------------------------------------------------------------------------
# input        $String | string to strip. | string
# return       input string stripped off all trailing slashes | string
#========================================================================================
sub _stripTrailingSlashes {
    my $self = shift;
    my ($String) = @_;

    my $Length = length($String);
    my $StripAmount = 0;

    for(my $i = $Length; $i > 0; $i--) {
        my $Char = substr($String, $i-1, 1);
        last unless ($Char eq '/');
        $StripAmount++;
    }

    my $ChopedString = substr($String, 0, $Length - $StripAmount);
    return $ChopedString;
}

#========================================================================================
# function     _stripLeadingSlashes
# state        private
#----------------------------------------------------------------------------------------
# description  strips all leading slashes of the passed string.
#----------------------------------------------------------------------------------------
# input        $String | string to strip. | string
# return       input string stripped off all leading slashes | string
#========================================================================================
sub _stripLeadingSlashes {
    my $self = shift;
    my ($String) = @_;

    my $Length = length($String);
    my $StripAmount = 0;

    for(my $i = 0; $i < $Length; $i++) {
        my $Char = substr($String, $i, 1);
        last unless ($Char eq '/');
        $StripAmount++;
    }

    my $ChopedString = substr($String, $StripAmount, $Length);
    return $ChopedString;
}

#========================================================================================
# function     _resolvePath
# state        private
#----------------------------------------------------------------------------------------
# description  resolves input vars in path template.
#----------------------------------------------------------------------------------------
# input        $PathTemplate | path template | string
# input        $hPathVars | hash containing variable values to resolve | refhash
# return       resolved path | string
#========================================================================================
sub _resolvePath {
    my $self = shift;
    my ($PathTemplate, $hPathVars) = @_;

    return '' unless isValid($PathTemplate);
    return $PathTemplate unless (ref $hPathVars eq 'HASH');

    my $UriString = $self->_stripLeadingSlashes($PathTemplate);

    foreach my $Key (keys %$hPathVars) {
        my $EncodedValue = uri_escape($hPathVars->{$Key});
        $UriString =~ s/\{$Key\}/$EncodedValue/g;
    }

    return $UriString;
}

sub _checkSslCertificateForPublicBaseUrl {
    my $self = shift;

    if(not (substr($self->{'PublicBaseUrl'}, 0, 5) eq 'https')) {
        Error('NoValidSSLCertificateError');
    }

    return;
}

sub _addQueryParameters {
    my $self = shift;
    my ($ResolvedPath, $hQueryParams) = @_;

    $hQueryParams //= {};
    # new perl 5.38.0 or higher makes a difference between undef and empty:
    # key => '' results in ...&key=&... whereas key => undef is ...&key&...
    # we expect ...&key=&... for both undef and empty
    map { $hQueryParams->{$_} = '' unless defined $hQueryParams->{$_} } keys %$hQueryParams;

    my $UriBuilder = URI->new($ResolvedPath);
    $UriBuilder->query_form($hQueryParams);

    return $UriBuilder->as_string();
}

1;
