#========================================================================================
# package      DE_EPAGES::Shop::API::TLE::ShopHandler
# state        public
#----------------------------------------------------------------------------------------
# description  Implements system-wide TLE variables for shops
#               <ul>
#                   <li>Shop</li>
#                   <li>ShopRoot</li>
#               </ul>
#========================================================================================
package DE_EPAGES::Shop::API::TLE::ShopHandler;

use strict;
use LWP::UserAgent;
use JSON;

use Scalar::Util qw ( blessed );
use DE_EPAGES::Core::API::Url qw ( StringIsURL );
use DE_EPAGES::Core::API::Log qw ( GetLog LogError );
use DE_EPAGES::Core::API::PerlTools qw ( ElementOf );
use DE_EPAGES::Core::API::TestParameter qw ( TestInteger TestObject TestString);
use DE_EPAGES::Object::API::Factory qw ( LoadObject );
use DE_EPAGES::Object::API::Table::Country qw ( ExistsCountry GetInfoCountry );
use DE_EPAGES::Shop::API::Shop qw ( GetNextNo );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Shop::API::Feature qw ( ExistsFeatureByAlias );

sub ErrorDefinition {
    return {
        'VIMEO_API_ERROR' => "GET request to the Vimeo API to \"#RequestURL\" failed with code #Status.\n"
        . "Response content: \"#Content\""
    };
}

#========================================================================================
# function     RegisterHandlerProc
# state        private
#----------------------------------------------------------------------------------------
# syntax       $RegisterHandlerProc( $hParams );
#----------------------------------------------------------------------------------------
# description  this function registers this attribute handle on the processor via hook
#               function on hook TLEProcessorRegistration
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments keys :
#               <ul>
#                   <li>Processor - tle processor - object
#               </ul> | ref.hash
#========================================================================================
sub RegisterHandlerProc {
    my ($Params) = @_;
    GetLog->debug( 'ShopHandler.RegisterHandlerProc' );

    __PACKAGE__->new()->register( $Params->{'Processor'} );
    return;
}

#========================================================================================
# function     new
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler = new DE_EPAGES::Object::API::TLE::ShopHandler;
# example      my $PrefixHandler  = new DE_EPAGES::Object::API::TLE::ShopHandler;
#----------------------------------------------------------------------------------------
# description  create a new handler
#----------------------------------------------------------------------------------------
# return       $Handler | blessed object | object
#========================================================================================
sub new { bless {}, shift; }

#========================================================================================
# function     register
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler->register( $Processor );
#----------------------------------------------------------------------------------------
# description  this function register this handle on the processor
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
#========================================================================================
sub register {
    my $self = shift;
    my ($Processor) = @_;
    GetLog->debug( 'ShopHandler.register' );

    $Processor->registerHandler('VariableHandler', $self, 'Shop');
    $Processor->registerHandler('VariableHandler', $self, 'ShopRoot');
    $Processor->registerHandler('FunctionHandler', $self, 'FeaturedActionAvailable');
    $Processor->registerHandler('FunctionHandler', $self, 'FEATURED_MENU_COUNT');
    $Processor->registerHandler('FunctionHandler', $self, 'VATID_EXAMPLE');
    $Processor->registerHandler('FunctionHandler', $self, 'SuggestNextNo');
    $Processor->registerHandler('FunctionHandler', $self, 'HasUntranslatedEntries');
    $Processor->registerHandler('FunctionHandler', $self, 'GetVideoTutorial');
    $Processor->registerHandler('FunctionHandler', $self, 'ParseAdditionalHtmlHead');
    $Processor->registerHandler('FunctionHandler', $self, 'ShowNextStepsTile');
    return;
}

#========================================================================================
# function     tle
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Value = $Handler->tle( $Processor, $TLEName);
# example      my $Value = $Handler->tle( $Processor, 'Name');
#----------------------------------------------------------------------------------------
# description  This function is called by tle processor to get the value of a tle
#               variable.
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
# input        $TLEName | tle name | string
# return       $Value | return value | integer
#========================================================================================
sub tle {
    my $self = shift;
    my ($Processor, $TLEName) = @_;
    my $ServletVars = $Processor->getInputContext;

    my $ShopHash = $ServletVars->{'TLEPROCESSOR_SHOP'};
    if (not defined $ShopHash) {
        my $Object = $Processor->tle('Object');
        return undef unless defined $Object;
        return undef unless $Object->instanceOf('Object');

        my $Shop = $Object->getSite();
        return undef unless $Shop->instanceOf('Shop');
        $ShopHash = $Shop->tleHash;
        $ServletVars->{'TLEPROCESSOR_SHOP'} = $ShopHash;
    }
    return $ShopHash->{'WebPath'} if $TLEName eq 'ShopRoot';
    return $ShopHash;
}

#========================================================================================
# function     existsTLE
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Value = $Handler->existsTLE( $Processor, $TLEName);
# example      my $Value = $Handler->existsTLE( $Processor, 'Products_Count');
#----------------------------------------------------------------------------------------
# description  This function is called by the tle processor to test if a TLE variable
#               is available.
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
# input        $TLEName | tle name | string
# return       $Value | yes or no | bool
#========================================================================================
sub existsTLE {
    my $self = shift;
    my ($Processor, $TLEName) = @_;
    return 1 if ElementOf($TLEName, qw (Shop ShopRoot));
    return undef;
}

#========================================================================================
# function     FeaturedActionAvailable
# state        public
#----------------------------------------------------------------------------------------
# syntax       #FeaturedActionAvailable(#ShopObject, #Object, #Feature)
# example      #IF( #FeaturedActionAvailable(#Shop.Object, #Action.Object) EQ "Allowed" ) ... #ENDIF
# example      #IF( #FeaturedActionAvailable(#Shop.Object, #UNDEF, "ProductRating") NE "NotVisible") ... #ENDIF
#----------------------------------------------------------------------------------------
# description  Returns the feature status for the action
#               If both, a regular and an excluding feature are defined,
#               'NotVisible' will be returned regardless of the state of the regular feature.
#               Applicable to every object with a 'feature' attribute.
#----------------------------------------------------------------------------------------
# input        #ShopObject | shop object | object
# input        #Object | object with 'feature' attribute (usually an action) | object or #UNDEF
# input        #Feature | feature object or feature name,
#                          can be omitted if avaible on #Object (default: #Object.Feature) | object or string or #UNDEF
# return       #Status |    'NotAllowed'   action not allowed for feature AND
#                                           no excluding feature active
#                         or 'NotVisible'   action not visible for feature OR
#                                           excluding feature is active
#                         or 'Allowed'      action allowed for feature (not active, but visible) AND
#                                           no excluding feature active
#                         or 'LimitReached' feature limit reached for this action AND
#                                           no excluding feature active
#                         or 'NotAssigned'  action not assigned to a feature |string
#========================================================================================
sub FeaturedActionAvailable { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParameter) = @_;
    my ($Shop, $Object, $Feature) = @$aParameter;
    TestObject( 'Shop', $Shop );

    if (not defined $Feature) {
        return 'NotAssigned' unless defined $Object;
        $Feature = $Object->get('Feature');
    }

    if (defined $Object && $Object->class->existsAttribute('ExcludingFeature')){
        my $ExcludingFeature = $Object->get('ExcludingFeature');
        if (defined $ExcludingFeature and $Shop->featureMaxValue($ExcludingFeature)){
            return 'NotVisible';
        }
    }

    return 'NotAssigned' unless defined $Feature;
    return 'NotVisible' if (!blessed($Feature) && !ExistsFeatureByAlias($Feature));

    my $MaxValue = $Shop->featureMaxValue($Feature);
    if (not $MaxValue) {
        return 'NotAllowed' if $Shop->featureIsVisible($Feature);
        return 'NotVisible';
    }

    my $CurrentValue = $Shop->featureCurrentValue($Feature);
    return 'Allowed' if $CurrentValue < $MaxValue;
    return 'LimitReached';
}

#========================================================================================
# function     FeaturedActionAvailable
# state        public
#----------------------------------------------------------------------------------------
# syntax       #FeaturedActionAvailable(#ShopObject, #ActionObject, #FeatureObject)
# example      #LOCAL("DisplayFeature", #FeaturedActionAvailable(#Shop.Object, #Action.Object))
#                 #DisplayFeature
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Returns the the feature status for the action
#----------------------------------------------------------------------------------------
# input        #ShopObject | shop object | object
# input        #ActionObject | action object | object
# input        #FeatureObject | feature object (default: Action.Feature) | object
# return       #Status |    'NotAllowed'   action not allowed for feature
#                         or 'NotVisible'   action not visible for feature
#                         or 'Allowed'      action allowed for feature
#                         or 'LimitReached' feature limit reached for this action
#                         or 'NotAssigned'  action not assigned to a feature
#                         or 'NoShop'       action called without shop   |string
#========================================================================================
sub FEATURED_MENU_COUNT {
    my $self = shift;
    my ($Processor, $raParams) = @_;
    my ($Shop, $Name) = @$raParams;
    return 'NoShop' unless defined $Shop;
    TestString( 'Name', $Name );

    # get PageType from ObjectPageType from Processor Context
    my $PageType = $Processor->getContext()->{'PageType'}->get('PageType');

    my $Count = 0;
    foreach my $SubMenu (@{$PageType->menu( $Name )->subMenus}) {
        my $Action = LoadObject($SubMenu->action->id);
        my $Feature = $Action->get('Feature');
        if (not defined $Feature) {
            $Count++;
            next;
        }
        my $MaxValue = $Shop->featureMaxValue($Feature);
        if (not $MaxValue) {
            $Count++ if $Shop->featureIsVisible($Feature);
            next;
        }
        $Count++;
    }

    # get count of menu items
    return $Count;
}

#========================================================================================
# function     ShowNextStepsTile
# state        public
#----------------------------------------------------------------------------------------
# syntax       #ShowNextStepsTile(#ShopObject)
# example      #LOCAL("ShowNextStepsTile", #ShowNextStepsTile(#Shop.Object))
#                 #ShowNextStepsTile
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Returns the value if ShowNextStepsTile should be displayed
#----------------------------------------------------------------------------------------
# input        #ShopObject | shop object | object
# return       #ShowNextStepsTile |  true if tile should be shown | boolean
#========================================================================================
sub ShowNextStepsTile {
    my $self = shift;
    my ($Processor, $raParams) = @_;
    my ($Shop) = @$raParams;
    return 'NoShop' unless defined $Shop;

    my $HideNextStepsTile = $Shop->get('HideNextStepsTile');
    my $ShowNextStepsTile = 0;
    my $ShopCreatedOnDate = $Shop->get('CreationDate')->ymd;
    my $Today = DateTime->now();
    my $CheckDateInPast = $Today->add(months => -6);

    if ((!$HideNextStepsTile) && (($ShopCreatedOnDate gt $CheckDateInPast) || (($ShopCreatedOnDate lt $CheckDateInPast) && ($Shop->get('ClosedByMerchant'))))) {
        $ShowNextStepsTile = 1;
    }

    return $ShowNextStepsTile;
}

#========================================================================================
# function     VATID_EXAMPLE
# state        public
#----------------------------------------------------------------------------------------
# syntax       #VATID_EXAMPLE(#CountryID)
# example      #LOCAL("VATID", #VATID_EXAMPLE(#Shop.Address.Country.CountryID))
#                 #VATID
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Returns an example for a valid vatid for the given country.
#----------------------------------------------------------------------------------------
# input        #CountryID | country identifier | integer
# return       #VATID | vatid example | string
#========================================================================================
sub VATID_EXAMPLE { ## no critic qw( RequireCommentForExportedFunctions )
    my $self = shift;
    my ($Processor, $raParams) = @_;
    my ($CountryID) = @$raParams;
    return 'XX 123456789' unless defined $CountryID;
    TestInteger( 'CountryCode', $CountryID );

    my $CountryCode = '';
    $CountryCode = GetInfoCountry($CountryID)->{'Code2'} if ExistsCountry($CountryID);

    my %VATIDs = (
        'CY' => 'CY 12345678X',
        'IE' => 'IE 1234567X',
        'GB' => 'GB 123456789',
        'NL' => 'NL 123456789B01',
        'AT' => 'AT U12345678',
        'BE' => 'BE 0123456789',
        'GR' => 'EL 123456789',
        'EE' => 'EE 123456789',
        'LV' => 'LV 12345678901',
        'DK' => 'DK 12345678',
        'PT' => 'PT 123456789',
        'FI' => 'FI 12345678',
        'DE' => 'DE 123456789',
        'SE' => 'SE 123456789012',
        'BG' => 'BG 123456789',
        'LU' => 'LU 12345678',
        'SI' => 'SI 12345678',
        'MT' => 'MT 12345678',
        'LT' => 'LT 123456789',
        'RO' => 'RO 0123456789',
        'IT' => 'IT 12345678901',
        'PL' => 'PL 0123456789',
        'CZ' => 'CZ 123456789',
        'ES' => 'ES X12345678',
        'SK' => 'SK 0123456789',
        'HU' => 'HU 12345678}',
        'FR' => 'FR X0123456789',
    );

    my $VATID = $VATIDs{$CountryCode};
    $VATID = $CountryCode . ' 123456789' unless defined $VATID;

    return $VATID;
}

#========================================================================================
# function     SuggestNextNo
# state        public
#----------------------------------------------------------------------------------------
# syntax       #SuggestNextNo(#ShopObject, #ClassName)
# syntax       #SuggestNextNo(#ShopObject, #ClassName, #Object)
#----------------------------------------------------------------------------------------
# example      #LOCAL("NewAlias", #SuggestNextNo(#Shop.Object, "Invoice", #Order.Object))
#                 #NewAlias
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Returns a suggestion for a new alias.
#----------------------------------------------------------------------------------------
# input        #ShopObject | shop object | DE_EPAGES::Shop::API::Object::Shop
# input        #ClassName | class name | String
# input        #Object | parent object (optional) | DE_EPAGES::Object::API::Object::Object
# return       #NewAlias | new alias suitable for given class name | String
#========================================================================================
sub SuggestNextNo {
    my $self = shift;
    my ($Processor, $aParameter) = @_;
    my ($Shop, $ClassName, $Object) = @$aParameter;
    TestString('ClassName', $ClassName);
    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');
    TestObject('Object', $Object, 'DE_EPAGES::Object::API::Object::Object') if $Object;
    return GetNextNo($Shop, $ClassName, $Object, 0);
}

#========================================================================================
# function     HasUntranslatedEntries
# state        public
#----------------------------------------------------------------------------------------
# syntax       #HasUntranslatedEntries(#Object, #AttributeName)
#----------------------------------------------------------------------------------------
# example      #IF(#HasUntranslatedEntries(#Object, "Name"))
#                 #INCLUDE("BubbleHelp")
#               #ENDIT
#----------------------------------------------------------------------------------------
# description  Checks for a language dependend attribute of a given shop object if there
#               are empty entries in any of the installed shop languages for an already
#               filled attribute
#----------------------------------------------------------------------------------------
# input        #Object | object | DE_EPAGES::Object::API::Object::Object
# input        #AttributeName | attribute name | String
# return       #HasUntranslatedEntries | result | Boolean
#========================================================================================
sub HasUntranslatedEntries {
    my $self = shift;
    my ($Processor, $aParameter) = @_;
    my ($Object, $AttributeName) = @$aParameter;
    TestObject('Object', $Object, 'DE_EPAGES::Object::API::Object::Object');
    TestString('AttributeName', $AttributeName);

    # only check existing attributes
    return unless $Object->class->existsAttribute($AttributeName);

    # TLE function is only valid for shop objects
    my $Shop = $Object->getSite();
    return unless $Shop->instanceOf('Shop');

    # check if there is at least one empty and one defined entry
    my $HasDefinedEntry = 0;
    my $HasUndefinedEntry = 0;

    # loop over all shop languages
    my @ShopLanguageIDs = map { $_->{'LanguageID'} } @{$Shop->get('Languages')};
    foreach my $LanguageID (@ShopLanguageIDs) {
        if (defined $Object->get($AttributeName, $LanguageID)) {
            $HasDefinedEntry = 1;
        } else {
            $HasUndefinedEntry = 1;
        }
    }

    # only return true if at least one entry is set and another is empty
    return ($HasDefinedEntry && $HasUndefinedEntry);
}

#========================================================================================
# function     GetVideoTutorial
# state        public
#----------------------------------------------------------------------------------------
# syntax       #BLOCK("GetVideoTutorial", #VideoID, #IframeWidth, #IframeHeight, #IframeClass)
#                   <HTML with TLEs>
#               #ENDBLOCK
#----------------------------------------------------------------------------------------
# description  Performs a GET request to Vimeo and writes the received data to the TLE
#               variables #ThumbnailURL and #Video.
#----------------------------------------------------------------------------------------
# input        #VideoID | video ID / Shortened Link | integer / string
# input        #IframeWidth | width of the embed iframe (default: 640) | integer
# input        #IframeHeight | height of the embed iframe (default: 360) | integer
# input        #IframeClass | Css class of the iframe | string
# return       $Content | processed template content with the replaced ThumbnailURL
#                   and the Video JSON object | string
#========================================================================================
sub GetVideoTutorial {
    my $self = shift;
    my ($Processor, $raParams, $cTemplate) = @_;
    my ($VideoID, $IframeWidth, $IframeHeight, $IframeClass) = @$raParams;
    $IframeWidth //= 640;
    $IframeHeight //= 360;

    TestInteger( 'IframeWidth', $IframeWidth );
    TestInteger( 'IframeHeight', $IframeHeight );
    TestString( 'IframeClass', $IframeClass ) if $IframeClass;

    GetLog->debug( "ShopHandler.GetVideoTutorial($VideoID)" );

    my $RealVideoURL = '';
    #-- If we have a short URL set for the video
    if ( StringIsURL($VideoID) ) {
        #-- Find out where it leads. It must be set up as a 301. Do NOT follow the 301.
        my $Ua = LWP::UserAgent->new(max_redirect => 0);
        my $Response = $Ua->get($VideoID);

        #-- This is the real location of the Video, should be a Vimeo link
        $RealVideoURL = $Response->header('location');
        if ( !StringIsURL($RealVideoURL) ) {
            my %Error = (
                Code => 'VIMEO_URL_ERROR',
                ErrorDefinitionPackage => __PACKAGE__,
                Vars => {
                    'ShortURL' => $VideoID,
                    'Location'  => $RealVideoURL,
                }
            );
            LogError(DE_EPAGES::Core::API::Error->new(%Error));
            return;
        }
    } else {
        #-- A normal Video ID
        TestInteger( 'VideoID', $VideoID );
        $RealVideoURL = 'https%3A//vimeo.com/' . $VideoID;
    }

    #-- Ask the Vimeo API for information on the video
    #-- Note that not all video URLs are accepted. This one, for e.g., at the time of writing, does not: https://vimeo.com/manage/videos/826789111/
    #-- So if providing a short url, the simplest format with just the video id is most fool proof
    my $RequestURL = 'https://vimeo.com/api/oembed.json?url='
        . $RealVideoURL . '&width=' . $IframeWidth . '&height=' . $IframeHeight;

    my $Response = LWP::UserAgent->new->get($RequestURL);

    # error handling
    unless ($Response->is_success) {
        my %Error = (
            Code => 'VIMEO_API_ERROR',
            ErrorDefinitionPackage => __PACKAGE__,
            Vars => {
                'RequestURL' => $RequestURL,
                'Status'  => $Response->code,
                'Content' => $Response->content
            }
        );
        LogError(DE_EPAGES::Core::API::Error->new(%Error));

        return;
    }

    my $hDecodedContent = decode_json($Response->decoded_content);

    my $ThumbnailURL = $hDecodedContent->{'thumbnail_url'};
    # remove size information
    $ThumbnailURL =~ s/\_[^\_]*(\.[^\.]+)$/$1/;
    $hDecodedContent->{'thumbnail_url'} = $ThumbnailURL;

    my $Iframe = $hDecodedContent->{'html'};
    # add iframe class
    if ($Iframe =~ m/(\<iframe)(.*<\/iframe>)/) {
        $Iframe = $IframeClass ? "$1 class=\"$IframeClass\"$2" : "$1$2";
    }

    if ($Iframe =~ m/src=\"([^\"]*)\"/) {
        my $VideoURL = $1;
        # call video url dereferred
        my $DereferredVideoURL = $Processor->callBlock('DEREFER', sub { return [ $VideoURL ] });
        $Iframe =~ s/src=\"([^\"]*)\"/src=\"$DereferredVideoURL\"/;
        $hDecodedContent->{'html'} = $Iframe;
    }

    my $hVars = {
        'ThumbnailURL' => $ThumbnailURL,
        'Video' => to_json($hDecodedContent),
        'RealVideoID' => $hDecodedContent->{video_id}, #-- The actual video id, for use in the template, in cases where it was taken from a URL
    };

    # concatenate parsed template content
    return join ('', @{$Processor->replaceTLEs($cTemplate, $hVars)});
}

#========================================================================================
# function     ParseAdditionalHtmlHead
# state        public
#----------------------------------------------------------------------------------------
# syntax       #ParseAdditionalHtmlHead(#AdditionalHtmlHead)
#----------------------------------------------------------------------------------------
# example      #ParseAdditionalHtmlHead(#Shop.AdditionalHtmlHead[0])
#----------------------------------------------------------------------------------------
# description  Parses the additional html head
#----------------------------------------------------------------------------------------
# input        #AdditionalHtmlHead | additional html head | string
# return       #ParsedAdditionalHtmlHead | parsed additional html head | string
#========================================================================================
sub ParseAdditionalHtmlHead {
    my $self = shift;
    my ($Processor, $aParameter) = @_;
    my ($AdditionalHtmlHead) = @$aParameter;

    my $hParams = {'AdditionalHtmlHead' => $AdditionalHtmlHead};

    TriggerHook('API_ParseAdditionalHtmlHead', $hParams);
    return $hParams->{'AdditionalHtmlHead'};
}


1;
