use strict;

use DE_EPAGES::Core::API::File qw (GetFileContent WriteFile AbsolutePath);
use DE_EPAGES::Core::API::Script qw ( RunScript );
use DE_EPAGES::Database::API::Connection qw ( ConnectTestSite DisconnectSite RunOnStore );
use DE_EPAGES::Object::API::Factory qw (ResetFactoryCaches JoinObjectPath ExistsObjectByPath LoadRootObject  LoadObjectByPath );
use DE_EPAGES::Object::API::XML::ObjectDeleteHandler qw (DeleteObjects);
use DE_EPAGES::Object::API::XML::ObjectExportDriver qw (ExportObjects);
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw (ImportObjects);
use DE_EPAGES::WebInterface::API::FakeRequest;
use DE_EPAGES::WebInterface::API::MessageCenter qw (SynchronizeCache);
use Test::More;
use JSON qw ( from_json to_json );

Readonly::Scalar my $EXPORT_SHOP_TIME            => 10;  # 6.14.1: 8.1 ; 6.13.2: 11.5; 6.13.3: 13; 6.13.4: 11.
Readonly::Scalar my $CREATE_SHOP_TIME            => 7;   # 6.14.1: 5.5 ; 6.13.2: 10  ; 6.13.3: 12. 6.13.4:  9.
Readonly::Scalar my $IMPORT_SHOP_TIME            => 46;  # 6.14.1: 45  ; 6.13.2: 36

my $ROOTPATH = '/Shops/TestShop';
my $NEWPATH  = '/Shops/Test';
local $ENV{'EPAGES_TESTSITE'} = 'Store' unless defined $ENV{'EPAGES_TESTSITE'};

sub _TimeOK {
    my ($MaxTime, $Comment, $Code) = @_;

    my $time0 = time();
    $Code->();
    my $Time = time() - $time0;
    SKIP: {
        skip( 'Forced skip of timing tests.', 1 ) if $ENV{'SKIP_TIMING'};
        cmp_ok( $Time, '<', $MaxTime, sprintf("$Comment: %0.1f s", $Time) );
    }
    return;
}

sub Main() {
    ConnectTestSite();
    ok( Init(), 'initialization' );

    my $ShopAlias = 'TestShop';
    my $NewAlias  = 'Test';
    my $User = '/Providers/Distributor/Users/admin';
    my $Password = 'admin';
    my $WebServiceProxy;

    RunOnStore( 'Store' => 'Site', 'Sub' => sub {
        $WebServiceProxy = LoadRootObject()->folder('Stores')->child('Site')->get('WebServiceProxy');
    });
    ok( 1, "get web service proxy ($WebServiceProxy)" );

    # create a web service client
    my $ShopConfigService = DE_EPAGES::WebService::API::Client
        ->uri('urn://epages.de/WebService/ShopConfigService/2009/09')
        ->proxy( $WebServiceProxy, timeout => 900 );
    $ShopConfigService->userinfo("$User:$Password");

    # delete the clone shop if it exists from a previous test
    if( $ShopConfigService->exists({ 'Alias' => $NewAlias })->result ) {
        note 'delete previously created clone shop';
        $ShopConfigService->deleteShopRef({ 'Alias' => $NewAlias });
    }

    my $TestShop = LoadObjectByPath($ROOTPATH);

    # Test1: Get value eq defaultvalue.
    my $hJSON  = $TestShop->get('JSONTest');
    my $Found = to_json($hJSON, {pretty => 1});
    my $Expected = to_json(from_json('{"Undef":null,"Test":"TestValue","Int":1}'),
                           {pretty => 1});

    is( join('', sort map { s/,//g; $_ } split(/\n/, $Found)),
        join('', sort map { s/,//g; $_ } split(/\n/, $Expected)),
        'Default json string is correct.' );

    # Test2: set correct value
    my $hNewValue = {
        'String' => 'foo',
        'Int' => 1,
        'Array' => ['foo','bar'],
        'Hash' => {
            'Test' => undef,
        },
    };
    $Expected = to_json($hNewValue, {pretty => 1});
    $TestShop->set({'JSONTest' => $hNewValue});
    my $hNewJSON = $TestShop->get('JSONTest');
    $Found = to_json($hNewJSON, {pretty => 1});

    is( join('', sort map { s/,//g; $_ } split(/\n/, $Found)),
        join('', sort map { s/,//g; $_ } split(/\n/, $Expected)),
        'New json attribute value is correct.' );

    # Test3: Export shop
    my $FileName = 'DemoShop.xml';
    _TimeOK($EXPORT_SHOP_TIME ,'export.xml', sub {
        ExportObjects( '/Shops', $FileName, { withdefaults => 1 }, [$ShopAlias], \&_TransForm );
    });

    my $Content = ${ GetFileContent( $FileName, { utf8 => 1 } ) };
    $Content =~ s/\Q$ShopAlias\E/$NewAlias/g;
    my $ImportFile = 'DemoShopXMLimport.xml';
    WriteFile( $ImportFile, \$Content, { utf8 => 1 } );
    ok( 1, 'transform shop alias' );

    my $NewPath = JoinObjectPath( [$NewAlias] );
    if( ExistsObjectByPath( $NewPath ) ) {
        note 'delete shop object of clone shop';
        LoadObjectByPath( $NewPath )->delete;
    }

    # Before creating a new shop, reset the default locale to en_GB
    is( LoadRootObject()->get('LocaleID'), 'en_GB', 'default system locale is en_GB, see DE_EPAGES::Object::API::Cartridge::_activateLocales' );

    my $EmptyFile = 'Empty.xml';
    $Content = '<epages />';
    WriteFile( $EmptyFile, \$Content, { utf8 => 1 } );

    # create an empty shop
    _TimeOK($CREATE_SHOP_TIME, 'create_shop', sub {
        $ShopConfigService->create( {
            Database                => $ENV{'EPAGES_TESTSITE'},
            ShopType                => 'MerchantCorporate',
            ShopAlias               => $NewAlias,
            Alias                   => $NewAlias,
            ImportFiles             => [ AbsolutePath($EmptyFile) ],
        });
    });

    # update cache for store database
    ResetFactoryCaches();

    # Test4: Import shop data from XML file
    _TimeOK($IMPORT_SHOP_TIME, 'import xml', sub {
        ImportObjects( $ImportFile, '/Shops');
    });

    # inform other application servers
    SynchronizeCache();

    my $NewShop = LoadObjectByPath($NEWPATH);
    $hNewJSON = $NewShop->get('JSONTest');
    $Found = to_json($hNewJSON, {pretty => 1});
    $Expected = to_json($hNewValue, {pretty => 1});

    is( join('', sort map { s/,//g; $_ } split(/\n/, $Found)),
        join('', sort map { s/,//g; $_ } split(/\n/, $Expected)),
        'imported json attribute value is correct.');

    ok( Cleanup($ShopConfigService, $NewAlias), 'cleanup');
    DisconnectSite();
    done_testing();
}

sub Init() {
    return 1 if LoadRootObject()->child('Shops')->existsChild('TestShop') && $ENV{'SKIP_CLEANUP'};
    Cleanup();
    ImportObjects('JSONAttributesShop.xml');
    #Create shop
    LoadRootObject()->child('Shops')->insertChild('Shop', {'Alias'=>'TestShop'});

    return 1;
}

sub Cleanup {
    my ($ShopConfigService, $NewAlias) = @_;

    return 1 if $ENV{'SKIP_CLEANUP'};
    LoadObjectByPath($ROOTPATH)->delete if ExistsObjectByPath($ROOTPATH);
    LoadObjectByPath($NEWPATH)->delete if ExistsObjectByPath($NEWPATH);
    DeleteObjects('JSONAttributesShop.xml');

    if(defined $ShopConfigService) {
        # delete the shopref on the site database
        if( $ShopConfigService->exists({ 'Alias' => $NewAlias })->result ) {
            note 'delete created shop reference';
            $ShopConfigService->deleteShopRef({ 'Alias' => $NewAlias });
        }
    }
    return 1;
}

sub _TransForm {
    my ($TagName, $hAttributes) = @_;
    delete $hAttributes->{'GUID'};
    delete $hAttributes->{'CreationDate'};
    delete $hAttributes->{'IsInternalTestShop'};

    return;
}

RunScript(Sub=>\&Main, Debug=>1);
