use strict;
use File::stat;
use File::Spec::Functions;
use Getopt::Long;
use POSIX;
use Time::HiRes qw(time);

use DE_EPAGES::Cartridge::API::Table::Cartridge qw(ExistsCartridgeByPackage);
use DE_EPAGES::Content::API::Table::ShortURL qw(GetPKeyShortURLsByContentObjectID DeleteShortURL);
use DE_EPAGES::Core::API::Error qw(ExistsError GetError Error);
use DE_EPAGES::Core::API::File qw(ExistsDirectory DeleteDirectory);
use DE_EPAGES::Core::API::Log qw(LogError LogContext);
use DE_EPAGES::Core::API::Script qw(RunScript);
use DE_EPAGES::Core::API::Warning qw(Warning);
use DE_EPAGES::Database::API::Connection qw(RunOnStore);
use DE_EPAGES::ShopMigration::API::Constants qw(SHOPMIGRATION_FINAL_DONE SHOPMIGRATION_FINAL_DONE_AND_PREP_DELETED);
use DE_EPAGES::WebInterface::API::MessageCenter qw(SynchronizeCache);
use DE_EPAGES::ShopMigration::API::MigrationShop qw(FindAllShopsWithMigrationStatus);

use constant ONE_YEAR_IN_DAYS => 365;
use constant SECONDS_PER_DAY => 86400;

sub Main {
    local $| = 1;
    my %Options = ();

    my @GetOpt = qw(help test passwd=s storename=s shop=s allstores! output! mindays=s);
    usage("Unkown parameter specified.\n") unless GetOptions(\%Options, @GetOpt);
    usage() if $Options{help};
    usage('Missing required parameter -storename or -allstores')
        unless ( defined $Options{storename} or defined $Options{allstores} );
    usage('Do not use -all and -storename together!')
        if ($Options{storename} && $Options{allstores});
    my $Stores = defined $Options{storename} ? [$Options{storename}] : getAllStores();
    usage('env EPAGES_BUSINESS_UNITS is not correctly set! Check your /etc/default/epages6.local!')
        unless scalar @$Stores;
    $Options{mindays}  //= ONE_YEAR_IN_DAYS;
    $Options{minseconds} = $Options{mindays} * SECONDS_PER_DAY;
    $Options{output} //= 1;

    my $time1 = time;
    foreach my $StoreName (@$Stores) {
        my $hCount = {};
        printf "##### Store '%s' #####\n", $StoreName if $Options{output};
        RunOnStore(
            Store      => $StoreName,
            DBPassword => $Options{passwd},
            Sub        => sub {
                return unless (ExistsCartridgeByPackage('DE_EPAGES::ShopMigration'));
                $Options{WantedStatus} = SHOPMIGRATION_FINAL_DONE;
                $Options{WantedString} = 'FINAL_DONE';
                $Options{storename}    = $StoreName;
                $Options{directory}    = catdir($ENV{EPAGES_WEBSECURE}, 'ShopExport', $Options{storename});
                $hCount = cleanupPrepShopsAndExportFolders(\%Options);
            },
        );
        if ($hCount->{Total} > 0) {
            printf "\n##### Statistics for Store '%s' #####\n", $StoreName if $Options{output};
            printf join('', map {"%-40s: %s\n"} keys %$hCount ),
                'Total Migration shops (FINAL_DONE)', $hCount->{Total},
                ' - with a PrepShop', $hCount->{WithPrep},
                sprintf(' - with ExportDir older than %d days', $Options{mindays}), $hCount->{Aged},
                ' - cleaned up PrepShops', $hCount->{CleanedPreps},
                ' - cleaned up ExportDirs', $hCount->{CleanedDirs},
                ' - failed', $hCount->{Failed} if $Options{output};
        } elsif (!scalar keys %$hCount) {
            print "DE_EPAGES::ShopMigration cartridge not installed in Store '$StoreName'.\n";
        }
    }
    my $time2 = time;
    printf "need %.3f seconds\n", $time2 - $time1 if $Options{output};
    eval {SynchronizeCache();};
    Warning(GetError()) if ExistsError();
}

sub getAllStores {
    my @StoreNames = grep { $_ !~ /Site/ } split( /\s+/, $ENV{'EPAGES_BUSINESS_UNITS'} );
    return \@StoreNames;
}

sub deletePrepShop {
    my ($PrepShop) = @_;

    return unless defined $PrepShop;
    if ($PrepShop->existsChild('LegalPagesUnity')) {
        foreach my $PageID (@{$PrepShop->child('LegalPagesUnity')->findChildren}) {
            DeleteShortURL($_) foreach @{GetPKeyShortURLsByContentObjectID($PageID)}
        };
    }
    $PrepShop->delete;
}

sub cleanupPrepShopsAndExportFolders {
    my ($hOptions) = @_;

    my %Count = (Total => 0, WithPrep => 0, Aged => 0, CleanedPreps => 0, CleanedDirs => 0, Failed => 0);
    my ($StoreName, $Test, $Directory, $Output, $WantedStatus, $WantedString, $MinAge, $ShopAlias) =
        @{$hOptions}{qw(storename test directory output WantedStatus WantedString minseconds shop)};
    my $aShopList =  FindAllShopsWithMigrationStatus($WantedStatus);
    if (defined $ShopAlias) {
        $aShopList = [ grep {$_->alias eq $ShopAlias} @$aShopList ];
    }

    $Count{Total} = scalar @$aShopList;
    if ($Count{Total}) {
        printf "Found %s shops in status %s\n", $Count{Total}, $WantedString if $Output;
        foreach my $Shop (@$aShopList) {
            my $ExportShopDir = catdir( $Directory, $Shop->alias);
            my $HasExportDir = ExistsDirectory($ExportShopDir);
            my $HasPrep      = 0;
            my $Age = 'n/a'; my $CleanUp = 0;
            eval {
                LogContext({ 'Shop' => $Shop->alias }, sub {
                    my $PrepShop = $Shop->get('MigrationPrepShop');
                    if (defined $PrepShop) {
                        $HasPrep = 1;
                        $Count{WithPrep}++;
                    }
                    if ($HasExportDir) {
                        my $DirTimeStamp = stat($ExportShopDir)->mtime;
                        $Age = time() - $DirTimeStamp;
                        if ($Age > $MinAge) {
                            $Count{Aged}++;
                            $CleanUp = 1;
                        } else {
                            next;
                        }
                    } else {
                        # If ExportDir is missing we assume it's old and/or half-deleted in which
                        # case we still want to trigger cleanup for the old PrepShop
                        $CleanUp = 1 if $HasPrep;
                    }
                    if ($CleanUp) {
                        printf "%sFound shop %s%s\n",
                            $Test ? '[TEST] ' : '',
                            $Shop->alias,
                            $HasExportDir ?
                                sprintf (' with ExportDir %s (age: %s days)',
                                    $ExportShopDir,
                                    floor($Age/SECONDS_PER_DAY))
                                : ' (ExportDir already deleted)'
                            if ($Output && ($HasExportDir || $HasPrep));
                        if ($HasPrep) {
                            printf " - %sDeleting PrepShop\n", $Test ? '[TEST] ' : '' if ($Output);
                            if ($PrepShop->get('IsUnityRegistered')) {
                                Error('PREPSHOP_HAS_UNITYREGISTERED');
                            } else {
                                deletePrepShop($PrepShop) unless $Test;
                            }
                            $Count{CleanedPreps}++;
                        }
                        if ($HasExportDir) {
                            printf " - %sDeleting ExportShopDir\n", $Test ? '[TEST] ' : '' if ($Output);
                            DeleteDirectory($ExportShopDir, {'recursive' => 1}) unless $Test;
                            $Count{CleanedDirs}++;
                        }
                        $Shop->set({ MigrationStatus => SHOPMIGRATION_FINAL_DONE_AND_PREP_DELETED }) unless $Test;
                    }
                });
            };
            # Set error state on BaseShop in case something goes wrong so we don't try over and over
            if (ExistsError()) {
                my $Error = GetError();
                my $Msg = sprintf "%sCleanup of prepshop for shop '%s' store '%s' failed with '%s'\n",
                    $Test ? '[TEST] ' : '', $Shop->alias, $StoreName, $Error->shortMessage;
                LogError($Error);
                print $Msg;
                $Count{Failed}++;
            }
        }
    } else {
        printf("No shops in status '%s' (%s) found!\n", $WantedStatus, $WantedString) if $Output;
    }
    return \%Count;
}

sub usage {
    my ($ErrorNote) = @_;

    print STDERR "\n$ErrorNote\n\n" if $ErrorNote;
    print <<END_USAGE;
Cleans up old ShopExport files and the PrepShops from Shops that have finished Migration.

Usage:
    perl $0 [options] flags objects

options:
    -passwd       database user password                                    ()
    -storename    name of store                                             required
    -mindays      minimum days since final migration for shop to clean up   (365)
    -shop         only run this script for a specific shop                  ()

flags:
    -help         show the command line options                             ()
    -test         run in verbose testmode, only print out what would happen (0)
    -allstores    run on all stores defined via env EPAGES_BUSINESS_UNITS

Example:
    perl $0 -storename Store -mindays 200
    perl $0 -allstores -test
    perl $0 -storename Store -test
    perl $0 -storename Store -shop DemoShop
END_USAGE
    exit 2;
}

RunScript(
    'Sub' => \&Main,
);

1;
