#========================================================================================
# package      DE_EPAGES::Spreedly::t::unit::API_StripeUrlBuilder
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for DE_EPAGES::Spreedly::t::unit::API_StripeUrlBuilder
#========================================================================================
package DE_EPAGES::Spreedly::t::unit::API_StripeUrlBuilder;
use base DE_EPAGES::Core::API::Test;

use DE_EPAGES::Core::API::Url qw ( ParseQueryString );
use DE_EPAGES::Spreedly::API::StripeUrlBuilder;
use DE_EPAGES::Test::Mock::Mockify;
use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use Test::Exception;
use Test::MockObject::Extends;
use Test::More;

use strict;

sub test {
    my $self = shift;

    $self->test_buildStripeBacklinkUrl_Fail_Empty();
    $self->test_buildStripeBacklinkUrl_Fail_WithoutState();
    $self->test_buildStripeBacklinkUrl_Fail_EmptyState();
    $self->test_buildStripeBacklinkUrl_Success();
    $self->test_buildStripeBacklinkUrl_Cancel();

    $self->test_ConnectAndReconnectUrl_Success();

    done_testing();
}

#----------------------------------------------------------------------------------------
# Tests
#----------------------------------------------------------------------------------------
sub test_buildStripeBacklinkUrl_Fail_Empty {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $hVars = {};

    my $SUT = $self->_createSUT();

    throws_ok( sub { $SUT->buildStripeBacklinkUrl($hVars); },
        qr/NoURLFound/, "$SubTestName - Throw a buildStripeBacklinkUrl exception when Parameter is not passed" );

}

#----------------------------------------------------------------------------------------
sub test_buildStripeBacklinkUrl_Fail_WithoutState {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $hVars = {
        'code' => [ 'AuthorizationCode' ],
        'scope' => [ 'read_write' ],
        'ChangeAction' => [ 'BacklinkSpreedlyStripeOnboarding' ]
    };

    my $SUT = $self->_createSUT();

    throws_ok( sub { $SUT->buildStripeBacklinkUrl($hVars); },
        qr/NoURLFound/, "$SubTestName - Throw a buildStripeBacklinkUrl exception when there is no state in Hash" );

}

#----------------------------------------------------------------------------------------
sub test_buildStripeBacklinkUrl_Fail_EmptyState {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $hVars = {
        'code' => [ 'AuthorizationCode' ],
        'scope' => [ 'read_write' ],
        'state' => [ ],
        'ChangeAction' => [ 'BacklinkSpreedlyStripeOnboarding' ]
    };

    my $SUT = $self->_createSUT();

    throws_ok( sub { $SUT->buildStripeBacklinkUrl($hVars); },
        qr/NoURLFound/, "$SubTestName - Throw a buildStripeBacklinkUrl exception when state is undef" );

    return;
}

#----------------------------------------------------------------------------------------
sub test_buildStripeBacklinkUrl_Success {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $hVars = {
        'code' => [ 'AuthorizationCode' ],
        'scope' => [ 'read_write' ],
        'state' => [ 'www.BackLinkUrl.com/?ViewAction=epages&ObjectId=53464' ],
        'ChangeAction' => [ 'BacklinkSpreedlyStripeOnboarding' ]
    };
    my $SUT = $self->_createSUT();

    my $URL = URI->new($SUT->buildStripeBacklinkUrl($hVars));
    my $hParsedState = ParseQueryString($URL->query(), 'utf-8');

    is($hParsedState->{'code'}->[0], 'AuthorizationCode',"$SubTestName - prove that the url parameters contains code");
    is($hParsedState->{'scope'}->[0], 'read_write',"$SubTestName - prove that the url parameters contains scope");
    is($hParsedState->{'ViewAction'}->[0], 'epages',"$SubTestName - prove that the url parameters contains ViewAction");
    is($hParsedState->{'ObjectId'}->[0], '53464',"$SubTestName - prove that the url parameters contains ObjectId");
    is($hParsedState->{'ChangeAction'}, undef,"$SubTestName - prove that the url parameters don't contain ChangeAction");
    is($hParsedState->{'state'}, undef,"$SubTestName - prove that the url parameters don't contain state");

    return;
}

#----------------------------------------------------------------------------------------
sub test_buildStripeBacklinkUrl_Cancel {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $hVars = {
        'error' => [ 'access_denied' ],
        'state' => [ 'www.BackLinkUrl.com/?ViewAction=epages&ObjectId=53464' ],
        'ChangeAction' => [ 'BacklinkSpreedlyStripeOnboarding' ]
    };
    my $SUT = $self->_createSUT();

    my $URL = URI->new($SUT->buildStripeBacklinkUrl($hVars));
    my $hParsedState = ParseQueryString($URL->query(), 'utf-8');

    is($hParsedState->{'error'}->[0], 'access_denied',"$SubTestName - prove that the url parameters contains error");
    is($hParsedState->{'ViewAction'}->[0], 'epages',"$SubTestName - prove that the url parameters contains ViewAction");
    is($hParsedState->{'ObjectId'}->[0], '53464',"$SubTestName - prove that the url parameters contains ObjectId");
    is($hParsedState->{'ChangeAction'}, undef,"$SubTestName - prove that the url parameters don't contain ChangeAction");
    is($hParsedState->{'state'}, undef,"$SubTestName - prove that the url parameters don't contain state");

    return;
}

#----------------------------------------------------------------------------------------
sub test_ConnectAndReconnectUrl_Success {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $PaymentMethod = $self->_createPaymentMethod();
    my $DummyStateUrl = 'https://encoded.state/url';
    my $SUT = $self->_createSUT();
    my $Url = $SUT->buildConnectAndReconnectUrl($PaymentMethod, $DummyStateUrl);

    my $ParsedUrl = URI::URL->new($Url);
    my %Query;
    map { if (/(.+?)=(.*)/) { $Query{$1} = $2; }
          else { $Query{$_} = undef } } split(/&/, $ParsedUrl->equery());

    my %ExpectedQuery = (
        'client_id' => 'ca_82tCJZgOteUhVj91ulpAR7TFF8m1UasG',
        'redirect_uri' => 'https%3A%2F%2Fdummy.de%2Fepages%2FSite.admin%2F%3FChangeAction%3DBacklinkSpreedlyStripeOnboarding',
        'response_type' => 'code',
        'scope' => 'read_write',
        'state' => 'https%3A%2F%2Fencoded.state%2Furl',
        'stripe_user%5Bcity%5D' => 'Emden',
        'stripe_user%5Bcountry%5D' => 'DE',
        'stripe_user%5Bemail%5D' => 'test%40epages.com',
        'stripe_user%5Bfirst_name%5D' => 'Lila',
        'stripe_user%5Blast_name%5D' => 'Launeb%C3%A4r',
        'stripe_user%5Bphone_number%5D' => '123456',
        'stripe_user%5Bstreet_address%5D' => 'Sesamstra%C3%9Fe+666',
        'stripe_user%5Bzip%5D' => '26721',
    );

    is($ParsedUrl->scheme(),'https',"$SubTestName - prove scheme");
    is($ParsedUrl->host(),'connect.stripe.com',"$SubTestName - prove host");
    is($ParsedUrl->epath(),'/oauth/authorize',"$SubTestName - prove path");
    is_deeply(\%Query ,\%ExpectedQuery, "$SubTestName - prove thats the connectUrl query is encoded and undefined address attributes are removed (e.g. State is undef)");
}

#----------------------------------------------------------------------------------------
# Helpers
#----------------------------------------------------------------------------------------
sub _createSUT {
    my $self = shift;

    my $aParameterList = [];
    my $Mockify = DE_EPAGES::Test::Mock::Mockify->new(
                   'DE_EPAGES::Spreedly::API::StripeUrlBuilder',
                   $aParameterList
              );
    $Mockify->mock('_getSiteUrlFromEpagesConf','https://dummy.de/epages/Site.admin/');
    return $Mockify->getMockObject();
}
#----------------------------------------------------------------------------------------
sub _createPaymentMethod {
    my $self = shift;

    my $hAttributes = {
        'Site' => $self->_createShop(),
        'PaymentType' => $self->_createPaymentType(),
    };
    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::Payment::API::Object::PaymentMethod',
        $hAttributes
    );

    return $Mockify->getMockObject();
}
#----------------------------------------------------------------------------------------
sub _createPaymentType {
    my $self = shift;

    my $hAttributes = {
        'Alias' => 'MyShop',
        'TestAccountName' => 'ca_82tCJZgOteUhVj91ulpAR7TFF8m1UasG',
        'OnBoardingUrl' => 'https://connect.stripe.com/oauth/authorize',
    };
    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::Payment::API::Object::PaymentType',
        $hAttributes
    );

    return $Mockify->getMockObject();
}
#----------------------------------------------------------------------------------------
sub _createAddress {
    my $self = shift;

    my $hAttributes = {
        'Country' => {'Code2' => 'DE'},
        'FirstName' => 'Lila',
        'LastName' => 'Launebr',
        'Street' => 'Sesamstrae 666',
        'City' => 'Emden',
        'State' => undef,
        'Zipcode' => '26721',
        'Phone' => '123456',
        'EMail' => 'test@epages.com',
    };
    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::Address::API::Object::Address',
        $hAttributes
    );

    return $Mockify->getMockObject();
}
#----------------------------------------------------------------------------------------
sub _createShop {
    my $self = shift;

    my $hAttributes = {
        'IsInternalTestShop' => 1,
        'Address' => $self->_createAddress(),
    };
    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::Shop::API::Object::Shop',
        $hAttributes
    );

    return $Mockify->getMockObject();
}
__PACKAGE__->runScript();
