use strict;

use Getopt::Long;
use JSON;
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::Error qw(ExistsError GetError );
use DE_EPAGES::Core::API::File qw(ExistsFile );
use DE_EPAGES::Core::API::Log qw(LogError );
use DE_EPAGES::Core::API::Warning qw (Warning);
use DE_EPAGES::Database::API::Connection qw( RunOnStore GetCurrentDBHandle Transaction );
use DE_EPAGES::Object::API::Factory qw( LoadRootObject LoadObject LoadClassByAlias ResetFactoryCaches);
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw (ImportObjects);
use DE_EPAGES::RemoteSearch::API::RemoteSearchImpl qw (GetOrCreateSearchIndex );
use DE_EPAGES::Shop::API::Constants qw(TAX_MODEL_GROSS TAX_MODEL_NET MAX_TRIES_FOR_PATCHING_UNITYBASEURL);
use DE_EPAGES::Shop::API::Feature qw ( LoadFeatureByAlias );
use DE_EPAGES::Tax::API::Cartridge;
use DE_EPAGES::WebInterface::API::MessageCenter qw( SynchronizeCache );

##no critic qw(ProhibitLongLines ProhibitMagicNumbers ProhibitExcessComplexity)

#Define global variables
my ($StoreName, $UseSQL, @TaxFileNames, $TaxClassIDNormal, $TaxClassIDReduced, $TaxClassIDNone, $TaxClassIDebooks, $TaxClassIDigitalservices, $TaxMatrixAlias,
    $SystemTaxMatrixID, $SystemTaxMatrixDigitalID, $ShopClass, $TaxMatrixAttributeID, $TaxMatrixDigitalAttributeID,
    $UpdateTaxCompletedAttributeID, $InputTaxModelAttributeID, $UnityFeatureID , $ClosedByProviderTemporarilyAttributeID);

#Define default values for taxes
my ($OldTaxNormal, $NewTaxNormal, $OldTaxReduced, $NewTaxReduced, $OldTaxEbooks, $NewTaxEbooks, $OldTaxDownload,$NewTaxDownload) = (1.077,1.081,1.025,1.026,undef,undef,undef,undef);

#define default values for inputfiles:
my $TaxCatridgeXMLs ='TaxesSwitzerland.xml';
my $ImportStatusAttributeFile;

RunScript( Sub => \&Main );

sub Main {
    my ($Help, $Password, $ShopAlias,$LastMerchantLogin,$SFOpenByMerchant, $RemoveBackupTables, $TestMode);
    my $ImportStatusAttribute=1;
    my $CloseShops = 1;
    my $BackupTables = 1;
    my $ImportTaxXMLs = 1;
    my $DoCalculate =1;
    my $DoCleanUp = 0;
    $UseSQL = 1;

    GetOptions(
        'help'                  => \$Help,
        'storename=s'           => \$StoreName,
        'password=s'            => \$Password,
        'shop=s'                => \$ShopAlias,
        'taxmatrix=s'           => \$TaxMatrixAlias,
        'oldtaxnormal=s'        => \$OldTaxNormal,
        'newtaxnormal=s'        => \$NewTaxNormal,
        'oldtaxreduced=s'       => \$OldTaxReduced,
        'newtaxreduced=s'       => \$NewTaxReduced,
        'sfopenbymerchant'      => \$SFOpenByMerchant,
        'lastmerchantlogin=s'   => \$LastMerchantLogin,
        'closeshops!'           => \$CloseShops,
        'importxml!'            => \$ImportTaxXMLs,
        'taxxml=s'              => \$TaxCatridgeXMLs,
        'backup!'               => \$BackupTables,
        'removebackup'          => \$RemoveBackupTables,
        'sql'                   => \$UseSQL,
        'calculate!'            => \$DoCalculate,
        'importstatusshopattr!' => \$ImportStatusAttribute,
        'cleanup'               => \$DoCleanUp,
        'test'                  => \$TestMode
    );

    usage() if $Help;

    unless ($StoreName) {
        print STDERR "Missing parameter storename\n\n";
        usage();
    }

    if ((not defined $TaxMatrixAlias) and not $DoCleanUp){
        print STDERR "\nMissing required parameter taxmatrix!  \n";
        usage();
    }

    my $Cartridge = DE_EPAGES::Tax::API::Cartridge->new;
    my $CartridgeDirectory = join('/', $ENV{EPAGES_CARTRIDGES}, $Cartridge->cartridgeDirectory);
    if($ImportTaxXMLs and defined $TaxCatridgeXMLs) {
        my $Directory = join('/', $CartridgeDirectory, 'Database','XML');
        my @FileNames = split(/,/,$TaxCatridgeXMLs);
        foreach my $FileName (@FileNames) {
            my $file="$Directory/$FileName";
            if(ExistsFile($file)) {
                push @TaxFileNames, $file ;
            } else {
                print STDERR "\nFile $file does not exists.\n";
                usage();
            }
        }
    } elsif($ImportTaxXMLs and not defined $TaxCatridgeXMLs) {
        print STDERR "\nNo tax import files defined.\n";
        usage();
    }
    if($ImportStatusAttribute) {
        $ImportStatusAttributeFile = join('/', $CartridgeDirectory, 'Scripts','imp_AttributesShop.xml');
    }
    RunOnStore(
        Store    => $StoreName,
        Password => $Password,
        Sub      => sub {

            my $time1 = time();

            if ($TestMode) {
                printf "$StoreName Script started in Testmode. Nothing will be saved in db.\n";
            }
            _defineStoreBasics();
            if($DoCleanUp) {
                _removeStatusAttribute($TestMode);
                _removeBackupTables($TestMode);
                return;
            }
            if ($ImportStatusAttribute) {
                _importStatusAttribute($TestMode);
            }
            if($CloseShops) {
                unless ($TestMode) {
                    # close all shops technically which have inputtaxmodel gross and no customized taxmatrix or customized digitaltax
                    _closeShops($ShopAlias);
                    # send also epages-ui an update for closed now shops which were not closed before
                    _closeSFUnityShops($ShopAlias);
                } else {
                    my $aShopIDs = _getShopsToClose($ShopAlias);
                    print "$StoreName TestMode: we would close " . scalar(@$aShopIDs) . " shops\n";
                }
            }
            if($BackupTables) {
                _backupTables($TestMode);
            } elsif ($RemoveBackupTables) {
                _removeBackupTables($TestMode);
            }
            if ($ImportTaxXMLs) {
                _importTaxXMLs($TestMode);
            }
            return if ($BackupTables or $RemoveBackupTables or $CloseShops or $ImportTaxXMLs or $ImportStatusAttribute) and (not $DoCalculate);

            my $timeShops = time;
            _updatePrices($TestMode, $ShopAlias,$SFOpenByMerchant,$LastMerchantLogin);

            my $time2   = time;
            my $seconds = $time2 - $time1;
            my $shopseconds = $timeShops - $time1;
            printf "\n$StoreName need %.3f seconds to load shops\n", $shopseconds;
            printf "\n$StoreName need %.3f seconds over all\n", $seconds;
        }
    );

    eval { SynchronizeCache(); };
    Warning( GetError() ) if ExistsError();

}

sub _defineStoreBasics {
    # Load the System object.
    my $System = LoadRootObject();

    $ShopClass =  LoadClassByAlias('Shop');

    #load attributeids:
    $TaxMatrixAttributeID = $ShopClass->attribute('TaxMatrix')->id;
    $TaxMatrixDigitalAttributeID=$ShopClass->attribute('TaxMatrixDigital')->id;
    $InputTaxModelAttributeID = $ShopClass->attribute('InputTaxModel')->id;
    $ClosedByProviderTemporarilyAttributeID =  $ShopClass->attribute('ClosedByProviderTemporarily')->id;
    $UnityFeatureID = LoadFeatureByAlias('Unity')->id;
    if($ShopClass->existsAttribute('UpdateTaxCompleted')) {
        $UpdateTaxCompletedAttributeID = $ShopClass->attribute('UpdateTaxCompleted')->id;
    } else {
        $UpdateTaxCompletedAttributeID =undef;
    }

    if (defined $TaxMatrixAlias) {
        # Load SystemTaxMatrix object
        my $SystemTaxMatrix = $System->child($TaxMatrixAlias);
        $SystemTaxMatrixID = $SystemTaxMatrix->id;
        my $SystemTaxMatrixDigital = $System->child('TaxMatrixDigital');
        $SystemTaxMatrixDigitalID = $SystemTaxMatrixDigital->id;

        # Get TaxClassID for $SystemTaxMatrix normal, none and reduced.
        $TaxClassIDNormal = $SystemTaxMatrix->child('normal')->id;
        $TaxClassIDReduced = $SystemTaxMatrix->child('reduced')->id;
        $TaxClassIDNone = $SystemTaxMatrix->child('none')->id;

        # Get TaxClassID for $SystemTaxMatrixDigital ebooks and digitalservices.
        $TaxClassIDebooks = $SystemTaxMatrixDigital->child('ebooks')->id;
        $TaxClassIDigitalservices = $SystemTaxMatrixDigital->child('digitalservices')->id;
    }
    return;
}

sub _backupTables {
    my ($TestMode) = @_;

    my $dbi= GetCurrentDBHandle();
    print "$StoreName start backup tables (TestMode:$TestMode)\n";
    foreach my $table (qw (
        productprice
        scaleprice
        valuediscount
        shippinglevel
        shippingvalues
        paymentlevel
    )) {
        if ($dbi->existsTable('backup_' . $table)) {
            print "$StoreName backup_$table already exists.\n";

        }else{
            unless ($TestMode) {
                print "$StoreName start copy table $table\n";
                my $sql = "CREATE TABLE backup_$table SELECT * FROM $table";
                $dbi->do($sql);
                print "$StoreName finish copy table $table to backup_$table \n";
            } else {
                print "$StoreName TestMode: we would copy table $table to backup_$table\n";
            }
        }
    }
    print "$StoreName end backup tables\n";
    return;
}

sub _removeBackupTables {
    my ($TestMode) = @_;

    my $dbi= GetCurrentDBHandle();
    print "$StoreName start remove backup tables (TestMode:$TestMode)\n";
    foreach my $table (qw (
        productprice
        scaleprice
        valuediscount
        shippinglevel
        shippingvalues
        paymentlevel
    )) {
        if ($dbi->existsTable('backup_' . $table)) {
            unless ($TestMode) {
                print "$StoreName start removing table backup_$table\n";
                $dbi->dropTable('backup_' . $table);
                print "$StoreName finish removing table backup_$table\n";
            } else {
                print "$StoreName TestMode: we would remove table backup_$table:\n";
            }
        } else{
            print "$StoreName backup_$table does not exist.\n";
        }
    }
    print "$StoreName end remove backup tables\n";
    return;
}

sub _importStatusAttribute {
    my ($TestMode) = @_;
    print "$StoreName start _importStatusAttribute (TestMode:$TestMode)\n";
    unless ($ShopClass->existsAttribute('UpdateTaxCompleted')) {
        if($TestMode) {
            print "$StoreName TestMode: we would import XML $ImportStatusAttributeFile .\n"
        } else {
            print "$StoreName import XML $ImportStatusAttributeFile \n";
            ImportObjects($ImportStatusAttributeFile);
            $UpdateTaxCompletedAttributeID = $ShopClass->attribute('UpdateTaxCompleted')->id;
        }
    } else {
        print "$StoreName Attribute 'UpdateTaxCompleted' already exists in database.\n"
    }
    print "$StoreName end _importStatusAttribute\n";
    return;
}

sub _removeStatusAttribute {
    my ($TestMode) = @_;

    print "$StoreName start _removeStatusAttribute (TestMode:$TestMode)\n";
    if ($ShopClass->existsAttribute('UpdateTaxCompleted') ) {
        if ($TestMode) {
            print "$StoreName TestMode: we would delete shop attribute 'UpdateTaxCompleted'.\n"
        } else {
            print "$StoreName delete shop attribute 'UpdateTaxCompleted'.\n";
            $ShopClass->attribute('UpdateTaxCompleted')->delete;
        }
    }
    print "$StoreName end _removeStatusAttribute\n";
    return;
}

sub _importTaxXMLs {
    my ($TestMode) = @_;

    print "$StoreName start _importTaxXMLs (TestMode:$TestMode)\n";
    foreach my $File (@TaxFileNames) {
        unless ($TestMode) {
            print "$StoreName import XML $File \n";
            ImportObjects($File);
        } else {
            print "$StoreName TestMode: we would import $File \n";
        }
    }
    print "$StoreName end _importTaxXMLs\n";
    return;
}

sub _getNowShopsToCloseSF {
    my ($ShopAlias) = @_;

    my $dbi= GetCurrentDBHandle();

    #get closed now shops where sf is not already closed
    my @Parameter = ($UnityFeatureID, $ClosedByProviderTemporarilyAttributeID);
    my $sql=<<END_SQL;
SELECT s.shopid FROM shop s
                         INNER JOIN site si ON si.siteid=s.shopid
                         INNER JOIN shopfeature sf ON sf.shopid=s.shopid AND sf.featureid=? AND sf.maxval=1
                         INNER JOIN object so on so.objectid=s.shopid
WHERE isclosed=1
  AND closedbyprovider=0 AND closedbymerchant=0 AND s.shopid NOT IN (
    SELECT ia.objectid FROM booleanattribute ia WHERE ia.attributeid=? AND ia.VALUE > 0
)
END_SQL
    my (@WhereClauses, @WhereParameter);
    if (defined $ShopAlias) {
        unshift @WhereClauses, 'so.alias = ?';
        unshift @WhereParameter, $ShopAlias;
    }
    if(scalar(@WhereClauses) ){
        #add additional parameters:
        $sql .= ' AND ' . join(' AND ', @WhereClauses);
        push @Parameter,@WhereParameter
    }
    my $dbResult = $dbi->execute($sql, \@Parameter, ['Integer']); #, [qw(ShopID)]);
    return [ map { $_->[0] } @$dbResult ];
}


sub _closeShops {
    my ($ShopAlias) = @_;
    my $dbi= GetCurrentDBHandle();

    printf "$StoreName start _closeShops($ShopAlias)\n";
    #close shops in db with systemtaxmatrix or systemdigitaltaxmatrix
    # which have taxmatrix alias= $TaxMatrixAlias
    # inputtaxmodel gross
    # and UpdateTaxCompletedAttribute is not set or 0
    my @Parameter = ($UpdateTaxCompletedAttributeID,$TaxMatrixAttributeID,$SystemTaxMatrixID,
        $UpdateTaxCompletedAttributeID,$TaxMatrixDigitalAttributeID,$SystemTaxMatrixDigitalID,
        $TaxMatrixAlias,$TaxMatrixAttributeID,);
    my $sql=<<END_SQL;
UPDATE site AS updatesite
INNER JOIN (
    SELECT s.shopid FROM shop s
    INNER JOIN site si ON si.siteid=s.shopid
    WHERE isclosed=0 and closedbyprovider=0
      AND s.shopid IN (
        #get shops with inputtaxmodel = gross and with assigned system taxmatrix
        SELECT shopid FROM shop
         WHERE inputtaxmodel = 1
           AND shopid NOT IN (
             #exclude shops which are already in progress or completed
             select ia.objectid from integerattribute ia WHERE ia.attributeid=? AND ia.VALUE > 0
          )
          AND shopid IN (select ra.objectid from referenceattribute ra where ra.attributeid=? AND ra.referencedobjectid=?)
        UNION
        #get shops with inputtaxmodel = gross and with assigned system taxmatrixdigital and taxmatrix has alias TaxMatrixAlias
        SELECT shopid FROM shop
         WHERE inputtaxmodel = 1
           AND shopid NOT IN (
                #exclude shops which are already in progress or completed
                select ia.objectid from integerattribute ia WHERE ia.attributeid=? AND ia.VALUE > 0
            )
            AND shopid IN (select ra.objectid from referenceattribute ra where ra.attributeid=? AND ra.referencedobjectid=?)
            AND shopid IN (select ra.objectid from referenceattribute ra inner join object rao on rao.objectid=ra.referencedobjectid AND rao.alias=? where ra.attributeid=? )
        )
) AS shopstoclose ON shopstoclose.shopid = updatesite.siteid
inner join object so on so.objectid=updatesite.siteid
SET updatesite.isclosed = 1
END_SQL
    my (@WhereClauses, @WhereParameter);
    if (defined $ShopAlias) {
        unshift @WhereClauses, 'so.alias = ?';
        unshift @WhereParameter, $ShopAlias;
    }
    if(scalar(@WhereClauses) ){
        #add additional parameters:
        $sql .= ' WHERE ' . join(' AND ', @WhereClauses);
        push @Parameter,@WhereParameter
    }

    my $sth = $dbi->executeStatement($sql, \@Parameter);
    printf "$StoreName Affected rows (shops closed) %s\n", $sth->rows();
    return;
}


sub _getShopsToClose {
    my ($ShopAlias) = @_;

    my $dbi= GetCurrentDBHandle();
    my (@WhereClauses, @WhereParameter);
    if (defined $ShopAlias) {
        unshift @WhereClauses, 'so.alias = ?';
        unshift @WhereParameter, $ShopAlias;
    }
    #get shops in db with systemtaxmatrix or systemdigitaltaxmatrix
    #which have taxmatrix alias= $TaxMatrixAlias
    #and inputtaxmodel gross
    #and UpdateTaxCompletedAttribute is not set or 0
    my @Parameter = ($UpdateTaxCompletedAttributeID,$TaxMatrixAttributeID,$SystemTaxMatrixID,
        $UpdateTaxCompletedAttributeID,$TaxMatrixDigitalAttributeID,$SystemTaxMatrixDigitalID,
        $TaxMatrixAlias,$TaxMatrixAttributeID,);
    my $sql=<<END_SQL;
SELECT s.shopid FROM shop s
                         INNER JOIN site si ON si.siteid=s.shopid
                         INNER JOIN object so on so.objectid=s.shopid
WHERE isclosed=0 and closedbyprovider=0
  AND s.shopid IN (
    #get shops with inputtaxmodel = gross and with assigned system taxmatrix
     SELECT shopid FROM shop
     WHERE inputtaxmodel = 1
     AND shopid NOT IN (
        #exclude shops which are already in progress or completed
        select ia.objectid from integerattribute ia WHERE ia.attributeid=? AND ia.VALUE > 0
        )
        AND shopid IN (select ra.objectid from referenceattribute ra where ra.attributeid=? AND ra.referencedobjectid=?)
        UNION
        #get shops with inputtaxmodel = gross and with assigned system taxmatrixdigital and taxmatrix with given taxmatrix alias
        SELECT shopid FROM shop
        WHERE inputtaxmodel = 1
        AND shopid NOT IN (
        #exclude shops which are already in progress or completed
        select ia.objectid from integerattribute ia WHERE ia.attributeid=? AND ia.VALUE > 0
        )
        AND shopid IN (select ra.objectid from referenceattribute ra where ra.attributeid=? AND ra.referencedobjectid=?)
        AND shopid IN (select ra.objectid from referenceattribute ra inner join object rao on rao.objectid=ra.referencedobjectid AND rao.alias=? where ra.attributeid=? )
    )
END_SQL


    if(scalar(@WhereClauses) ){
        #add additional parameters:
        $sql .= ' AND ' . join(' AND ', @WhereClauses);
        push @Parameter,@WhereParameter
    }
    my $dbResult = $dbi->execute($sql, \@Parameter, ['Integer']); #, [qw(ShopID)]);
    return [ map { $_->[0] } @$dbResult ];
}


sub _closeSFUnityShops {
    my ($ShopAlias) = @_;

    my ($aShopIDs) = _getNowShopsToCloseSF($ShopAlias);
    return unless scalar @$aShopIDs;

    # update isClosed parameters
    my $hContent->{'isClosed'} = JSON::true;
    foreach my $ShopID (@$aShopIDs) {
        my $Shop = LoadObject($ShopID);

        for (my $Try = 1; $Try <= MAX_TRIES_FOR_PATCHING_UNITYBASEURL; $Try++) {
            eval {
                # send PATCH call to Unity server
                $Shop->sendUnityRequest('PATCH', '/shop', undef, $hContent);
            };
            if (ExistsError()) {
                if ($Try == MAX_TRIES_FOR_PATCHING_UNITYBASEURL) {
                    # only log an error in case all tries have failed
                    LogError(GetError());
                } else {
                    # otherwise wait a second and try again
                    sleep 1;
                }
            } else {
                last;
            }
        }
    }
    return;
}


sub _getShopsToUpdate {
    my ($ShopAlias, $SFOpenByMerchant,$LastMerchantLogin)= @_;

    print "$StoreName _getShopsToUpdate(ShopAlias: $ShopAlias, SFOpenByMerchant:$SFOpenByMerchant,LastMerchantLogin:$LastMerchantLogin)\n";

    my (@WhereClauses, @WhereParameter);
    if (defined $ShopAlias) {
        unshift @WhereClauses, 'so.alias = ?';
        unshift @WhereParameter, $ShopAlias;
    }
    if (defined $SFOpenByMerchant) {
        unshift @WhereClauses, 'closedbymerchant = ?';
        unshift @WhereParameter, 0;
    }
    if (defined $LastMerchantLogin) {
        unshift @WhereClauses, 'lastmerchantlogin >= ?';
        unshift @WhereParameter, $LastMerchantLogin;
    }
    my $dbi= GetCurrentDBHandle();
    my @Parameter = ($UpdateTaxCompletedAttributeID,$TaxMatrixAttributeID,$SystemTaxMatrixID,
        $UpdateTaxCompletedAttributeID,$TaxMatrixDigitalAttributeID,$SystemTaxMatrixDigitalID,
        $TaxMatrixAlias,$TaxMatrixAttributeID,);

    my $sql=<<END_SQL;
SELECT s.shopid FROM shop s
                         INNER JOIN site si ON si.siteid=s.shopid
                         INNER JOIN object so ON so.objectid=s.shopid
WHERE closedbyprovider=0
  AND s.shopid IN (
    #get shops with inputtaxmodel = gross and with assigned system taxmatrix
    SELECT shopid FROM shop
      WHERE inputtaxmodel = 1
      AND shopid NOT IN (
        #exclude shops which are already in progress or completed
        select ia.objectid from integerattribute ia WHERE ia.attributeid=? AND ia.VALUE > 0
        )
        AND shopid IN (select ra.objectid from referenceattribute ra where ra.attributeid=? AND ra.referencedobjectid=?)
        UNION
        #get shops with inputtaxmodel = gross and with assigned system taxmatrixdigital germany
        SELECT shopid FROM shop
        WHERE inputtaxmodel = 1
        AND shopid NOT IN (
        #exclude shops which are already in progress or completed
        select ia.objectid from integerattribute ia WHERE ia.attributeid=? AND ia.VALUE > 0
        )
        AND shopid IN (select ra.objectid from referenceattribute ra where ra.attributeid=? AND ra.referencedobjectid=?)
        AND shopid IN (select ra.objectid from referenceattribute ra inner join object rao on rao.objectid=ra.referencedobjectid AND rao.alias=? where ra.attributeid=? )
    )
END_SQL
    if(scalar(@WhereClauses) ){
        #add additional parameters:
        $sql .= ' AND ' . join(' AND ', @WhereClauses);
        push @Parameter,@WhereParameter
    }

    my $dbResult = $dbi->execute($sql, \@Parameter, ['Integer']);
    return [ map { $_->[0] } @$dbResult ];
}

sub _updatePrices {
    my ($TestMode, $ShopAlias, $SFOpenByMerchant, $LastMerchantLogin) = @_;

    print "$StoreName update prices\n";
    print "$StoreName start _updatePrices(TestMode:$TestMode, ShopAlias:$ShopAlias, SFOpenByMerchant: $SFOpenByMerchant, LastMerchantLogin: $LastMerchantLogin )\n";

    # Define array for shops.
    my $aShopIDs = _getShopsToUpdate($ShopAlias, $SFOpenByMerchant, $LastMerchantLogin);
    if ($TestMode) {
        print "$StoreName TestMode: we would update " . scalar(@$aShopIDs) . " shops.\n";
    } else {
        print "$StoreName We will update " . scalar(@$aShopIDs) . " shops.\n";
    }
    foreach my $ShopID (@$aShopIDs) {
        my $Shop = LoadObject($ShopID);

        printf "================================================================================ \n";
        # Try the whole updateShopPrices at best effort, if it fails continue with next shop
        eval {
            Transaction(sub {
                _updateShopPrices($Shop, $TestMode);
                die "$StoreName $ShopAlias rollback everything because it is a test\n" if $TestMode;
            });
        };
        if (ExistsError()) {
            my $Error = GetError();
            LogError($Error);
            printf "$StoreName There was an issue with shop %s, changes have been rolled back.\n", $Shop->alias;
            printf "$StoreName The reported error was '%s' (Details in error log)\n", $Error->shortMessage;
        } else {
            if ($UseSQL) {
                unless ($TestMode) {
                    #update search index, sync cache
                    eval {
                        GetOrCreateSearchIndex($Shop)->initialize();
                        SynchronizeCache();
                    };
                    Warning(GetError()) if ExistsError();
                    printf "$StoreName Update Cache, update SearchIndex for shop %s.\n", $Shop->alias;
                } else {
                    printf "$StoreName TestMode: Update Cache, update SearchIndex for shop %s.\n", $Shop->alias;
                }
            }
        }
    }
    ResetFactoryCaches() if $UseSQL;
    printf "================================================================================ \n";
}

sub _updateShopPrices {
    my ( $Shop, $TestMode ) = @_;

    my $ShopAlias = $Shop->alias();
    print "$StoreName _updateShopPrices(Shop:$ShopAlias, TestMode:$TestMode)\n";

    # Run script only for shops with $TaxMatrix parameter value
    unless ($Shop->get('TaxMatrix')->alias eq $TaxMatrixAlias) {
        printf "$StoreName Skip " .$ShopAlias. " Wrong TaxMatrix.\n";
        $Shop->set({'IsClosed'=>0});
        return;
    }

    # Get ShopTaxMatrixID used for $Shop.
    my $ShopTaxMatrixID = $Shop->get('TaxMatrix')->id();
    my $ShopTaxMatrixDigitalID;
    if (defined ($Shop->get('TaxMatrixDigital'))) {
        $ShopTaxMatrixDigitalID = $Shop->get('TaxMatrixDigital')->id();
    }
    # Next if shop has its own TaxMatrix and own taxmatrixdigital
    if(($SystemTaxMatrixID != $ShopTaxMatrixID) && ($SystemTaxMatrixDigitalID != $ShopTaxMatrixDigitalID)) {
        printf "$StoreName Skip " . $ShopAlias . " Customized TaxMatrix.\n";
        $Shop->set({'IsClosed'=>0});
        return;
    }

    # Next if Shop InputTaxModel is not TAX_MODEL_GROSS.
    if ($Shop->get('InputTaxModel') != TAX_MODEL_GROSS) {
        printf "$StoreName Skip " . $Shop->alias() . " Taxmodel not gross. \n";
        $Shop->set({'IsClosed'=>0});
        return;
    }
    # Next if UpdateTaxCompleted is set.
    if ($Shop->get('UpdateTaxCompleted')) {
        printf "$StoreName Skip " .$ShopAlias. ": Already updated prices. \n";
        return;
    } else {
        if (!$TestMode){
            $Shop->set({ 'UpdateTaxCompleted' => 1 });
        } else {
            printf "$StoreName TestMode: " . $ShopAlias . ": would set UpdateTaxCompleted=1. \n";
        }
    }
    printf "Start updates for: $ShopAlias \n";
    if($UseSQL) {
        _updateProductPricesSQL($Shop, $TestMode);
        _updateShippingMethodPricesSQL($Shop, $TestMode);
        _updatePaymentMethodPricesSQL($Shop, $TestMode);
    } else {
        _updateProductPrices($Shop, $TestMode);
        _updateShippingMethodPrices($Shop, $TestMode);
        _updatePaymentMethodPrices($Shop, $TestMode);
    }
    printf "$StoreName Finished updates for: $ShopAlias \n";

    if (!$TestMode) {
        $Shop->set({ 'UpdateTaxCompleted' => 2, 'IsClosed' => 0 });
        printf "$StoreName Finished for shop: $ShopAlias set UpdateTaxCompleted: " . $Shop->get('UpdateTaxCompleted') . "\n";
    } else {
        printf "$StoreName TestMode: $ShopAlias : would set UpdateTaxCompleted=2.\n";
    }
    return;
}

sub _updateProductPricesWithTaxClassSQL {
    my ($ShopID, $TaxClassID, $OldTaxRate, $NewTaxRate, $TestMode) = @_;

    my $dbi= GetCurrentDBHandle();

    if (!$TestMode) {
        print("$StoreName _updateProductPricesWithTaxClassSQL ($ShopID,$TaxClassID, $OldTaxRate,$NewTaxRate, $TestMode ) \n");

        #update productprices with taxclass:
        my @Parameter = ($OldTaxRate,$NewTaxRate,$OldTaxRate,$NewTaxRate,$OldTaxRate,$NewTaxRate,$TaxClassID,$ShopID );
        #update products/subproducts with prices by taxclass
        my $sth = $dbi->executeStatement(<<END_SQL, \@Parameter);
UPDATE productprice AS updateprice
    INNER JOIN (
        SELECT pp.productid, currencyid,listprice, ROUND(listprice * ? / ?,3) as newlistprice , manufacturerprice, ROUND(manufacturerprice * ? / ?,3) as newmanufacturerprice, deposit, ROUND(deposit * ? / ?,3) as newdeposit FROM productprice pp
       INNER JOIN product p ON p.productid=pp.productid AND taxclassid=? AND taxclassdigitalid IS NULL
       WHERE shopid=?
    ) AS originalprice ON originalprice.productid=updateprice.productid AND originalprice.currencyid=updateprice.currencyid
SET updateprice.listprice=originalprice.newlistprice, updateprice.manufacturerprice=originalprice.newmanufacturerprice, updateprice.deposit=originalprice.newdeposit;
END_SQL
        printf "$StoreName Affected rows (productprice with taxclass) %s\n", $sth->rows();

        #update subproducts without own taxclassid (use parent taxclassid)
        $sth = $dbi->executeStatement(<<END_SQL, \@Parameter);
UPDATE productprice AS updateprice
    INNER JOIN (
        SELECT p.taxclassid,p.hassubownprices,p.superproductid, pp.productid, currencyid,listprice, ROUND(listprice * ? / ?,3) as newlistprice , manufacturerprice, ROUND(manufacturerprice * ? / ?,3) as newmanufacturerprice, deposit, ROUND(deposit * ? / ?,3) as newdeposit FROM productprice pp
        INNER JOIN product p ON p.productid=pp.productid AND taxclassid IS null AND taxclassdigitalid IS NULL AND hassubownprices=1
        INNER JOIN product superp on superp.productid=p.superproductid AND superp.taxclassid=? AND superp.taxclassdigitalid IS NULL
        WHERE p.shopid=?
    )  AS originalprice ON originalprice.productid=updateprice.productid AND originalprice.currencyid=updateprice.currencyid
SET updateprice.listprice=originalprice.newlistprice, updateprice.manufacturerprice=originalprice.newmanufacturerprice, updateprice.deposit=originalprice.newdeposit;
END_SQL
        printf "$StoreName Affected rows (productprice with super product taxclass) %s\n", $sth->rows();

        #update scaleprices with taxclass:
        my @ParameterScalePrice = ($OldTaxRate,$NewTaxRate,$TaxClassID,$ShopID );
        #update scale prices of products by taxclass
        $sth = $dbi->executeStatement(<<END_SQL, \@ParameterScalePrice);
UPDATE scaleprice AS updatescaleprice
    INNER JOIN (
        SELECT sc.scalepriceid,sc.price,p.taxclassid, ROUND(sc.price * ? / ?,4) as newprice
        FROM scaleprice sc
        INNER JOIN product p ON p.productid=sc.productid AND taxclassid=? AND taxclassdigitalid IS NULL
        WHERE shopid=?
    ) AS originalprice ON originalprice.scalepriceid=updatescaleprice.scalepriceid
SET updatescaleprice.price=originalprice.newprice
END_SQL
        printf "$StoreName Affected rows (scaleprice with taxclass) %s\n", $sth->rows();

        #update scale prices of subproducts without own taxclass:
        $sth = $dbi->executeStatement(<<END_SQL, \@ParameterScalePrice);
UPDATE scaleprice AS updatescaleprice
    INNER JOIN (
        SELECT sc.scalepriceid,sc.price, ROUND(sc.price * ? / ?,4) as newprice
        FROM scaleprice sc
        INNER JOIN product p ON p.productid=sc.productid AND taxclassid IS null AND taxclassdigitalid IS NULL AND hassubownprices=1
        INNER JOIN product superp on superp.productid=p.superproductid AND superp.taxclassid=? AND superp.taxclassdigitalid IS NULL
        WHERE p.shopid=?
    ) AS originalprice ON originalprice.scalepriceid=updatescaleprice.scalepriceid
 SET updatescaleprice.price=originalprice.newprice
END_SQL
        #update valuediscounts with taxclass:
        my @ParameterValueDiscount = ($OldTaxRate,$NewTaxRate,$TaxClassID,$ShopID );
        #update valuediscounts of products by taxclass
        $sth = $dbi->executeStatement(<<END_SQL, \@ParameterValueDiscount);
UPDATE valuediscount AS updatevaluediscount
    INNER JOIN (
        SELECT vd.valuediscountid,vd.value,p.taxclassid, ROUND(vd.value * ? / ?,4) as newvalue
        FROM valuediscount vd
        INNER JOIN product p ON p.productid=vd.productid AND taxclassid=? AND taxclassdigitalid IS NULL
        WHERE shopid=?
    ) AS originalvalue ON originalvalue.valuediscountid=updatevaluediscount.valuediscountid
SET updatevaluediscount.value=originalvalue.newvalue
END_SQL
        printf "$StoreName Affected rows (valuediscount with taxclass) %s\n", $sth->rows();

        #update valuediscounts of subproducts without own taxclass:
        $sth = $dbi->executeStatement(<<END_SQL, \@ParameterValueDiscount);
UPDATE valuediscount AS updatevaluediscount
    INNER JOIN (
        SELECT vd.valuediscountid,vd.value, ROUND(vd.value * ? / ?,4) as newvalue
        FROM valuediscount vd
        INNER JOIN product p ON p.productid=vd.productid AND taxclassid IS null AND taxclassdigitalid IS NULL AND hassubownprices=1
        INNER JOIN product superp on superp.productid=p.superproductid AND superp.taxclassid=? AND superp.taxclassdigitalid IS NULL
        WHERE p.shopid=?
    ) AS originalvalue ON originalvalue.valuediscountid=updatevaluediscount.valuediscountid
 SET updatevaluediscount.value=originalvalue.newvalue
END_SQL
        printf "$StoreName Affected rows (scaleprice  with super product taxclass) %s\n", $sth->rows();
    } else {
        print("$StoreName TestMode: _updateProductPricesWithTaxClassSQL ($ShopID, $TaxClassID, $OldTaxRate, $NewTaxRate, $TestMode)\n");
    }
    return;
}

sub _updateProductPricesSQL {
    my ( $Shop, $TestMode ) = @_;

    my $ShopAlias = $Shop->alias();
    my $ShopID =  $Shop->id;

    printf "$StoreName Start update product prices (SQL): $ShopAlias\n";
    _updateProductPricesWithTaxClassSQL($ShopID,$TaxClassIDNormal,$OldTaxNormal,$NewTaxNormal, $TestMode);
    _updateProductPricesWithTaxClassSQL($ShopID,$TaxClassIDReduced,$OldTaxReduced,$NewTaxReduced, $TestMode);
    printf "$StoreName Finish update product prices (SQL): $ShopAlias\n";
    return;
}


sub _updateShippingMethodsWithTaxClassSQL {
    my ($ShopID,$TaxClassID, $OldTaxRate,$NewTaxRate, $TestMode ) = @_;
    my $dbi= GetCurrentDBHandle();

    if (!$TestMode) {
        print("$StoreName _updateShippingMethodsWithTaxClassSQL ($ShopID, $TaxClassID, $OldTaxRate, $NewTaxRate, $TestMode)\n");
        my @Parameter = ($OldTaxRate,$NewTaxRate,$OldTaxRate,$NewTaxRate,$TaxClassID,$ShopID );
        #update shippinglevel (basevalue,multiplier) with taxclass not type PriceOfBasket:
        my $sth = $dbi->executeStatement(<<END_SQL, \@Parameter);
UPDATE shippinglevel AS updateshipping
    INNER JOIN (
        SELECT shippinglevelid,basevalue,ROUND(basevalue * ? / ?,3) AS newbasevalue,multiplier,ROUND(multiplier * ? / ?,9) AS newmultiplier FROM shippinglevel sl
        INNER JOIN shippingmethod sm ON sm.shippingmethodid=sl.shippingmethodid and sm.taxclassid=?
        INNER JOIN object sto on sm.shippingtypeid=sto.objectid and sto.alias != 'PriceOfBasket'
        WHERE shopid=?
    )  AS originalshipping ON originalshipping.shippinglevelid=updateshipping.shippinglevelid
SET updateshipping.basevalue=originalshipping.newbasevalue, updateshipping.multiplier=originalshipping.newmultiplier;
END_SQL
        printf "$StoreName Affected rows (shipping levels with taxclass not type PriceOfBasket) %s\n", $sth->rows();

        #update shippinglevel ((only basevalue) with taxclass type PriceOfBasket:
        my @ParameterPriceOfBasket = ($OldTaxRate,$NewTaxRate,$TaxClassID,$ShopID );
        $sth = $dbi->executeStatement(<<END_SQL, \@ParameterPriceOfBasket);
UPDATE shippinglevel AS updateshipping
    INNER JOIN (
        SELECT shippinglevelid,basevalue,ROUND(basevalue * ? / ?,3) AS newbasevalue FROM shippinglevel sl
        INNER JOIN shippingmethod sm ON sm.shippingmethodid=sl.shippingmethodid and sm.taxclassid=?
        INNER JOIN object sto on sm.shippingtypeid=sto.objectid and sto.alias = 'PriceOfBasket'
        WHERE shopid=?
    )  AS originalshipping ON originalshipping.shippinglevelid=updateshipping.shippinglevelid
SET updateshipping.basevalue=originalshipping.newbasevalue;
END_SQL
        printf "$StoreName Affected rows (shipping level with taxclass type PriceOfBasket) %s\n", $sth->rows();

        #update shippingvalues with taxclass:
        my @ParameterShippingValues = ($OldTaxRate,$NewTaxRate,$TaxClassID,$ShopID);
        $sth = $dbi->executeStatement(<<END_SQL, \@ParameterShippingValues);
UPDATE shippingvalues AS updateshipping
    INNER JOIN (
        SELECT shippingvaluesid,freeshippingprice, ROUND(freeshippingprice * ? / ?,3) AS newfreeshippingprice FROM shippingvalues sv
        INNER JOIN shippingmethod sm ON sm.shippingmethodid=sv.shippingmethodid and sm.taxclassid=?
        WHERE freeshippingprice IS NOT NULL
        AND shopid=?
    )  AS originalshipping ON originalshipping.shippingvaluesid=updateshipping.shippingvaluesid
SET updateshipping.freeshippingprice=originalshipping.newfreeshippingprice;
END_SQL
        printf "$StoreName Affected rows (shipping values with taxclass) %s\n", $sth->rows();

    } else {
        print("$StoreName TestMode: _updateShippingMethodsWithTaxClassSQL ($ShopID, $TaxClassID, $OldTaxRate, $NewTaxRate, $TestMode)\n");
    }
    return;
}

sub _updateShippingMethodPricesSQL {
    my ($Shop, $TestMode) = @_;

    my $ShopAlias = $Shop->alias();
    my $ShopID =  $Shop->id;

    printf "$StoreName Start update shippingmethod prices (SQL): $ShopAlias\n";
    _updateShippingMethodsWithTaxClassSQL($ShopID,$TaxClassIDNormal,$OldTaxNormal,$NewTaxNormal, $TestMode);
    _updateShippingMethodsWithTaxClassSQL($ShopID,$TaxClassIDReduced,$OldTaxReduced,$NewTaxReduced, $TestMode);
    printf "$StoreName Finish update shippingmethod prices (SQL): $ShopAlias\n";
    return;
}

sub _updatePaymentMethodsWithTaxClassSQL {
    my ($ShopID,$TaxClassID, $OldTaxRate,$NewTaxRate, $TestMode ) = @_;
    my $dbi= GetCurrentDBHandle();

    if (!$TestMode) {
        print("$StoreName _updatePaymentMethodsWithTaxClassSQL ($ShopID, $TaxClassID, $OldTaxRate, $NewTaxRate, $TestMode)\n");
        my @Parameter = ($OldTaxRate,$NewTaxRate,$OldTaxRate,$NewTaxRate,$TaxClassID,$ShopID );
        #update paymentlevel with taxclass:
        my $sth = $dbi->executeStatement(<<END_SQL, \@Parameter);
UPDATE paymentlevel AS updatepayment
    INNER JOIN (
        SELECT paymentlevelid,basevalue,ROUND(basevalue * ? / ?,3) AS newbasevalue,multiplier,ROUND(multiplier * ? / ?,9) AS newmultiplier FROM paymentlevel sl
        INNER JOIN paymentmethod sm ON sm.paymentmethodid=sl.paymentmethodid and sm.taxclassid=?
        WHERE shopid=?
    )  AS originalpayment ON originalpayment.paymentlevelid=updatepayment.paymentlevelid
SET updatepayment.basevalue=originalpayment.newbasevalue, updatepayment.multiplier=originalpayment.newmultiplier;
END_SQL
        printf "$StoreName Affected rows (payment level with taxclass) %s\n", $sth->rows();

    } else {
        print("$StoreName TestMode: _updatePaymentMethodsWithTaxClassSQL ($ShopID, $TaxClassID, $OldTaxRate, $NewTaxRate, $TestMode)\n");
    }
    return;
}

sub _updatePaymentMethodPricesSQL {
    my ($Shop, $TestMode) = @_;

    my $ShopAlias = $Shop->alias();
    my $ShopID =  $Shop->id;

    printf "\n$StoreName Start update paymentmethod prices: $ShopAlias\n";
    _updatePaymentMethodsWithTaxClassSQL($ShopID,$TaxClassIDNormal,$OldTaxNormal,$NewTaxNormal, $TestMode);
    _updatePaymentMethodsWithTaxClassSQL($ShopID,$TaxClassIDReduced,$OldTaxReduced,$NewTaxReduced, $TestMode);
    printf "$StoreName Finish update paymentmethod prices (SQL): $ShopAlias\n";
    return;
}


sub _updateProductPrices {
    my ( $Shop, $TestMode ) = @_;

    printf "$StoreName Update product prices.\n";

    # Get all products.
    my $aaProducts = GetCurrentDBHandle()->execute(<<END_SQL, [ $Shop->id ]);
                SELECT p.productid, pp.listprice, pp.manufacturerprice, pp.currencyid, pp.deposit, p.taxclassid
                FROM product p
                left outer join productprice pp on p.productid=pp.productid
                where p.shopid= ?
                ORDER BY productid;
END_SQL

    my $ScalePricesDoneProductID;
    my $ScalePriceClass = LoadClassByAlias('ScalePrice');
    my $ValueDiscountClass = LoadClassByAlias('ValueDiscount');

    foreach my $aProduct (@$aaProducts) {
        my $ListPrice;
        my $ProductCurrencyID;
        my $ProductTaxClass;
        my $ProductTaxClassDigital;
        my $Product = LoadObject($aProduct->[0]);
        my $ProductPath = $Product->pathString();
        my $ProductListPrice = $aProduct->[1];
        my $ManufacturePrice = $aProduct->[2];
        my $DepositPrice = $aProduct->[4];
        $ProductCurrencyID = $aProduct->[3];

        # Check if we have own prices on sub product.
        if ($Product->get('SuperProduct')) {
            next unless $Product->get('HasSubOwnPrices');
        }

        $ProductTaxClass = $Product->getWithParent('TaxClass');

        #for swiss shops we don't change digital stuff because this is not available for swiss shops
        # (we dont't have to change the tax values for an other country)
        $ProductTaxClassDigital = undef; #$Product->getWithParent('TaxClassDigital');

        unless (defined $ProductTaxClass || defined $ProductTaxClassDigital) {
            printf "\n$StoreName Product: " . $aProduct->[0] . ' has no TaxClass';
            next;
        }

        my $ProductTaxClassID = defined $ProductTaxClass ? $ProductTaxClass->id() : 0;
        my $ProductTaxClassDigitalID = defined $ProductTaxClassDigital ? $ProductTaxClassDigital->id() : 0;

        if ($ProductListPrice) {
            # Update listprice by using API.
            if ($ProductTaxClassDigitalID) {
                if ($ProductTaxClassDigitalID == $TaxClassIDigitalservices) {
                    $ListPrice = $ProductListPrice * $OldTaxDownload / $NewTaxDownload;
                }
                elsif ($ProductTaxClassDigitalID == $TaxClassIDebooks) {
                    $ListPrice = $ProductListPrice * $OldTaxEbooks / $NewTaxEbooks;
                }
            }
            else {
                if ($ProductTaxClassID == $TaxClassIDNormal) {
                    $ListPrice = $ProductListPrice * $OldTaxNormal / $NewTaxNormal;
                }
                elsif ($ProductTaxClassID == $TaxClassIDReduced) {
                    $ListPrice = $ProductListPrice * $OldTaxReduced / $NewTaxReduced;
                }
            }

            # Save listprice as net like saved in db.
            if ($TestMode && defined $ListPrice) {
                printf "$StoreName Set listprice: " . $ProductPath . '[' . $aProduct->[0] . "]( $ProductCurrencyID, TAX_MODEL_NET, $ListPrice ) oldprice: " . $ProductListPrice . "\n";
            }
            elsif (defined $ListPrice) {
                $Product->setListPrice($ProductCurrencyID, TAX_MODEL_NET, $ListPrice);
            }
        }

        if ($ManufacturePrice) {
            # Update manufacturerprice by using API if ManufacturerPrice > 0.
            if ($ProductTaxClassDigitalID) {
                if ($ProductTaxClassID == $TaxClassIDigitalservices) {
                    $ManufacturePrice = $ManufacturePrice * $OldTaxDownload / $NewTaxDownload;
                }
                elsif ($ProductTaxClassID == $TaxClassIDebooks) {
                    $ManufacturePrice = $ManufacturePrice * $OldTaxEbooks / $NewTaxEbooks;
                }
            }
            else {
                if ($ProductTaxClassID == $TaxClassIDNormal) {
                    $ManufacturePrice = $ManufacturePrice * $OldTaxNormal / $NewTaxNormal;
                }
                elsif ($ProductTaxClassID == $TaxClassIDReduced) {
                    $ManufacturePrice = $ManufacturePrice * $OldTaxReduced / $NewTaxReduced;
                }
            }

            # Save manufacturerprice as net like saved in db.
            if ($TestMode) {
                printf "$StoreName Set ManufacturerPrice: " . $ProductPath . '[' . $aProduct->[0] . "]( $ProductCurrencyID, TAX_MODEL_NET, $ManufacturePrice) oldprice: " . $aProduct->[2] . "\n";
            }
            else {
                $Product->setManufacturerPrice($ProductCurrencyID, TAX_MODEL_NET, $ManufacturePrice);
            }
        }

        # Update deposit price by using API if DepositPrice > 0.
        if ($DepositPrice) {
            if ($ProductTaxClassDigitalID) {
                if ($ProductTaxClassID == $TaxClassIDigitalservices) {
                    $DepositPrice = $DepositPrice * $OldTaxDownload / $NewTaxDownload;
                }
                elsif ($ProductTaxClassID == $TaxClassIDebooks) {
                    $DepositPrice = $DepositPrice * $OldTaxEbooks / $NewTaxEbooks;
                }
            }
            else {
                if ($ProductTaxClassID == $TaxClassIDNormal) {
                    $DepositPrice = $DepositPrice * $OldTaxNormal / $NewTaxNormal;
                }
                elsif ($ProductTaxClassID == $TaxClassIDReduced) {
                    $DepositPrice = $DepositPrice * $OldTaxReduced / $NewTaxReduced;
                }
            }

            # Save deposit price as net like saved in db.
            if ($TestMode) {
                printf "$StoreName Set deposit price: " . $ProductPath . '[' . $aProduct->[0] . "]( $ProductCurrencyID, TAX_MODEL_NET, $DepositPrice) oldprice: " . $aProduct->[4] . "\n";
            }
            else {
                $Product->setDeposit($ProductCurrencyID, TAX_MODEL_NET, $DepositPrice);
            }
        }


        # Update scale prices for product
        my $ahProductScalePrices = $ScalePriceClass->getObjects('ByProductID', { 'ProductID' => $Product->id });
        my $Price;

        my $ahValueDiscountPrices = $ValueDiscountClass->getObjects('ByProductID', { 'ProductID' => $Product->id });
        my $ValueDiscountPrice;

        if ($ScalePricesDoneProductID != $Product->id) {
            foreach my $ScalePrice (@{$ahProductScalePrices}) {
                if ($ProductTaxClassDigitalID) {
                    if ($ProductTaxClassDigitalID == $TaxClassIDigitalservices) {
                        $Price = $ScalePrice->get('Price') * $OldTaxDownload / $NewTaxDownload;
                    }
                    elsif ($ProductTaxClassDigitalID == $TaxClassIDebooks) {
                        #$Price = $ScalePrice->get('Price') * $OldTaxEbooks / $NewTaxEbooks;
                        # Workaround UNITY-6467
                        $Price = $ScalePrice->get('Price') * $OldTaxNormal / $NewTaxNormal;
                    }
                }
                elsif ($ProductTaxClassID == $TaxClassIDNormal) {
                    $Price = $ScalePrice->get('Price') * $OldTaxNormal / $NewTaxNormal;
                }
                elsif ($ProductTaxClassID == $TaxClassIDReduced) {
                    $Price = $ScalePrice->get('Price') * $OldTaxReduced / $NewTaxReduced;
                }
                if ($TestMode) {
                    printf "$StoreName Set ScalePrice for : ".$Product->id.' old: '.$ScalePrice->get('Price').' -> new: '.$Price."\n";
                }
                else {
                    $ScalePrice->set({ 'Price' => $Price });
                }
            }

            # Update value discounts for product

            foreach my $ValueDiscount (@{$ahValueDiscountPrices}) {
                if ($ProductTaxClassDigitalID) {
                    if ($ProductTaxClassDigitalID == $TaxClassIDigitalservices) {
                        $ValueDiscountPrice = $ValueDiscount->get('Value') * $OldTaxDownload / $NewTaxDownload;
                    }
                    elsif ($ProductTaxClassDigitalID == $TaxClassIDebooks) {
                        #$Price = $ScalePrice->get('Price') * $OldTaxEbooks / $NewTaxEbooks;
                        # Workaround UNITY-6467
                        $ValueDiscountPrice = $ValueDiscount->get('Value') * $OldTaxNormal / $NewTaxNormal;
                    }
                }
                elsif ($ProductTaxClassID == $TaxClassIDNormal) {
                    $ValueDiscountPrice = $ValueDiscount->get('Value') * $OldTaxNormal / $NewTaxNormal;
                }
                elsif ($ProductTaxClassID == $TaxClassIDReduced) {
                    $ValueDiscountPrice = $ValueDiscount->get('Value') * $OldTaxReduced / $NewTaxReduced;
                }
                if ($TestMode) {
                    printf "$StoreName Set Discount value for : ".$Product->id.' old: '.$ValueDiscount->get('Value').' -> new: '.$ValueDiscountPrice."\n";
                }
                else {
                    $ValueDiscount->set({ 'Value' => $ValueDiscountPrice });
                }
            }
        }
        $ScalePricesDoneProductID = $Product->id;
    }
    printf "$StoreName Finish update product prices.\n";
    return;
}

sub _updateShippingMethodPrices {
    my ($Shop, $TestMode) = @_;

    my $OldTax;
    my $NewTax;
    my $BaseValue;
    my $LowerBound;
    my $UpperBound;
    my $Multiplier;
    my $FreeShippingPrice;
    my $ShippingLevelID;
    my $ahCurrencies = $Shop->get('Currencies');

    my $ahShippingMethods = $Shop->get('ShippingMethods');

    printf "\n$StoreName Start update shippingmethod prices.\n";

    foreach my $ShippingMethod (@{$ahShippingMethods}) {

        next if ($ShippingMethod->get('Class')->alias() eq 'ShippingMethodFree');

        my $ShippingMethodTaxClassID = $ShippingMethod->get('TaxClass')->id;
        # Ogone has TaxClass none
        next if ($ShippingMethodTaxClassID == $TaxClassIDNone);

        if ($ShippingMethodTaxClassID == $TaxClassIDNormal) {
            $OldTax = $OldTaxNormal;
            $NewTax = $NewTaxNormal;
        } elsif ($ShippingMethodTaxClassID == $TaxClassIDReduced) {
            $OldTax = $OldTaxReduced;
            $NewTax = $NewTaxReduced;
        } else {
            #may be system taxclass id
            next;
        }

        foreach my $Currency (@{$ahCurrencies}) {
            my $CurrencyID = $Currency->{'CurrencyID'};
            my $ahLevels = [];

            # Get all Levels for current Shippingmethod
            my $aLevelsShippingMethod = $ShippingMethod->getLevels($CurrencyID, TAX_MODEL_NET, {'ExtraPrecision' => 1 });

            foreach my $level (@{$aLevelsShippingMethod}) {

                $BaseValue = $level->{'BaseValue'} * $OldTax / $NewTax;
                $LowerBound = $level->{'LowerBound'};
                $UpperBound = $level->{'UpperBound'};
                $Multiplier = $level->{'Multiplier'};
                $Multiplier = $ShippingMethod->get('ShippingType')->alias ne 'PriceOfBasket' ? $level->{'Multiplier'} * $OldTax / $NewTax : $Multiplier;
                $ShippingLevelID = $level->{'ShippingLevelID'};

                push(@{$ahLevels}, {
                    'BaseValue'       => $BaseValue,
                    'LowerBound'      => $LowerBound,
                    'UpperBound'      => $UpperBound,
                    'Multiplier'      => $Multiplier,
                    'ShippingLevelID' => $ShippingLevelID });
            }

            if ($TestMode) {
                printf "$StoreName Update price for: " . $ShippingMethod->get('Class')->alias() . "\n";
            }
            else {
                # Update shipping prices
                $ShippingMethod->setLevels($CurrencyID, TAX_MODEL_NET, $ahLevels);
            }

            # Get all Shipping values for current Shippingmethod
            my $aGetShippingValues = $ShippingMethod->getShippingValues($CurrencyID, TAX_MODEL_NET, 1);
            my $ahShippingValues = [];

            foreach my $ShippingValue ($aGetShippingValues) {
                if ($ShippingValue->{'FreeShippingPrice'}){

                    $FreeShippingPrice = $ShippingValue->{'FreeShippingPrice'} * $OldTax / $NewTax;

                    push(@{$ahShippingValues}, {
                        'MaximumWeight'     => $ShippingValue->{'MaximumWeight'},
                        'CurrencyID'        => $ShippingValue->{'CurrencyID'},
                        'FreeShippingPrice' => $FreeShippingPrice,
                    });
                    if ($TestMode) {
                        printf "$StoreName Update shipping value for: " . $ShippingMethod->get('Class')->alias() . "\n";
                    }
                    else {
                        # Update shipping values
                        $ShippingMethod->setShippingValues($CurrencyID, TAX_MODEL_NET, @{$ahShippingValues});
                    }
                }
            }
        }
    }
    printf "$StoreName Finish update shippingmethods. \n";
    return;
}

sub _updatePaymentMethodPrices {
    my ($Shop, $TestMode) = @_;

    my $OldTax;
    my $NewTax;
    my $BaseValue;
    my $LowerBound;
    my $UpperBound;
    my $Multiplier;
    my $PaymentLevelID;
    my $ahCurrencies = $Shop->get('Currencies');

    printf "\n$StoreName Start update paymentmethod prices.\n";

    foreach my $PaymentMethod (@{$Shop->get('PaymentMethods')}) {

        next unless (defined $PaymentMethod->get('TaxClass'));
        next unless $PaymentMethod->get('IsActivated');

        my $PaymentMethodTaxClassID= $PaymentMethod->get('TaxClass')->id;
        if ($PaymentMethodTaxClassID == $TaxClassIDNormal) {
            $OldTax = $OldTaxNormal;
            $NewTax = $NewTaxNormal;
        } elsif ($PaymentMethodTaxClassID == $TaxClassIDReduced) {
            $OldTax = $OldTaxReduced;
            $NewTax = $NewTaxReduced;
        } else {
            #may be system taxclass id
            next;
        }

        foreach my $Currency (@{$ahCurrencies}) {
            my $CurrencyID = $Currency->{'CurrencyID'};

            my $ahLevels = [];

            # Get all Levels for current payment method
            my $aLevelsPaymentMethod = $PaymentMethod->getLevels($CurrencyID, TAX_MODEL_NET, {'ExtraPrecision' => 1 });

            foreach my $Level (@{$aLevelsPaymentMethod}) {

                $BaseValue = $Level->{'BaseValue'} * $OldTax / $NewTax;
                $LowerBound = $Level->{'LowerBound'};
                $UpperBound = $Level->{'UpperBound'};
                $Multiplier = $Level->{'Multiplier'};
                $PaymentLevelID = $Level->{'PaymentLevelID'};

                push(@{$ahLevels}, {
                    'BaseValue'       => $BaseValue,
                    'LowerBound'      => $LowerBound,
                    'UpperBound'      => $UpperBound,
                    'Multiplier'      => $Multiplier,
                    'ShippingLevelID' => $PaymentLevelID });
            }

            if ($TestMode) {
                printf "$StoreName Update price for: " . $PaymentMethod->get('Class')->alias() . "\n";
            }
            else {
                # Update payment method
                $PaymentMethod->setLevels($CurrencyID, TAX_MODEL_NET, $ahLevels);
            }
        }
    }
    printf "$StoreName Finish update paymentmethod prices.\n";
    return;
}

sub usage {
    print <<END_USAGE;
Description:
    Recalculates the net prices of shops which have input taxmodel gross and not customized taxmatrix
    Used formula: newnetprice = netprice * oldtax / newtax
Usage:
    \$PERL $0 [options] [flags]
Options:
    -storename      store name for the database connection                              required
    -shop           name of the shop                                                    optional
    -taxmatrix      name of the TaxMatrix to be used                                    required
Optional parameters:
    -oldtaxnormal    value of the old taxrate e.g 1.077 (default 7.7%)                     optional
    -newtaxnormal    value of the new taxrate e.g 1.081 (default 8.1%)                     optional
    -oldtaxreduced   value of the old taxrate e.g 1.025 (default 2.5%)                      optional
    -newtaxreduced   value of the new taxrate e.g 1.026 (default 2.6%)                      optional
    -lastmerchantlogin  lastmerchantlogin should be newer than this (e.g. 2020-06-01)    optional
    -taxxml          import tax file names from Tax/Database/XML
                        (default  TaxesSwitzerland.xml)                                  optional
flags:
    -sfopenbymerchant   select only shops where ClosedByMerchant=0                       optional
    -closeshops         close shops which have inputtaxmodel=gross and use system taxmatrix
                                                            (default:yes)                optional
    -backup             backup tables                       (default:yes)
        (productprice, scaleprice, valuediscount, shippinglevel, shippingvalue, paymentlevel)
                                                                                            optional
    -removebackup       removed backuped tables                                             optional
    -sql                calculate prices only in db(is faster, but no hooks will be triggered) (default:yes) optional
    -importxml          import taxxmls to store            (default:yes)                       optional
    -calculate          runs calculation also if importstatusshopattr, closeshops, backup or importxml is set
                                                            (default:yes)                      optional
    -importstatusshopattr imports imp_AttributesShop.xml    (default:yes)                       optional
    -cleanup            removes backuped tables and attribute UpdateTaxCompleted               optional
    -test               Run script in test mode                                                optional
Example:
    \$PERL $0 -storename Store -taxmatrix TaxMatrixSwitzerland
    \$PERL $0 -storename Store -taxmatrix TaxMatrixSwitzerland -backup -closeshops -importxml -importstatusshopattr -nocalculate
    \$PERL $0 -storename Store -taxmatrix TaxMatrixSwitzerland -sfopenbymerchant -lastmerchantlogin 2020-06-01
    \$PERL $0 -storename Store -taxmatrix TaxMatrixSwitzerland -sfopenbymerchant
    \$PERL $0 -storename Store -taxmatrix TaxMatrixSwitzerland
    \$PERL $0 -storename Store -taxmatrix TaxMatrixSwitzerland -backup -closeshops -importxml
    \$PERL $0 -storename Store -taxmatrix TaxMatrixSwitzerland -oldtaxnormal 1.077 -newtaxnormal 1.081 -oldtaxreduced 1.025 -newtaxreduced 1.026
    \$PERL $0 -storename Store -shop DemoShop -taxmatrix TaxMatrixSwitzerland -oldtaxnormal 1.077 -newtaxnormal 1.081 -oldtaxreduced 1.025 -newtaxreduced 1.026
END_USAGE
    exit(0);
}
