package Test::Mock::LWP::Dispatch;
use strict;
our $mock_ua; # avoid warning: Name "Test::Mock::LWP::Dispatch::mock_ua" used only once: possible typo

## no critic qw(ProhibitMultiplePackages)
package DE_EPAGES::Twenga::t::TwengaRequest;
use base 'DE_EPAGES::Database::API::Test';
use strict;
use warnings;

use HTTP::Status qw( HTTP_OK );
use Test::More;

use DE_EPAGES::Core::API::Object::BaseObject;
use DE_EPAGES::Core::API::PerlTools qw( UsePackage );
use DE_EPAGES::Twenga::API::TwengaRequest qw( CallTwenga GetRedirectUrl GetSiteExists GetTrackingScript );

plan tests => 7;

# creates a mocked HTTP::Response object
sub MakeResponse {
    my ($Message) = @_;

    my $Response = HTTP::Response->new( HTTP_OK );
    $Response->content_type( 'application/xml' );
    $Response->content( <<"END_CONTENT" );
<?xml version="1.0" encoding="utf-8"?>
<response><message>$Message</message></response>
END_CONTENT

    return $Response;
}

sub init {
    my $self = shift;

    $self->{PortalConfig} = DE_EPAGES::Core::API::Object::BaseObject->new(
        PortalMerchantID => '98bcc090709d98402d172623970aa1ec',
        TwengaMerchantUsername => 'john.doe@twenga.com',
        TwengaMerchantPassword => 'h1kid3',
    );

    if( $ENV{SKIP_EXTERNAL_TESTS} ) { # use mock data instead
        UsePackage( 'Test::Mock::LWP::Dispatch' ); #exports $mock_ua
        my %ResponseMap = (
            '/api/Order/Cancel' => 'true',
            '/api/Site/Exist' => 'true',
            '/api/Site/GetSubscriptionLink' => 'https://rts.twenga.fr/signup/subscriptionkey/3215455299',
            '/api/Site/GetTrackingScript' => <<'END_SCRIPT',
&lt;script language="javascript" type="text/javascript" src="
http://tracker.twenga.co.uk/2d2c8394e31101a261abf1784302bf75"&gt;&lt;/script&gt;

&lt;script language="javascript" type="text/javascript"&gt;
var track = new twengaTrack();
track.totalHT = "88.79";
...
track.twSend();
&lt;/script&gt;
END_SCRIPT
        );

        # global mappings for requests and responses for LWP::UserAgent
        my $ua = $Test::Mock::LWP::Dispatch::mock_ua;
        $ua->map( qr{$_} => MakeResponse( $ResponseMap{$_} ) ) foreach keys %ResponseMap;
    }
}

sub portalConfig {
    return shift->{PortalConfig};
}

sub testCallTwenga {
    my $self = shift;

    my %Parms = (
        key         => $self->portalConfig->get('PortalMerchantID'),
        basket_id   => '12345',
        cli_email   => 'customer@example.com',
        user_id     => '12345',
    );

    my $hResult = CallTwenga $self->portalConfig, Order => 'Cancel', \%Parms;
    is $hResult->{error}, '', 'CallTwenga.error';
    is $hResult->{errormessage}, '', 'CallTwenga.errormessage';
    is $hResult->{content}, 'true', 'CallTwenga.content';
}

sub testGetRedirectUrl {
    my $self = shift;

    my %Params = (
        'site_id' => 'MY_SHOP_GUID',
        'country' => 'FR',
        'site_url' => 'http://myshop.de/',
        'feed_url' => 'http://myshop.de/Portals/Twenga/MyShop/feed.txt',
    );
    my $hResult = GetRedirectUrl( $self->portalConfig, \%Params );
    is $hResult->{error}, '', 'GetRedirectUrl.error';
    is $hResult->{errormessage}, '', 'GetRedirectUrl.errormessage';
    like $hResult->{content}, qr{^https\://(rts.twenga.fr|app.twenga-solutions.com/fr/)/signup/subscriptionkey/\d+}, 'GetRedirectUrl.content';
}

sub testGetSiteExists {
    my $self = shift;

    my $hValues = $self->portalConfig->get([qw(TwengaMerchantUsername TwengaMerchantPassword PortalMerchantID)]);
    my $hAuth = {
        merchant_login      => $hValues->{TwengaMerchantUsername},
        merchant_password   => $hValues->{TwengaMerchantPassword},
        merchant_hashkey    => $hValues->{PortalMerchantID},
    };

    my $hResult = GetSiteExists( $self->portalConfig, $hAuth );

    is $hResult->{content}, 'true', 'GetSiteExists';
}

sub test {
    my $self = shift;

    $self->testCallTwenga;
    $self->testGetRedirectUrl;
    $self->testGetSiteExists;
}

__PACKAGE__->runScript;
