use strict;
use warnings;

use Getopt::Std;
use HTTP::Status qw( HTTP_OK );
use LWP::UserAgent;
use Test::More;
use Test::Exception;

use DE_EPAGES::Core::API::PerlTools qw( UsePackage );
use DE_EPAGES::Twenga::API::Impl;

# old contenttype text/xml has changed to this:
use constant CTXML => 'application/xml';

plan tests => 5;

sub ERRBS {
    my ($Response) = @_;
    return "Request:\n" . $Response->request->as_string
        ."\nResponse:\n". $Response->as_string ."\n"
        .'Error talking to Twenga over the net; might be temporary; giving up.';
}

# ==== What is this supposed to do? ====================================
# Simple script to test Twenga API calls.
# Just test them one after the other.
# There is a little bit of configurability thanks to getopts.
# And there's some quirkiness with some of the Twenga API details.
# Just take it as it is, it has to be that way.



# ==== utility methods =================================================

sub xml {
    my( $twenga, $rsp, $msg_rex ) = @_;
    $msg_rex ||= qr/^true$/;

    my $hash = $twenga->data( $rsp );
    0 and Test::More::diag explain $hash;
    ok ! $hash->{error}, 'no error code present';
    ok ! $hash->{errormessage}, 'no error message present. '.$hash->{errormessage};
    like $hash->{content}, $msg_rex, 'content matches expected result: '.$msg_rex;
    return;
}

# ==== main ============================================================

my $aAccount; # [ login, password, hash_key, country ]
{
    getopts 'hvK:', \my %opt;
    if ( $opt{h} ) {
        print <<'EOS';
  -h            help
  -K            account (default=test)
EOS
        exit 0;
    }
    # which account?
    my %merchant_accounts = (
        tmetzinger  => [qw[tmetzinger@epages.de   8uakpusr    aa541ef3422583424d4ad9bb4681b2f0 FR]],
        balu        => [qw[balu@epages.de         n/oxklbb    0ffaca95e3e5242ba1097ad8a9a6e95d DE]],
        test        => [qw[test@epages.de         @4hbnatv    2d2c8394e31101a261abf1784302bf75 GB]],
        cartridge   => [qw[cartridge@epages.com   9jbnuox2    f3baf4b37d667198fe58da763b3d8976 ES]],
        johndoe     => [qw[john.doe@twenga.com    h1kid3      98bcc090709d98402d172623970aa1ec DE]],
    );
    $opt{K} //= 'johndoe'; # default account
    $aAccount = $merchant_accounts{ $opt{K} } or die "invalid account [$opt{K}]!";
    Test::More::diag "Using account: @$aAccount";
}

# Partner accounts
# ============================
# prod_epages   D5vuV9Z98n
# dev_epages    sxwoxhfp
# ============================

my $twenga = Twenga->new(
    partner_auth_key => '1f2027e88b6964f3b952212b6d15397ae2a7d74a',
#   test_mode        => 1,
);

my $api_url = $twenga->api_url;
my $twenga_url = qr{^$api_url\w+/\w+/?\?.*PARTNER_AUTH_KEY}x;
my $ua = LWP::UserAgent->new( keep_alive => 1 );

# creates a mocked HTTP::Response object
sub MakeResponse {
    my ($Message) = @_;

    my $Response = HTTP::Response->new( HTTP_OK );
    $Response->content_type( CTXML );
    $Response->content( <<"END_CONTENT" );
<?xml version="1.0" encoding="utf-8"?>
<response><message>$Message</message></response>
END_CONTENT

    return $Response;
}

sub MakeErrorResponse {
    my ($Message) = @_;

    my $Response = HTTP::Response->new( HTTP_OK );
    $Response->content_type( CTXML );
    $Response->content( <<"END_CONTENT" );
<?xml version="1.0" encoding="utf-8"?>
<error><message>$Message</message></error>
END_CONTENT

    return $Response;
}

if( $ENV{SKIP_EXTERNAL_TESTS} ) { # use mock data instead
    UsePackage( 'Test::Mock::LWP::Dispatch' );
    my %ResponseMap = (
        '/api/Site/GetSubscriptionLink' => 'https://rts.twenga.fr/signup/subscriptionkey/3215455299',
        '/api/Site/Exist' => 'true',
    );
    $ua->map( qr{$_} => MakeResponse( $ResponseMap{$_} ) ) foreach keys %ResponseMap;
    $ua->map( qr{/api/Site/Activate} => MakeErrorResponse( 'Tracking disable.' ) );
}

subtest 'General' => sub {
    plan tests => 3;

    isa_ok $twenga, 'Twenga';

    # md5 as hex, then as Base64 (doesn't make much sense, but it's like that)
    is $twenga->partner_auth_key_b64,
        'MWYyMDI3ZTg4YjY5NjRmM2I5NTIyMTJiNmQxNTM5N2FlMmE3ZDc0YQ==', 'PARTNER_AUTH_KEY correct'
        or die 'wrong PARTNER_AUTH_KEY';

    is $api_url, 'https://rts.twenga.com/api/', 'API URL ok';
};

subtest 'GetSubscriptionLink' => sub {
    plan tests => 8;

    # test Site/GetSubscriptionLink call w/o auth (will work)
    my $Request = $twenga->make_req( Site => GetSubscriptionLink => {
            country     => 'FR',
            feed_url    => 'http://bla.de/feed.txt',
            site_url    => 'http://bla.de/',
        } );
    isa_ok $Request, 'HTTP::Request';
    like $Request->uri, $twenga_url, 'URL ~ ok';
    ok ! $Request->header( 'Authorization' ), 'no Authorization header';

    my $Response = $ua->request( $Request );
    ok $Response->is_success, 'response success' or die ERRBS($Response);
    is $Response->content_type, CTXML, 'response has correct Content-Type' or die ERRBS($Response);

    xml( $twenga, $Response, qr{^https\://(rts.twenga.fr|app.twenga-solutions.com/fr/)/signup/subscriptionkey/\d+} );
};

subtest 'no_authentication' => sub {
    plan tests => 1;

    # test Site/Exist call w/o auth (won't work)
    throws_ok { $twenga->make_req( Site => Exist => {} ) } qr/\bAttribute\b.*\bkey\b/;
};

# add credentials for auth and retry
$twenga->merchant_login(    $aAccount->[0] );
$twenga->merchant_password( $aAccount->[1] );
$twenga->merchant_hashkey(  $aAccount->[2] );

subtest 'with_athentication' => sub {
    plan tests => 6;

    # test Site/Exist call with auth
    my $Request = $twenga->make_req( Site => Exist => { } );
    isa_ok $Request, 'HTTP::Request';
    like $Request->uri, $twenga_url, 'URL ~ ok';
    my $auth = $Request->header( 'Authorization' );
    ok $auth, 'has Authorization header';
    like $auth, qr/^Basic /, 'uses "Basic" authentication';

    my $Response = $ua->request( $Request );
    ok $Response->is_success, 'response success' or die ERRBS($Response);
    is $Response->content_type, CTXML, 'correct contenttype ' or die ERRBS($Response);
};

subtest 'Activate' => sub {
    plan tests => 3;

    my $Request = $twenga->make_req( Site => Activate => { } );
    ok $Request->header( 'Authorization' ), 'request has Authorization header';

    my $Response = $ua->request( $Request );
    ok $Response->is_success, 'response success' or die ERRBS($Response);
    is $Response->content_type, CTXML, 'response has correct Content-Type' or die ERRBS($Response);
};
